!
! Copyright (c) 2000-2013,2015 Yoshihide Yoshimoto
!
! for rotwfn
!
module arch_rotwfn
contains
  subroutine ltrovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    complex(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    complex(kind=8):: g(m,m)

    call tpf_begin(tpf_arch)

    if (m.gt.0) then
       call zgemm('C','N',m,m,e-b+1,(1.0d0,0.0d0),x,e-b+1,y,e-b+1, &
            (1.0d0,0.0d0),g,m)
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ltrovlmat

  subroutine ovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    complex(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    complex(kind=8):: g(m,m)

    call tpf_begin(tpf_arch)

    if (m.gt.0) then
       call zgemm('C','N',m,m,e-b+1,(1.0d0,0.0d0),x,e-b+1,y,e-b+1, &
            (1.0d0,0.0d0),g,m)
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotvecs(b,e,m,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    complex(kind=8),intent(in):: a(m,m)
    complex(kind=8),intent(in):: x(b:e,m)
    complex(kind=8):: y(b:e,m)

    call tpf_begin(tpf_arch)

    if (m.gt.0) then
       call zgemm('N','N',e-b+1,m,m,(1.0d0,0.0d0),x,e-b+1,a,m, &
            (1.0d0,0.0d0),y,e-b+1)
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine rotvecs
end module arch_rotwfn
!
! for rotwfni
!
module arch_rotwfni
contains
  subroutine ltrovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    real(kind=8):: g(m,m)

    call tpf_begin(tpf_arch)

    call dgemm('T','N',m,m,e-b+1,1.0d0,x,e-b+1,y,e-b+1,1.0d0,g,m)

    call tpf_end(tpf_arch)
    return
  end subroutine ltrovlmat

  subroutine ovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    real(kind=8):: g(m,m)

    call tpf_begin(tpf_arch)

    call dgemm('T','N',m,m,e-b+1,1.0d0,x,e-b+1,y,e-b+1,1.0d0,g,m)

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotvecs(b,e,m,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: a(m,m), x(b:e,m)
    real(kind=8):: y(b:e,m)

    call tpf_begin(tpf_arch)

    call dgemm('N','N',e-b+1,m,m,1.0d0,x,e-b+1,a,m,1.0d0,y,e-b+1)

    call tpf_end(tpf_arch)
    return
  end subroutine rotvecs
end module arch_rotwfni
!
! for rotwfng
!
module arch_rotwfng
contains
  subroutine ltrovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    complex(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    real(kind=8):: g(m,m)

    integer:: j,k
    integer:: b2

    call tpf_begin(tpf_arch)

    b2 = max(b,2)

    ! 
    ! intentionally passing x,y as a real(kind=8)*2 array
    !
    call dgemm('T','N',m,m,2*(e-b2+1),1.0d0, &
         x(b2,1),2*(e-b+1),y(b2,1),2*(e-b+1),1.0d0,g,m)

    g = 2.0d0*g
    if (b.eq.1) then
       do k=1,m
          do j=k,m
             g(j,k) = g(j,k) + dble(x(1,j))*dble(y(1,k))
          end do
       end do
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ltrovlmat

  subroutine ovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    complex(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    real(kind=8):: g(m,m)

    integer:: j,k
    integer:: b2

    call tpf_begin(tpf_arch)

    b2 = max(b,2)

    ! 
    ! intentionally passing x,y as a real(kind=8)*2 array
    !
    call dgemm('T','N',m,m,2*(e-b2+1),1.0d0, &
         x(b2,1),2*(e-b+1),y(b2,1),2*(e-b+1),1.0d0,g,m)

    g = 2.0d0*g
    if (b.eq.1) then
       do k=1,m
          do j=1,m
             g(j,k) = g(j,k) + dble(x(1,j))*dble(y(1,k))
          end do
       end do
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotvecs(b,e,m,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: a(m,m)
    complex(kind=8),intent(in):: x(b:e,m)
    complex(kind=8):: y(b:e,m)

    call tpf_begin(tpf_arch)

    ! 
    ! intentionally passing x,y as a real(kind=8)*2 array
    !
    call dgemm('N','N',2*(e-b+1),m,m,1.0d0, &
         x,2*(e-b+1),a,m,1.0d0,y,2*(e-b+1))

    call tpf_end(tpf_arch)
    return
  end subroutine rotvecs
end module arch_rotwfng
!
! for nlprj
!
module arch_nlprj
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: x(b:e,m1),y(b:e,m2)
    complex(kind=8):: g(m1,m2)

    call tpf_begin(tpf_arch)

    if ( (m1.gt.0) .and. (m2.gt.0) ) then
       call zgemm('C','N',m1,m2,e-b+1,(1.0d0,0.0d0), &
            x,e-b+1,y,e-b+1,(0.0d0,0.0d0),g,m1)
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotprjs(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: a(m1,m2)
    complex(kind=8),intent(in):: x(b:e,m1)
    complex(kind=8):: y(b:e,m2)

    call tpf_begin(tpf_arch)

    if ( (m1.gt.0) .and. (m2.gt.0) ) then
       call zgemm('N','N',e-b+1,m2,m1,(1.0d0,0.0d0),x,e-b+1,a,m1, &
            (1.0d0,0.0d0),y,e-b+1)
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine rotprjs
end module arch_nlprj

module arch_nlprji
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: x(b:e,m1,2)
    real(kind=8),intent(in):: y(b:e,m2)
    complex(kind=8):: g(m1,m2)

    integer:: j,k
    real(kind=8),allocatable:: gr(:,:), gi(:,:)

    call tpf_begin(tpf_arch)

    allocate(gr(m1,m2),gi(m1,m2))

    call dgemm('T','N',m1,m2,e-b+1,1.0d0,x(b,1,1),e-b+1,y,e-b+1,0.0d0,gr,m1)

    call dgemm('T','N',m1,m2,e-b+1,1.0d0,x(b,1,2),e-b+1,y,e-b+1,0.0d0,gi,m1)

!$omp parallel do private(j,k)
    do k=1,m2
       do j=1,m1
          g(j,k) = dcmplx(gr(j,k),- gi(j,k))
       end do
    end do

    deallocate(gr,gi)

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotprjs(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: a(m1,m2)
    real(kind=8),intent(in):: x(b:e,m1,2)
    real(kind=8):: y(b:e,m2)

    integer:: j,k
    real(kind=8),allocatable:: ar(:,:),ai(:,:)

    call tpf_begin(tpf_arch)

    allocate(ar(m1,m2),ai(m1,m2))

!$omp parallel do private(j,k)
    do k=1,m2
       do j=1,m1
          ar(j,k) = dble(a(j,k))
          ai(j,k) = aimag(a(j,k))
       end do
    end do

    call dgemm('N','N',e-b+1,m2,m1,1.0d0,x(b,1,1),e-b+1,ar,m1,1.0d0,y,e-b+1)

    call dgemm('N','N',e-b+1,m2,m1,-1.0d0,x(b,1,2),e-b+1,ai,m1,1.0d0,y,e-b+1)

    deallocate(ar,ai)

    call tpf_end(tpf_arch)
    return
  end subroutine rotprjs
end module arch_nlprji

module arch_nlprjg
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: x(b:e,m1),y(b:e,m2)
    real(kind=8):: g(m1,m2)

    integer:: j,k
    integer:: b2

    call tpf_begin(tpf_arch)

    g = 0.0d0

    b2 = max(b,2)

    ! 
    ! intentionally passing x,y as a real(kind=8)*2 array
    !
    call dgemm('T','N',m1,m2,2*(e-b2+1),1.0d0, &
         x(b2,1),2*(e-b+1),y(b2,1),2*(e-b+1),1.0d0,g,m1)

    g = 2.0d0*g
    if (b.eq.1) then
       do k=1,m2
          do j=1,m1
             g(j,k) = g(j,k) + dble(x(1,j))*dble(y(1,k))
          end do
       end do
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotprjs(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: a(m1,m2)
    complex(kind=8),intent(in):: x(b:e,m1)
    complex(kind=8):: y(b:e,m2)

    call tpf_begin(tpf_arch)

    ! 
    ! intentionally passing x,y as a real(kind=8)*2 array
    !
    call dgemm('N','N',2*(e-b+1),m2,m1,1.0d0, &
         x,2*(e-b+1),a,m1,1.0d0,y,2*(e-b+1))

    call tpf_end(tpf_arch)
    return
  end subroutine rotprjs
end module arch_nlprjg
!
! for sorthcl
!
module arch_sorthcl
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: x(b:e,m1),y(b:e,m2)
    complex(kind=8):: g(m1,m2)

    call tpf_begin(tpf_arch)

    if ( (m1.gt.0) .and. (m2.gt.0) ) then
       call zgemm('C','N',m1,m2,e-b+1,(1.0d0,0.0d0),x,e-b+1,y,e-b+1, &
            (0.0d0,0.0d0),g,m1)
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine subovlpt(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: a(m1,m2)
    complex(kind=8),intent(in):: x(b:e,m1)
    complex(kind=8):: y(b:e,m2)

    call tpf_begin(tpf_arch)

    if ( (m1.gt.0) .and. (m2.gt.0) ) then 
       call zgemm('N','N',e-b+1,m2,m1,(-1.0d0,0.0d0),x,e-b+1,a,m1, &
            (1.0d0,0.0d0),y,e-b+1)
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine subovlpt
end module arch_sorthcl
!
! for sorthcli
!
module arch_sorthcli
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: x(b:e,m1),y(b:e,m2)
    real(kind=8):: g(m1,m2)

    call tpf_begin(tpf_arch)

    call dgemm('T','N',m1,m2,e-b+1,1.0d0,x,e-b+1,y,e-b+1,0.0d0,g,m1)

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine subovlpt(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: a(m1,m2)
    real(kind=8),intent(in):: x(b:e,m1)
    real(kind=8):: y(b:e,m2)

    call tpf_begin(tpf_arch)

    call dgemm('N','N',e-b+1,m2,m1,-1.0d0,x,e-b+1,a,m1,1.0d0,y,e-b+1)

    call tpf_end(tpf_arch)
    return
  end subroutine subovlpt
end module arch_sorthcli
!
! for sorthclg
!
module arch_sorthclg
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: x(b:e,m1),y(b:e,m2)
    real(kind=8):: g(m1,m2)

    integer:: j,k
    integer:: b2

    call tpf_begin(tpf_arch)

    g = 0.0d0

    b2 = max(b,2)

    ! 
    ! intentionally passing x,y as a real(kind=8)*2 array
    !
    call dgemm('T','N',m1,m2,2*(e-b2+1),1.0d0, &
         x(b2,1),2*(e-b+1),y(b2,1),2*(e-b+1),0.0d0,g,m1)

    g = 2.0d0*g
    if (b.eq.1) then
       do k=1,m2
          do j=1,m1
             g(j,k) = g(j,k) + dble(x(1,j))*dble(y(1,k))
          end do
       end do
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine subovlpt(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: a(m1,m2)
    complex(kind=8),intent(in):: x(b:e,m1)
    complex(kind=8):: y(b:e,m2)

    call tpf_begin(tpf_arch)

    ! 
    ! intentionally passing x,y as a real(kind=8)*2 array
    !
    call dgemm('N','N',2*(e-b+1),m2,m1,-1.0d0, &
         x,2*(e-b+1),a,m1,1.0d0,y,2*(e-b+1))

    call tpf_end(tpf_arch)
    return
  end subroutine subovlpt
end module arch_sorthclg

module arch_engdffr
  integer,parameter:: blkpw = 3000, blkprj = 300
end module arch_engdffr
