!
! Copyright (c) 2000-2013,2015,2017 Yoshihide Yoshimoto
!
! for rotwfn
!
module arch_rotwfn
contains
  subroutine ltrovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    complex(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    complex(kind=8):: g(m,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do schedule(static,1)
!POPTION CYCLIC
    do k=1,m
!SOPTION NOLOOPINTERCHANGE, UNROLL(4)
!CDIR NOLOOPCHG
       do j=k,m
          do i=b,e
             g(j,k) = g(j,k) + conjg(x(i,j))*y(i,k)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ltrovlmat

  subroutine ovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    complex(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    complex(kind=8):: g(m,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(3)
    do k=1,m
!SOPTION NOLOOPINTERCHANGE, UNROLL(3)
!CDIR NOLOOPCHG
       do j=1,m
          do i=b,e
             g(j,k) = g(j,k) + conjg(x(i,j))*y(i,k)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotvecs(b,e,m,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    complex(kind=8),intent(in):: a(m,m)
    complex(kind=8),intent(in):: x(b:e,m)
    complex(kind=8):: y(b:e,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(6)
    do k=1,m
!SOPTION UNROLL(6)
       do j=1,m
          do i=b,e
             y(i,k) = y(i,k) + a(j,k)*x(i,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine rotvecs
end module arch_rotwfn
!
! for rotwfni
!
module arch_rotwfni
contains
  subroutine ltrovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    real(kind=8):: g(m,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do schedule(static,1)
!POPTION CYCLIC
    do k=1,m
!SOPTION NOLOOPINTERCHANGE, UNROLL(4)
!CDIR NOLOOPCHG
       do j=k,m
          do i=b,e
             g(j,k) = g(j,k) + x(i,j)*y(i,k)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ltrovlmat

  subroutine ovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    real(kind=8):: g(m,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(5)
    do k=1,m
!SOPTION NOLOOPINTERCHANGE, UNROLL(5)
!CDIR NOLOOPCHG
       do j=1,m
          do i=b,e
             g(j,k) = g(j,k) + x(i,j)*y(i,k)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotvecs(b,e,m,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: a(m,m), x(b:e,m)
    real(kind=8):: y(b:e,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(5)
    do k=1,m
!SOPTION UNROLL(5)
       do j=1,m
          do i=b,e
             y(i,k) = y(i,k) + a(j,k)*x(i,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine rotvecs
end module arch_rotwfni
!
! for rotwfng
!
module arch_rotwfng
contains
  subroutine ltrovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    complex(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    real(kind=8):: g(m,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do schedule(static,1)
!POPTION CYCLIC
    do k=1,m
!SOPTION NOLOOPINTERCHANGE, UNROLL(4)
!CDIR NOLOOPCHG
       do j=k,m
          do i=max(b,2),e
             g(j,k) = g(j,k) + dble(x(i,j))*dble(y(i,k)) &
                  + aimag(x(i,j))*aimag(y(i,k))
          end do
       end do
    end do
    g = 2.0d0*g
    if (b.eq.1) then
       do k=1,m
          do j=k,m
             g(j,k) = g(j,k) + dble(x(1,j))*dble(y(1,k))
          end do
       end do
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ltrovlmat

  subroutine ovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    complex(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    real(kind=8):: g(m,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(3)
    do k=1,m
!SOPTION NOLOOPINTERCHANGE, UNROLL(3)
!CDIR NOLOOPCHG
       do j=1,m
          do i=max(b,2),e
             g(j,k) = g(j,k) + dble(x(i,j))*dble(y(i,k)) &
                  + aimag(x(i,j))*aimag(y(i,k))
          end do
       end do
    end do
    g = 2.0d0*g
    if (b.eq.1) then
       do k=1,m
          do j=1,m
             g(j,k) = g(j,k) + dble(x(1,j))*dble(y(1,k))
          end do
       end do
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotvecs(b,e,m,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: a(m,m)
    complex(kind=8),intent(in):: x(b:e,m)
    complex(kind=8):: y(b:e,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(6)
    do k=1,m
!SOPTION UNROLL(6)
       do j=1,m
          do i=b,e
             y(i,k) = y(i,k) + a(j,k)*x(i,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine rotvecs
end module arch_rotwfng
!
! for nlprj
!
module arch_nlprj
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: x(b:e,m1),y(b:e,m2)
    complex(kind=8):: g(m1,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

    g = (0.0d0, 0.0d0)
!$omp parallel do private(tr,ti)
!SOPTION UNROLL(3)
    do k=1,m2
!SOPTION NOLOOPINTERCHANGE,UNROLL(3)
!CDIR NOLOOPCHG
       do j=1,m1
          do i=b,e
             g(j,k) = g(j,k) + conjg(x(i,j))*y(i,k)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotprjs(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: a(m1,m2)
    complex(kind=8),intent(in):: x(b:e,m1)
    complex(kind=8):: y(b:e,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(6)
    do k=1,m2
!SOPTION UNROLL(6)
       do j=1,m1
          do i=b,e
             y(i,k) = y(i,k) + a(j,k)*x(i,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine rotprjs
end module arch_nlprj

module arch_nlprji
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: x(b:e,m1,2)
    real(kind=8),intent(in):: y(b:e,m2)
    complex(kind=8):: g(m1,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

    g = (0.0d0, 0.0d0)
!$omp parallel do private(tr,ti)
!SOPTION UNROLL(4)
    do k=1,m2
!SOPTION NOLOOPINTERCHANGE, UNROLL(2)
!CDIR NOLOOPCHG
       do j=1,m1
          do i=b,e
             g(j,k) = g(j,k) + conjg(dcmplx(x(i,j,1),x(i,j,2)))*y(i,k)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotprjs(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: a(m1,m2)
    real(kind=8),intent(in):: x(b:e,m1,2)
    real(kind=8):: y(b:e,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(4)
    do k=1,m2
!SOPTION UNROLL(2)
       do j=1,m1
          do i=b,e
             y(i,k) = y(i,k) &
                  + dble(a(j,k))*x(i,j,1) - aimag(a(j,k))*x(i,j,2)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine rotprjs
end module arch_nlprji

module arch_nlprjg
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: x(b:e,m1),y(b:e,m2)
    real(kind=8):: g(m1,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

    g = 0.0d0
!$omp parallel do
!SOPTION UNROLL(3)
    do k=1,m2
!SOPTION NOLOOPINTERCHANGE,UNROLL(3)
!CDIR NOLOOPCHG
       do j=1,m1
          do i=max(b,2),e
             g(j,k) = g(j,k) + dble(x(i,j))*dble(y(i,k)) &
                  + aimag(x(i,j))*aimag(y(i,k))
          end do
       end do
    end do
    g = 2.0d0*g
    if (b.eq.1) then
       do k=1,m2
          do j=1,m1
             g(j,k) = g(j,k) + dble(x(1,j))*dble(y(1,k))
          end do
       end do
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotprjs(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: a(m1,m2)
    complex(kind=8),intent(in):: x(b:e,m1)
    complex(kind=8):: y(b:e,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(6)
    do k=1,m2
!SOPTION UNROLL(6)
       do j=1,m1
          do i=b,e
             y(i,k) = y(i,k) + a(j,k)*x(i,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine rotprjs
end module arch_nlprjg
!
! for sorthcl
!
module arch_sorthcl
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: x(b:e,m1),y(b:e,m2)
    complex(kind=8):: g(m1,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

    g = (0.0d0, 0.0d0)
!$omp parallel do private(tr,ti)
!SOPTION UNROLL(3)
    do k=1,m2
!SOPTION NOLOOPINTERCHANGE,UNROLL(3)
!CDIR NOLOOPCHG
       do j=1,m1
          do i=b,e
             g(j,k) = g(j,k) + conjg(x(i,j))*y(i,k)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine subovlpt(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: a(m1,m2)
    complex(kind=8),intent(in):: x(b:e,m1)
    complex(kind=8):: y(b:e,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(2)
    do k=1,m2
!SOPTION UNROLL(4)
       do j=1,m1
          do i=b,e
             y(i,k) = y(i,k) - a(j,k)*x(i,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine subovlpt
end module arch_sorthcl
!
! for sorthcli
!
module arch_sorthcli
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: x(b:e,m1),y(b:e,m2)
    real(kind=8):: g(m1,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

    g = 0.0d0
!$omp parallel do
!SOPTION UNROLL(4)
    do k=1,m2
!SOPTION UNROLL(4)
       do j=1,m1
!SOPTION NOLOOPINTERCHANGE
!CDIR NOLOOPCHG
          do i=b,e
             g(j,k) = g(j,k) + x(i,j)*y(i,k)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine subovlpt(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: a(m1,m2)
    real(kind=8),intent(in):: x(b:e,m1)
    real(kind=8):: y(b:e,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(4)
    do k=1,m2
!SOPTION UNROLL(4)
       do j=1,m1
          do i=b,e
             y(i,k) = y(i,k) - a(j,k)*x(i,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine subovlpt
end module arch_sorthcli
!
! for sorthclg
!
module arch_sorthclg
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: x(b:e,m1),y(b:e,m2)
    real(kind=8):: g(m1,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

    g = 0.0d0
!$omp parallel do
!SOPTION UNROLL(3)
    do k=1,m2
!SOPTION NOLOOPINTERCHANGE,UNROLL(3)
!CDIR NOLOOPCHG
       do j=1,m1
          do i=max(b,2),e
             g(j,k) = g(j,k) + dble(x(i,j))*dble(y(i,k)) &
                  + aimag(x(i,j))*aimag(y(i,k))
          end do
       end do
    end do
    g = 2.0d0*g
    if (b.eq.1) then
       do k=1,m2
          do j=1,m1
             g(j,k) = g(j,k) + dble(x(1,j))*dble(y(1,k))
          end do
       end do
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine subovlpt(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: a(m1,m2)
    complex(kind=8),intent(in):: x(b:e,m1)
    complex(kind=8):: y(b:e,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(2)
    do k=1,m2
!SOPTION UNROLL(4)
       do j=1,m1
          do i=b,e
             y(i,k) = y(i,k) - a(j,k)*x(i,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine subovlpt
end module arch_sorthclg

module arch_engdffr
  integer,parameter:: blkpw = 3000, blkprj = 300
end module arch_engdffr
