!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015-2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module atomicchrg
contains
  subroutine mkatmchrg(bspotfc,nkd,bspots,nkfxa,kf,aff,va, &
       nsi,asi,kd,nrx2,nry2,nrz2,nspin,ndens,udirchr,rho)
    use bsplpot
    use fft_3d
    implicit none
    type (bspotfc_struct):: bspotfc
    integer,intent(in):: nkd
    type (bspot_for_atm):: bspots(nkd)
    integer,intent(in):: nkfxa,nsi,nrx2,nry2,nrz2,nspin,ndens
    integer,intent(in):: kf(3,nkfxa),kd(nsi)
    real(kind=8),intent(in):: aff(nkfxa),va,asi(3,nsi)
    real(kind=8),intent(out):: FFT_A3M_DEFSZ(rho,nrx2,nry2,nrz2,nspin*ndens)
    real(kind=8),pointer:: udirchr(:,:)

    real(kind=8),parameter:: p2 = 6.28318530717958648d0

    integer:: it,i1,i2,i3,j1,j2,j3,k,is,ierr
    real(kind=8):: rlp(2),imgp(2),cc,ss,pp
    real(kind=8),allocatable:: ztmp(:,:,:,:),achg(:,:,:)
    integer,allocatable:: flipchr(:)
    real(kind=8):: pol
    complex(kind=8):: ctemp,chr(4)
    type (fft3_struct):: fs

    integer:: ic1, ic2

    ! from the atomic charge of solps.
    allocate(achg(nkfxa,2,nkd), stat=ierr)
    call chkalloc('mkatmchrg','achg',ierr)
    call evachg(nkd,bspotfc,bspots,nkfxa,aff,va,achg)

    rho = 0.0d0

    if (ndens.eq.1) then
       if (nspin.eq.1) then
!$omp parallel do private(rlp,imgp,it,i1,i2,i3,j1,j2,j3,pp,cc,ss)
!POPTION INDEP(rho)
!CDIR NOSYNC
          do k=1,nkfxa
             i1=kf(1,k)
             i2=kf(2,k)
             i3=kf(3,k)
             rlp(1)=0.0d0
             imgp(1)=0.0d0
             do it=1,nsi
                pp=-p2*(asi(1,it)*i1+asi(2,it)*i2+asi(3,it)*i3)
                cc=cos(pp)
                ss=sin(pp)
                rlp(1) = rlp(1) + cc*( achg(k,1,kd(it))+achg(k,2,kd(it)) )
                imgp(1) = imgp(1) + ss*( achg(k,1,kd(it))+achg(k,2,kd(it)) )
             end do
             j1=mod(nrx2+i1,nrx2)+1
             j2=mod(nry2+i2,nry2)+1
             j3=mod(nrz2+i3,nrz2)+1
             FFT_A3M_RL(rho,j1,j2,j3,1)=rlp(1)
             FFT_A3M_IM(rho,j1,j2,j3,1)=imgp(1)
          end do
       else
          allocate(flipchr(nsi))
          do it=1,nsi
             if (udirchr(3,nsi).gt.0.0d0) then
                flipchr(it) = 0
             else
                flipchr(it) = 1
             end if
          end do
!$omp parallel do private(rlp,imgp,it,i1,i2,i3,j1,j2,j3,pp,cc,ss,ic1,ic2)
!POPTION INDEP(rho)
!CDIR NOSYNC
          do k=1,nkfxa
             i1=kf(1,k)
             i2=kf(2,k)
             i3=kf(3,k)
             rlp=0.0d0
             imgp=0.0d0
             do it=1,nsi
                pp=-p2*(asi(1,it)*i1+asi(2,it)*i2+asi(3,it)*i3)
                cc=cos(pp)
                ss=sin(pp)
                ic1 = 1+flipchr(it)
                ic2 = 2-flipchr(it)
                rlp(1) = rlp(1) + cc*achg(k,ic1,kd(it))
                imgp(1) = imgp(1) + ss*achg(k,ic1,kd(it))
                rlp(2) = rlp(2) + cc*achg(k,ic2,kd(it))
                imgp(2) = imgp(2) + ss*achg(k,ic2,kd(it))
             end do
             j1=mod(nrx2+i1,nrx2)+1
             j2=mod(nry2+i2,nry2)+1
             j3=mod(nrz2+i3,nrz2)+1
             FFT_A3M_RL(rho,j1,j2,j3,1)=rlp(1)
             FFT_A3M_IM(rho,j1,j2,j3,1)=imgp(1)
             FFT_A3M_RL(rho,j1,j2,j3,2)=rlp(2)
             FFT_A3M_IM(rho,j1,j2,j3,2)=imgp(2)
          end do
          deallocate(flipchr)
       end if
    else
!$omp parallel do private(it,i1,i2,i3,j1,j2,j3,pp,ctemp,chr,pol)
!POPTION INDEP(rho)
!CDIR NOSYNC
       do k=1,nkfxa
          i1=kf(1,k)
          i2=kf(2,k)
          i3=kf(3,k)
          chr = 0.0d0
          do it=1,nsi
             pp = -p2*(asi(1,it)*i1+asi(2,it)*i2+asi(3,it)*i3)
             ctemp = dcmplx(cos(pp),sin(pp))
             chr(1) = chr(1) + ctemp*( achg(k,1,kd(it))+achg(k,2,kd(it)) )
             pol = achg(k,1,kd(it)) - achg(k,2,kd(it))
             chr(2:4) = chr(2:4) + ctemp*pol*udirchr(1:3,it)
          end do
          j1 = mod(nrx2+i1,nrx2)+1
          j2 = mod(nry2+i2,nry2)+1
          j3 = mod(nrz2+i3,nrz2)+1
          FFT_A3M_RL(rho,j1,j2,j3,1:4) = dble(chr(1:4))
          FFT_A3M_IM(rho,j1,j2,j3,1:4) = aimag(chr(1:4))
       end do
    end if

    call fft3_init(nrx2,nry2,nrz2,nrx2,nry2,nrz2,fs)
    allocate(FFT_A3D_DEFSZ(ztmp,nrx2,nry2,nrz2), stat=ierr)
    call chkalloc('mkatmchrg','ztmp',ierr)
    do is=1,nspin*ndens
       call fft3_bw(fs,FFT_A3M_RL(rho,1,1,1,is),ztmp)
    end do
    call fft3_done(fs)
    deallocate(ztmp,achg)
    return
  end subroutine mkatmchrg

  subroutine mkionchrg(bspotfc,nkd,bspots,nkfxa,kf,aff,va, &
       nsi,asi,kd,nrx2,nry2,nrz2,rho)
    use bsplpot
    use fft_3d
    implicit none
    type (bspotfc_struct):: bspotfc
    integer,intent(in):: nkd
    type (bspot_for_atm):: bspots(nkd)
    integer,intent(in):: nkfxa,nsi,nrx2,nry2,nrz2
    integer,intent(in):: kf(3,nkfxa),kd(nsi)
    real(kind=8),intent(in):: aff(nkfxa),va,asi(3,nsi)
    real(kind=8),intent(out):: FFT_A3D_DEFSZ(rho,nrx2,nry2,nrz2)

    real(kind=8),parameter:: p2 = 6.28318530717958648d0

    integer:: it,i1,i2,i3,j1,j2,j3,k,ierr
    real(kind=8):: rlp,imgp,cc,ss,pp
    real(kind=8),allocatable:: ztmp(:,:,:,:),ionchg(:,:)
    type (fft3_struct):: fs

    ! from the atomic charge of solps.
    allocate(ionchg(nkfxa,nkd), stat=ierr)
    call chkalloc('mkatmchrg','ionchg',ierr)
    call evionchg(nkd,bspotfc,bspots,nkfxa,aff,va,ionchg)

    rho = 0.0d0

!$omp parallel do private(rlp,imgp,it,i1,i2,i3,j1,j2,j3,pp,cc,ss)
!POPTION INDEP(rho)
!CDIR NOSYNC
    do k=1,nkfxa
       i1=kf(1,k)
       i2=kf(2,k)
       i3=kf(3,k)
       rlp=0.0d0
       imgp=0.0d0
       do it=1,nsi
          pp=-p2*(asi(1,it)*i1+asi(2,it)*i2+asi(3,it)*i3)
          cc=cos(pp)
          ss=sin(pp)
          rlp = rlp + cc*ionchg(k,kd(it))
          imgp = imgp + ss*ionchg(k,kd(it))
       end do
       j1=mod(nrx2+i1,nrx2)+1
       j2=mod(nry2+i2,nry2)+1
       j3=mod(nrz2+i3,nrz2)+1
       FFT_A3D_RL(rho,j1,j2,j3)=rlp
       FFT_A3D_IM(rho,j1,j2,j3)=imgp
    end do

    call fft3_init(nrx2,nry2,nrz2,nrx2,nry2,nrz2,fs)
    allocate(FFT_A3D_DEFSZ(ztmp,nrx2,nry2,nrz2), stat=ierr)
    call chkalloc('mkatmchrg','ztmp',ierr)
    call fft3_bw(fs,FFT_A3D_RL(rho,1,1,1),ztmp)

    call fft3_done(fs)
    deallocate(ztmp,ionchg)
    return
  end subroutine mkionchrg
end module atomicchrg
