!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015-2017 Yoshihide Yoshimoto
!
module bsplpot
  integer,private,parameter:: bspot_for_atm_lmax = 9
  integer,private,parameter:: bspot_for_atm_maxr = 3
  type bspotfc_struct
     integer:: ibsrk,nbsktf,nbsf
     real(kind=8),pointer:: bsxif(:),bsaf(:)
  end type bspotfc_struct
  abstract interface
     function l2lq_func(il) result(ilq)
       integer, intent(in):: il
       integer:: ilq
     end function l2lq_func

     function l2nm_func(il) result(nm)
       integer, intent(in):: il
       integer:: nm
     end function l2nm_func
  end interface
  type bspot_for_atm
     character(len=256):: name
     integer:: nob,na2,ldnr,nl
     integer:: nr(bspot_for_atm_lmax)
     integer:: nprj
     integer:: ip2l(bspot_for_atm_lmax**2 * bspot_for_atm_maxr)
     integer:: ip2r(bspot_for_atm_lmax**2 * bspot_for_atm_maxr)
     integer:: ip2m(bspot_for_atm_lmax**2 * bspot_for_atm_maxr)
     integer:: l2ph(bspot_for_atm_lmax)
     integer:: l2pt(bspot_for_atm_lmax)
     logical:: ifncpp,ifachg,ifionchg
     integer:: ifcr(bspot_for_atm_maxr,bspot_for_atm_lmax)
     real(kind=8):: zo,zn
     real(kind=8):: a1,a2,a3,a4
     real(kind=8):: rc
     integer:: nsmpl
     real(kind=8),pointer:: vgsr(:),ccg(:),achg(:,:),ionchg(:)
     real(kind=8),pointer:: ddi(:,:,:),qqr(:,:,:),qq(:,:,:)
     real(kind=8),pointer:: vpj(:,:,:),qj(:,:,:)
     integer:: ncomp
     procedure(l2lq_func),pointer,nopass:: l2lq
     procedure(l2nm_func),pointer,nopass:: l2nm
     real(kind=8):: engatom
  end type bspot_for_atm
contains
  function l2lq_one(il) result(ilq)
    implicit none
    integer,intent(in):: il
    integer:: ilq
    ilq = il - 1
  end function l2lq_one

  function l2lq_two(il) result(ilq)
    implicit none
    integer,intent(in):: il
    integer:: ilq
    ilq = il/2
  end function l2lq_two

  function l2nm_one(il) result(nm)
    implicit none
    integer,intent(in):: il
    integer:: nm
    nm = 2*il-1
  end function l2nm_one

  function l2nm_two(il) result(nm)
    implicit none
    integer,intent(in):: il
    integer:: nm

    integer:: pty
    
    pty = mod(il,2)
    if (pty.eq.0) then
       nm = il
    else
       nm = il+1
    end if
  end function l2nm_two

  subroutine loadbspot(nfile,nkd,xctp,bspotfc,bspots,myrank,comm)
    use mpi
    use iounit, only: xrewind
    implicit none
    integer,intent(in):: nfile,nkd
    integer,intent(out):: xctp
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    integer,intent(in):: myrank,comm

    integer:: irnk,nbsktf,nbsf,ldnr,nl,na2,ikd,ir,il,ierr

    if (myrank.eq.0) then
       call xrewind(nfile)
       read(nfile,iostat=ierr) bspotfc%ibsrk,bspotfc%nbsktf,bspotfc%nbsf
       if (ierr.ne.0) then
          write(6,'(1x,a)') 'error in opening P.P. fitting data'
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
    end if
    call MPI_BCAST(bspotfc%ibsrk,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(bspotfc%nbsktf,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(bspotfc%nbsf,1,MPI_INTEGER,0,comm,ierr)
    irnk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    nbsf = bspotfc%nbsf
    allocate(bspotfc%bsxif(1-irnk:nbsktf+irnk),bspotfc%bsaf(nbsf),stat=ierr)
    call chkalloc('loadbspot','bsxif,bsaf',ierr)
    if (myrank.eq.0) then
       read(nfile) bspotfc%bsxif, bspotfc%bsaf
    end if
    call MPI_BCAST(bspotfc%bsxif,nbsktf+2*irnk,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(bspotfc%bsaf,nbsf,MPI_DOUBLE_PRECISION,0,comm,ierr)

    do ikd=1,nkd
       if (myrank.eq.0) then
          read(nfile) bspots(ikd)%name
          read(nfile) bspots(ikd)%nob,bspots(ikd)%na2, &
               bspots(ikd)%ldnr,bspots(ikd)%nl
       end if
       call MPI_BCAST(bspots(ikd)%name,256,MPI_CHARACTER,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%nob,1,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%na2,1,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%ldnr,1,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%nl,1,MPI_INTEGER,0,comm,ierr)

       na2 = bspots(ikd)%na2
       ldnr = bspots(ikd)%ldnr
       if (ldnr.gt.bspot_for_atm_maxr) then
          write(6,'(1x,a)') 'loadbspot: too small bspot_for_atm_maxr'
          stop
       end if
       nl = bspots(ikd)%nl
       if (nl.gt.bspot_for_atm_lmax) then
          write(6,'(1x,a)') 'loadbspot: too small bspot_for_atm_lmax'
          stop
       end if
       if (myrank.eq.0) then
          read(nfile) bspots(ikd)%nr(:nl)
          read(nfile) bspots(ikd)%ifncpp,bspots(ikd)%ifachg,bspots(ikd)%ifionchg
          read(nfile) ((bspots(ikd)%ifcr(ir,il),ir=1,bspots(ikd)%ldnr), &
               il=1,bspots(ikd)%nl)
          read(nfile) bspots(ikd)%zo,bspots(ikd)%zn
          read(nfile) bspots(ikd)%a1,bspots(ikd)%a2, &
               bspots(ikd)%a3,bspots(ikd)%a4
          read(nfile) bspots(ikd)%rc,bspots(ikd)%nsmpl
       end if
       call MPI_BCAST(bspots(ikd)%nr,bspot_for_atm_lmax,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%ifncpp,1,MPI_LOGICAL,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%ifachg,1,MPI_LOGICAL,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%ifionchg,1,MPI_LOGICAL,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%ifcr, &
            bspot_for_atm_maxr*bspot_for_atm_lmax,MPI_LOGICAL,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%zo,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%zn,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%a1,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%a2,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%a3,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%a4,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%rc,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%nsmpl,1,MPI_INTEGER,0,comm,ierr)

       allocate(bspots(ikd)%vgsr(nbsf),bspots(ikd)%ccg(nbsf), &
            bspots(ikd)%achg(nbsf,2),bspots(ikd)%ionchg(nbsf),stat=ierr)
       call chkalloc('loadbspot','vgsr,ccg,achg,ionchg',ierr)
       if (myrank.eq.0) then
          read(nfile) bspots(ikd)%vgsr, bspots(ikd)%ccg, bspots(ikd)%achg, &
               bspots(ikd)%ionchg
          read(nfile) bspots(ikd)%engatom
       end if
       call MPI_BCAST(bspots(ikd)%vgsr,nbsf,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%ccg,nbsf,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%achg,nbsf*2,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%ionchg,nbsf,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%engatom,1,MPI_DOUBLE_PRECISION,0,comm,ierr)

       allocate(bspots(ikd)%ddi(ldnr,ldnr,nl), &
            bspots(ikd)%qqr(ldnr,ldnr,nl), &
            bspots(ikd)%qq(ldnr,ldnr,nl),stat=ierr)
       call chkalloc('loadbspot','ddi,qqr,qq',ierr)
       if (myrank.eq.0) then
          read(nfile) bspots(ikd)%ddi,bspots(ikd)%qqr,bspots(ikd)%qq
       end if
       call MPI_BCAST(bspots(ikd)%ddi,ldnr*ldnr*nl,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%qqr,ldnr*ldnr*nl,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%qq,ldnr*ldnr*nl,MPI_DOUBLE_PRECISION,0,comm,ierr)

       allocate(bspots(ikd)%vpj(nbsf,ldnr,nl), &
            bspots(ikd)%qj(nbsf,na2,nl),stat=ierr)
       call chkalloc('loadbspot','vpj,qj',ierr)
       if (myrank.eq.0) then
          read(nfile) bspots(ikd)%vpj,bspots(ikd)%qj
       end if
       call MPI_BCAST(bspots(ikd)%vpj,nbsf*ldnr*nl,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%qj,nbsf*na2*nl,MPI_DOUBLE_PRECISION,0,comm,ierr)

       if (myrank.eq.0) then
          read(nfile) bspots(ikd)%nprj
          read(nfile) bspots(ikd)%ip2l, bspots(ikd)%ip2r, bspots(ikd)%ip2m
          read(nfile) bspots(ikd)%l2ph, bspots(ikd)%l2pt
          read(nfile) bspots(ikd)%ncomp
       end if
       call MPI_BCAST(bspots(ikd)%nprj,1,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%ip2l,bspots(ikd)%nprj,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%ip2r,bspots(ikd)%nprj,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%ip2m,bspots(ikd)%nprj,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%l2ph,bspots(ikd)%nl,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%l2pt,bspots(ikd)%nl,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(bspots(ikd)%ncomp,1,MPI_INTEGER,0,comm,ierr)
       if (bspots(ikd)%ncomp .eq. 1) then
          bspots(ikd)%l2lq => l2lq_one
          bspots(ikd)%l2nm => l2nm_one
       else if (bspots(ikd)%ncomp .eq. 2) then
          bspots(ikd)%l2lq => l2lq_two
          bspots(ikd)%l2nm => l2nm_two
       else
          write(6,'(1x,a)') 'loadbspot: invalid ncomp'
          stop
       end if
    end do
    if (myrank.eq.0) then
       read(nfile) xctp
       close(nfile)
    end if
    call MPI_BCAST(xctp,1,MPI_INTEGER,0,comm,ierr)

    if (maxval(bspots(:)%ncomp) .ne. minval(bspots(:)%ncomp)) then
       write(6,'(1x,a)') 'loadbspot: inconsistent number_component ' &
            // 'among potentials'
       call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
       stop
    end if
    return
  end subroutine loadbspot

  subroutine tellbspot(nkd,bspots)
    implicit none
    integer,intent(in):: nkd
    type (bspot_for_atm):: bspots(nkd)

    integer:: ikd

    write(6,'(1x,a)') 'bspot detail:'
    write(6,'(3x,''number of P.P. ='',i4)') nkd
    do ikd=1,nkd
       write(6,'(3x,''kind ='',i4,'', name = '',a)') &
            ikd, trim(bspots(ikd)%name)
    end do
    return
  end subroutine tellbspot

  subroutine storebspot(nfile,nkd,xctp,bspotfc,bspots)
    use iounit, only: xrewind
    implicit none
    integer,intent(in):: nfile,nkd
    integer,intent(in):: xctp
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)

    integer:: ikd,ir,il

    call xrewind(nfile)
    write(nfile) bspotfc%ibsrk,bspotfc%nbsktf,bspotfc%nbsf
    write(nfile) bspotfc%bsxif,bspotfc%bsaf

    do ikd=1,nkd
       write(nfile) bspots(ikd)%name
       write(nfile) bspots(ikd)%nob,bspots(ikd)%na2, &
            bspots(ikd)%ldnr,bspots(ikd)%nl
       write(nfile) bspots(ikd)%nr(:bspots(ikd)%nl)
       write(nfile) bspots(ikd)%ifncpp,bspots(ikd)%ifachg,bspots(ikd)%ifionchg
       write(nfile) ((bspots(ikd)%ifcr(ir,il),ir=1,bspots(ikd)%ldnr), &
            il=1,bspots(ikd)%nl)
       write(nfile) bspots(ikd)%zo,bspots(ikd)%zn
       write(nfile) bspots(ikd)%a1,bspots(ikd)%a2, &
            bspots(ikd)%a3,bspots(ikd)%a4
       write(nfile) bspots(ikd)%rc,bspots(ikd)%nsmpl
       write(nfile) bspots(ikd)%vgsr, bspots(ikd)%ccg, bspots(ikd)%achg, &
            bspots(ikd)%ionchg
       write(nfile) bspots(ikd)%engatom
       write(nfile) bspots(ikd)%ddi,bspots(ikd)%qqr,bspots(ikd)%qq
       write(nfile) bspots(ikd)%vpj,bspots(ikd)%qj

       write(nfile) bspots(ikd)%nprj
       write(nfile) bspots(ikd)%ip2l, bspots(ikd)%ip2r, bspots(ikd)%ip2m
       write(nfile) bspots(ikd)%l2ph, bspots(ikd)%l2pt
       write(nfile) bspots(ikd)%ncomp
    end do
    write(nfile) xctp
    close(nfile)
    return
  end subroutine storebspot

  subroutine freebspot(nkd,bspotfc,bspots)
    implicit none
    integer,intent(in):: nkd
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)

    integer:: ikd

    deallocate(bspotfc%bsxif,bspotfc%bsaf)
    do ikd=1,nkd
       deallocate(bspots(ikd)%vgsr,bspots(ikd)%ccg)
       deallocate(bspots(ikd)%achg,bspots(ikd)%ionchg)
       deallocate(bspots(ikd)%ddi,bspots(ikd)%qqr,bspots(ikd)%qq)
       deallocate(bspots(ikd)%vpj,bspots(ikd)%qj)
    end do

    return
  end subroutine freebspot

  subroutine potheader(xctp,nkd,zoin,znin,nfpot1,bspots)
    use xcf_types
    use mpi
    use iounit, only: xrewind
    implicit none
    integer:: nkd,nfpot1
    integer:: xctp
    real(kind=8):: zoin(nkd),znin(nkd)
    type (bspot_for_atm):: bspots(nkd)

    integer:: ikd,l,nfpot
    real(kind=8):: temp
    character(len=256):: cbuf, xcname
    integer:: inxctp
    integer:: ip, im, ir
    integer:: imb, imd
    integer:: ierr

    write(6,30)
30  format(' ATOMIC KIND DATA'/)
    do ikd=1,nkd
       bspots(ikd)%nr = 0
       nfpot=nfpot1+ikd-1
       call xrewind(nfpot)
       read(nfpot,'(a256)',iostat=ierr) cbuf
       if (ierr.ne.0) then
          write(6,'(1x,a,i3)') 'error in opening P.P. file for # = ', ikd
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
       if (index(cbuf,'#xPSV-1').eq.1) then
          read(nfpot,'(a256)') cbuf
          bspots(ikd)%name = 'Unknown'
          do while (index(cbuf,'#').eq.1)
             if (index(cbuf,'# ATOM NAME:').eq.1) then
                read(cbuf,'(t12,a)') bspots(ikd)%name
             end if
             read(nfpot,'(a256)') cbuf
          end do
          read(cbuf,*) xcname
          if (trim(xcname).eq.xcf_pbe_name) then
             inxctp = xcf_pbe
          else if (trim(xcname).eq.xcf_pbesol_name) then
             inxctp = xcf_pbesol
          else if (trim(xcname).eq.xcf_pw91_name) then
             inxctp = xcf_pw91
          else if (trim(xcname).eq.xcf_capz_name) then
             inxctp = xcf_capz
          else if (trim(xcname).eq.xcf_ldapw92_name) then
             inxctp = xcf_ldapw92
          else if (trim(xcname).eq.xcf_hartree_fock_name) then
             inxctp = xcf_hartree_fock
          else if (trim(xcname).eq.xcf_pbe0_name) then
             inxctp = xcf_pbe0
          else if (trim(xcname).eq.xcf_b3lyp_name) then
             inxctp = xcf_b3lyp
          else
             inxctp = -1
          end if
          if (inxctp.ne.xctp) then
             write(6,*) 'inconsistent XC type for kind = ', ikd
             write(6,*) 'XC for pseudopotential = ', trim(xcname)
             stop
          end if
       else
          ! COMMENT LINE SKIP
          read(nfpot,*)
          write(6,*) 'XC type for kind = ', ikd, ' is not available'
	      bspots(ikd)%name = 'Unknown'
       end if
       read(nfpot,'(a256)') cbuf
       read(cbuf,*,iostat=ierr) bspots(ikd)%zo,bspots(ikd)%zn
       if (ierr.ne.0) then
          bspots(ikd)%zn = 0.0d0
          read(cbuf,*) bspots(ikd)%zo
       end if
       if (zoin(ikd).gt.0.0d0) then
          if (abs(zoin(ikd)-bspots(ikd)%zo).gt.1.0d-8) then
             write(6,*) 'inconsistent zo for ikd = ', ikd
             stop
          end if
       end if
       if (znin(ikd).gt.0.0d0) then
          if (bspots(ikd)%zn.gt.0.0d0) then
             if (abs(znin(ikd)-bspots(ikd)%zn).gt.1.0d-8) then
                write(6,*) 'inconsistent zn for ikd = ', ikd
                stop
             end if
          else
             write(6,*) 'zn check is not possible for ikd = ', ikd
          end if
       end if
       read(nfpot,'(a256)') cbuf
       read(cbuf,*,iostat=ierr) bspots(ikd)%nl, bspots(ikd)%ncomp
       if (ierr.ne.0) bspots(ikd)%ncomp = 1

       read(nfpot,*)(bspots(ikd)%nr(l),l=1,bspots(ikd)%nl)
       read(nfpot,*) bspots(ikd)%nsmpl,bspots(ikd)%rc
       read(nfpot,*) bspots(ikd)%a1,bspots(ikd)%a2, &
            bspots(ikd)%a3,bspots(ikd)%a4
       ! DATA CHECK
       temp=abs(bspots(ikd)%a1+bspots(ikd)%a3-1.0d0)
       if (temp.gt.1.0d-10) then
          write(6,*)' LOCAL PARAMETER FITTING ERROR IKD=',ikd
          write(6,*)'A1,A2,A3,A4',bspots(ikd)%a1,bspots(ikd)%a2, &
               bspots(ikd)%a3,bspots(ikd)%a4
          stop
       end if
       write(6,300)ikd,bspots(ikd)%zo,bspots(ikd)%zn, &
            bspots(ikd)%nl,bspots(ikd)%nsmpl
300    format(' IKD=',i3,5x,'ZO=',f10.7,5x,'ZN=',f10.7,5x, &
            'NL=',i3,'  NSMPL=',i7)
       write(6,350)(l,bspots(ikd)%nr(l),l=1,bspots(ikd)%nl)
350    format((3x,'NR(',i1,')=',i3))
       write(6,'(1x,''A1,A2,A3,A4='',4(1x,e15.7))') &
            bspots(ikd)%a1,bspots(ikd)%a2,bspots(ikd)%a3,bspots(ikd)%a4
       write(6,'(1x,''NAME= '',a)') trim(bspots(ikd)%name)
       write(6,*)

       if (bspots(ikd)%ncomp.eq.1) then
          bspots(ikd)%l2lq => l2lq_one
          bspots(ikd)%l2nm => l2nm_one
       else if (bspots(ikd)%ncomp.eq.2) then
          bspots(ikd)%l2lq => l2lq_two
          bspots(ikd)%l2nm => l2nm_two
       else
          write(6,*) 'potheader(): invalid ncomp = ', bspots(ikd)%ncomp
          stop
       end if

       ip = 0
       imb = 0
       do l=1,bspots(ikd)%nl
          bspots(ikd)%l2ph(l) = ip + 1
          do ir=1,bspots(ikd)%nr(l)
             do imd=1,bspots(ikd)%l2nm(l)
                im = imb + imd
                ip = ip + 1
                bspots(ikd)%ip2l(ip) = l
                bspots(ikd)%ip2r(ip) = ir
                bspots(ikd)%ip2m(ip) = im
             end do
          end do
          bspots(ikd)%l2pt(l) = ip
          imb = imb + bspots(ikd)%l2nm(l)
       end do
       bspots(ikd)%nprj = ip

    end do
    return
  end subroutine potheader

  subroutine potfitc(irnk,qmax,n,bspotfc)
    use bspline
    implicit none
    integer:: irnk,n
    real(kind=8):: qmax
    type(bspotfc_struct):: bspotfc

    integer:: nbsktf,nbsf,i,ierr

    bspotfc%ibsrk = irnk
    nbsf = n
    bspotfc%nbsf = nbsf
    nbsktf = n - irnk
    if (nbsktf.le.0) then
       write(6,'(1x,a,i3)') 'potfitc: too small n', n
       stop
    end if
    bspotfc%nbsktf = nbsktf
    allocate(bspotfc%bsxif(1-irnk:nbsktf+irnk),bspotfc%bsaf(nbsf), &
         stat=ierr)
    call chkalloc('formfitc','dat',ierr)

    do i=0,nbsf-1
       bspotfc%bsaf(i+1)=qmax/nbsf*i
    end do
    call formgzai(n,irnk,bspotfc%bsaf,bspotfc%nbsktf,bspotfc%bsxif,ierr)
    if (ierr.ne.0) then
       write(6,'(1x,a,i3)') 'error in potfitc = ', ierr
       stop
    end if

    return
  end subroutine potfitc

  subroutine fitlpt(bspotfc,nkd,nfpot1,bspots)
    use bspline
    implicit none
    integer,intent(in):: nkd
    type (bspotfc_struct):: bspotfc
    integer,intent(in):: nfpot1
    type (bspot_for_atm):: bspots(nkd)

!SOPTION CONTAREA(bsxif,bsaf)
    real(kind=8),pointer:: bsxif(:),bsaf(:)
    real(kind=8),allocatable:: bsafsq(:),bsvgsr(:,:),bsccg(:,:)
    integer:: ibsrk,nbsktf,nbsf
    integer:: nfpot, ikd, ndlc
    integer:: ierr

    ibsrk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    nbsf = bspotfc%nbsf
    bsxif => bspotfc%bsxif
    bsaf => bspotfc%bsaf

    do ikd=1,nkd
       allocate(bspots(ikd)%vgsr(nbsf),bspots(ikd)%ccg(nbsf),stat=ierr)
       call chkalloc('fitlpot','bspots',ierr)
    end do

    allocate(bsafsq(nbsf),bsvgsr(nbsf,nkd),bsccg(nbsf,nkd),stat=ierr)
    call chkalloc('fitlpot','tmp',ierr)
    bsafsq = bsaf*bsaf

    !  MAKE \INT V_LOC(R)*EXP(-IGR) D^3R
    nfpot=nfpot1-1
    do ikd=1,nkd
       nfpot=nfpot+1
       read(nfpot,*) ndlc
       call makevgsr(bsafsq,ndlc,bsvgsr(1,ikd), &
            bsccg(1,ikd),nbsf,nfpot,bspots(ikd)%zo, &
            bspots(ikd)%a1,bspots(ikd)%a2,bspots(ikd)%a3,bspots(ikd)%a4,ikd)
    end do

    call bintrpl(nbsf,ibsrk,nbsktf,bsxif,bsaf,nkd,bsvgsr,ierr)
    if (ierr.ne.0) then
       write(6,'(1x,a,i3)') 'error in fitlpt = ', ierr
       stop
    end if
    call bintrpl(nbsf,ibsrk,nbsktf,bsxif,bsaf,nkd,bsccg,ierr)
    if (ierr.ne.0) then
       write(6,'(1x,a,i3)') 'error in fitlpt = ', ierr
       stop
    end if

    do ikd=1,nkd
       bspots(ikd)%vgsr = bsvgsr(:,ikd)
       bspots(ikd)%ccg = bsccg(:,ikd)
    end do

    deallocate(bsafsq,bsvgsr,bsccg)
    return
  end subroutine fitlpt

  subroutine evlpot(nkd,bspotfc,bspots,nkf,akk,va,vg,ccg)
    use bspline
#ifdef ESM
    use global_esm
#endif
    implicit none
    integer,intent(in):: nkd
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    integer,intent(in):: nkf
    real(kind=8),intent(in):: akk(nkf),va
    real(kind=8),intent(out):: vg(nkf,nkd),ccg(nkf,nkd)

    real(kind=8),parameter:: pi=3.14159265358979323846d0
    real(kind=8),parameter:: cnst=4.0d0*pi

!SOPTION CONTAREA(bsxif,bsvgsr,bsccg)
    real(kind=8),pointer:: bsxif(:),bsvgsr(:),bsccg(:)
    real(kind=8),allocatable:: g(:)

    integer:: ibsrk,nbsktf
    integer:: ikd,ierr
    real(kind=8):: a1,a2,a3,a4,zo
    integer:: ig

    ibsrk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    bsxif => bspotfc%bsxif

    allocate(g(nkf),stat=ierr)
    call chkalloc('evlpot','tmp',ierr)
    g = sqrt(akk)

    do ikd=1,nkd
       bsvgsr => bspots(ikd)%vgsr
       bsccg => bspots(ikd)%ccg
       zo = bspots(ikd)%zo
       a1 = bspots(ikd)%a1
       a2 = bspots(ikd)%a2
       a3 = bspots(ikd)%a3
       a4 = bspots(ikd)%a4
       call evspline(nbsktf,bsxif,ibsrk,0,bsvgsr,g,nkf,vg(1,ikd))
#ifdef ESM
       if (esm%enable) then
!       write(*,*) "Constant and long range part is not contained."
       vg(:,ikd) = vg(:,ikd)/va
       else
#endif
       vg(1,ikd) = vg(1,ikd) + pi*zo*(a1/a2+a3/a4)
       vg(1,ikd) = vg(1,ikd)/va
       do ig=2,nkf
          vg(ig,ikd) = vg(ig,ikd) - cnst*zo/akk(ig)*( &
               a1*exp(-akk(ig)/4.0d0/a2) + a3*exp(-akk(ig)/4.0d0/a4) )
          vg(ig,ikd) = vg(ig,ikd)/va
       end do
#ifdef ESM
       end if
#endif

       call evspline(nbsktf,bsxif,ibsrk,0,bsccg,g,nkf,ccg(1,ikd))
       ccg(:,ikd) = ccg(:,ikd)/va
    end do

    deallocate(g)
    return
  end subroutine evlpot

  subroutine evslpot(nkd,bspotfc,bspots,nkf,akk,va,svg,sccg)
    use bspline
    implicit none
    integer,intent(in):: nkd
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    integer,intent(in):: nkf
    real(kind=8),intent(in):: akk(nkf),va
    real(kind=8),intent(out):: svg(nkf,nkd),sccg(nkf,nkd)

    real(kind=8),parameter:: pi=3.14159265358979323846d0
    real(kind=8),parameter:: cnst=4.0d0*pi

!SOPTION CONTAREA(bsxif,bsvgsr,bsccg)
    real(kind=8),pointer:: bsxif(:),bsvgsr(:),bsccg(:)
    real(kind=8),allocatable:: g(:),scoef(:)

    integer:: ibsrk,nbsktf,nbsf
    integer:: ikd,ierr
    real(kind=8):: a1,a2,a3,a4,zo
    integer:: ig

    ibsrk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    nbsf = bspotfc%nbsf
    bsxif => bspotfc%bsxif

    allocate(g(nkf),scoef(nbsf),stat=ierr)
    call chkalloc('evslpot','tmp',ierr)
    g = sqrt(akk)

    do ikd=1,nkd
       bsvgsr => bspots(ikd)%vgsr
       bsccg => bspots(ikd)%ccg
       zo = bspots(ikd)%zo
       a1 = bspots(ikd)%a1
       a2 = bspots(ikd)%a2
       a3 = bspots(ikd)%a3
       a4 = bspots(ikd)%a4
       scoef = bsvgsr
       call cderiv(nbsktf,bsxif,ibsrk,1,scoef)
       call evspline(nbsktf,bsxif,ibsrk,1,scoef,g,nkf,svg(1,ikd))
       svg(1,ikd) = 0.0d0
       do ig=2,nkf
          svg(ig,ikd) = svg(ig,ikd)*g(ig) + 2.0d0*cnst*zo/akk(ig)*( &
               a1*exp(-akk(ig)/4.0d0/a2) + a3*exp(-akk(ig)/4.0d0/a4) ) &
           + 0.5d0*cnst*zo*( &
             + a1/a2*exp(-akk(ig)/4.0d0/a2) + a3/a4*exp(-akk(ig)/4.0d0/a4) )
          svg(ig,ikd) = svg(ig,ikd)/va
       end do
       scoef = bsccg
       call cderiv(nbsktf,bsxif,ibsrk,1,scoef)
       call evspline(nbsktf,bsxif,ibsrk,1,scoef,g,nkf,sccg(1,ikd))
       sccg(:,ikd) = sccg(:,ikd)*g/va
    end do
    deallocate(g,scoef)
    return
  end subroutine evslpot

  !
  ! imported from p10vb.F at 2016/1/9
  !
  subroutine vbnlgn(akg,nkm,nakm,nl,nr,lpnum,maxr &
       ,bet,r,nsmpl,nasmpl,vpj,l2lq)
    implicit none
    integer nkm,nakm,nl,lpnum,maxr,nsmpl,nasmpl
    double precision akg(nakm)
    integer nr(lpnum)
    double precision bet(nasmpl,maxr,lpnum),r(nasmpl)
    double precision vpj(nakm,maxr,lpnum)
    external l2lq
    integer l2lq
    !
    double precision eps,pi,pi4
    parameter(eps=1.0d-12,pi=3.14159265358979323846d0,pi4=pi*4.0d0)
    !
    external smpln
    double precision smpln
    !
    real(kind=8) work(:)
    allocatable work
    integer ist,ir,i,il,ig,ilq
    double precision q,z,sb
    integer ierr
    !
    allocate(work(nasmpl),stat=ierr)
    call chkalloc('vbnlgn','work',ierr)

    do il=1,nl
       ilq = l2lq(il)
       if (ilq.eq.0) then
          !     
          ! START ILQ=0 (S)
          !     
          do ir=1,nr(il)
             !
             ! TREATMENT FOR SK(3)=0
             if (akg(1).lt.eps) then
                ist=2
                do i=1,nsmpl
                   work(i)=r(i)*bet(i,ir,il)
                end do
                vpj(1,ir,il)=pi4*smpln(work,r,nsmpl)
             else
                ist=1
             end if
!POPTION PARALLEL, TLOCAL(q,z,sb,work(nsmpl))
             do ig=ist,nkm
                q=sqrt(akg(ig))
                do i=1,nsmpl
                   z=q*r(i)
                   if (z.lt.1.d-1) then
                      sb=-(1.d0/39916800.d0*z**10-1.d0/362880.d0*z**8 &
                           +1.d0/5040.d0*z**6-1.d0/120.d0*z**4 &
                           +1.d0/6.d0*z**2-1.d0)
                   else
                      sb=sin(z)/z
                   end if
                   work(i)=r(i)*sb*bet(i,ir,il)
                end do
                vpj(ig,ir,il)=pi4*smpln(work,r,nsmpl)
             end do
          end do
       else if (ilq.eq.1) then
          !     
          ! START ILQ=1 (P)
          !     
          do ir=1,nr(il)
             !
             ! TREATMENT FOR SK(3)=0
             if (akg(1).lt.eps) then
                ist=2
                vpj(1,ir,il)=0.0d0
             else
                ist=1
             end if
!POPTION PARALLEL, TLOCAL(q,z,sb,work(nsmpl))
             do ig=ist,nkm
                q=sqrt(akg(ig))
                do i=1,nsmpl
                   z=q*r(i)
                   if (z.lt.1.d-1) then
                      sb=1.d0/3991680.d0*z**9-1.d0/45360.d0*z**7 &
                           +1.d0/840.d0*z**5-1.d0/30.d0*z**3 &
                           +1.d0/3.d0*z
                   else
                      sb=sin(z)/z**2-cos(z)/z
                   end if
                   work(i)=r(i)*sb*bet(i,ir,il)
                end do
                vpj(ig,ir,il)=pi4*smpln(work,r,nsmpl)
             end do
          end do
       else if (ilq.eq.2) then
          !     
          ! START ILQ=2 (D)
          !     
          do ir=1,nr(il)
             !
             ! TREATMENT FOR SK(3)=0
             if (akg(1).lt.eps) then
                ist=2
                vpj(1,ir,il)=0.0d0
             else
                ist=1
             end if
!POPTION PARALLEL, TLOCAL(q,z,sb,work(nsmpl))
             do ig=ist,nkm
                q=sqrt(akg(ig))
                do i=1,nsmpl
                   z=q*r(i)
                   if (z.lt.1.d-1) then
                      sb=1.d0/51891840.d0*z**10-1.d0/498960.d0*z**8 &
                           +1.d0/7560.d0*z**6-1.d0/210.d0*z**4 &
                           +1.d0/15.d0*z**2
                   else
                      sb=(3.d0-1.d0*z**2)/z**3*sin(z)-3.d0/z**2*cos(z)
                   end if
                   work(i)=r(i)*sb*bet(i,ir,il)
                end do
                vpj(ig,ir,il)=pi4*smpln(work,r,nsmpl)
             end do
          end do
       else if (ilq.eq.3) then
          !     
          ! Imported from TAPP-3.3.0 2013/4/30
          ! Originally developed by T. Kosugi 2007/6/22
          !     
          ! START ILQ=3 (F)
          !
          do ir=1,nr(il)
             !
             ! TREATMENT FOR SK(3)=0
             if (akg(1).lt.eps) then
                ist=2
                vpj(1,ir,il)=0.0d0
             else
                ist=1
             end if
!POPTION PARALLEL, TLOCAL(q,z,sb,work(nsmpl))
             do ig=ist,nkm
                q=sqrt(akg(ig))
                do i=1,nsmpl
                   z=q*r(i)
                   if (z.lt.1.d-1) then
                      sb=-1.d0/6486480.d0*z**9+1.d0/83160.d0*z**7 &
                           -1.d0/1890.d0*z**5+1.d0/105.d0*z**3
                   else
                      sb=(15.d0-6.d0*z**2)/z**4*sin(z) &
                           -(15.d0-z**2)/z**3*cos(z)
                   end if
                   work(i)=r(i)*sb*bet(i,ir,il)
                end do
                vpj(ig,ir,il)=pi4*smpln(work,r,nsmpl)
             end do
          end do
       else
          write(6,*)'ERROR ILQ',ilq
          stop 'VBNLGN'
       end if
    end do
    deallocate(work)
    return
  end subroutine vbnlgn

  subroutine fitnlpot(nfpot,bspotfc,bsapot,fwqcom,qm,qc)
    use uspp_qg
    use bspline
    use util
    implicit none
    integer,intent(in):: nfpot
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bsapot
    real(kind=8):: fwqcom
    real(kind=8),intent(in):: qm,qc
    !
    integer:: nl,nr(bspot_for_atm_lmax),nsmpl
    integer:: nob
    ! NACO:PARAMETER MAX(# OF V.PSEUDIZATION TERM)
    integer:: naco
    parameter(naco=8)
    real(kind=8),parameter:: qgk0(4) = (/0.0d0, 0.0d0, 0.0d0, 0.0d0/)
    !
!SOPTION CONTAREA(bsxif,bsaf)
    real(kind=8),pointer:: bsxif(:),bsaf(:)
!SOPTION CONTAREA(bsqj,bsvpj,qqr,qq)
    real(kind=8),pointer:: bsqj(:,:,:),bsvpj(:,:,:)
    real(kind=8),pointer:: qqr(:,:,:),qq(:,:,:)
    real(kind=8),allocatable:: ddi(:,:,:)
    real(kind=8),allocatable:: psi(:,:,:),phi(:,:,:)
    real(kind=8),allocatable:: bet(:,:,:),r(:)
    real(kind=8),allocatable:: bsafsq(:)
    integer:: ibsrk,nbsktf,nbsf
    integer:: il,ndata,i,j,mr,ir,nnn,mmm,ldnr,na2
    real(kind=8):: h,x1,x2,xend
    real(kind=8):: maxdif
    integer:: ierr
    real(kind=8),allocatable:: qj0(:,:)
    complex(kind=8),allocatable:: qg0(:)

    ibsrk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    nbsf = bspotfc%nbsf
    bsxif => bspotfc%bsxif
    bsaf => bspotfc%bsaf

    nsmpl = bsapot%nsmpl
    nl = bsapot%nl
    nr(:) = bsapot%nr(:)
    if (nl.gt.bspot_for_atm_lmax) then
       write(6,'(1x,a)') 'fitnlpot: nl too large'
       stop
    end if
    ldnr = 0
    do il=1,nl
       ldnr = max(ldnr,bsapot%nr(il))
    end do
    bsapot%ldnr = ldnr

    allocate(psi(nsmpl,ldnr,nl),phi(nsmpl,ldnr,nl), &
         bet(nsmpl,ldnr,nl),r(nsmpl),ddi(ldnr,ldnr,nl), stat=ierr)
    call chkalloc('fitnlpot','tmp',ierr)
    allocate(bsapot%qqr(ldnr,ldnr,nl),bsapot%qq(ldnr,ldnr,nl),stat=ierr)
    call chkalloc('fitnlpot','qqr,qq',ierr)
    qqr => bsapot%qqr
    qq => bsapot%qq
    !
    ! NONLOCAL POTENTIAL READ
    qqr = 0.0d0
    qq = 0.0d0
    ddi = 0.0d0
    !
    ! COMENT LINE
    ! LOG_MESH
    read(nfpot,*)
    read(nfpot,*)ndata,x1,xend,x2
    ! MAKE R(I)
    if (ndata.ne.nsmpl) then
       write(6,*)' ERROR ON # OF DATA  NSMPL,NDATA=',nsmpl,ndata
       stop ' MKQG0'
    end if
    h=log(xend/x1)/dble(ndata-1)
    h=exp(h)
    write(6,7)ndata,x1,xend,h
7   format(' NDATA=',i5,' X1=',e15.7,' XEND=',e15.7,' H=',e15.7)
    r(1)=x1
    do i=2,ndata
       r(i)=r(i-1)*h
    end do
    if ((abs(r(2)-x2).gt.1.d-6).or.(abs(r(ndata)-xend).gt.1.d-8)) then
       write(6,*)'ERROR  NDATA=',ndata
       write(6,15)r(1),x1,r(2),x2,r(ndata),xend
15     format(' R(1),X1=',2e15.7,' R(2),X2=',2e15.7/ &
            ,' R(NDATA),XEND=',2e15.7)
       stop ' MKQG0'
    end if
    !
    bsapot%ifncpp = .true.
    !
    do il=1,nl
       ! COMMENT LINE SKIP
       read(nfpot,*)
       mr=nr(il)
       read(nfpot,300)((ddi(i,j,il),i=1,j),j=1,mr)
300    format(3e20.12)
       do j=1,mr
          do i=j+1,mr
             ddi(i,j,il)=ddi(j,i,il)
          end do
       end do
       ! READ QQ FOR CHECK
       read(nfpot,*)((qqr(i,j,il),i=1,j),j=1,mr)
       do j=1,mr
          do i=j+1,mr
             qqr(i,j,il)=qqr(j,i,il)
          end do
       end do
       do ir=1,mr
          ! COMMENT LINE SKIP
          read(nfpot,*)
          !
          maxdif=0.0d0
          ! DATA OF R IGNORED
          do i=1,nsmpl
             read(nfpot,*)psi(i,ir,il),phi(i,ir,il),bet(i,ir,il)
             maxdif=max(abs(psi(i,ir,il)-phi(i,ir,il)),maxdif)
          end do
#ifdef COMPAT_PSV_2_3
          if (maxdif.lt.1.0d-16) then
#else
          if (maxdif.eq.0.0d0) then
#endif
             bsapot%ifcr(ir,il) = -1
          else
             bsapot%ifcr(ir,il) = 1
             bsapot%ifncpp = .false.
          end if
       end do
    end do
    ! NUMERICAL DATA ARE ON LOG_MESH
    nnn=0
    mmm=0
    do il=1,nl
       nnn=nnn+nr(il)
       mmm=mmm+nr(il)*bsapot%l2nm(il)
    end do
    na2=nnn*(nnn+1)/2
    nob=mmm

    allocate(bsafsq(nbsf),stat=ierr)
    call chkalloc('fitnlpot','bsafsq',ierr)
    bsafsq = bsaf*bsaf

    bsapot%nob = nob
    bsapot%na2 = na2

    allocate(bsapot%qj(nbsf,na2,nl),bsapot%vpj(nbsf,ldnr,nl), &
         bsapot%ddi(ldnr,ldnr,nl),stat=ierr)
    call chkalloc('fitnlpot','qj,vpj,ddi',ierr)
    bsapot%ddi = ddi(:ldnr,:ldnr,:nl)
    bsqj => bsapot%qj
    bsvpj => bsapot%vpj
    bsqj = 0.0d0
    bsvpj = 0.0d0
    call getqj(nfpot,bsqj,r,bsaf,psi,phi,nbsf &
         ,nsmpl,nsmpl,ldnr,nl,nl,nr,na2,naco,bsapot%l2lq)
    call vbnlgn(bsafsq,nbsf,nbsf,nl,nr,nl,ldnr &
         ,bet,r,nsmpl,nsmpl,bsvpj,bsapot%l2lq)
    
    if (fwqcom.gt.0.0d0) then
       write(6,'(1x,a,f5.2)') &
            'Apply Hanning window for Qc at G = ', fwqcom*qm
       call hanningwind(nbsf,na2*nl,bsaf,bsqj,fwqcom*qm,qc)
    end if

    allocate(qj0(na2,nl),qg0(nob*(nob+1)/2),stat=ierr)
    call chkalloc('fitnlpot','qj0,qg0',ierr)
    qj0 = bsqj(1,:,:)
    !
    ! va = 1.0d0 will be canceled.
    call makeqg(bsapot%ncomp,1,qj0,qgk0,qg0,1.0d0,nl,nr,1,na2,nob, &
         bsapot%l2lq,bsapot%l2nm)
    call getqq(qq,qg0,1.0d0,nl,nr,1,nob,ldnr,nl,bsapot%l2nm)
    deallocate(qj0,qg0)
    write(6,'(1x,a,1x,e12.4)') 'QQ-ERROR MAX=',maxval(abs(qq-qqr))

    call bintrpl(nbsf,ibsrk,nbsktf,bsxif,bsaf,na2*nl,bsqj,ierr)
    if (ierr.ne.0) then
       write(6,'(1x,a,i3)') 'error in fitnlpot = ', ierr
       stop
    end if
    call bintrpl(nbsf,ibsrk,nbsktf,bsxif,bsaf,ldnr*nl,bsvpj,ierr)
    if (ierr.ne.0) then
       write(6,'(1x,a,i3)') 'error in fitnlpot = ', ierr
       stop
    end if
    !
    deallocate(bsafsq)
    deallocate(psi,phi,bet,r,ddi)

    return
  end subroutine fitnlpot

  subroutine evnlpot(nkd,lmax,maxr, &
       bspotfc,bspots, &
       nkc,nks,nkm,qgk,va,ldakk,akk,qq,ifcr,nkmmax,vpj)
    use ioqg
    use uspp_qg
    use bspline
    implicit none
    integer,intent(in):: nkd,lmax,maxr
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    integer,intent(in):: nkc,nks,nkm(nks),ldakk
    real(kind=8),intent(in):: qgk(4,nkc),va,akk(ldakk,nks)
    real(kind=8),intent(out):: qq(maxr,maxr,lmax,nkd)
    integer:: ifcr(maxr,lmax,nkd)
    integer,intent(in):: nkmmax
    real(kind=8):: vpj(nkmmax,maxr,lmax,nkd,nks)
    !
    real(kind=8),allocatable:: g(:)
!SOPTION CONTAREA(qg,qqr)
    complex(kind=8),pointer:: qg(:,:)
    real(kind=8),pointer:: qqr(:,:,:)
    !
    integer:: iks,il,ir,nob,nl,ikd,nnn,i1,i2
    integer:: ierr
    real(kind=8):: ermax

    call openqg(nkc,nkd)

    do ikd=1,nkd
       nl = bspots(ikd)%nl
       nob = bspots(ikd)%nob
       call newqg(ikd,nob,qg)
       call evqg(bspotfc,bspots(ikd),nkc,qgk,va,nob,1,qg)
       call getqq(qq(1,1,1,ikd),qg,va,nl,bspots(ikd)%nr,nkc,nob,maxr,lmax, &
            bspots(ikd)%l2nm)
       ermax=0.0d0
       qqr => bspots(ikd)%qqr
       do il=1,bspots(ikd)%nl
          do i1=1,bspots(ikd)%nr(il)
             do i2=1,bspots(ikd)%nr(il)
                ermax=max(ermax,abs(qq(i1,i2,il,ikd)-qqr(i1,i2,il)))
             end do
          end do
       end do
       write(6,'(1x,a,1x,e12.4)') 'QQ-ERROR MAX=',ermax
       call putqg(ikd,qg)
    end do

    ! MAKE IFCR:FLAG INDICATES CORE CHARGE EXISTENCE.
    !            1: \PSI.NE.\PHI
    !           -1: \PSI.EQ.\PHI
    write(6,'(1x,a,a)') 'IFCR( IR,  L,IKD): ', &
         'FLAG WHICH INDICATES NO CORE CHARGE.'
    nnn=0
    do ikd=1,nkd
       do il=1,bspots(ikd)%nl
          do ir=1,bspots(ikd)%nr(il)
             if (abs(qq(ir,ir,il,ikd)).lt.1.0d-16) then
                if (bspots(ikd)%ifcr(ir,il) .ne. -1) then
                   write(6,'(1x,a,3(1x,i3))') &
                        'ifcr definition error 1',ikd,il,ir
                   stop
                end if
                nnn=nnn+1
                ifcr(ir,il,ikd)=-1
                write(6,9300)ir,il,ikd,ifcr(ir,il,ikd)
9300            format(' ','IFCR(',i3,',',i3,',',i3,')=',i3)
             else
                if (bspots(ikd)%ifcr(ir,il) .ne. 1) then
                   write(6,'(1x,a,3(1x,i3))') &
                        'ifcr definition error 2',ikd,il,ir
                   stop
                end if
                ifcr(ir,il,ikd)= 1
             end if
          end do
       end do
    end do
    if (nnn.eq.0) then
       write(6,'(1x,a)') 'NONE'
    end if

    allocate(g(nkmmax),stat=ierr)
    call chkalloc('evlpot','tmp 2',ierr)
    do iks=1,nks
       g(:nkm(iks))=sqrt(akk(:nkm(iks),iks))
       do ikd=1,nkd
          call evbetg(bspotfc,bspots(ikd),nkm(iks),g, &
               nkmmax,maxr,bspots(ikd)%nl,vpj(1,1,1,ikd,iks))
       end do
    end do
    deallocate(g)

    call closeqg()

    return
  end subroutine evnlpot

  subroutine evqg(bspotfc,bsapot,nkc,qgk,va,nob,idens,qg)
    use uspp_qg
    use bspline
    implicit none
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bsapot
    integer,intent(in):: nkc
    real(kind=8),intent(in):: qgk(4,nkc),va
    integer,intent(in):: nob,idens
    complex(kind=8):: qg(nkc,nob*(nob+1)/2)
    !
    integer:: ibsrk,nbsktf,nbsf
!SOPTION CONTAREA(bsxif)
    real(kind=8),pointer:: bsxif(:)
!SOPTION CONTAREA(bsqj)
    real(kind=8),pointer:: bsqj(:,:,:)
    real(kind=8),allocatable:: g(:),qj(:,:,:)
    integer:: nl,na2,ia2,il
    integer:: ierr

    ibsrk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    nbsf = bspotfc%nbsf
    bsxif => bspotfc%bsxif

    if (nbsf.ne.nbsktf+ibsrk) then
       write(6,'(1x,a)') 'evqg: invalid nbsf,nbsktf,ibsrk'
       stop
    end if

    bsqj => bsapot%qj
    nl = bsapot%nl
    na2 = bsapot%na2

    allocate(g(nkc),qj(nkc,na2,nl),stat=ierr)
    call chkalloc('evqg','tmp',ierr)

    g = qgk(4,:)

    do il=1,nl
       do ia2=1,na2
          call evspline(nbsktf,bsxif,ibsrk,0,bsqj(:,ia2,il), &
               g,nkc,qj(1,ia2,il))
       end do
    end do
    call makeqg(bsapot%ncomp,idens,qj,qgk,qg,va,nl,bsapot%nr,nkc,na2,nob, &
         bsapot%l2lq,bsapot%l2nm)

    deallocate(g,qj)

    return
  end subroutine evqg

  subroutine evbetg(bspotfc,bsapot,nkm,g,ldvpj1,ldvpj2,ldvpj3,vpj)
    use bspline
    implicit none
    type(bspotfc_struct):: bspotfc
    type(bspot_for_atm):: bsapot
    integer,intent(in):: nkm
    real(kind=8),intent(in):: g(nkm)
    integer,intent(in):: ldvpj1,ldvpj2,ldvpj3
    real(kind=8),intent(out):: vpj(ldvpj1,ldvpj2,ldvpj3)
    !
    integer:: ibsrk,nbsktf,nbsf
!SOPTION CONTAREA(bsxif)
    real(kind=8),pointer:: bsxif(:)
!SOPTION CONTAREA(bsvpj)
    real(kind=8),pointer:: bsvpj(:,:,:)
    integer:: il,ir

    ibsrk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    nbsf = bspotfc%nbsf
    bsxif => bspotfc%bsxif

    if (nbsf.ne.nbsktf+ibsrk) then
       write(6,'(1x,a)') 'evbetg: invalid nbsf,nbsktf,ibsrk'
       stop
    end if

    bsvpj => bsapot%vpj

    do il=1,bsapot%nl
       do ir=1,bsapot%nr(il)
          call evspline(nbsktf,bsxif,ibsrk,0,bsvpj(:,ir,il),g, &
               nkm,vpj(1,ir,il))
       end do
    end do

    return
  end subroutine evbetg

  subroutine evsqg(bspotfc,bsapot,nkc,qgk,va,nob,idens,sqg)
    use uspp_qg
    use bspline
    implicit none
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bsapot
    integer,intent(in):: nkc
    real(kind=8),intent(in):: qgk(4,nkc),va
    integer,intent(in):: nob,idens
    complex(kind=8):: sqg(nkc,6,nob*(nob+1)/2)
    !
    integer:: ibsrk,nbsktf,nbsf
!SOPTION CONTAREA(bsxif)
    real(kind=8),pointer:: bsxif(:)
!SOPTION CONTAREA(bsqj)
    real(kind=8),pointer:: bsqj(:,:,:)
    real(kind=8),allocatable:: g(:),sbsqj(:),qj(:,:,:),sqj(:,:,:)
    integer:: nl,na2,ia2,il
    integer:: ierr

    ibsrk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    nbsf = bspotfc%nbsf
    bsxif => bspotfc%bsxif

    if (nbsf.ne.nbsktf+ibsrk) then
       write(6,'(1x,a)') 'evsqg: invalid nbsf,nbsktf,ibsrk'
       stop
    end if

    bsqj => bsapot%qj
    nl = bsapot%nl
    na2 = bsapot%na2

    allocate(g(nkc),sbsqj(nbsf),qj(nkc,na2,nl),sqj(nkc,na2,nl),stat=ierr)
    call chkalloc('evsqg','tmp',ierr)

    g = qgk(4,:)

    do il=1,nl
       do ia2=1,na2
          call evspline(nbsktf,bsxif,ibsrk,0,bsqj(:,ia2,il), &
               g,nkc,qj(1,ia2,il))
          sbsqj = bsqj(:,ia2,il)
          call cderiv(nbsktf,bsxif,ibsrk,1,sbsqj)
          call evspline(nbsktf,bsxif,ibsrk,1,sbsqj, &
               g,nkc,sqj(1,ia2,il))
          sqj(:,ia2,il) = g*sqj(:,ia2,il)
       end do
    end do
    call makesqg(bsapot%ncomp,idens,qj,sqj,qgk,sqg,va, &
         nl,bsapot%nr,nkc,na2,nob, &
         bsapot%l2lq,bsapot%l2nm)

    deallocate(g,sbsqj,qj,sqj)

    return
  end subroutine evsqg

  subroutine evsbetg(bspotfc,bsapot,nkm,g,ldvpj1,ldvpj2,ldvpj3,svpj)
    use bspline
    implicit none
    type(bspotfc_struct):: bspotfc
    type(bspot_for_atm):: bsapot
    integer,intent(in):: nkm
    real(kind=8),intent(in):: g(nkm)
    integer,intent(in):: ldvpj1,ldvpj2,ldvpj3
    real(kind=8),intent(out):: svpj(ldvpj1,ldvpj2,ldvpj3)
    !
    integer:: ibsrk,nbsktf,nbsf
!SOPTION CONTAREA(bsxif)
    real(kind=8),pointer:: bsxif(:)
!SOPTION CONTAREA(bsvpj)
    real(kind=8),pointer:: bsvpj(:,:,:)

    real(kind=8),allocatable:: sbsvpj(:)
    integer:: il,ir,ierr


    ibsrk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    nbsf = bspotfc%nbsf
    bsxif => bspotfc%bsxif

    if (nbsf.ne.nbsktf+ibsrk) then
       write(6,'(1x,a)') 'evsbetg: invalid nbsf,nbsktf,ibsrk'
       stop
    end if

    bsvpj => bsapot%vpj

    allocate(sbsvpj(nbsf),stat=ierr)
    call chkalloc('evsbetg','sbsvpj',ierr)

    do il=1,bsapot%nl
       do ir=1,bsapot%nr(il)
          sbsvpj = bsvpj(:,ir,il)
          call cderiv(nbsktf,bsxif,ibsrk,1,sbsvpj)
          call evspline(nbsktf,bsxif,ibsrk,1,sbsvpj,g, &
               nkm,svpj(1,ir,il))
          svpj(:,ir,il) = g*svpj(:,ir,il)
       end do
    end do

    deallocate(sbsvpj)
    return
  end subroutine evsbetg

  subroutine fitachg_core0(nfrhi,ndlc,r,rhoi)
    implicit none
    integer,intent(in):: ndlc, nfrhi
    real(kind=8),intent(out):: r(ndlc),rhoi(ndlc,2)
    character(len=256):: inbuff

    integer:: i, ierr
    real(kind=8):: h,temp
    
    read(nfrhi,'(a256)') inbuff
    read(inbuff,*,iostat=ierr) r(1),rhoi(1,1),rhoi(1,2)
    if (ierr .eq. 0) then
       do i=2,ndlc
          read(nfrhi,*) r(i),rhoi(i,1),rhoi(i,2)
       end do
    else
       do i=2,ndlc
          read(nfrhi,*) r(i),rhoi(i,1)
       end do
       do i=1,ndlc
          rhoi(i,1) = 0.5d0*rhoi(i,1)
          rhoi(i,2) = rhoi(i,1)
       end do
    end if
    h=exp(log(r(ndlc)/r(1))/dble(ndlc-1))
    temp=h*r(1)
    if (abs(temp-r(2)).gt.1.d-12) then
       write(6,*)' Error: initial local potential'
       write(6,*)' ndlc=',ndlc,'h=',h
       write(6,*)' r(1),r(2),r(ndlc)=',r(1),r(2),r(ndlc)
       stop
    end if
    return
  end subroutine fitachg_core0

  subroutine fitachg_core1(nspin,ndlc,r,rhoi,nbsf,bsaf,bsachg)
    implicit none
    real(kind=8),parameter:: pi = 3.14159265358979323846d0
    interface
       real(kind=8) function smpln(f,x,n)
         integer:: n
         real(kind=8):: f(n), x(n)
       end function smpln
    end interface
    integer,intent(in):: nspin,ndlc,nbsf
    real(kind=8),pointer:: bsaf(:)
    real(kind=8),intent(in):: r(ndlc),rhoi(ndlc,nspin)
    real(kind=8),intent(out):: bsachg(nbsf,nspin)

    integer:: is,i,ip,ierr
    real(kind=8):: g,z,sb,cnst
    real(kind=8),allocatable:: work(:)
    
    cnst=4.0d0*pi

    allocate(work(ndlc),stat=ierr)
    call chkalloc('fitachg','work',ierr)
    do is=1,nspin
       do ip=1,ndlc
          work(ip)=r(ip)*r(ip)*rhoi(ip,is)
       end do
       bsachg(1,is)=cnst*smpln(work,r,ndlc)
!POPTION PARALLEL, TLOCAL(g,z,sb,work(ndlc))
       do i=2,nbsf
          g=bsaf(i)
          do ip=1,ndlc
             z=g*r(ip)
             if (z.lt.1.d-1) then
                sb=-(1.d0/39916800.d0*z**10-1.d0/362880.d0*z**8 &
                     +1.d0/5040.d0*z**6-1.d0/120.d0*z**4 &
                     +1.d0/6.d0*z**2-1.d0)
             else
                sb=sin(z)/z
             end if
             work(ip)=r(ip)*r(ip)*sb*rhoi(ip,is)
          end do
          bsachg(i,is)=cnst*smpln(work,r,ndlc)
       end do
    end do
    deallocate(work)
    return
  end subroutine fitachg_core1
  
  subroutine fitachg(bspotfc,nkd,nfrhi1,bspots)
    use bspline
    use iounit, only: xrewind
    implicit none
    integer,intent(in):: nkd
    integer,intent(in):: nfrhi1
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)

    real(kind=8),parameter:: pi = 3.14159265358979323846d0

    interface
       real(kind=8) function smpln(f,x,n)
         integer:: n
         real(kind=8):: f(n), x(n)
       end function smpln
    end interface

!SOPTION CONTAREA(bsxif,bsaf)
    real(kind=8),pointer:: bsxif(:),bsaf(:)
    real(kind=8),allocatable:: bsachg(:,:,:)
    integer:: ibsrk,nbsktf,nbsf
    integer:: ikd, ndlc, nfrhi
    integer:: ierr
    real(kind=8),allocatable:: r(:),rhoi(:,:)

    ibsrk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    nbsf = bspotfc%nbsf
    bsxif => bspotfc%bsxif
    bsaf => bspotfc%bsaf

    do ikd=1,nkd
       allocate(bspots(ikd)%achg(nbsf,2),stat=ierr)
       call chkalloc('fitachg','bspots',ierr)
    end do

    allocate(bsachg(nbsf,2,nkd),stat=ierr)
    call chkalloc('fitachg','tmp',ierr)

    do ikd=1,nkd
       nfrhi=nfrhi1+ikd-1
       call xrewind(nfrhi)
       read(nfrhi,*,iostat=ierr) ndlc
       if (ierr.ne.0) then
          write(6,'(1x,a,i3)') 'no atomic charge for ikd = ', ikd
          bsachg(:,:,ikd) = 0.0d0
          bspots(ikd)%ifachg = .false.
          cycle
       end if
       bspots(ikd)%ifachg = .true.
       allocate(r(ndlc),rhoi(ndlc,2),stat=ierr)
       call chkalloc('fitachg','r,rhoi',ierr)

       call fitachg_core0(nfrhi,ndlc,r,rhoi)

       call fitachg_core1(2,ndlc,r,rhoi,nbsf,bsaf,bsachg(:,:,ikd))

       deallocate(r,rhoi)
    end do

    call bintrpl(nbsf,ibsrk,nbsktf,bsxif,bsaf,2*nkd,bsachg,ierr)
    if (ierr.ne.0) then
       write(6,'(1x,a,i3)') 'error in fitachg = ', ierr
       stop
    end if

    do ikd=1,nkd
       bspots(ikd)%achg = bsachg(:,:,ikd)
    end do

    deallocate(bsachg)
    return
  end subroutine fitachg

  subroutine fitionchg(bspotfc,nkd,nfionc1,bspots)
    use bspline
    use iounit, only: xrewind
    implicit none
    integer,intent(in):: nkd
    integer,intent(in):: nfionc1
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)

    real(kind=8),parameter:: pi = 3.14159265358979323846d0

    interface
       real(kind=8) function smpln(f,x,n)
         integer:: n
         real(kind=8):: f(n), x(n)
       end function smpln
    end interface

!SOPTION CONTAREA(bsxif,bsaf)
    real(kind=8),pointer:: bsxif(:),bsaf(:)
    real(kind=8),allocatable:: bsionchg(:,:)
    integer:: ibsrk,nbsktf,nbsf
    integer:: ikd, ndlc, nfionc, i
    integer:: ierr
    real(kind=8),allocatable:: r(:),rhoi(:)
    real(kind=8):: h,temp

    ibsrk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    nbsf = bspotfc%nbsf
    bsxif => bspotfc%bsxif
    bsaf => bspotfc%bsaf

    do ikd=1,nkd
       allocate(bspots(ikd)%ionchg(nbsf),stat=ierr)
       call chkalloc('fitionchg','bspots',ierr)
    end do

    allocate(bsionchg(nbsf,nkd),stat=ierr)
    call chkalloc('fitionchg','tmp',ierr)

    do ikd=1,nkd
       nfionc=nfionc1+ikd-1
       call xrewind(nfionc)
       read(nfionc,*,iostat=ierr) ndlc, bspots(ikd)%engatom
       if (ierr.ne.0) then
          write(6,'(1x,a,i3)') 'no ion core charge for ikd = ', ikd
          bsionchg(:,ikd) = 0.0d0
          bspots(ikd)%ifionchg = .false.
          bspots(ikd)%ionchg = 0.0d0
          bspots(ikd)%engatom = 0.0d0
          cycle
       end if
       write(6,'(1x,a,i3)')     'ion core charge for ikd = ', ikd
       write(6,'(1x,a,e18.11)') '         atomic energy  = ', &
            bspots(ikd)%engatom
       bspots(ikd)%ifionchg = .true.
       allocate(r(ndlc),rhoi(ndlc),stat=ierr)
       call chkalloc('fitioncchg','r,rhoi',ierr)

       do i=1,ndlc
          read(nfionc,*) r(i),rhoi(i)
       end do
       h=exp(log(r(ndlc)/r(1))/dble(ndlc-1))
       temp=h*r(1)
       if (abs(temp-r(2)).gt.1.d-12) then
          write(6,*)' Error: initial local potential'
          write(6,*)' ndlc=',ndlc,'h=',h
          write(6,*)' r(1),r(2),r(ndlc)=',r(1),r(2),r(ndlc)
          stop
       end if

       call fitachg_core1(1,ndlc,r,rhoi,nbsf,bsaf,bsionchg(:,ikd))

       deallocate(r,rhoi)
    end do

    call bintrpl(nbsf,ibsrk,nbsktf,bsxif,bsaf,nkd,bsionchg,ierr)
    if (ierr.ne.0) then
       write(6,'(1x,a,i3)') 'error in fitionchg = ', ierr
       stop
    end if

    do ikd=1,nkd
       bspots(ikd)%ionchg = bsionchg(:,ikd)
    end do

    deallocate(bsionchg)
    return
  end subroutine fitionchg
  
  subroutine evachg(nkd,bspotfc,bspots,nkf,akk,va,achg)
    use bspline
    implicit none
    integer,intent(in):: nkd
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    integer,intent(in):: nkf
    real(kind=8),intent(in):: akk(nkf),va
    real(kind=8),intent(out):: achg(nkf,2,nkd)

!SOPTION CONTAREA(bsxif,bsachg)
    real(kind=8),pointer:: bsxif(:),bsachg(:,:)
    real(kind=8),allocatable:: g(:)

    integer:: ibsrk,nbsktf
    integer:: ikd,ierr

    ibsrk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    bsxif => bspotfc%bsxif

    allocate(g(nkf),stat=ierr)
    call chkalloc('evachg','tmp',ierr)
    g = sqrt(akk)

    do ikd=1,nkd
       bsachg => bspots(ikd)%achg
       call evspline(nbsktf,bsxif,ibsrk,0,bsachg(:,1),g,nkf,achg(1,1,ikd))
       call evspline(nbsktf,bsxif,ibsrk,0,bsachg(:,2),g,nkf,achg(1,2,ikd))
       achg(:,:,ikd) = achg(:,:,ikd)/va
    end do

    deallocate(g)
    return
  end subroutine evachg

  subroutine evionchg(nkd,bspotfc,bspots,nkf,akk,va,ionchg)
    use bspline
    implicit none
    integer,intent(in):: nkd
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    integer,intent(in):: nkf
    real(kind=8),intent(in):: akk(nkf),va
    real(kind=8),intent(out):: ionchg(nkf,nkd)

!SOPTION CONTAREA(bsxif,bsachg)
    real(kind=8),pointer:: bsxif(:),bsionchg(:)
    real(kind=8),allocatable:: g(:)

    integer:: ibsrk,nbsktf
    integer:: ikd,ierr

    ibsrk = bspotfc%ibsrk
    nbsktf = bspotfc%nbsktf
    bsxif => bspotfc%bsxif

    allocate(g(nkf),stat=ierr)
    call chkalloc('evachg','tmp',ierr)
    g = sqrt(akk)

    do ikd=1,nkd
       bsionchg => bspots(ikd)%ionchg
       call evspline(nbsktf,bsxif,ibsrk,0,bsionchg,g,nkf,ionchg(1,ikd))
       ionchg(:,ikd) = ionchg(:,ikd)/va
    end do

    deallocate(g)
    return
  end subroutine evionchg

  subroutine loadchkpots(nkd,nsi,aa0,qf,rf,qm,qc,ng,nnp,rg,pg,srg, &
       nk,nsk,mmm,idat,nvx,nvy,nvz,nvxyz,nks,nkmesh,irdkpt, &
       nng,nrcell,nkr,nkfxa,nkc,myrank,comm)
    use iounit
    use subr_kpoint
    use mpi
    implicit none
    integer,intent(in):: nkd,nsi,ng,nnp
    real(kind=8),intent(in):: aa0(3,3),qf,rf,qm,qc
    integer,intent(in):: rg(3,3,ng),pg(3,ng),srg(ng),nk,nsk,mmm(3,2,nsk)
    integer,intent(in):: idat(3)
    integer,intent(out):: nvx,nvy,nvz,nvxyz,nks,nkmesh,nng,nrcell
    type (irdkpt_struct),pointer,intent(out):: irdkpt(:)
    integer,intent(out):: nkr,nkfxa,nkc
    integer,intent(in):: myrank,comm

    integer:: nkdt,nsit,ngt,nnpt
    real(kind=8):: aat(3,3),qft,rft,qmt,qct
    integer:: rgt(3,3,ng),pgt(3,ng),srgt(ng),nkt,nskt,mmmt(3,2,nsk),idatt(3)

    integer:: i,j,ig,isk,ierr
    integer:: iks

    if (myrank.eq.0) then
       call xrewind(NFPOTS)
       read(NFPOTS,iostat=ierr) nkdt, nsit, aat, qft, rft, qmt, qct, ngt, nnpt
       if (ierr.ne.0) then
          write(6,'(1x,a)') 'error in opening P.P. init data'
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
       if ((nkdt .ne. nkd) .or. (nsit .ne. nsi)) then
          write(6,*) 'Error: invalid potential data'
          write(6,*) 'nkd, nsi = ', nkdt, nsit
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
       do j=1,3
          do i=1,3
             if (aa0(i,j) .ne. aat(i,j)) then
                write(6,*) 'Error: invalid potential data'
                write(6,*) 'i, j, aa0(i,j) = ', i, j, aa0(i,j)
                call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
                stop
             end if
          end do
       end do
       if ((qft .ne. qf) .or. (rft .ne. rf) .or. &
            (qmt .ne. qm) .or. (qct .ne. qc) .or. &
            (ngt .ne. ng) .or. (nnpt .ne. nnp)) then
          write(6,*) 'Error: invalid potential data'
          write(6,*) 'qf, rf, qm, qc, ng, nnp = ', &
               qft, rft, qmt, qct, ng, nnp
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
       read(NFPOTS) (((rgt(i,j,ig),i=1,3),j=1,3),ig=1,ng)
       read(NFPOTS) ((pgt(i,ig),i=1,3),ig=1,ng),(srgt(ig),ig=1,ng)
       do ig=1,ng
          do j=1,3
             do i=1,3
                if (rgt(i,j,ig) .ne. rg(i,j,ig)) then
                   write(6,*) 'Error: invalid potential data'
                   write(6,*) 'i, j, ig, rg = ', i, j, ig, rgt(i,j,ig)
                   call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
                   stop
                end if
             end do
             if (pgt(j,ig) .ne. pg(j,ig)) then
                write(6,*) 'Error: invalid potential data'
                write(6,*) 'j, ig, pg = ', j, ig, pgt(j,ig)
                call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
                stop
             end if
          end do
          if (srgt(ig) .ne. srg(ig)) then
             write(6,*) 'Error: invalid potential data'
             write(6,*) 'ig, srg = ', ig, srgt(ig)
             call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
             stop
          end if
       end do
       read(NFPOTS) nkt, nskt
       if ((nkt .ne. nk) .or. (nskt .ne. nsk)) then
          write(6,*) 'Error: invalid potential data'
          write(6,*) 'nk, nsk = ', nk, nsk
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
       read(NFPOTS) (((mmmt(i,j,isk),i=1,3),j=1,2),isk=1,nsk)
       do isk=1,nsk
          do j=1,2
             do i=1,3
                if (mmmt(i,j,isk) .ne. mmm(i,j,isk)) then
                   write(6,*) 'Error: invalid potential data'
                   write(6,*) 'i, j, isk, mmm = ', i, j, isk, mmm(i,j,isk)
                   call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
                   stop
                end if
             end do
          end do
       end do
       read(NFPOTS) (idatt(i),i=1,3)
       if (nsk.ge.2 .and. maxval(abs(idatt - idat)) .gt. 0) then
          write(6,*) 'Error: invalid potential data'
          write(6,*) 'i, idat = ', idat(i)
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
       
       read(NFPOTS) nvx,nvy,nvz,nvxyz
       read(NFPOTS) nng,nrcell,nkr,nkfxa
       read(NFPOTS) nkc
       call load_irdkpt(NFPOTS,nkmesh,nks,irdkpt)
    end if

    call MPI_BCAST(nvx,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nvy,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nvz,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nvxyz,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nks,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nkmesh,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nng,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nrcell,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nkr,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nkfxa,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nkc,1,MPI_INTEGER,0,comm,ierr)

    if (myrank.ne.0) then
       allocate(irdkpt(nks))
    end if
    do iks=1,nks
       call MPI_BCAST(irdkpt(iks)%nsg,1,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(irdkpt(iks)%ngir,1,MPI_INTEGER,0,comm,ierr)
       if (myrank.ne.0) then
          allocate(irdkpt(iks)%sgrg(irdkpt(iks)%nsg), &
               irdkpt(iks)%irrg(irdkpt(iks)%ngir))
       end if
       call MPI_BCAST(irdkpt(iks)%sgrg,irdkpt(iks)%nsg,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(irdkpt(iks)%irrg,irdkpt(iks)%ngir,MPI_INTEGER,0,comm,ierr)

       call MPI_BCAST(irdkpt(iks)%kp,3,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(irdkpt(iks)%sk,3,MPI_DOUBLE_PRECISION,0,comm,ierr)
    end do

    return
  end subroutine loadchkpots

  integer function getlaprj(nkd,bspots)
    implicit none
    integer,intent(in):: nkd
    type(bspot_for_atm):: bspots(nkd)

    integer:: ikd
    integer:: retval

    retval = 0
    do ikd=1,nkd
       retval = max(retval,bspots(ikd)%nprj)
    end do
    getlaprj = retval
  end function getlaprj

  integer function getlmax(nkd,bspots)
    implicit none
    integer,intent(in):: nkd
    type(bspot_for_atm):: bspots(nkd)

    integer:: ikd
    integer:: retval

    retval = 0
    do ikd=1,nkd
       retval = max(retval,bspots(ikd)%l2lq(bspots(ikd)%nl))
    end do
    getlmax = retval+1
  end function getlmax

  integer function getlpnum(nkd,bspots)
    implicit none
    integer,intent(in):: nkd
    type(bspot_for_atm):: bspots(nkd)

    integer:: ikd
    integer:: retval

    retval = 0
    do ikd=1,nkd
       retval = max(retval,bspots(ikd)%nl)
    end do

    getlpnum = retval
    return
  end function getlpnum

  integer function getmaxr(nkd,bspots)
    implicit none
    integer,intent(in):: nkd
    type(bspot_for_atm):: bspots(nkd)

    integer:: ikd,il
    integer:: retval

    retval = 0
    do ikd=1,nkd
       do il=1,bspots(ikd)%nl
          retval = max(retval,bspots(ikd)%nr(il))
       end do
    end do
    getmaxr = retval
  end function getmaxr
end module bsplpot
