!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015-2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module uspp_cr
contains
  subroutine crgen(nspin,bkc,ekc,kqg,pby,ndens,crhoio &
       ,bzwk,wgtk,nsi,nkd,kd &
       ,laprj,ng,nnp,rg,pg,srg,aa,nrcell,nkmesh,nks,asi &
       ,nrx,nry,nrz,nrxyz,nb1,nb2,nw,nwk,idcwk &
       ,dqgs,bspots,iflg,mssg,comm)
    use fft_3d
    use bsplpot
    use types
    implicit none
    integer:: nspin,nsi,nkd,laprj,ng,nrcell,nkmesh,nks,ndens
    integer:: nrx,nry,nrz,nrxyz,nb1,nb2,nw,mssg
    integer,intent(in):: nnp,rg(3,3,ng),pg(3,ng),srg(ng)
    real(kind=8):: aa(3,3)
    real(kind=8):: bzwk(nks)
    real(kind=8):: wgtk(nb1:nb2,nks,nspin),asi(3,nsi)
    integer:: kd(nsi)
    real(kind=8):: crhoio(nrxyz,nspin*ndens)
    complex(kind=8):: pby(laprj,nsi,nw,nks,nspin)
    integer bkc,ekc,kqg(3,bkc:ekc),nwk(nrxyz,3),idcwk(nrcell,2)
    type (qg_for_atm):: dqgs(nkd,ndens)
    type (bspot_for_atm):: bspots(nkd)
    integer iflg,comm

    integer:: is
    
    if (nspin.eq.1) then
       call crgen_core(bkc,ekc,kqg,pby,ndens,crhoio &
            ,bzwk,wgtk,nsi,nkd,kd &
            ,laprj,ng,nnp,rg,pg,srg,aa,nrcell,nkmesh,nks,asi &
            ,nrx,nry,nrz,nrxyz,nb1,nb2,nw,nwk,idcwk &
            ,dqgs,bspots,iflg,mssg,comm)
    else
       do is=1,nspin
          call crgen_core(bkc,ekc,kqg,pby(1,1,1,1,is),ndens,crhoio(1,is) &
               ,bzwk,wgtk(nb1,1,is),nsi,nkd,kd &
               ,laprj,ng,nnp,rg,pg,srg,aa,nrcell,nkmesh,nks,asi &
               ,nrx,nry,nrz,nrxyz,nb1,nb2,nw,nwk,idcwk &
               ,dqgs,bspots,iflg,mssg,comm)
       end do
    end if
    return
  end subroutine crgen

  subroutine crgen_core(bkc,ekc,kqg,pby,ndens,crhoio &
       ,bzwk,wgtk,nsi,nkd,kd &
       ,laprj,ng,nnp,rg,pg,srg,aa,nrcell,nkmesh,nks,asi &
       ,nrx,nry,nrz,nrxyz,nb1,nb2,nw,nwk,idcwk &
       ,dqgs,bspots,iflg,mssg,comm)
    use fft_3d
    use bsplpot
    use util
    use xmpi
    use symchrg
    use types
    use timeprof
    implicit none
    integer:: nsi,nkd,laprj,ng,nrcell,nkmesh,nks,ndens
    integer:: nrx,nry,nrz,nrxyz,nb1,nb2,nw,mssg
    integer,intent(in):: nnp,rg(3,3,ng),pg(3,ng),srg(ng)
    real(kind=8):: aa(3,3)
    real(kind=8):: bzwk(nks)
    real(kind=8):: wgtk(nb1:nb2,nks),asi(3,nsi)
    integer:: kd(nsi)
    real(kind=8):: crhoio(nrxyz,ndens)
    complex(kind=8):: pby(laprj,nsi,nw,nks)
    integer bkc,ekc,kqg(3,bkc:ekc),nwk(nrxyz,3),idcwk(nrcell,2)
    type(fft3_struct) fs
    type (qg_for_atm):: dqgs(nkd,ndens)
    type (bspot_for_atm):: bspots(nkd)
    integer iflg,comm
    logical cnd
    !
    real(kind=8),allocatable:: ztmp(:,:), crho(:,:)
    complex(kind=8),allocatable:: qco(:,:)
    integer:: nrx2,nry2,nrz2
    integer:: ikd, ks, naob, ierr
    integer:: idens
    real(kind=8):: fac

    call tpf_begin(tpf_cchrg)
    !
    ! MAKE CORE CHARGE
    !
    cnd = .true.
    do ikd=1, nkd
       cnd = cnd .and. bspots(ikd)%ifncpp
    end do
    if (cnd) then
       if (mssg .gt. 0) then
          write(6,*) 'norm conserving p.p. only'
          write(6,*) 'bypass crgen.'
       end if
       crhoio = 0.0d0
       call tpf_end(tpf_cchrg)
       return
    end if
    nrx2=nrx*2
    nry2=nry*2
    nrz2=nrz*2
    naob =0
    do ikd=1,nkd
       naob =max(naob,bspots(ikd)%nob)
    end do
    !
    allocate(qco(naob*(naob+1)/2,nsi), stat=ierr)
    call chkalloc('crgen','qco',ierr)
    qco = 0.0d0
    do ks=1,nks
       call crgn0(qco,pby(1,1,1,ks) &
            ,bzwk(ks),wgtk(nb1,ks),nb1,nb2 &
            ,kd,naob,laprj,nkd,nsi,nw,bspots,iflg)
    end do
    !
    ! MAKE CORE CHARGE IN RECIPROCAL SPACE
    allocate(FFT_A1D_DEFSZ(ztmp,nrxyz), stat=ierr)
    call chkalloc('crgen','ztmp',ierr)

    allocate(FFT_A1D_DEFSZ(crho,nrxyz),stat=ierr)
    call chkalloc('crgen','crho',ierr)

    call fft3_init(nrx2,nry2,nrz2,nrx2,nry2,nrz2,fs)

    do idens=1,ndens
       ztmp = 0.0d0
       call crgn1(qco,ztmp,kqg,nrx2,nry2,nrz2,kd,asi &
            ,bkc,ekc,naob,nkd,nsi,dqgs(:,idens),bspots)

       ! call MPI_ALLREDUCE(ztmp,crho,nrxyz*2, &
       !    MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
       !
       ! work around for SR8000
       !
       call xmpi_alldsum(ztmp,crho,nrxyz*2,comm,ierr)
       call fft3_bw(fs,crho,ztmp)

       fac=1.0d0/dble(nkmesh)
       crhoio(:,idens) = fac*FFT_A1D_RL(crho,:)
    end do

    call fft3_done(fs)
    deallocate(ztmp,crho,qco)

    ! SYMMETRIZE CHARGE
    call ysymch(crhoio,nwk,idcwk,nrx2,nry2,nrz2,nrxyz,ng,nrcell)
    if (ndens.eq.4) then
       call symmag(crhoio(:,2:4),nrx2,nry2,nrz2,nrx2,nry2,nrz2, &
            ng,nnp,rg,pg,srg,aa)
    end if
    !
    call tpf_end(tpf_cchrg)
    return
  end subroutine crgen_core
  !******************************************************************
  subroutine crgn0(qco,pby,bzwk,wgtk,nb1,nb2 &
       ,kd,naob,laprj,nkd,nsi,nw,bspots,iflg)
    !  NAOB:MAX OF NOB
    use bsplpot
    implicit none
    integer nb1,nb2,naob,laprj,nkd,nsi,nw
    real(kind=8):: bzwk, wgtk(nb1:nb2)
    integer kd(nsi)
    complex(kind=8):: qco(naob*(naob+1)/2,nsi)
    complex(kind=8):: pby(laprj,nsi,nw)
    type (bspot_for_atm):: bspots(nkd)
    !
    ! flag to sum completely filled bands ib=1,nb-1
    !
    integer iflg
    !
    integer it,ikd,k1,l1,i1,m1,l2,i2,m2,ib,ip1,ip2
    !
!$omp parallel do private(ikd,k1,l1,i1,m1,l2,i2,m2,ib,ip1,ip2)
    do it=1,nsi
       k1=0
       ikd=kd(it)
       do ip1 = 1, bspots(ikd)%nprj
          l1 = bspots(ikd)%ip2l(ip1)
          i1 = bspots(ikd)%ip2r(ip1)
          m1 = bspots(ikd)%ip2m(ip1)
          do ip2 = 1, bspots(ikd)%l2pt(l1)
             l2 = bspots(ikd)%ip2l(ip2)
             i2 = bspots(ikd)%ip2r(ip2)
             m2 = bspots(ikd)%ip2m(ip2)
             if (.not.( (l1.eq.l2).and. &
                  ((i2.gt.i1) .or. ((i1.eq.i2).and.(m2.gt.m1))))) then
                k1=k1+1
                if (iflg .eq. 1) then
                   do ib=1,nb1-1
                      qco(k1,it)=qco(k1,it) + &
                           bzwk*conjg(pby(ip1,it,ib))*pby(ip2,it,ib)
                   end do
                endif
                do ib=nb1,nb2
                   qco(k1,it)=qco(k1,it) &
                        +wgtk(ib)*conjg(pby(ip1,it,ib))*pby(ip2,it,ib)
                end do
             end if
          end do
       end do
    end do
    return
  end subroutine crgn0

  subroutine crgn1(qco,crho,kqg,nrx2,nry2,nrz2,kd,asi &
       ,bkc,ekc,naob,nkd,nsi,dqgs,bspots)
    ! NAOB:MAX OF NOB
    use bsplpot
    use types
    use timeprof
#ifdef _OPENMP
    use omp_lib
#endif
    implicit none
    integer:: nrx2,nry2,nrz2
    integer:: nkd,nsi,bkc,ekc,naob
    complex(kind=8):: qco(naob*(naob+1)/2,nsi)
    integer:: kqg(3,bkc:ekc)

    integer:: kd(nsi)
    real(kind=8):: asi(3,nsi)

    type (bspot_for_atm):: bspots(nkd)
    type (qg_for_atm):: dqgs(nkd)

    real(kind=8):: FFT_A3D_DEFSZ(crho,nrx2,nry2,nrz2)
    !
    real(kind=8),parameter:: pi2 = 6.283185307179586d0
    integer,parameter:: blkkc = 100, blkthr = 200
    !
!SOPTION CONTAREA(qg)
    complex(kind=8),pointer:: qg(:,:)
    complex(kind=8):: sfct(bkc:ekc)
    complex(kind=8):: cdwrk(bkc:ekc), cowrk(bkc:ekc)
    !
    integer:: itt, it, ig, k1, l1, i1, m1, l2, i2, m2, ip1, ip2
    integer:: j1, j2, j3
    integer:: nsikd, itk, itk2it(nsi)
    real(kind=8):: t1, t2, t3, ph
    integer:: numk1, bbkc, bekc
    integer:: flag(naob*(naob+1)/2)
    logical:: ifblock
    !
#ifdef _OPENMP
    if (ekc-bkc+1 .gt. omp_get_max_threads()*blkthr) then
       ifblock = .true.
    else
       ifblock = .false.
    end if
#else
    ifblock = .true.
#endif

    cdwrk = (0.0d0, 0.0d0)
    cowrk = (0.0d0, 0.0d0)

    do itt=1,nkd
       qg => dqgs(itt)%qg

       nsikd = 0
       do it=1, nsi
          if (kd(it).eq.itt) then
             nsikd = nsikd + 1
             itk2it(nsikd) = it
          end if
       end do

       numk1=0
       do ip1 = 1,bspots(itt)%nprj
          l1 = bspots(itt)%ip2l(ip1)
          i1 = bspots(itt)%ip2r(ip1)
          m1 = bspots(itt)%ip2m(ip1)
          do ip2 = 1,bspots(itt)%l2pt(l1)
             l2 = bspots(itt)%ip2l(ip2)
             i2 = bspots(itt)%ip2r(ip2)
             m2 = bspots(itt)%ip2m(ip2)
             if (.not.( (l1.eq.l2).and. &
                  ((i2.gt.i1) .or. ((i1.eq.i2).and.(m2.gt.m1))))) then
                numk1=numk1+1
                if ( (bspots(itt)%ifcr(i1,l1).gt.0) &
                     .or. (bspots(itt)%ifcr(i2,l2).gt.0) ) then
                   if ((l1.eq.l2).and.(i1.eq.i2).and.(m1.eq.m2)) then
                      flag(numk1) = 1
                   else
                      flag(numk1) = 2
                   end if
                else
                   flag(numk1) = 0
                endif
             end if
          end do
       end do
       if (ifblock) then
!$omp parallel do private(itk,it,ig,k1,t1,t2,t3,ph,bbkc,bekc)
         do bbkc=bkc,ekc,blkkc
           bekc = min(bbkc+blkkc-1,ekc)
           do itk=1,nsikd
             it=itk2it(itk)
             t1=asi(1,it)
             t2=asi(2,it)
             t3=asi(3,it)
             do ig=bbkc,bekc
               ph= - pi2*( t1*kqg(1,ig)+t2*kqg(2,ig)+t3*kqg(3,ig) )
               sfct(ig) = dcmplx(cos(ph),sin(ph))
             end do
             do k1=1,numk1
                if (flag(k1).eq.1) then
                   do ig=bbkc,bekc
                      cdwrk(ig) = cdwrk(ig) &
                           + dble(qco(k1,it))*qg(ig,k1)*sfct(ig)
                   end do
                else if (flag(k1).eq.2) then
                   do ig=bbkc,bekc
                      cowrk(ig) = cowrk(ig) + qco(k1,it)*qg(ig,k1)*sfct(ig)
                   end do
                end if
             end do
           end do
         end do
       else
!$omp parallel do private(itk,it,ig,k1,t1,t2,t3,ph,sfct), &
!$omp reduction(+:cdwrk),reduction(+:cowrk)
         do itk=1,nsikd
           it=itk2it(itk)
           t1=asi(1,it)
           t2=asi(2,it)
           t3=asi(3,it)
           do ig=bkc,ekc
              ph = - pi2*( t1*kqg(1,ig)+t2*kqg(2,ig)+t3*kqg(3,ig) )
              sfct(ig) = dcmplx(cos(ph),sin(ph))
           end do
           do k1=1,numk1
              if (flag(k1).eq.1) then
                 do ig=bkc,ekc
                    cdwrk(ig) = cdwrk(ig) &
                         + dble(qco(k1,it))*qg(ig,k1)*sfct(ig)
                 end do
              else if (flag(k1).eq.2) then
                 do ig=bkc,ekc
                    cowrk(ig) = cowrk(ig) + qco(k1,it)*qg(ig,k1)*sfct(ig)
                 end do
              end if
           end do
         end do
       end if
    end do

    call tpf_begin(tpf_scgt)
!$omp parallel do private(ig,j1,j2,j3)
    do ig=bkc,ekc
       j1=mod(kqg(1,ig)+nrx2,nrx2)+1
       j2=mod(kqg(2,ig)+nry2,nry2)+1
       j3=mod(kqg(3,ig)+nrz2,nrz2)+1
       FFT_A3D_RL(crho,j1,j2,j3) = FFT_A3D_RL(crho,j1,j2,j3) &
            + dble(cdwrk(ig)) + dble(cowrk(ig))
       FFT_A3D_IM(crho,j1,j2,j3) = FFT_A3D_IM(crho,j1,j2,j3) &
            + aimag(cdwrk(ig)) + aimag(cowrk(ig))
    end do

!$omp parallel do private(ig,j1,j2,j3)
    do ig=bkc,ekc
       j1=mod(-kqg(1,ig)+nrx2,nrx2)+1
       j2=mod(-kqg(2,ig)+nry2,nry2)+1
       j3=mod(-kqg(3,ig)+nrz2,nrz2)+1
       FFT_A3D_RL(crho,j1,j2,j3) = FFT_A3D_RL(crho,j1,j2,j3) &
            + dble(cowrk(ig))
       FFT_A3D_IM(crho,j1,j2,j3) = FFT_A3D_IM(crho,j1,j2,j3) &
            - aimag(cowrk(ig))
    end do
    call tpf_end(tpf_scgt)

    return
  end subroutine crgn1

  subroutine calc_cratmstrs(aa,ncomp,nspin,ndens,nb1,nb2,nw &
       ,lpt,wgtk,nsi,nkd,kd,asi,laprj,pby &
       ,seatm,mgs,bspotfc,bspots,mssg,comm)
    use types
    use util
    use meshgrid
    use bsplpot
    use subr_kpoint
    use converter
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: ncomp,nspin,ndens,nb1,nb2,nw,nsi,nkd,laprj
    type (meshgrid_struct):: mgs
    real(kind=8),intent(in):: wgtk(nb1:nb2,mgs%nks,nspin),asi(3,nsi)
    integer,intent(in):: kd(nsi)
    complex(kind=8),intent(in):: lpt(mgs%nvxyz,nspin*ndens)
    complex(kind=8),intent(in):: pby(laprj,nsi,nw,mgs%nks,nspin)
    real(kind=8):: seatm(6,nsi)
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    type (irdkpt_struct),pointer:: irdkpt(:)
    integer:: mssg,comm

    integer:: nkmesh
    integer:: nvx,nvy,nvz,nvxyz,nks
    integer:: bkc,ekc
    integer,pointer:: kqg(:,:)
    real(kind=8):: bzwk(mgs%nks)
    complex(kind=8),allocatable:: xseatm(:)
    type (qg_for_atm),allocatable:: dqgs(:,:)
    integer:: is,it,ierr
    integer:: idens,ikd,nob
    real(kind=8),allocatable:: qgk(:,:)
    real(kind=8):: va,bb(3,3)
    complex(kind=8),pointer:: qg(:,:)

    nks = mgs%nks
    nkmesh = mgs%nkmesh

    irdkpt => mgs%irdkpt
    kqg => mgs%kqg
    bkc = mgs%bkc
    ekc = mgs%ekc

    nvx = mgs%nvx
    nvy = mgs%nvy
    nvz = mgs%nvz
    nvxyz = mgs%nvxyz

    va = cellvol(aa)
    call recp_uvec(aa,bb)
    
    bzwk(:) = irdkpt(:)%ngir
    if (nspin .eq. 1) then
       bzwk = 2.0d0/ncomp*bzwk
    end if

    allocate(dqgs(nkd,ndens),stat=ierr)
    call chkalloc('calc_cratmstrs','qgs',ierr)
    do idens=1,ndens
       do ikd=1,nkd
          nob=bspots(ikd)%nob
          allocate(dqgs(ikd,idens)%qg(bkc:ekc,nob*(nob+1)/2),stat=ierr)
          call chkalloc('calc_cratmstrs','qg',ierr)
       end do
    end do
    allocate(qgk(4,bkc:ekc),stat=ierr)
    call chkalloc('calc_cratmstrs','qgk',ierr)
    call kqg2qgk(bb,ekc-bkc+1,kqg,qgk)
    do idens=1,ndens
       do ikd=1,nkd
          qg => dqgs(ikd,idens)%qg
          nob=bspots(ikd)%nob
          call evqg(bspotfc,bspots(ikd),ekc-bkc+1,qgk,va,nob,idens,qg)
       end do
    end do
    deallocate(qgk)

    allocate(xseatm(nsi),stat=ierr)
    call chkalloc('calc_cratmstrs','xseatm',ierr)
    xseatm = 0.0d0
    do is=1,nspin
       call cratmstrs_core(ndens,nb1,nb2,nw &
            ,nvx,nvy,nvz,nvxyz,lpt(:,is),bzwk,wgtk(:,:,is),nsi,nkd,kd,asi &
            ,laprj,pby(:,:,:,:,is),xseatm &
            ,bkc,ekc,kqg,nks &
            ,dqgs,bspots,mssg,comm)
    end do
    do it=1,nsi
       seatm(1,it) = va/dble(nkmesh)*dble(xseatm(it))
       seatm(2,it) = va/dble(nkmesh)*dble(xseatm(it))
       seatm(3,it) = va/dble(nkmesh)*dble(xseatm(it))
       seatm(4,it) = 0.0d0
       seatm(5,it) = 0.0d0
       seatm(6,it) = 0.0d0
    end do

    do idens=1,ndens
       do ikd=1,nkd
          deallocate(dqgs(ikd,idens)%qg)
       end do
    end do
    deallocate(dqgs)

    return
  end subroutine calc_cratmstrs

  subroutine cratmstrs_core(ndens,nb1,nb2,nw &
       ,nvx,nvy,nvz,nvxyz,lpt,bzwk,wgtk,nsi,nkd,kd,asi &
       ,laprj,pby,seatm &
       ,bkc,ekc,kqg,nks &
       ,dqgs,bspots,mssg,comm)
    use bsplpot
    use util
    use xmpi
    use types
    use timeprof
    implicit none
    integer,intent(in):: nvx,nvy,nvz,nvxyz
    complex(kind=8),intent(in):: lpt(nvxyz,ndens)
    integer,intent(in):: ndens,nb1,nb2,nw,nsi,nkd,laprj
    real(kind=8),intent(in):: bzwk(nks),wgtk(nb1:nb2,nks),asi(3,nsi)
    integer,intent(in):: kd(nsi)
    complex(kind=8),intent(in):: pby(laprj,nsi,nw,nks)
    complex(kind=8):: seatm(nsi)
    integer,intent(in):: nks
    integer,intent(in):: bkc,ekc,kqg(3,bkc:ekc)
    type (qg_for_atm):: dqgs(nkd,ndens)
    type (bspot_for_atm):: bspots(nkd)
    integer:: mssg,comm
    logical:: cnd
    !
    complex(kind=8),allocatable:: pxseatm(:), xseatm(:)
    complex(kind=8),allocatable:: qco(:,:)
    integer:: ikd, ks, naob, ierr
    integer:: idens

    call tpf_begin(tpf_cchrg)
    !
    ! make core stress per atom
    !
    cnd = .true.
    do ikd=1, nkd
       cnd = cnd .and. bspots(ikd)%ifncpp
    end do
    if (cnd) then
       if (mssg .gt. 0) then
          write(6,*) 'norm conserving p.p. only'
          write(6,*) 'bypass cratmstrs.'
       end if
       call tpf_end(tpf_cchrg)
       return
    end if
    naob =0
    do ikd=1,nkd
       naob = max(naob,bspots(ikd)%nob)
    end do
    !
    allocate(qco(naob*(naob+1)/2,nsi), &
         pxseatm(nsi),xseatm(nsi),stat=ierr)
    call chkalloc('cratmstrs','qco,pxseatm,xseatm',ierr)
    qco = 0.0d0
    do ks=1,nks
       call crgn0(qco,pby(1,1,1,ks) &
            ,bzwk(ks),wgtk(nb1,ks),nb1,nb2 &
            ,kd,naob,laprj,nkd,nsi,nw,bspots,1)
    end do
    !
    do idens=1,ndens
       pxseatm = 0.0d0
       call cratmstrs1(nvx,nvy,nvz,lpt(:,idens),naob,qco &
            ,bkc,ekc,kqg,nkd,nsi,kd,asi &
            ,pxseatm,dqgs(:,idens),bspots)

       call xmpi_allzsum(pxseatm,xseatm,nsi,comm,ierr)

       seatm = seatm + xseatm
    end do

    deallocate(pxseatm,xseatm,qco)
    !
    call tpf_end(tpf_cchrg)
    return
  end subroutine cratmstrs_core

  subroutine cratmstrs1(nvx,nvy,nvz,lpt,naob,qco &
       ,bkc,ekc,kqg,nkd,nsi,kd,asi &
       ,seatm,dqgs,bspots)
    ! NAOB:MAX OF NOB
    use bsplpot
    use types
    use timeprof
#ifdef _OPENMP
    use omp_lib
#endif
    implicit none
    integer,intent(in):: nvx,nvy,nvz
    complex(kind=8),intent(in):: lpt(-nvx:nvx,-nvy:nvy,-nvz:nvz)
    integer,intent(in):: naob
    complex(kind=8),intent(in):: qco(naob*(naob+1)/2,nsi)
    integer,intent(in):: bkc,ekc,nkd,nsi
    integer,intent(in):: kqg(3,bkc:ekc)
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: asi(3,nsi)
    complex(kind=8):: seatm(nsi)

    type (bspot_for_atm):: bspots(nkd)
    type (qg_for_atm):: dqgs(nkd)

    !
    real(kind=8),parameter:: pi2 = 6.283185307179586d0
    !
!SOPTION CONTAREA(qg)
    complex(kind=8),pointer:: qg(:,:)
    complex(kind=8):: sfct(bkc:ekc)
    !
    integer:: itt, it, ig, k1, l1, i1, m1, l2, i2, m2, ip1, ip2
    integer:: j1, j2, j3
    integer:: nsikd, itk, itk2it(nsi)
    real(kind=8):: t1, t2, t3, ph
    integer:: numk1
    integer:: flag(naob*(naob+1)/2)
    complex(kind=8):: ztmp
    integer:: bbkc, bekc, num_thr, my_thr, blkkc
    !

    do itt=1,nkd
       qg => dqgs(itt)%qg

       nsikd = 0
       do it=1, nsi
          if (kd(it).eq.itt) then
             nsikd = nsikd + 1
             itk2it(nsikd) = it
          end if
       end do

       numk1=0
       do ip1 = 1,bspots(itt)%nprj
          l1 = bspots(itt)%ip2l(ip1)
          i1 = bspots(itt)%ip2r(ip1)
          m1 = bspots(itt)%ip2m(ip1)
          do ip2 = 1,bspots(itt)%l2pt(l1)
             l2 = bspots(itt)%ip2l(ip2)
             i2 = bspots(itt)%ip2r(ip2)
             m2 = bspots(itt)%ip2m(ip2)
             if (.not.( (l1.eq.l2).and. &
                  ((i2.gt.i1) .or. ((i1.eq.i2).and.(m2.gt.m1))))) then
                numk1=numk1+1
                if ( (bspots(itt)%ifcr(i1,l1).gt.0) &
                     .or. (bspots(itt)%ifcr(i2,l2).gt.0) ) then
                   if ((l1.eq.l2).and.(i1.eq.i2).and.(m1.eq.m2)) then
                      flag(numk1) = 1
                   else
                      flag(numk1) = 2
                   end if
                else
                   flag(numk1) = 0
                endif
             end if
          end do
       end do
!$omp parallel private(itk,it,ig,k1,t1,t2,t3,ph,ztmp), &
!$omp private(num_thr,my_thr,blkkc,bbkc,bekc), &
!$omp reduction(+:seatm)
#ifdef _OPENMP
       num_thr = omp_get_num_threads()
       my_thr = omp_get_thread_num()
#else
       num_thr = 1
       my_thr = 0
#endif
       blkkc = (ekc-bkc+1)/num_thr
       bbkc = bkc + blkkc * my_thr
       bekc = bbkc + blkkc - 1
       if (my_thr .eq. num_thr-1) then
          bekc = ekc
       end if

       do itk = 1,nsikd
          it = itk2it(itk)
          t1 = asi(1,it)
          t2 = asi(2,it)
          t3 = asi(3,it)
          do ig = bbkc,bekc
             ph = - pi2*( t1*kqg(1,ig)+t2*kqg(2,ig)+t3*kqg(3,ig) )
             sfct(ig) = dcmplx(cos(ph),sin(ph))
          end do
          do k1 = 1,numk1
             if (flag(k1).eq.1) then
                do ig = bbkc,bekc
                   j1 = kqg(1,ig)
                   j2 = kqg(2,ig)
                   j3 = kqg(3,ig)
                   seatm(it) = seatm(it) + lpt(-j1,-j2,-j3) &
                        *dble(qco(k1,it))*qg(ig,k1)*sfct(ig)
                end do
             else if (flag(k1).eq.2) then
                do ig = bbkc,bekc
                   j1 = kqg(1,ig)
                   j2 = kqg(2,ig)
                   j3 = kqg(3,ig)
                   ztmp = dble(qco(k1,it))*qg(ig,k1)*sfct(ig)
                   seatm(it) = seatm(it) + lpt(-j1,-j2,-j3)*ztmp &
                        + lpt(j1,j2,j3)*conjg(ztmp)
                end do
             end if
          end do
       end do
!$omp end parallel
    end do
    return
  end subroutine cratmstrs1
end module uspp_cr
