!
! Copyright (c) 2014 Yoshihide Yoshimoto
!
! REF: S. Grimme, J. Compt. Chem. 27 (15) 1787-1799 (2006)
!
module dftd2
  public :: fdftd2

  ! 
  real(kind=8),parameter :: c6si(0:54) = &
       (/ 0.0d0, &
        ! none
          0.14d0,  0.08d0, &
        ! H,       He &
          1.61d0,  1.61d0,  3.13d0,  1.75d0, 1.23d0, 0.70d0, 0.75d0, 0.63d0, &
        ! Li,      Be,      B,       C,      N,      O,      F,      Ne &
          5.71d0,  5.71d0,  10.79d0, 9.23d0, 7.84d0, 5.57d0, 5.07d0, 4.61d0, &
        ! Na,      Mg,      Al,      Si,     P,      S,      Cl,     Ar &
          10.80d0, 10.80d0, &
        ! K,       Ca, &
          10.80d0, 10.80d0, 10.80d0, 10.80d0, 10.80d0, 10.80d0, 10.80d0, 10.80d0, 10.80d0, 10.80d0, &
        ! Sc,      Ti,      V,       Cr,      Mn,      Fe,      Co,      Ni,      Cu,      Zn, &
          16.99d0, 17.10d0, 16.37d0, 12.64d0, 12.47d0, 12.01d0, &
        ! Ga,      Ge,      As,      Se,      Br,      Kr, &
          24.67d0, 24.67d0, &
        ! Rb,      Sr, &
          24.67d0, 24.67d0, 24.67d0, 24.67d0, 24.67d0, 24.67d0, 24.67d0, 24.67d0, 24.67d0, 24.67d0, &
        ! Y,       Zr,      Nb,      Mo,      Tc,      Ru,      Rh,      Pd,      Ag,      Cd, &
          37.32d0, 38.71d0, 38.44d0, 31.74d0, 31.50d0, 29.99d0 /)
        ! In,      Sn,      Sb,      Te,      I,       Xe

  real(kind=8),parameter :: r0si(0:54) = &
       (/ 1.0d0, &
        ! none
          1.001d0, 1.012d0, &
        ! H,       He &
          0.825d0, 1.408d0, 1.485d0, 1.452d0, 1.397d0, 1.342d0, 1.287d0, 1.243d0, &
        ! Li,      Be,      B,       C,       N,       O,       F,       Ne &
          1.144d0, 1.364d0, 1.639d0, 1.716d0, 1.705d0, 1.683d0, 1.639d0, 1.595d0, &
        ! Na,      Mg,      Al,      Si,      P,       S,       Cl,      Ar &
          1.485d0, 1.474d0, &
        ! K,       Ca, &
          1.562d0, 1.562d0, 1.562d0, 1.562d0, 1.562d0, 1.562d0, 1.562d0, 1.562d0, 1.562d0, 1.562d0, &
        ! Sc,      Ti,      V,       Cr,      Mn,      Fe,      Co,      Ni,      Cu,      Zn, &
          1.650d0, 1.727d0, 1.760d0, 1.771d0, 1.749d0, 1.727d0, &
        ! Ga,      Ge,      As,      Se,      Br,      Kr, &
          1.628d0, 1.606d0, &
        ! Rb,      Sr, &
          1.639d0, 1.639d0, 1.639d0, 1.639d0, 1.639d0, 1.639d0, 1.639d0, 1.639d0, 1.639d0, 1.639d0, &
        ! Y,       Zr,      Nb,      Mo,      Tc,      Ru,      Rh,      Pd,      Ag,      Cd, &
          1.672d0, 1.804d0, 1.881d0, 1.892d0, 1.892d0, 1.881d0 /)
        ! In,      Sn,      Sb,      Te,      I,       Xe
   
   real(kind=8),parameter :: s6(11) = &
        (/ 0.75d0, 0.0d0,   0.0d0, 0.0d0,  0.0d0,   0.0d0,  0.6d0,  0.0d0,  1.05d0, 0.0d0,   0.0d0 /)
   !       PBE,    PBEsol,  PW91,  CAPZ,   LDAPW92, HF,     PBE0,   PBE0_X, B3LYP,  B3LYP_X, B3LYP_loc

   real(kind=8),parameter :: delta = 1.0d-9
   real(kind=8),parameter :: onebohr = 0.052917721067d0 ! nm
   real(kind=8),parameter :: onemol = 6.0221415d23
   real(kind=8),parameter :: onehart = 4.35974417d-18 ! J
   real(kind=8),parameter :: oneang = 0.52917721067d0 ! Angstrom CODATA 2014
   real(kind=8),parameter :: alp6 = 20.0d0

   real(kind=8),parameter :: c6(0:54) = c6si*(1.0d0/onehart/(onebohr**6)/onemol)
   real(kind=8),parameter :: r0(0:54) = r0si/oneang
   !real(kind=8),parameter :: c6(0:54) = c6si*17.3452656005951d0
   !real(kind=8),parameter :: r0(0:54) = r0si/0.52917726d0

contains
  !
  subroutine fdftd2(xctp,nkd,bspots,aa,nsi,kd,asi,eng,seng,cforce)
    use bsplpot
    use util
    implicit none
    integer:: xctp, nkd
    type (bspot_for_atm):: bspots(nkd)
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: nsi,kd(nsi)
    real(kind=8),intent(in):: asi(3,nsi)
    real(kind=8),intent(out):: eng,seng(6),cforce(3,nsi)

    integer:: i,j,k1,k2,k3
    integer:: izkd,jzkd
    real(kind=8):: dst(3),cdst(4),zdst(3),tcd(4),r(4)
    real(kind=8):: v2,dv2r
    real(kind=8):: tmp1,tmp2
    real(kind=8):: h(3), rmax
    integer:: ndl(3)

    !
    ! quick return if appropriate
    !
    if (s6(xctp).eq.0.0d0) return

    call latticeinterval(aa,h)
    tmp1 = 0.0d0
    do i=1,nsi
       izkd = nint(bspots(kd(i))%zn)
       tmp1 = max(tmp1,c6(izkd))
    end do
    rmax = (tmp1/delta)**(1.0d0/6.0d0)
    ndl(:) = ceiling(rmax/h(:))+1

    if (maxval(nint(bspots(kd(:))%zo)).gt.54) then
       write(6,*) 'Unsupported Atoms are specified for DFT-D2'
       stop
    end if
    if (xctp.lt.1 .or. xctp.gt.11) then
       write(6,*) 'Unsupported XC type for DFT-D2'
       stop
    end if

    eng = 0.0d0
    seng = 0.0d0
    cforce = 0.0d0

!$omp parallel do private(dst,cdst,zdst,tcd,r,i,j,k1,k2,k3), &
!$omp private(izkd,jzkd,v2,dv2r,tmp1,tmp2), &
!$omp reduction(+:eng,seng,cforce)
!poption cyclic,tlocal(cdst,izkd,jzkd,tmp1,tmp2),psum(eng,seng,cforce)
    do i=1,nsi
       !
       ! include the self images for energy
       !
       do j=i,nsi

          zdst = asi(:,i) - asi(:,j)
          zdst = zdst - nint(zdst)
          do k3 = -ndl(3),ndl(3)
             do k2 = -ndl(2),ndl(2)
                do k1 = -ndl(1),ndl(1)

                   dst(1) = zdst(1) + k1
                   dst(2) = zdst(2) + k2
                   dst(3) = zdst(3) + k3
                   tcd(1) = aa(1,1)*dst(1) + aa(1,2)*dst(2) + aa(1,3)*dst(3)
                   tcd(2) = aa(2,1)*dst(1) + aa(2,2)*dst(2) + aa(2,3)*dst(3)
                   tcd(3) = aa(3,1)*dst(1) + aa(3,2)*dst(2) + aa(3,3)*dst(3)
                   tcd(4) = tcd(1)*tcd(1)+tcd(2)*tcd(2)+tcd(3)*tcd(3)

                   cdst = tcd

                   tmp1 = sqrt(cdst(4))
                   if (tmp1 .gt. rmax) cycle
                   cdst(4) = tmp1
                   cdst(1:3) = cdst(1:3)/tmp1
                   r(1:4) = cdst

                   izkd = nint(bspots(kd(i))%zn)
                   jzkd = nint(bspots(kd(j))%zn)

                   v2 = 0.0d0
                   dv2r = 0.0d0

                   !
                   ! include the self images for energy
                   !
                   if ( r(4) .gt. 100*epsilon(rmax) ) then
                      !
                      ! van der Waals attraction
                      !
                      tmp2 = exp(-alp6*(r(4)/(r0(izkd)+r0(jzkd)) - 1.0d0))
                      tmp1 = s6(xctp)*sqrt(c6(izkd)*c6(jzkd))/r(4)**6
                      ! v2(ikd,jkd; rij)
                      v2 = v2 - tmp1/(1.0d0+tmp2)

                      ! dv2r(ikd,jkd; rij)
                      dv2r = dv2r + 6.0d0*tmp1/r(4)/(1.0d0+tmp2) + tmp1/(1.0d0+tmp2)**2*tmp2*( -alp6/(r0(izkd)+r0(jzkd)) )

                      !
                      if (i.ne.j) then
                         eng = eng + v2
                         !
                         ! df/ds_xx =   df/dr_x * dr_x/ds_xx + df/dr_y * dr_y/ds_xx
                         !            + df/dr_y * dr_y/ds_xx
                         ! df/dr_x = df/d|r| * r_x/|r|
                         ! dr_x/ds_xx = r_x
                         ! dr_y/ds_xx = 0
                         ! dr_z/ds_xx = 0
                         ! df/ds_xx = df/d|r| * r_x*r_x/|r|
                         !
                         ! CODE: r(1:3) = (r_x, r_y, r_z)/|r|
                         !
                         seng(1) = seng(1) - dv2r*r(1)*r(1)*r(4)
                         seng(2) = seng(2) - dv2r*r(2)*r(2)*r(4)
                         seng(3) = seng(3) - dv2r*r(3)*r(3)*r(4)
                         seng(4) = seng(4) - 2.0d0*dv2r*r(2)*r(3)*r(4)
                         seng(5) = seng(5) - 2.0d0*dv2r*r(3)*r(1)*r(4)
                         seng(6) = seng(6) - 2.0d0*dv2r*r(1)*r(2)*r(4)
                         cforce(:,i) = cforce(:,i) - dv2r*r(1:3)
                         cforce(:,j) = cforce(:,j) + dv2r*r(1:3)
                      else
                         !
                         ! self images do not contribute forces
                         ! energy contribution per a cell is 1/2
                         !
                         eng = eng + 0.5d0*v2
                         !
                         ! df/ds_xx =   df/dr_x * dr_x/ds_xx + df/dr_y * dr_y/ds_xx
                         !            + df/dr_y * dr_y/ds_xx
                         ! df/dr_x = df/d|r| * r_x/|r|
                         ! dr_x/ds_xx = r_x
                         ! dr_y/ds_xx = 0
                         ! dr_z/ds_xx = 0
                         ! df/ds_xx = df/d|r| * r_x*r_x/|r|
                         !
                         ! CODE: r(1:3) = (r_x, r_y, r_z)/|r|
                         !
                         seng(1) = seng(1) - 0.5d0*dv2r*r(1)*r(1)*r(4)
                         seng(2) = seng(2) - 0.5d0*dv2r*r(2)*r(2)*r(4)
                         seng(3) = seng(3) - 0.5d0*dv2r*r(3)*r(3)*r(4)
                         seng(4) = seng(4) - dv2r*r(2)*r(3)*r(4)
                         seng(5) = seng(5) - dv2r*r(3)*r(1)*r(4)
                         seng(6) = seng(6) - dv2r*r(1)*r(2)*r(4)
                      end if
                   end if
                end do
             end do
          end do
       end do
    end do

    return
  end subroutine fdftd2

end module dftd2
