!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015,2016 Yoshihide Yoshimoto
!
!*********************************************************************
!  SCGDG3: Sophistcated version (1995-09-08)
!
!   CONJUGATE GRADIENT FOR FIXED CHARGE DENSITY RHO
!               WITH OVERLAP MATRIX S FOR VANDERBILT PSEUDOPOTENTIAL
!    ORIGINAL VERSION WRITTEN BY       (1990-11-28) OSAMU SUGINO
!    MODIFIED FOR SHIRAISHI FORM       (1992-03-18) JUN YAMAUCHI
!    S-MATRIX VERSION FOR VANDERBILT   (1992-10-01)
!
!    ZZEF2M(MSL2) -> ZHEEV(LAPACk)     (1995-07-04)
!    Correction for PBY                (1995-08-08)
!    OrthoNormalize Switch(ifon)       (1995-08-10)
!    Removed partitioned band          (1995-08-25)
!    CFILE is deleted
!    Work areas are unified to Y       (1995-09-07)
!     Last modified at 1995-09-08
!
!********************************************************************
module diag_dvdsn
contains
  !
  ! for general case
  !
  subroutine dvdsn1(ncomp,ndiag,itsb,nwx2,nwy2,nwz2,nlx,nly,nlz,nlxyz, &
       nkm,bkm,ekm, &
       nw,iks,sk,bb,kg,ee,ndens,vr, &
       out,lpnum,maxr,vpj,laprj,dd,asi, &
       ifoeps,eps,ier,p,va,nsi,nkd,kd,bspots,pby,qbtp,ifothnm, &
       xctp,dpf,mgsf,nb1,nb2,bzwk,wgtk, &
       nproc,myrank,g_count,g_displ,comm)
    !
    use mpi
    use bsplpot
    use fft_3d
    use converter
    use util
    use nonlocal_proj
    use rotate_wavefun
    use orthg_seavec
    use types
    use xcf_types
    use meshgrid
    implicit none
    integer,intent(in):: ndiag,itsb,nwx2,nwy2,nwz2,nlx,nly,nlz,nlxyz, &
         nkm,bkm,ekm,nw,lpnum,maxr,laprj,nsi,nkd
    integer,intent(in):: ncomp,ndens
    real(kind=8),intent(in):: va
    logical:: ifoeps,ifothnm
    real(kind=8):: eps
    integer,intent(in):: iks
    real(kind=8),intent(in):: sk(3),bb(3,3)
    real(kind=8),intent(in):: vr(nlxyz,ndens)
    integer,intent(in)::  kg(3,nkm),kd(nsi)
    real(kind=8),intent(in):: &
         asi(3,nsi),vpj(bkm:ekm,maxr,lpnum,nkd)
    complex(kind=8),intent(in):: dd(laprj,laprj,nsi)
    complex(kind=8):: p(ncomp,bkm:ekm,nw)
    real(kind=8),intent(in):: qbtp(3)
    type (bspot_for_atm):: bspots(nkd)
    !
    integer:: ier
    real(kind=8):: out(nw,3),ee(nw)
    complex(kind=8):: pby(laprj,nsi,nw)
    !
    integer,intent(in):: xctp
    type (meshgrid_struct):: mgsf
    type (wfn_for_k):: dpf(mgsf%nks)
    integer,intent(in):: nb1,nb2
    real(kind=8),intent(in):: bzwk(mgsf%nks),wgtk(nb1:nb2,mgsf%nks)
    !
    integer:: nproc,g_count(0:nproc-1),g_displ(0:nproc-1)
    integer:: myrank,comm
    !
    complex(kind=8),dimension(:,:,:),allocatable:: hp,sp
    complex(kind=8),dimension(:,:,:),allocatable:: cl,hcl,scl
    real(kind=8),allocatable:: akk(:)
    complex(kind=8),allocatable:: dcprj(:,:), dcprjcl(:,:)
    integer,allocatable:: pj_count(:),pj_displ(:)
    integer:: bprj,eprj,irnk
    integer,allocatable:: j2gs(:,:)
    type(nlprj_struct):: nlpjs

    real(kind=8),allocatable:: oldee(:),pee(:)
    !
    real(kind=8):: epsmax
    integer:: ig,ib,jb,idiag,nwxyz,ibc,nwc
    type(fft3_struct):: fs
    integer:: ntask,nblock
    integer:: ista
    integer:: bkms, ekms
    !
    call getntask(ntask)
    nblock = max(ntask,nproc)
    !
    nwxyz = nwx2*nwy2*nwz2

    bkms = (bkm-1)*ncomp+1
    ekms = ekm*ncomp
    !
    allocate(hp(ncomp,bkm:ekm,nw),sp(ncomp,bkm:ekm,nw), &
         akk(bkm:ekm),j2gs(nkm,2),stat=ista)
    call chkalloc('dvdsn1','tmp',ista)
    ! INITIALIZE
    call kg2akk(bb,sk,ekm-bkm+1,kg(1,bkm),akk)
    hp(:,:,:) = 0.0d0
    call fft3_init(nwx2,nwy2,nwz2,nlx,nly,nlz,fs)
    !
    if (fs%ifsft) then
       call kg2lj2gs(nwx2,nwy2,nwz2,nlx,nly,nlz,nkm,kg,j2gs)
    else
       call kg2lj2g(nwx2,nwy2,nwz2,nlx,nly,nlz,nkm,kg,j2gs(:,1))
       j2gs(:,2) = j2gs(:,1)
    end if
    !
    call nlprj_init(lpnum,maxr,nkd,bspots,nsi,kd,asi, &
         bkm,ekm,bb,sk,kg(1,bkm),vpj,nlpjs)

    allocate(pj_count(0:nproc-1),pj_displ(0:nproc-1),stat=ista)
    call chkalloc('dvdsn1','pj_count',ista)
    pj_count = nlpjs%nprj/nproc
    do irnk=0,mod(nlpjs%nprj,nproc)-1
       pj_count(irnk) = pj_count(irnk)+1
    end do
    pj_displ(0) = 0
    do irnk=0,nproc-2
       pj_displ(irnk+1) = pj_displ(irnk) + pj_count(irnk)
    end do
    bprj = pj_displ(myrank)+1
    eprj = pj_displ(myrank)+pj_count(myrank)

    allocate(dcprj(bprj:eprj,nw),stat=ista)
    call chkalloc('dvdsn1','dcprj',ista)

#ifdef TIMING
    write(6,'(1x,a,f10.2)') 'TIME DVDSN INIT= ', xetime(1)
#endif
    if (ndiag.le.0) ifoeps = .false.
    !     
    !  XXX: we should check scf state by an another variable
    !
    if (ifoeps) then
       do jb=1,nw-1
          if (ee(jb+1) .lt. ee(jb)) then
             ifoeps = .false.
          end if
       end do
    end if
    if (ifoeps) then
       allocate(oldee(nw),stat=ista)
       call chkalloc('dvdsn1','oldee',ista)
       oldee = ee
    end if
    !
    ! Zero clear variables outside for vectorization
    dcprj(:,:) = (0.0d0,0.0d0)
    !
    ! \Psi := p
    ! H \Psi -> hp
    ! S \Psi -> sp
    !
    call mulths(ncomp,nkm,bkm,ekm,nw,laprj,nsi,nkd,kd,bspots,va, &
         nlpjs,dd,nlxyz, &
         j2gs,ndens,vr,akk,p,hp,sp,bprj,eprj,dcprj,qbtp, &
         xctp,bb,sk,kg,dpf,mgsf,nb1,nb2,bzwk,wgtk, &
         fs,nproc,g_count,g_displ,comm)
#ifdef TIMING
    write(6,'(1x,a,f10.2)') 'TIME DVDSN MAKE HP,SP= ', xetime(1)
#endif

    if (.not. ifothnm) then
       !
       !  sub-space diagonalization
       !
       call ssdiag(bkms,ekms,nw,bprj,eprj,ee,p,hp,sp,dcprj,comm)
       
#ifdef TIMING
       write(6,'(1x,a,f10.2)') 'TIME DVDSN SSDIAG= ', xetime(1)
#endif
       ifothnm = .true.
    else if (ndiag.gt.0) then
       !     
       ! CALCULATE EIGENVALUE
       !     
       allocate(pee(nw),stat=ista)
       call chkalloc('dvdsn1','pee',ista)
       do jb=1,nw
          pee(jb) = 0.0d0
          if (ncomp.eq.1) then
             do ig=bkm,ekm
                pee(jb) = pee(jb) + conjg(p(1,ig,jb))*hp(1,ig,jb)
             end do
          else
             do ig=bkm,ekm
                pee(jb) = pee(jb) + conjg(p(1,ig,jb))*hp(1,ig,jb) &
                     + conjg(p(2,ig,jb))*hp(2,ig,jb)
             end do
          end if
       end do
       call MPI_ALLREDUCE(pee,ee,nw,MPI_DOUBLE_PRECISION,MPI_SUM, &
            comm,ista)
       deallocate(pee)
    end if
    !
    ! IF NDIAG=0 return
    if (ndiag.le.0) then
       out(:,:) = 0.0d0
       ier = 0
       !
       call nlprj_unify(bprj,eprj,nw,laprj,nsi, &
            dcprj,pby,nlpjs,nproc,myrank,pj_count,pj_displ,comm)
       deallocate(dcprj,pj_count,pj_displ)
       !
       call fft3_done(fs)
       call nlprj_done(nlpjs)
       !
       ! Be aware when you add new variable to deallocate.
       ! There are one more deallocate in this routine.
       !
       deallocate(hp,sp,akk,j2gs)
       return
    end if

    if (ifoeps) then
       eps = 0.0d0
       do jb=1,nw
          eps = eps + abs(ee(jb) - oldee(jb))
       end do
       eps = eps/nw
       if (itsb .gt. 0) then
          eps = 10.0d0**(-itsb/10.0d0)*eps
       else
          eps = 0.1d0*eps
       end if
       deallocate(oldee)
    end if
    !
    ! MAIN LOOP INCLUDING SUBSPACE DIAGONALIZATION
    !
    allocate(cl(ncomp,bkm:ekm,nw),hcl(ncomp,bkm:ekm,nw),scl(ncomp,bkm:ekm,nw), &
         dcprjcl(bprj:eprj,nw),stat=ista)
    call chkalloc('dvdsn1','tmp cg',ista)
    ibc=1
    nwc=nw
    ier=0
    do idiag=1,ndiag
       dcprjcl = (0.0d0,0.0d0)
       do jb=ibc,nw
          out(jb,1)=ee(jb)
       end do
       call gencl(ncomp,bkm,ekm,nwc,ee(ibc) &
            ,p(1,bkm,ibc),hp(1,bkm,ibc),sp(1,bkm,ibc) &
            ,akk,cl(1,bkm,ibc),out(ibc,2),comm)
       call sorthcl(bkms,ekms,ibc-1,nwc,p,sp,cl(1,bkm,ibc),comm)
       call mulths(ncomp,nkm,bkm,ekm,nwc,laprj,nsi,nkd,kd,bspots,va, &
            nlpjs,dd,nlxyz, &
            j2gs,ndens,vr,akk, &
            cl(1,bkm,ibc),hcl(1,bkm,ibc),scl(1,bkm,ibc), &
            bprj,eprj,dcprjcl(bprj,ibc),qbtp, &
            xctp,bb,sk,kg,dpf,mgsf,nb1,nb2,bzwk,wgtk, &
            fs,nproc,g_count,g_displ,comm)
       call rotwfn(bkms,ekms,nw,bprj,eprj,ee(ibc),p,hp,sp,dcprj, &
            ibc,nwc, &
            cl(1,bkm,ibc),hcl(1,bkm,ibc),scl(1,bkm,ibc), &
            dcprjcl(bprj,ibc), &
            .false.,(idiag.eq.1),(idiag.eq.ndiag),comm)

       !
       ! update wave function for exchange if iks > 0
       if (iks .gt. 0 .and. xcf_exch_factor(xctp) .ne. 0.0d0) then
          dpf(iks)%wf = p
       end if
       !
       ! CONVERGENCE CHECK OF EIGEN VALUE
       !
       do ib=ibc,nw
          out(ib,1)=out(ib,1)-ee(ib)
       end do
       ib = ibc
       do while((abs(out(ib,1)) .lt. eps) .and. (ib .lt. nw))
          ib = ib + 1
       end do
       !
       ! epsmax must be defined here since ibc can be smaller than this ib
       ! when nwc is aligned to a multiple of nblock.
       !
       epsmax = abs(out(ib,1))
       if ((eps.gt.0.0d0).and.(epsmax.lt.eps)) then
          if (ib .lt. nw) then
             write(6,*) 'invalid short-cut condition in dvdsn1()'
             stop
          end if
          ier=idiag
          exit
       end if
       !
       nwc = nw - ib + 1
       if (mod(nwc, nblock) .ne. 0) then
          nwc = min(nwc - mod(nwc, nblock) + nblock, nw)
       end if
       ibc = nw - nwc + 1
    end do
    !
    ! XXX: use out(1,3) to store iteration count for bands
    out(:,3) = 0.0
    out(1,3) = nwc
    !
    deallocate(cl,hcl,scl,dcprjcl)

    call nlprj_unify(bprj,eprj,nw,laprj,nsi, &
         dcprj,pby,nlpjs,nproc,myrank,pj_count,pj_displ,comm)
    deallocate(dcprj,pj_count,pj_displ)
    !
    call fft3_done(fs)
    call nlprj_done(nlpjs)
    !
    ! Be aware when you add new variable to deallocate.
    ! There are one more deallocate in this routine.
    !
    deallocate(hp,sp,akk,j2gs)
    !
    return
  end subroutine dvdsn1

  subroutine gencl(ncomp,bkm,ekm,nw,ee,p,hp,sp,akk,cl,out,comm)
    use mpi
    implicit none
    integer,intent(in):: ncomp
    integer,intent(in):: bkm,ekm,nw
    real(kind=8),intent(in):: ee(nw)
    complex(kind=8),dimension(ncomp,bkm:ekm,nw),intent(in):: p,hp,sp
    real(kind=8),intent(in):: akk(bkm:ekm)
    complex(kind=8),dimension(ncomp,bkm:ekm,nw),intent(out):: cl
    real(kind=8),intent(out):: out(nw)
    integer:: comm

    integer:: ib,ig,ierr
    real(kind=8):: x,pre,tmp(nw),fac(nw)

    do ib=1,nw
       tmp(ib) = 0.0d0
       if (ncomp.eq.1) then
          do ig=bkm,ekm
             tmp(ib) = tmp(ib) + akk(ig)*conjg(p(1,ig,ib))*p(1,ig,ib)
          end do
       else
          do ig=bkm,ekm
             tmp(ib) = tmp(ib) + akk(ig)*( &
                    conjg(p(1,ig,ib))*p(1,ig,ib) &
                  + conjg(p(2,ig,ib))*p(2,ig,ib) )
          end do
       end if
    end do

    call MPI_ALLREDUCE(tmp,fac,nw,MPI_DOUBLE_PRECISION, &
         MPI_SUM,comm,ierr)
    fac=1.d0/fac

    do ib=1,nw
       if (ncomp.eq.1) then
          cl(1,:,ib) = hp(1,:,ib) - ee(ib)*sp(1,:,ib)
       else
          do ig=bkm,ekm
             cl(1,ig,ib) = hp(1,ig,ib) - ee(ib)*sp(1,ig,ib)
             cl(2,ig,ib) = hp(2,ig,ib) - ee(ib)*sp(2,ig,ib)
          end do
       end if
       tmp(ib) = 0.0
       if (ncomp.eq.1) then
          do ig = bkm,ekm
             tmp(ib) = tmp(ib) + conjg(cl(1,ig,ib))*cl(1,ig,ib)
          end do
       else
          do ig = bkm,ekm
             tmp(ib) = tmp(ib) + conjg(cl(1,ig,ib))*cl(1,ig,ib) &
                  + conjg(cl(2,ig,ib))*cl(2,ig,ib)
          end do
       end if
       !
       ! PRECONDITIONING OF THE SEARCHING VECTOR HP
       !
       do ig=bkm,ekm
          x=akk(ig)*fac(ib)
          pre=(27.d0+18.d0*x+12.d0*x**2+8.d0*x**3)/ &
               (27.d0+18.d0*x+12.d0*x**2+8.d0*x**3+16.d0*x**4)
          !
          ! not use array expression to suppress stupid vectorization.
          !
          cl(:,ig,ib)=cl(:,ig,ib)*pre
       end do
    end do

    call MPI_ALLREDUCE(tmp,out,nw,MPI_DOUBLE_PRECISION, &
         MPI_SUM,comm,ierr)

    return
  end subroutine gencl

  subroutine mulths(ncomp,nkm,bkm,ekm,nw,laprj,nsi,nkd,kd, &
       bspots,va,nlpjs,dd,nlxyz, &
       j2gs,ndens,vr,akk,cl,hcl,scl,bprj,eprj,dcprjcl,qbtp, &
       xctp,bb,sk,kg,dpf,mgsf,nb1,nb2,bzwk,wgtk, &
       fs,nproc,g_count,g_displ,comm)
    use fft_3d
    use nonlocal_proj
    use mult_loc_pot
    use bsplpot
    use types
    use exchange
    use xcf_types
    use meshgrid
    implicit none
    integer,intent(in):: ncomp,ndens
    integer,intent(in):: nkm,bkm,ekm,nw,laprj,nsi
    integer,intent(in):: nkd,kd(nsi)
    real(kind=8),intent(in):: va
    type(nlprj_struct),intent(in):: nlpjs

    complex(kind=8),intent(in):: dd(laprj,laprj,nsi)
    integer,intent(in):: nlxyz
    integer,intent(in)::j2gs(nkm,2)
    real(kind=8),intent(in):: vr(nlxyz,ndens)
    real(kind=8),intent(in):: akk(bkm:ekm)
    complex(kind=8),dimension(ncomp,bkm:ekm,nw),intent(in):: cl
    complex(kind=8),dimension(ncomp,bkm:ekm,nw),intent(out):: hcl,scl
    integer,intent(in):: bprj,eprj
    complex(kind=8),intent(out):: dcprjcl(bprj:eprj,nw)
    real(kind=8),intent(in):: qbtp(3)
    type (bspot_for_atm),intent(in):: bspots(nkd)

    integer,intent(in):: xctp
    real(kind=8),intent(in):: bb(3,3),sk(3)
    integer,intent(in):: kg(3,nkm)
    type (meshgrid_struct):: mgsf
    type (wfn_for_k):: dpf(mgsf%nks)
    integer,intent(in):: nb1,nb2
    real(kind=8),intent(in):: bzwk(mgsf%nks),wgtk(nb1:nb2,mgsf%nks)

    type(fft3_struct):: fs
    integer,intent(in):: nproc,comm
    integer,intent(in):: g_count(0:nproc-1),g_displ(0:nproc-1)

    integer:: ig,ib
    integer:: bkms,ekms

    real(kind=8):: a,qsq,sgm

    real(kind=8):: awgtkmax

    a=qbtp(1)
    qsq=qbtp(2)*qbtp(2)
    sgm=qbtp(3)

    !
    ! local part
    !
    call hploc(ncomp,nlxyz,nkm,bkm,ekm,nw,cl,hcl,ndens,vr,j2gs,fs, &
         nproc,g_count,g_displ,comm)

    if (ncomp.eq.1 .and. xcf_exch_factor(xctp) .ne. 0.0d0) then
       awgtkmax = maxval(abs(wgtk))
       if (awgtkmax .gt. 0.0d0) then
          call potexch(va,bb,mgsf,nw,nb1,nb2,wgtk,bzwk,dpf,sk,nkm,bkm,ekm, &
               kg,cl,scl,nproc,g_count,g_displ,comm,0)
          !
          ! scl contains V_x[cl]
          !
          hcl = hcl + xcf_exch_factor(xctp)*scl
       end if
    end if

    !
    ! kinetic part and unity
    !
    do ib=1,nw
       do ig=bkm,ekm
          hcl(:,ig,ib) = hcl(:,ig,ib) + 0.5d0*cl(:,ig,ib)*( &
               akk(ig) + a*( 1.0d0 + erf(0.5d0*(akk(ig)-qsq)/sgm) ) )
          scl(:,ig,ib) = cl(:,ig,ib)
       end do
    end do
    !
    ! non-local part
    !
    bkms = (bkm-1)*ncomp+1
    ekms = ekm*ncomp
    call nlprj_apply(laprj,nkd,bspots,kd,nsi,bkms,ekms, &
         va,dd,nw,cl,hcl,scl,bprj,eprj,dcprjcl,nlpjs,comm)

    return
  end subroutine mulths

  subroutine ssdiag(bkm,ekm,nw,bprj,eprj,ee,p,hp,sp,dcprj,comm)
    use xmpi
    use eigsystm
    use arch_rotwfn
    use timeprof
    implicit none
    integer,intent(in):: bkm,ekm,nw,bprj,eprj
    real(kind=8),intent(inout):: ee(nw)
    complex(kind=8),dimension(bkm:ekm,nw):: p,hp,sp
    complex(kind=8):: dcprj(bprj:eprj,nw)
    integer,intent(in):: comm

    complex(kind=8),allocatable:: mat(:,:),smat(:,:),tmat(:,:),g(:,:)
    complex(kind=8),allocatable:: cpjtmp(:,:)
    complex(kind=8),allocatable:: wftmp(:,:)

    integer:: ib,jb,lnm,ierr

    call tpf_begin(tpf_rwf)

    lnm = getldm('Z',nw)

    allocate(mat(lnm,nw), &
         smat(lnm,nw), g(nw,nw), &
         wftmp(bkm:ekm,nw),cpjtmp(bprj:eprj,nw), &
         tmat(lnm,nw),stat=ierr)
    call chkalloc('rotwfn','tmp',ierr)
    !
    !     determine hamiltonian in sub-space
    !
    tmat = 0.0d0
    g = 0.0d0
    call ltrovlmat(bkm,ekm,nw,p,hp,g)
    tmat(1:nw, 1:nw) = g

    !call MPI_ALLREDUCE(tmat,mat,lnm*nw, &
    !     MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_allzsum(tmat,mat,lnm*nw,comm,ierr)

    tmat = 0.0d0
    g = 0.0d0
    call ltrovlmat(bkm,ekm,nw,sp,p,g)
    tmat(1:nw, 1:nw) = g

    !call MPI_ALLREDUCE(tmat,smat,lnm*nw, &
    !     MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_allzsum(tmat,smat,lnm*nw,comm,ierr)

    deallocate(tmat)

    call zhegv1vl(nw,mat,smat,lnm,ee,8,comm,ierr)
    if (ierr.ne.0) then
       write(6,*)'ssdiag:zhegv1vl info =',ierr
       call MPI_FINALIZE(ierr)
       stop
    end if

    deallocate(smat)

    g = mat(1:nw,1:nw)
    !
    ! rotate p
    !
    wftmp = 0.0d0
    call rotvecs(bkm,ekm,nw,g,p,wftmp)
!$omp parallel do private(jb)
    do ib=1,nw
       cpjtmp(:,ib) = (0.0d0, 0.0d0)
       do jb=1,nw
          cpjtmp(:,ib) = cpjtmp(:,ib) + dcprj(:,jb)*mat(jb,ib)
       end do
    end do
!$omp parallel do
    do ib=1,nw
       p(:,ib) = wftmp(:,ib)
       dcprj(:,ib) = cpjtmp(:,ib)
    end do

    !
    ! rotate hp
    !
    wftmp = 0.0d0
    call rotvecs(bkm,ekm,nw,g,hp,wftmp)
!$omp parallel do
    do ib=1,nw
       hp(:,ib) = wftmp(:,ib)
    end do

    !
    ! rotate sp
    !
    wftmp = 0.0d0
    call rotvecs(bkm,ekm,nw,g,sp,wftmp)
!$omp parallel do
    do ib=1,nw
       sp(:,ib) = wftmp(:,ib)
    end do

    deallocate(g,mat,wftmp,cpjtmp)

    call tpf_end(tpf_rwf)
    return
  end subroutine ssdiag

  !
  ! for inversion symmetry
  !
  !
  subroutine dvdsn1i(ndiag,itsb,nwx2,nwy2,nwz2,nlx,nly,nlz,nlxyz, &
       nkm,bkm,ekm,nw,sk,bb,kg,ee,vr, &
       out,lpnum,maxr,vpj,laprj,dd,asi, &
       ifoeps,eps,ier,p,va,nsi,nkd,kd,bspots,pby,qbtp,ifothnm, &
       nproc,myrank,g_count,g_displ,comm)
    !
    use mpi
    use bsplpot
    use fft_3d
    use converter
    use util
    use nonlocal_proj
    use rotate_wavefun
    use orthg_seavec
    implicit none
    integer,intent(in):: ndiag,itsb,nwx2,nwy2,nwz2,nlx,nly,nlz,nlxyz, &
         nkm,bkm,ekm,nw,lpnum,maxr,laprj,nsi,nkd
    real(kind=8),intent(in):: va
    logical:: ifoeps,ifothnm
    real(kind=8):: eps
    real(kind=8),intent(in):: sk(3),bb(3,3)
    real(kind=8),intent(in):: vr(nlxyz)
    integer,intent(in)::  kg(3,nkm),kd(nsi)
    real(kind=8),intent(in):: &
         asi(3,nsi),vpj(bkm:ekm,maxr,lpnum,nkd)
    complex(kind=8),intent(in):: dd(laprj,laprj,nsi)
    real(kind=8):: p(bkm:ekm,nw)
    real(kind=8),intent(in):: qbtp(3)
    type (bspot_for_atm):: bspots(nkd)
    !
    integer:: ier
    real(kind=8):: out(nw,3),ee(nw)
    complex(kind=8):: pby(laprj,nsi,nw)
    integer:: nproc,g_count(0:nproc-1),g_displ(0:nproc-1)
    integer:: myrank,comm
    !
    real(kind=8),dimension(:,:),allocatable:: hp,sp
    real(kind=8),dimension(:,:),allocatable:: cl,hcl,scl
    real(kind=8),allocatable:: akk(:)
    complex(kind=8),allocatable:: dcprj(:,:), dcprjcl(:,:)
    integer,allocatable:: pj_count(:),pj_displ(:)
    integer:: bprj,eprj,irnk
    integer,allocatable:: j2gs(:,:)

    type(nlprj_struct):: nlpjs

    real(kind=8),allocatable:: oldee(:),pee(:)
    !
    real(kind=8):: epsmax
    integer:: ig,ib,jb,idiag,nwxyz,ibc,nwc
    type(fft3_struct):: fs
    integer:: ntask,nblock
    integer:: ista
    !
    call getntask(ntask)
    nblock = max(ntask,nproc)
    !
    nwxyz = nwx2*nwy2*nwz2
    !
    allocate(hp(bkm:ekm,nw),sp(bkm:ekm,nw), &
         akk(bkm:ekm),j2gs(nkm,2),stat=ista)
    call chkalloc('dvdsn1i','tmp',ista)
    ! INITIALIZE
    call kg2akk(bb,sk,ekm-bkm+1,kg(1,bkm),akk)
    hp(:,:) = 0.0d0
    call fft3_init(nwx2,nwy2,nwz2,nlx,nly,nlz,fs)
    !
    if (fs%ifsft) then
       call kg2lj2gs(nwx2,nwy2,nwz2,nlx,nly,nlz,nkm,kg,j2gs)
    else
       call kg2lj2g(nwx2,nwy2,nwz2,nlx,nly,nlz,nkm,kg,j2gs(:,1))
       j2gs(:,2) = j2gs(:,1)
    end if
    !
    call nlprj_init(lpnum,maxr,nkd,bspots,nsi,kd,asi, &
         bkm,ekm,bb,sk,kg(1,bkm),vpj,nlpjs,.true.)

    allocate(pj_count(0:nproc-1),pj_displ(0:nproc-1),stat=ista)
    call chkalloc('dvdsn1','pj_count',ista)
    pj_count = nlpjs%nprj/nproc
    do irnk=0,mod(nlpjs%nprj,nproc)-1
       pj_count(irnk) = pj_count(irnk)+1
    end do

    pj_displ(0) = 0
    do irnk=0,nproc-2
       pj_displ(irnk+1) = pj_displ(irnk) + pj_count(irnk)
    end do
    bprj = pj_displ(myrank)+1
    eprj = pj_displ(myrank)+pj_count(myrank)

    allocate(dcprj(bprj:eprj,nw),stat=ista)
    call chkalloc('dvdsn1','dcprj',ista)

    if (ndiag.le.0) ifoeps = .false.
    !     
    ! XXX: we should check scf state by an another variable
    !
    if (ifoeps) then
       do jb=1,nw-1
          if (ee(jb+1) .lt. ee(jb)) then
             ifoeps = .false.
          end if
       end do
    end if
    if (ifoeps) then
       allocate(oldee(nw),stat=ista)
       call chkalloc('dvdsn1i','oldee',ista)
       oldee = ee
    end if
    !
    ! Zero clear variables outside for vectorization
    dcprj(:,:) = (0.0d0,0.0d0)
    !
    ! \Psi := p
    ! H \Psi -> hp
    ! S \Psi -> sp
    !
    call multhsi(nkm,bkm,ekm,nw,laprj,nsi,nkd,kd,bspots,va, &
         nlpjs,dd,nlxyz, &
         j2gs,vr,akk,p,hp,sp,bprj,eprj,dcprj,qbtp, &
         fs,nproc,g_count,g_displ,comm)

    if (.not. ifothnm) then
       !
       !  sub-space diagonalization
       !
       call ssdiagi(bkm,ekm,nw,bprj,eprj,ee,p,hp,sp,dcprj,comm)

       ifothnm = .true.
    else if (ndiag.gt.0) then
       !
       ! CALCULATE EIGENVALUE
       !     
       allocate(pee(nw),stat=ista)
       call chkalloc('dvdsn1i','pee',ista)
       do jb=1,nw
          pee(jb) = 0.0
          do ig=bkm,ekm
             pee(jb) = pee(jb) + p(ig,jb)*hp(ig,jb)
          end do
       end do
       call MPI_ALLREDUCE(pee,ee,nw,MPI_DOUBLE_PRECISION,MPI_SUM, &
            comm,ista)
       deallocate(pee)
    end if
    !
    ! IF NDIAG=0 return
    if (ndiag.le.0) then
       out(:,:) = 0.0d0
       ier = 0
       !
       call nlprj_unify(bprj,eprj,nw,laprj,nsi, &
            dcprj,pby,nlpjs,nproc,myrank,pj_count,pj_displ,comm)
       deallocate(dcprj,pj_count,pj_displ)
       !
       call fft3_done(fs)
       call nlprj_done(nlpjs)
       !
       ! Be aware when you add new variable to deallocate.
       ! There are one more deallocate in this routine.
       !
       deallocate(hp,sp,akk,j2gs)
       return
    end if

    if (ifoeps) then
       eps = 0.0d0
       do jb=1,nw
          eps = eps + abs(ee(jb) - oldee(jb))
       end do
       eps = eps/nw
       if (itsb .gt. 0) then
          eps = 10.0d0**(-itsb/10.0d0)*eps
       else
          eps = 0.1d0*eps
       end if
       deallocate(oldee)
    end if
    !
    ! MAIN LOOP INCLUDING SUBSPACE DIAGONALIZATION
    !
    allocate(cl(bkm:ekm,nw),hcl(bkm:ekm,nw),scl(bkm:ekm,nw), &
         dcprjcl(bprj:eprj,nw),stat=ista)
    call chkalloc('dvdsn1i','tmp cg',ista)
    ibc=1
    nwc=nw
    ier=0
    do idiag=1,ndiag
       dcprjcl = (0.0d0,0.0d0)
       do jb=ibc,nw
          out(jb,1)=ee(jb)
       end do
       call gencli(bkm,ekm,nwc,ee(ibc) &
            ,p(bkm,ibc),hp(bkm,ibc),sp(bkm,ibc) &
            ,akk,cl(bkm,ibc),out(ibc,2),comm)
       call sorthcli(bkm,ekm,ibc-1,nwc,p,sp,cl(bkm,ibc),comm)
       call multhsi(nkm,bkm,ekm,nwc,laprj,nsi,nkd,kd,bspots,va, &
            nlpjs,dd,nlxyz, &
            j2gs,vr,akk, &
            cl(bkm,ibc),hcl(bkm,ibc),scl(bkm,ibc), &
            bprj,eprj,dcprjcl(bprj,ibc),qbtp, &
            fs,nproc,g_count,g_displ,comm)

       call rotwfni(bkm,ekm,nw,bprj,eprj,ee(ibc),p,hp,sp,dcprj, &
            ibc,nwc, &
            cl(bkm,ibc),hcl(bkm,ibc),scl(bkm,ibc), &
            dcprjcl(bprj,ibc), &
            (idiag.eq.1),(idiag.eq.ndiag),comm)
       !
       ! CONVERGENCE CHECK OF EIGEN VALUE
       !
       do ib=ibc,nw
          out(ib,1)=out(ib,1)-ee(ib)
       end do
       ib = ibc
       do while((abs(out(ib,1)) .lt. eps) .and. (ib .lt. nw))
          ib = ib + 1
       end do
       !
       ! epsmax must be defined here since ibc can be smaller than this ib
       ! when nwc is aligned to a multiple of nblock.
       !
       epsmax = abs(out(ib,1))
       if ((eps.gt.0.0d0).and.(epsmax.lt.eps)) then
          if (ib .lt. nw) then
             write(6,*) 'invalid short-cut condition in dvdsn1i()'
             stop
          end if
          ier=idiag
          exit
       end if
       !
       nwc = nw - ib + 1
       if (mod(nwc, nblock) .ne. 0) then
          nwc = min(nwc - mod(nwc, nblock) + nblock, nw)
       end if
       ibc = nw - nwc + 1
    end do
    !
    ! XXX: use out(1,3) to store iteration count for bands
    out(:,3) = 0.0
    out(1,3) = nwc
    !
    deallocate(cl,hcl,scl,dcprjcl)

    call nlprj_unify(bprj,eprj,nw,laprj,nsi, &
         dcprj,pby,nlpjs,nproc,myrank,pj_count,pj_displ,comm)
    deallocate(dcprj,pj_count,pj_displ)
    !
    call fft3_done(fs)
    call nlprj_done(nlpjs)
    !
    ! Be aware when you add new variable to deallocate.
    ! There are one more deallocate in this routine.
    !
    deallocate(hp,sp,akk,j2gs)
    !
    return
  end subroutine dvdsn1i

  subroutine gencli(bkm,ekm,nw,ee,p,hp,sp,akk,cl,out,comm)
    use mpi
    implicit none
    integer,intent(in):: bkm,ekm,nw
    real(kind=8),intent(in):: ee(nw)
    real(kind=8),dimension(bkm:ekm,nw),intent(in):: p,hp,sp
    real(kind=8),intent(in):: akk(bkm:ekm)
    real(kind=8),dimension(bkm:ekm,nw),intent(out):: cl
    real(kind=8),intent(out):: out(nw)
    integer:: comm

    integer:: ib,ig,ierr
    real(kind=8):: x,pre,tmp(nw),fac(nw)

    do ib=1,nw
       tmp(ib) = 0.0d0
       do ig=bkm,ekm
          tmp(ib) = tmp(ib) + akk(ig)*p(ig,ib)*p(ig,ib)
       end do
    end do

    call MPI_ALLREDUCE(tmp,fac,nw,MPI_DOUBLE_PRECISION, &
         MPI_SUM,comm,ierr)
    fac=1.d0/fac

    do ib=1,nw
       cl(:,ib) = hp(:,ib) - ee(ib)*sp(:,ib)
       tmp(ib) = 0.0
       do ig = bkm,ekm
          tmp(ib) = tmp(ib) + cl(ig,ib)*cl(ig,ib)
       end do
       !
       ! PRECONDITIONING OF THE SEARCHING VECTOR HP
       !
       do ig=bkm,ekm
          x=akk(ig)*fac(ib)
          pre=(27.d0+18.d0*x+12.d0*x**2+8.d0*x**3)/ &
               (27.d0+18.d0*x+12.d0*x**2+8.d0*x**3+16.d0*x**4)
          !
          ! not use array expression to suppress stupid vectorization.
          !
          cl(ig,ib)=cl(ig,ib)*pre
       end do
    end do

    call MPI_ALLREDUCE(tmp,out,nw,MPI_DOUBLE_PRECISION, &
         MPI_SUM,comm,ierr)

    return
  end subroutine gencli

  subroutine multhsi(nkm,bkm,ekm,nw,laprj,nsi,nkd,kd, &
       bspots,va,nlpjs,dd,nlxyz, &
       j2gs,vr,akk,cl,hcl,scl,bprj,eprj,dcprjcl,qbtp,fs, &
       nproc,g_count,g_displ,comm)
    use fft_3d
    use nonlocal_proj
    use mult_loc_pot
    use bsplpot
    implicit none
    integer,intent(in):: nkm,bkm,ekm,nw,laprj,nsi
    integer,intent(in):: nkd,kd(nsi)
    real(kind=8),intent(in):: va
    type(nlprj_struct),intent(in):: nlpjs

    complex(kind=8),intent(in):: dd(laprj,laprj,nsi)
    integer,intent(in):: nlxyz
    integer,intent(in)::j2gs(nkm,2)
    real(kind=8),intent(in):: vr(nlxyz)
    real(kind=8),intent(in):: akk(bkm:ekm)
    real(kind=8),dimension(bkm:ekm,nw),intent(in):: cl
    real(kind=8),dimension(bkm:ekm,nw),intent(out):: hcl,scl
    integer,intent(in):: bprj,eprj
    complex(kind=8),intent(out):: dcprjcl(bprj:eprj,nw)
    real(kind=8),intent(in):: qbtp(3)
    type (bspot_for_atm):: bspots(nkd)

    type(fft3_struct):: fs
    integer,intent(in):: nproc,comm
    integer,intent(in):: g_count(0:nproc-1),g_displ(0:nproc-1)
    integer:: ig,ib

    real(kind=8):: a,qsq,sgm

    a=qbtp(1)
    qsq=qbtp(2)*qbtp(2)
    sgm=qbtp(3)

    !
    ! local part
    !
    call hploci(nlxyz,nkm,bkm,ekm,nw,cl,hcl,vr,j2gs,fs, &
         nproc,g_count,g_displ,comm)
    !
    ! kinetic part and unity
    !
    do ib=1,nw
       do ig=bkm,ekm
          hcl(ig,ib) = hcl(ig,ib) + 0.5d0*cl(ig,ib)*( &
               akk(ig) + a*( 1.0d0 + erf(0.5d0*(akk(ig)-qsq)/sgm) ) )
          scl(ig,ib) = cl(ig,ib)
       end do
    end do
    !
    ! non-local part
    !
    call nlprj_apyi(laprj,nkd,bspots,kd,nsi,bkm,ekm, &
         va,dd,nw,cl,hcl,scl,bprj,eprj,dcprjcl,nlpjs,comm)

    return
  end subroutine multhsi

  subroutine ssdiagi(bkm,ekm,nw,bprj,eprj,ee,p,hp,sp,dcprj,comm)
    use xmpi
    use eigsystm
    use arch_rotwfni
    use timeprof
    implicit none
    integer,intent(in):: bkm,ekm,nw,bprj,eprj
    real(kind=8),intent(inout):: ee(nw)
    real(kind=8),dimension(bkm:ekm,nw):: p,hp,sp
    complex(kind=8):: dcprj(bprj:eprj,nw)
    integer,intent(in):: comm

    real(kind=8),allocatable:: mat(:,:),smat(:,:),tmat(:,:),g(:,:)
    complex(kind=8),allocatable:: cpjtmp(:,:)
    real(kind=8),allocatable:: wftmp(:,:)

    integer:: ib,jb,lnm,ierr

    call tpf_begin(tpf_rwf)

    lnm = getldm('D',nw)

    allocate(mat(lnm,nw), &
         smat(lnm,nw), g(nw,nw), &
         wftmp(bkm:ekm,nw),cpjtmp(bprj:eprj,nw), &
         tmat(lnm,nw),stat=ierr)
    call chkalloc('rotwfng','tmp',ierr)
    !
    !     determine hamiltonian in sub-space
    !
    tmat = 0.0d0
    g = 0.0d0
    call ltrovlmat(bkm,ekm,nw,p,hp,g)
    tmat(1:nw, 1:nw) = g

    !call MPI_ALLREDUCE(tmat,mat,lnm*nw, &
    !     MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_alldsum(tmat,mat,lnm*nw,comm,ierr)

    tmat = 0.0d0
    g = 0.0d0
    call ltrovlmat(bkm,ekm,nw,sp,p,g)
    tmat(1:nw, 1:nw) = g

    !call MPI_ALLREDUCE(tmat,smat,lnm*nw, &
    !     MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_alldsum(tmat,smat,lnm*nw,comm,ierr)

    deallocate(tmat)

    call dsygv1vl(nw,mat,smat,lnm,ee,8,comm,ierr)
    if (ierr.ne.0) then
       write(6,*)'ssdiag:dsygv1vl info =',ierr
       call MPI_FINALIZE(ierr)
       stop
    end if

    deallocate(smat)

    g = mat(1:nw,1:nw)
    !
    ! rotate p
    !
    wftmp = 0.0d0
    call rotvecs(bkm,ekm,nw,g,p,wftmp)
!$omp parallel do private(jb)
    do ib=1,nw
       cpjtmp(:,ib) = (0.0d0, 0.0d0)
       do jb=1,nw
          cpjtmp(:,ib) = cpjtmp(:,ib) + dcprj(:,jb)*mat(jb,ib)
       end do
    end do
!$omp parallel do
    do ib=1,nw
       p(:,ib) = wftmp(:,ib)
       dcprj(:,ib) = cpjtmp(:,ib)
    end do

    !
    ! rotate hp
    !
    wftmp = 0.0d0
    call rotvecs(bkm,ekm,nw,g,hp,wftmp)
!$omp parallel do
    do ib=1,nw
       hp(:,ib) = wftmp(:,ib)
    end do

    !
    ! rotate sp
    !
    wftmp = 0.0d0
    call rotvecs(bkm,ekm,nw,g,sp,wftmp)
!$omp parallel do
    do ib=1,nw
       sp(:,ib) = wftmp(:,ib)
    end do

    deallocate(g,mat,wftmp,cpjtmp)

    call tpf_end(tpf_rwf)
    return
  end subroutine ssdiagi

  !
  ! for gamma point
  !
  !
  subroutine dvdsn1g(ndiag,itsb,nwx2,nwy2,nwz2,nlx,nly,nlz,nlxyz, &
       nkm,bkm,ekm, &
       nw,iks,sk,bb,kg,ee,vr, &
       out,lpnum,maxr,vpj,laprj,dd,asi, &
       ifoeps,eps,ier,p,va,nsi,nkd,kd,bspots,pby,qbtp,ifothnm, &
       xctp,dpgf,mgsf,nb1,nb2,bzwk,wgtk, &
       nproc,myrank,g_count,g_displ,comm)
    !
    use mpi
    use bsplpot
    use fft_3d
    use converter
    use util
    use nonlocal_proj
    use rotate_wavefun
    use orthg_seavec
    use types
    use xcf_types
    use meshgrid
    implicit none
    integer,intent(in):: ndiag,itsb,nwx2,nwy2,nwz2,nlx,nly,nlz,nlxyz, &
         nkm,bkm,ekm,nw,lpnum,maxr,laprj,nsi,nkd
    real(kind=8),intent(in):: va
    logical:: ifoeps,ifothnm
    real(kind=8):: eps
    integer,intent(in):: iks
    real(kind=8),intent(in):: sk(3),bb(3,3)
    real(kind=8),intent(in):: vr(nlxyz)
    integer,intent(in)::  kg(3,nkm),kd(nsi)
    real(kind=8),intent(in):: &
         asi(3,nsi),vpj(bkm:ekm,maxr,lpnum,nkd)
    !
    ! dvdsn1g() assumes that dd are real numbers.
    !
    real(kind=8),intent(in):: dd(laprj,laprj,nsi)
    complex(kind=8):: p(bkm:ekm,nw)
    real(kind=8),intent(in):: qbtp(3)
    type (bspot_for_atm):: bspots(nkd)
    !
    integer:: ier
    real(kind=8):: out(nw,3),ee(nw)
    complex(kind=8):: pby(laprj,nsi,nw)

    integer,intent(in):: xctp
    type (meshgrid_struct):: mgsf
    complex(kind=8):: dpgf(bkm:ekm,nw)
    integer,intent(in):: nb1,nb2
    real(kind=8),intent(in):: bzwk,wgtk(nb1:nb2)

    integer:: nproc,g_count(0:nproc-1),g_displ(0:nproc-1)
    integer:: myrank,comm
    !
    complex(kind=8),dimension(:,:),allocatable:: hp,sp
    complex(kind=8),dimension(:,:),allocatable:: cl,hcl,scl
    real(kind=8),allocatable:: akk(:)
    complex(kind=8),allocatable:: dcprj(:,:), dcprjcl(:,:)
    integer,allocatable:: pj_count(:),pj_displ(:)
    integer:: bprj,eprj,irnk
    integer,allocatable:: j2gs(:,:)
    type(nlprj_struct):: nlpjs

    real(kind=8),allocatable:: oldee(:),pee(:)
    !
    real(kind=8):: epsmax
    integer:: ig,ib,jb,idiag,nwxyz,ibc,nwc
    type(fft3_struct):: fs
    integer:: ntask,nblock
    integer:: ista
    !
    call getntask(ntask)
    nblock = max(ntask,nproc)
    !
    nwxyz = nwx2*nwy2*nwz2
    !
    allocate(hp(bkm:ekm,nw),sp(bkm:ekm,nw), &
         akk(bkm:ekm),j2gs(nkm,4),stat=ista)
    call chkalloc('dvdsn1g','tmp',ista)
    ! INITIALIZE
    call kg2akk(bb,sk,ekm-bkm+1,kg(1,bkm),akk)
    hp(:,:) = 0.0d0
    call fft3_init(nwx2,nwy2,nwz2,nlx,nly,nlz,fs)
    !
    if (fs%ifsft) then
       call kg2lj2gsg(nwx2,nwy2,nwz2,nlx,nly,nlz,nkm,kg,j2gs)
    else
       call kg2lj2gg(nwx2,nwy2,nwz2,nlx,nly,nlz,nkm,kg,j2gs(1,1))
       j2gs(:,3) = j2gs(:,1)
       j2gs(:,4) = j2gs(:,2)
    end if
    !
    call nlprj_init(lpnum,maxr,nkd,bspots,nsi,kd,asi, &
         bkm,ekm,bb,sk,kg(1,bkm),vpj,nlpjs)

    allocate(pj_count(0:nproc-1),pj_displ(0:nproc-1),stat=ista)
    call chkalloc('dvdsn1g','pj_count',ista)
    pj_count = nlpjs%nprj/nproc
    do irnk=0,mod(nlpjs%nprj,nproc)-1
       pj_count(irnk) = pj_count(irnk)+1
    end do

    pj_displ(0) = 0
    do irnk=0,nproc-2
       pj_displ(irnk+1) = pj_displ(irnk) + pj_count(irnk)
    end do
    bprj = pj_displ(myrank)+1
    eprj = pj_displ(myrank)+pj_count(myrank)

    allocate(dcprj(bprj:eprj,nw),stat=ista)
    call chkalloc('dvdsn1g','dcprj',ista)

#ifdef TIMING
    write(6,'(1x,a,f10.2)') 'TIME DVDSN INIT= ', xetime(1)
#endif
    if (ndiag.le.0) ifoeps = .false.
    !     
    ! XXX: we should check scf state by an another variable
    !
    if (ifoeps) then
       do jb=1,nw-1
          if (ee(jb+1) .lt. ee(jb)) then
             ifoeps = .false.
          end if
       end do
    end if
    if (ifoeps) then
       allocate(oldee(nw),stat=ista)
       call chkalloc('dvdsn1g','oldee',ista)
       oldee = ee
    end if
    !
    ! Zero clear variables outside for vectorization
    dcprj(:,:) = (0.0d0,0.0d0)
    !
    ! \Psi := p
    ! H \Psi -> hp
    ! S \Psi -> sp
    !
    call multhsg(nkm,bkm,ekm,nw,laprj,nsi,nkd,kd,bspots,va, &
         nlpjs,dd,nlxyz, &
         j2gs,vr,akk,p,hp,sp,bprj,eprj,dcprj,qbtp, &
         xctp,bb,kg,dpgf,mgsf,nb1,nb2,bzwk,wgtk, &
         fs,nproc,g_count,g_displ,comm)
#ifdef TIMING
    write(6,'(1x,a,f10.2)') 'TIME DVDSN MAKE HP,SP= ', xetime(1)
#endif

    if (.not. ifothnm) then
       !
       !  sub-space diagonalization
       !
       call ssdiagg(bkm,ekm,nw,bprj,eprj,ee,p,hp,sp,dcprj,comm)

#ifdef TIMING
       write(6,'(1x,a,f10.2)') 'TIME DVDSN SSDIAG= ', xetime(1)
#endif
       ifothnm = .true.
    else if (ndiag.gt.0) then
       !     
       ! CALCULATE EIGENVALUE
       !     
       allocate(pee(nw),stat=ista)
       call chkalloc('dvdsn1g','pee',ista)
       do jb=1,nw
          pee(jb) = 0.0
          do ig=max(bkm,2),ekm
             pee(jb) = pee(jb) + conjg(p(ig,jb))*hp(ig,jb)
          end do
          pee(jb) = 2.0d0*pee(jb)
          if (bkm.eq.1) then
             pee(jb) = pee(jb) + dble(p(1,jb))*dble(hp(1,jb))
          end if
       end do
       call MPI_ALLREDUCE(pee,ee,nw,MPI_DOUBLE_PRECISION,MPI_SUM, &
            comm,ista)
       deallocate(pee)
    endif
    !
    ! IF NDIAG=0 return
    if (ndiag.le.0) then
       out(:,:) = 0.0d0
       ier = 0
       !
       call nlprj_unify(bprj,eprj,nw,laprj,nsi, &
            dcprj,pby,nlpjs,nproc,myrank,pj_count,pj_displ,comm)
       deallocate(dcprj,pj_count,pj_displ)
       !
       call fft3_done(fs)
       call nlprj_done(nlpjs)

       !
       ! Be aware when you add new variable to deallocate.
       ! There are one more deallocate in this routine.
       !
       deallocate(hp,sp,akk,j2gs)
       return
    end if

    if (ifoeps) then
       eps = 0.0d0
       do jb=1,nw
          eps = eps + abs(ee(jb) - oldee(jb))
       end do
       eps = eps/nw
       if (itsb .gt. 0) then
          eps = 10.0d0**(-itsb/10.0d0)*eps
       else
          eps = 0.1d0*eps
       end if
       deallocate(oldee)
    end if
    !
    !  MAIN LOOP INCLUDING SUBSPACE DIAGONALIZATION
    !
    allocate(cl(bkm:ekm,nw),hcl(bkm:ekm,nw),scl(bkm:ekm,nw), &
         dcprjcl(bprj:eprj,nw),stat=ista)
    call chkalloc('dvdsn1g','tmp cg',ista)
    ibc=1
    nwc=nw
    ier=0
    do idiag=1,ndiag
       dcprjcl = (0.0d0,0.0d0)
       do jb=ibc,nw
          out(jb,1)=ee(jb)
       end do
       call genclg(bkm,ekm,nwc,ee(ibc) &
            ,p(bkm,ibc),hp(bkm,ibc),sp(bkm,ibc) &
            ,akk,cl(bkm,ibc),out(ibc,2),comm)
       call sorthclg(bkm,ekm,ibc-1,nwc,p,sp,cl(bkm,ibc),comm)
       call multhsg(nkm,bkm,ekm,nwc,laprj,nsi,nkd,kd,bspots,va, &
            nlpjs,dd,nlxyz, &
            j2gs,vr,akk, &
            cl(bkm,ibc),hcl(bkm,ibc),scl(bkm,ibc), &
            bprj,eprj,dcprjcl(bprj,ibc),qbtp, &
            xctp,bb,kg,dpgf,mgsf,nb1,nb2,bzwk,wgtk, &
            fs,nproc,g_count,g_displ,comm)
       call rotwfng(bkm,ekm,nw,bprj,eprj,ee(ibc),p,hp,sp,dcprj, &
            ibc,nwc, &
            cl(bkm,ibc),hcl(bkm,ibc),scl(bkm,ibc), &
            dcprjcl(bprj,ibc), &
            .false.,(idiag.eq.1),(idiag.eq.ndiag),comm)

       !
       ! update wave function for exchange if iks > 0
       if (iks .gt. 0 .and. xcf_exch_factor(xctp) .ne. 0.0d0) then
          dpgf = p
       end if
       !
       ! CONVERGENCE CHECK OF EIGEN VALUE
       !
       do ib=ibc,nw
          out(ib,1)=out(ib,1)-ee(ib)
       end do
       ib = ibc
       do while((abs(out(ib,1)) .lt. eps) .and. (ib .lt. nw))
          ib = ib + 1
       end do
       !
       ! epsmax must be defined here since ibc can be smaller than this ib
       ! when nwc is aligned to a multiple of nblock.
       !
       epsmax = abs(out(ib,1))
       if ((eps.gt.0.0d0).and.(epsmax.lt.eps)) then
          if (ib .lt. nw) then
             write(6,*) 'invalid short-cut condition in dvdsn1g()'
             stop
          end if
          ier=idiag
          exit
       end if
       !
       nwc = nw - ib + 1
       if (mod(nwc, nblock) .ne. 0) then
          nwc = min(nwc - mod(nwc, nblock) + nblock, nw)
       end if
       ibc = nw - nwc + 1
    end do
    !
    ! XXX: use out(1,3) to store iteration count for bands
    out(:,3) = 0.0
    out(1,3) = nwc
    !
    deallocate(cl,hcl,scl,dcprjcl)

    call nlprj_unify(bprj,eprj,nw,laprj,nsi, &
         dcprj,pby,nlpjs,nproc,myrank,pj_count,pj_displ,comm)
    deallocate(dcprj,pj_count,pj_displ)
    !
    call fft3_done(fs)
    call nlprj_done(nlpjs)
    !
    ! Be aware when you add new variable to deallocate.
    ! There are one more deallocate in this routine.
    !
    deallocate(hp,sp,akk,j2gs)
    !
    return
  end subroutine dvdsn1g

  subroutine genclg(bkm,ekm,nw,ee,p,hp,sp,akk,cl,out,comm)
    use mpi
    implicit none
    integer,intent(in):: bkm,ekm,nw
    real(kind=8),intent(in):: ee(nw)
    complex(kind=8),dimension(bkm:ekm,nw),intent(in):: p,hp,sp
    real(kind=8),intent(in):: akk(bkm:ekm)
    complex(kind=8),dimension(bkm:ekm,nw),intent(out):: cl
    real(kind=8),intent(out):: out(nw)
    integer:: comm

    integer:: ib,ig,ierr
    real(kind=8):: x,pre,tmp(nw),fac(nw)

    do ib=1,nw
       tmp(ib) = 0.0d0
       do ig=max(bkm,2),ekm
          tmp(ib) = tmp(ib) + akk(ig)*conjg(p(ig,ib))*p(ig,ib)
       end do
       tmp(ib) = 2.0d0*tmp(ib)
       if (bkm.eq.1) then
          tmp(ib) = tmp(ib) + akk(1)*dble(p(1,ib))*dble(p(1,ib))
       end if
    end do

    call MPI_ALLREDUCE(tmp,fac,nw,MPI_DOUBLE_PRECISION, &
         MPI_SUM,comm,ierr)
    fac=1.d0/fac

    do ib=1,nw
       cl(:,ib) = hp(:,ib) - ee(ib)*sp(:,ib)
       tmp(ib) = 0.0
       do ig = max(bkm,2),ekm
          tmp(ib) = tmp(ib) &
               + conjg(cl(ig,ib))*cl(ig,ib)
       end do
       tmp(ib) = 2.0d0*tmp(ib)
       if (bkm.eq.1) then
          tmp(ib) = tmp(ib) + dble(cl(1,ib))*dble(cl(1,ib))
       end if
       !
       ! PRECONDITIONING OF THE SEARCHING VECTOR HP
       !
       do ig=bkm,ekm
          x=akk(ig)*fac(ib)
          pre=(27.d0+18.d0*x+12.d0*x**2+8.d0*x**3)/ &
               (27.d0+18.d0*x+12.d0*x**2+8.d0*x**3+16.d0*x**4)
          !
          ! not use array expression to suppress stupid vectorization.
          !
          cl(ig,ib)=cl(ig,ib)*pre
       end do
    end do

    call MPI_ALLREDUCE(tmp,out,nw,MPI_DOUBLE_PRECISION, &
         MPI_SUM,comm,ierr)

    return
  end subroutine genclg

  subroutine multhsg(nkm,bkm,ekm,nw,laprj,nsi,nkd,kd, &
       bspots,va,nlpjs,dd,nlxyz, &
       j2gs,vr,akk,cl,hcl,scl,bprj,eprj,dcprjcl,qbtp, &
       xctp,bb,kg,dpgf,mgsf,nb1,nb2,bzwk,wgtk, &
       fs,nproc,g_count,g_displ,comm)
    use fft_3d
    use nonlocal_proj
    use mult_loc_pot
    use bsplpot
    use exchange
    use xcf_types
    use meshgrid
    implicit none
    integer,intent(in):: nkm,bkm,ekm,nw,laprj,nsi
    integer,intent(in):: nkd,kd(nsi)
    real(kind=8),intent(in):: va
    type(nlprj_struct),intent(in):: nlpjs

    real(kind=8),intent(in):: dd(laprj,laprj,nsi)
    integer,intent(in):: nlxyz
    integer,intent(in)::j2gs(nkm,2)
    real(kind=8),intent(in):: vr(nlxyz)
    real(kind=8),intent(in):: akk(bkm:ekm)
    complex(kind=8),dimension(bkm:ekm,nw),intent(in):: cl
    complex(kind=8),dimension(bkm:ekm,nw),intent(out):: hcl,scl
    integer,intent(in):: bprj,eprj
    complex(kind=8),intent(out):: dcprjcl(bprj:eprj,nw)
    real(kind=8),intent(in):: qbtp(3)
    type (bspot_for_atm),intent(in):: bspots(nkd)

    integer,intent(in):: xctp
    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: kg(3,nkm)
    type (meshgrid_struct):: mgsf
    complex(kind=8),intent(in):: dpgf(bkm:ekm,nw)
    integer,intent(in):: nb1,nb2
    real(kind=8),intent(in):: bzwk,wgtk(nb1:nb2)

    type(fft3_struct):: fs
    integer,intent(in):: nproc,comm
    integer,intent(in):: g_count(0:nproc-1),g_displ(0:nproc-1)

    integer:: ig,ib
    real(kind=8):: a,qsq,sgm

    real(kind=8):: awgtkmax

    a=qbtp(1)
    qsq=qbtp(2)*qbtp(2)
    sgm=qbtp(3)

    !
    ! local part
    !
    call hplocg(nlxyz,nkm,bkm,ekm,nw,cl,hcl,vr,j2gs,fs, &
         nproc,g_count,g_displ,comm)

    if (xcf_exch_factor(xctp) .ne. 0.0d0) then
       awgtkmax = maxval(abs(wgtk))
       if (awgtkmax .gt. 0.0d0) then
          call potexchg(va,bb,mgsf,nw,nb1,nb2,wgtk,bzwk,dpgf,nkm,bkm,ekm, &
               kg,cl,scl,nproc,g_count,g_displ,comm,0)
          !
          ! scl contains V_x[cl]
          !
          hcl = hcl + xcf_exch_factor(xctp)*scl
       end if
    end if
    !
    ! kinetic part and unity
    !
    do ib=1,nw
       do ig=bkm,ekm
          hcl(ig,ib) = hcl(ig,ib) + 0.5d0*cl(ig,ib)*( &
               akk(ig) + a*( 1.0d0 + erf(0.5d0*(akk(ig)-qsq)/sgm) ) )
          scl(ig,ib) = cl(ig,ib)
       end do
    end do
    !
    ! non-local part
    !
    call nlprj_apyg(laprj,nkd,bspots,kd,nsi,bkm,ekm, &
         va,dd,nw,cl,hcl,scl,bprj,eprj,dcprjcl,nlpjs,comm)
    return
  end subroutine multhsg

  subroutine ssdiagg(bkm,ekm,nw,bprj,eprj,ee,p,hp,sp,dcprj,comm)
    use xmpi
    use eigsystm
    use arch_rotwfng
    use timeprof
    implicit none
    integer,intent(in):: bkm,ekm,nw,bprj,eprj
    real(kind=8),intent(inout):: ee(nw)
    complex(kind=8),dimension(bkm:ekm,nw):: p,hp,sp
    complex(kind=8):: dcprj(bprj:eprj,nw)
    integer,intent(in):: comm

    real(kind=8),allocatable:: mat(:,:),smat(:,:),tmat(:,:),g(:,:)
    complex(kind=8),allocatable:: cpjtmp(:,:)
    complex(kind=8),allocatable:: wftmp(:,:)

    integer:: ib,jb,lnm,ierr

    call tpf_begin(tpf_rwf)

    lnm = getldm('D',nw)

    allocate(mat(lnm,nw), &
         smat(lnm,nw), g(nw,nw), &
         wftmp(bkm:ekm,nw),cpjtmp(bprj:eprj,nw), &
         tmat(lnm,nw),stat=ierr)
    call chkalloc('rotwfng','tmp',ierr)
    !
    !     determine hamiltonian in sub-space
    !
    tmat = 0.0d0
    g = 0.0d0
    call ltrovlmat(bkm,ekm,nw,p,hp,g)
    tmat(1:nw, 1:nw) = g

    !call MPI_ALLREDUCE(tmat,mat,lnm*nw, &
    !     MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_alldsum(tmat,mat,lnm*nw,comm,ierr)

    tmat = 0.0d0
    g = 0.0d0
    call ltrovlmat(bkm,ekm,nw,sp,p,g)
    tmat(1:nw, 1:nw) = g

    !call MPI_ALLREDUCE(tmat,smat,lnm*nw, &
    !     MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_alldsum(tmat,smat,lnm*nw,comm,ierr)

    deallocate(tmat)

    call dsygv1vl(nw,mat,smat,lnm,ee,8,comm,ierr)
    if (ierr.ne.0) then
       write(6,*)'ssdiag:dsygv1vl info =',ierr
       call MPI_FINALIZE(ierr)
       stop
    end if

    deallocate(smat)

    !
    ! rotate p
    !
    wftmp = 0.0d0
    g = mat(1:nw,1:nw)
    call rotvecs(bkm,ekm,nw,g,p,wftmp)

!$omp parallel do private(jb)
    do ib=1,nw
       cpjtmp(:,ib) = (0.0d0, 0.0d0)
       do jb=1,nw
          cpjtmp(:,ib) = cpjtmp(:,ib) + dcprj(:,jb)*mat(jb,ib)
       end do
    end do

!$omp parallel do
    do ib=1,nw
       p(:,ib) = wftmp(:,ib)
       dcprj(:,ib) = cpjtmp(:,ib)
    end do

    !
    ! rotate hp
    !
    wftmp = 0.0d0
    g = mat(1:nw,1:nw)
    call rotvecs(bkm,ekm,nw,g,hp,wftmp)

!$omp parallel do
    do ib=1,nw
       hp(:,ib) = wftmp(:,ib)
    end do

    !
    ! rotate sp
    !
    wftmp = 0.0d0
    g = mat(1:nw,1:nw)
    call rotvecs(bkm,ekm,nw,g,sp,wftmp)

!$omp parallel do
    do ib=1,nw
       sp(:,ib) = wftmp(:,ib)
    end do

    deallocate(g,mat,wftmp,cpjtmp)

    call tpf_end(tpf_rwf)
    return
  end subroutine ssdiagg

end module diag_dvdsn
