! Copyright (c) 2012, Minoru Otani <minoru.otani@aist.go.jp>
! 
! Permission is hereby granted, free of charge, to any person 
! obtaining a copy of this software and associated documentation 
! files (the "Software"), to deal in the Software without restriction, 
! including without limitation the rights to use, copy, modify, merge, 
! publish, distribute, sublicense, and/or sell copies of the Software, 
! and to permit persons to whom the Software is furnished to do so, 
! subject to the following conditions:
 
! The above copyright notice and this permission notice shall be 
! included in all copies or substantial portions of the Software.
 
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
! OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
! NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
! HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
! WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
! OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
! DEALINGS IN THE SOFTWARE.
#include "fftgrid.h"
module esm_manage
use global_esm
use m_2d
use esm_util
use m_deb
use meshgrid
use bsplpot
use iounit
use mpi

type(esm_2d)  ::var2d
type(cg2esm)  ::cginp
!
integer,private ::ideb(50)
contains
  subroutine esm_probe(mssg)
    implicit none
    integer,intent(in):: mssg
    logical ::exist

    inquire(FILE='esmprm.dat',EXIST=exist)
    if(exist)then
       if (mssg.gt.-1) then
          write(6,*)"ESM PARAMETER FILE was found"
          write(6,*)"ESM is enabled"
       end if
       esm%enable = .true.
    else
       if (mssg.gt.-1) then
          write(6,*)"ESM PARAMETER FILE was not found"
          write(6,*)"ESM is disabled"
       end if
       esm%enable = .false.
    endif
    return
  end subroutine esm_probe

!!!!__ANDO___ 
  subroutine esm_initialize_xTAPP(nsi,nkd,mgs &
       ,ng,aa,bb,kd,qf,va,bspots,cginp,esm,mssg)
!xTAPP
  use bsplpot
  use meshgrid
!
  implicit none
  integer,intent(in)        ::nsi,nkd,kd(nsi),ng
  real(kind=8),intent(in)   ::aa(3,3),bb(3,3),qf,va
  type(bspot_for_atm),intent(in) ::bspots(nkd)
  type(meshgrid_struct),intent(in) ::mgs
  type(cg2esm),intent(out)  ::cginp
  type(esm_prm),intent(out) ::esm
  integer,intent(in) ::mssg
  !local
  integer ::it,err
  logical ::exist
  real(kind=8) :: tmp
  !
  if (.not. esm%enable) then
     return
  end if

  inquire(FILE='esmprm.dat',EXIST=exist)
  if(exist)then
     open(27,file='esmprm.dat',STATUS='old')
  else
     if (mssg.gt.-1) write(6,*)"ESM PARAMETER FILE was not found"
     stop
  endif

!  read(27,*)esm%ngau
  esm%ngau=2 ! for TAPP
  read(27,*)esm%normal
  if (esm%normal .ne. 'X' .and. esm%normal .ne. 'Z') then
     if (mssg.gt.-1) write(6,*) 'unknown normal axis = ', esm%normal
     stop
  end if
  if (esm%normal .eq. 'X') then
     tmp = aa(1,1)*aa(1,2)+aa(2,1)*aa(2,2)+aa(3,1)*aa(3,2)
     if (abs(tmp).gt.1.0d-8) then
        if (mssg.gt.-1) write(6,*) "cell form is invalid for ESM: axis 1, 2"
        stop
     end if
     tmp = aa(1,1)*aa(1,3)+aa(2,1)*aa(2,3)+aa(3,1)*aa(3,3)
     if (abs(tmp).gt.1.0d-8) then
        if (mssg.gt.-1) write(6,*) "cell form is invalid for ESM: axis 1, 3"
        stop
     end if
  else if (esm%normal .eq. 'Z') then
     tmp = aa(1,3)*aa(1,1)+aa(2,3)*aa(2,1)+aa(3,3)*aa(3,1)
     if (abs(tmp).gt.1.0d-8) then
        if (mssg.gt.-1) write(6,*) "cell form is invalid for ESM: axis 3, 1"
        stop
     end if
     tmp = aa(1,3)*aa(1,2)+aa(2,3)*aa(2,2)+aa(3,3)*aa(3,2)
     if (abs(tmp).gt.1.0d-8) then
        if (mssg.gt.-1) write(6,*) "cell form is invalid for ESM: axis 3, 2"
        stop
     end if
  end if
  read(27,*)esm%bctype

!  read(27,*)esm%nfit
!  read(27,*)esm%idebug
!  read(27,*)esm%gp_s,esm%gp_e
  esm%nfit = 4
  esm%idebug = 0
  esm%gp_s = 1
  esm%gp_e = 1

  read(27,*)esm%z1
  read(27,*)esm%iffix
  if(esm%iffix.eq.1)then
    read(27,*)esm%add_elec
  elseif(esm%iffix.eq.2)then
    if (mssg.gt.-1) write(6,*)' ef=const. mode does not support'
    stop ' stop, esm_initialize'
    read(27,*)esm%fix_ef
  else
    if (mssg.gt.-1) write(6,'(A,I2,A)')' iffix=',esm%iffix,' dose not support'
    stop ' stop, esm_initialize'
  endif
  if (esm%bctype .eq. 4) then
     read(27,*)esm%v0
     !covert from electric field to vias voltage
     esm%v0=esm%v0*2.d0*esm%z1
  else
     esm%v0=0.d0
  end if
  close(27)
  !
  allocate(cginp%kd(nsi),cginp%zv(nsi) &
!added by ANDO
         ,cginp%kf(3,mgs%nkfxa) &
!
         ,cginp%agauss(nsi,esm%ngau) &
          ,cginp%bgauss(nsi,esm%ngau),stat=err)
  call chkalloc('esm_init','kd,zv,abgauss',err)
  !
  cginp%nsi=nsi
  cginp%nkd=nkd
!reading from mgs(mesh_grid_struct)
  cginp%nrx=mgs%nrx;cginp%nry=mgs%nry;cginp%nrz=mgs%nrz
  cginp%nrx2=2*mgs%nrx;cginp%nry2=2*mgs%nry;cginp%nrz2=2*mgs%nrz
  cginp%nvx=mgs%nvx;cginp%nvy=mgs%nvy;cginp%nvz=mgs%nvz
!
!Here added for local potential data of 1d-arrey
  cginp%nkfxa=mgs%nkfxa
  cginp%kf=mgs%kf
!
  cginp%ng=ng
  cginp%qf=qf
  cginp%va=va
  cginp%aa(:,:)=aa(:,:)
  cginp%bb(:,:)=bb(:,:)
  cginp%kd(:)=kd(:)
  do it=1,nsi
    cginp%zv(it) = bspots(kd(it))%zo
    cginp%agauss(it,1)= bspots(kd(it))%a2
    cginp%agauss(it,2)= bspots(kd(it))%a4
    cginp%bgauss(it,1)= bspots(kd(it))%a1
    cginp%bgauss(it,2)= bspots(kd(it))%a3
    !if esm%ngau is larger than 2, we define more a(b)gauss
    !below this line.
  enddo
  !
  end subroutine esm_initialize_xTAPP
!!!__E_ANDO

!__ANDO
subroutine  esm_init_xTAPP(nsi,nkd,mgs,ng,aa,bb,kd,qf,va,bspots,mssg)
  implicit none
  !
  integer,intent(in)                ::nsi,nkd,kd(nsi),ng,mssg
  real(kind=8),intent(in)           ::aa(3,3),bb(3,3),qf,va
  type(bspot_for_atm),intent(in)    ::bspots(nkd)
  type(meshgrid_struct),intent(in) ::mgs
  !
  integer ::nkf, naf

  if (.not. esm%enable) then
     return
  end if
  !
  call esm_initialize_xTAPP(nsi,nkd,mgs,ng,aa,bb,kd,qf,va,bspots,cginp, &
       esm,mssg)
  call esm_2d_initialize(cginp,var2d)
  call esm_2d_make_nkf_2d_akk_2d(cginp,var2d)
  call esm_deb_initialize(esm%idebug)
  call esm_get_debug_info(ideb)
  if (esm%bctype .eq. 1) then
     esm%boundary_condition=' boundary condition (i)'
  else if (esm%bctype .eq. 2) then
     esm%boundary_condition=' boundary condition (ii)'
  else if (esm%bctype .eq. 3) then
     esm%boundary_condition=' boundary condition (iii)'
  else if (esm%bctype .eq. 4) then
     esm%boundary_condition=' boundary condition (iv)'
  else
     if (mssg.gt.-1) then
        write(6,*) 'unknown boundary condition type = ', esm%bctype
     end if
     stop
  end if

  if(mssg .gt. -1) then
     call esm_write_esm_info(cginp,var2d,esm)
     write(*,*) "esm_init is finished."
  end if
  return
end subroutine esm_init_xTAPP

!!!!!!!!!!!!!!!!!!!!!!!!__E_ANDO!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!___ANDO___
subroutine esm_vblpt2_xTAPP(asi,vl,comm,mssg)
  implicit none
  real(kind=8),intent(in)       ::asi(3,cginp%nsi)
  integer,intent(in)            ::mssg, comm

  complex(kind=8),intent(inout) ::vl(cginp%nkfxa)
  !local
  integer                     ::err,k,isy,k1,k2,k3,ks1,ks2,ks3
  complex(kind=8),allocatable ::v(:,:,:),vg(:)

  if (.not. esm%enable) then
     return
  end if

  allocate(v(cginp%nrx2,cginp%nry2,cginp%nrz2),vg(var2d%n0),stat=err)
  call chkalloc('esm_vblpt2_xTAPP','v,vg',err)
  v(1:cginp%nrx2,1:cginp%nry2,1:cginp%nrz2)=(0.d0,0.d0)
if (esm%normal .eq. 'X') then
  var2d%asi_z   (1:cginp%nsi)=asi(1,1:cginp%nsi)
  var2d%asi_2d(1,1:cginp%nsi)=asi(2,1:cginp%nsi)
  var2d%asi_2d(2,1:cginp%nsi)=asi(3,1:cginp%nsi)
else if (esm%normal .eq. 'Z') then
  var2d%asi_2d(1,1:cginp%nsi)=asi(1,1:cginp%nsi)
  var2d%asi_2d(2,1:cginp%nsi)=asi(2,1:cginp%nsi)
  var2d%asi_z   (1:cginp%nsi)=asi(3,1:cginp%nsi)
endif
  call m_2d_make_long_range_part(cginp,var2d,v,vg,comm,mssg)
  !
  do k=1, cginp%nkfxa
     k1=cginp%kf(1,k)
     k2=cginp%kf(2,k)
     k3=cginp%kf(3,k)
     ks1=mod(cginp%nrx2+k1,cginp%nrx2)+1
     ks2=mod(cginp%nry2+k2,cginp%nry2)+1
     ks3=mod(cginp%nrz2+k3,cginp%nrz2)+1
     vl(k)=vl(k)+v(ks1,ks2,ks3)
  end do
  !
  if(ideb(4)==1)call print_2d_average_riffle('iloc',cginp,var2d,esm,v)
!  if(ideb(5)==1)write(100)v
! ANDO
  if(ideb(5)==1)then;rewind 100;write(100)v;endif
! e_ANDO
  deallocate(v,vg)
  return
end subroutine esm_vblpt2_xTAPP

subroutine esm_ewald(asi,for,dfor,ewald,mssg)
!___ANDO
  use util
!___E_ANDO
  implicit none
  real(kind=8),intent(in)  ::asi(3,cginp%nsi)
  real(kind=8),intent(out) ::for(3,cginp%nsi),dfor(3,cginp%nsi),ewald
  !local
  integer                  ::err,it,mssg
  real(kind=8),allocatable ::f_g_z(:),f_g_2d(:,:),f_r_z(:),f_r_2d(:,:)
  !
  allocate(f_g_z(cginp%nsi),f_g_2d(2,cginp%nsi) &
          ,f_r_z(cginp%nsi),f_r_2d(2,cginp%nsi),stat=err)
  call chkalloc('esm_ewald','f_g_z,f_g_2d,f_r_z,f_r_2d',err)
  !initialize
  for(:,:)=0.d0; dfor(:,:)=0.d0; ewald=0.d0
if (esm%normal .eq. 'X') then
  var2d%asi_z   (1:cginp%nsi)=asi(1,1:cginp%nsi)
  var2d%asi_2d(1,1:cginp%nsi)=asi(2,1:cginp%nsi)
  var2d%asi_2d(2,1:cginp%nsi)=asi(3,1:cginp%nsi)
else if (esm%normal .eq. 'Z') then
  var2d%asi_2d(1,1:cginp%nsi)=asi(1,1:cginp%nsi)
  var2d%asi_2d(2,1:cginp%nsi)=asi(2,1:cginp%nsi)
  var2d%asi_z   (1:cginp%nsi)=asi(3,1:cginp%nsi)
endif
  call m_2d_ewald(cginp,var2d,ewald,f_g_z,f_g_2d,f_r_z,f_r_2d,mssg)
if (esm%normal .eq. 'X') then
  for(1,1:cginp%nsi)=-f_g_z   (1:cginp%nsi)-f_r_z   (1:cginp%nsi)
  for(2,1:cginp%nsi)=-f_g_2d(1,1:cginp%nsi)-f_r_2d(1,1:cginp%nsi)
  for(3,1:cginp%nsi)=-f_g_2d(2,1:cginp%nsi)-f_r_2d(2,1:cginp%nsi)
else if (esm%normal .eq. 'Z') then
  for(1,1:cginp%nsi)=-f_g_2d(1,1:cginp%nsi)-f_r_2d(1,1:cginp%nsi)
  for(2,1:cginp%nsi)=-f_g_2d(2,1:cginp%nsi)-f_r_2d(2,1:cginp%nsi)
  for(3,1:cginp%nsi)=-f_g_z   (1:cginp%nsi)-f_r_z   (1:cginp%nsi)
endif
  call formod(for,dfor,cginp%bb,cginp%nsi)
!  write(6,'(A,6x,A,12x,A)')' IN','POSITION OF ATOM','FORCE OF EWALD PART'
!  do it=1,cginp%nsi
!     write(6,'(I3,A,3F9.5,A,3F22.16,A)')it,',(',asi(:,it) &
!          ,'),(',dfor(:,it),')'
!  enddo
  deallocate(f_g_z,f_g_2d,f_r_z,f_r_2d)
  return
end subroutine esm_ewald

subroutine esm_yhtgen(nlx,nly,nlz,rhotmp,eht,comm,mssg)
  implicit none
  integer,intent(in) ::nlx,nly,nlz,comm, mssg
  real(kind=8),intent(out) ::eht
  real(kind=8),intent(inout) ::rhotmp(nlx,nly,nlz,2)
  !local
  integer ::ix,iy,iz,err
  complex(kind=8),allocatable ::rho(:,:,:),rg(:),vg(:),w2(:,:,:)
  !
  allocate(rho(cginp%nrx2,cginp%nry2,cginp%nrz2),rg(var2d%n0) &
          ,vg(var2d%n0),stat=err)
  call chkalloc('esm_yhtgen','rho,rg,vg',err)
             rho(1:cginp%nrx2,1:cginp%nry2,1:cginp%nrz2) &
  =dcmplx(rhotmp(1:cginp%nrx2,1:cginp%nry2,1:cginp%nrz2,1) &
         ,rhotmp(1:cginp%nrx2,1:cginp%nry2,1:cginp%nrz2,2))
             
  if(ideb(2)==1)call print_2d_average_riffle('chrg',cginp,var2d,esm,rho)
!  if(mssg .gt. -1) then
!     write(6,'(A)')'Solve Poisson equation......'
!  end if
  call m_2d_make_hartree_part(cginp,var2d,eht,rho,rg,vg,comm)
  do ix=1,cginp%nrx2; do iy=1,cginp%nry2; do iz=1,cginp%nrz2
  rhotmp(ix,iy,iz,1)= dble(rho(ix,iy,iz))
  rhotmp(ix,iy,iz,2)=dimag(rho(ix,iy,iz))
  enddo; enddo; enddo
  if(mssg .gt. -1) then
     if(ideb(3)==1)call print_2d_average_riffle('hart',cginp,var2d,esm,rho)
     if(ideb(5)==1)allocate(w2(cginp%nrx2,cginp%nry2,cginp%nrz2))
     if(ideb(5)==1)then; rewind 100;read(100)w2; endif
     if(ideb(5)==1)rho(:,:,:)=rho(:,:,:)+w2(:,:,:)
     if(ideb(5)==1)deallocate(w2)
     if(ideb(5)==1)call print_2d_average_riffle('elec',cginp,var2d,esm,rho)
     if(ideb(40)==1)call m_2d_average_all_riffle(cginp,var2d,esm)
  end if
  deallocate(rho,rg,vg)
  return
end subroutine esm_yhtgen

subroutine esm_vblfor_xTAPP(asi,rhotmp,for,epotl,myrank,comm)
  implicit none
  real(kind=8),intent(in)    ::asi(3,cginp%nsi)
  real(kind=8),intent(in)    ::FFT_A3D_DEFSZ(rhotmp,cginp%nrx2,cginp%nry2,cginp%nrz2)
  real(kind=8),intent(inout) ::for(3,cginp%nsi),epotl
  integer, intent(in)        ::myrank,comm
  !local
  integer ::err
  real(kind=8),allocatable    ::wf_z(:),wf_2d(:,:)
  complex(kind=8),allocatable ::rho(:,:,:),rg(:),w_f(:,:)
  integer:: i1,i2,i3,mssg

  if (.not. esm%enable) then
     return
  end if

  if (myrank == 0)  then
     mssg = 1
  else
     mssg = -1
  end if
  !
  allocate(rho(cginp%nrx2,cginp%nry2,cginp%nrz2),rg(var2d%n0) &
          ,w_f(var2d%n0,3),wf_z(cginp%nsi),wf_2d(2,cginp%nsi),stat=err)
  call chkalloc('esm_vblfor','rho,rg,w,w_f,wf_z,wf_2d',err)
  wf_z(1:cginp%nsi)=0.d0
  wf_2d(1:2,1:cginp%nsi)=0.d0
  do i3=1,cginp%nrz2
     do i2=1,cginp%nry2
        do i1=1,cginp%nrx2
           rho(i1,i2,i3) = dcmplx( FFT_A3D_RL(rhotmp,i1,i2,i3), &
                FFT_A3D_IM(rhotmp,i1,i2,i3) )
        end do
     end do
  end do
if (esm%normal .eq. 'X') then
  var2d%asi_z   (1:cginp%nsi)=asi(1,1:cginp%nsi)
  var2d%asi_2d(1,1:cginp%nsi)=asi(2,1:cginp%nsi)
  var2d%asi_2d(2,1:cginp%nsi)=asi(3,1:cginp%nsi)
else if (esm%normal .eq. 'Z') then
  var2d%asi_2d(1,1:cginp%nsi)=asi(1,1:cginp%nsi)
  var2d%asi_2d(2,1:cginp%nsi)=asi(2,1:cginp%nsi)
  var2d%asi_z   (1:cginp%nsi)=asi(3,1:cginp%nsi)
endif
  call m_2d_make_force_long_range_part(cginp,var2d,rho,rg,epotl &
                                      ,w_f,wf_z,wf_2d,comm,mssg)
if (esm%normal .eq. 'X') then
  for(1,1:cginp%nsi)=for(1,1:cginp%nsi)+wf_z   (1:cginp%nsi)
  for(2,1:cginp%nsi)=for(2,1:cginp%nsi)+wf_2d(1,1:cginp%nsi)
  for(3,1:cginp%nsi)=for(3,1:cginp%nsi)+wf_2d(2,1:cginp%nsi)
else if (esm%normal .eq. 'Z') then
  for(1,1:cginp%nsi)=for(1,1:cginp%nsi)+wf_2d(1,1:cginp%nsi)
  for(2,1:cginp%nsi)=for(2,1:cginp%nsi)+wf_2d(2,1:cginp%nsi)
  for(3,1:cginp%nsi)=for(3,1:cginp%nsi)+wf_z   (1:cginp%nsi)
endif
  deallocate(rho,rg,w_f,wf_z,wf_2d)
  return
end subroutine esm_vblfor_xTAPP

end module esm_manage
