! Copyright (c) 2012, Minoru Otani <minoru.otani@aist.go.jp>
! 
! Permission is hereby granted, free of charge, to any person 
! obtaining a copy of this software and associated documentation 
! files (the "Software"), to deal in the Software without restriction, 
! including without limitation the rights to use, copy, modify, merge, 
! publish, distribute, sublicense, and/or sell copies of the Software, 
! and to permit persons to whom the Software is furnished to do so, 
! subject to the following conditions:
 
! The above copyright notice and this permission notice shall be 
! included in all copies or substantial portions of the Software.
 
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
! OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
! NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
! HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
! WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
! OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
! DEALINGS IN THE SOFTWARE.
module esm_util
use global_esm
contains
integer function lat_points(r,v)
  implicit none
  integer,parameter:: margin=0
  real(kind=8):: pi,r,v
  !local
  integer:: n
  real(kind=8):: second
  pi=4*atan(1.d0)
  second=1.d0/2.d0
  n=int(r/(v)**second)+1
  n=n+margin
  lat_points=int(pi*n**2)+1
  return
end function lat_points

subroutine m_2d_average_all_riffle(cginp,var2d,esm)
  implicit none
  type(esm_2d)  ::var2d
  type(cg2esm)  ::cginp
  type(esm_prm) ::esm
  integer       ::ig,igs,ige
  character(17) ::gpfile='plt_all_riffle.gp'
  !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  open(2,file=gpfile)
  !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  igs=esm%gp_s
  if(esm%gp_e<0)then
    ige=var2d%nkf_2d
  else
    ige=esm%gp_e
  endif
  do ig=igs,ige
    !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    write(2,'(A)')'reset'
    write(2,'(A)')'set multiplot'
    write(2,'(A)')'set size 0.5,0.5'
    write(2,'(A)')'set nokey'
    write(2,'(A,F8.4,A,F8.4,A)') &
    'set xrange[',-var2d%L/2.d0,':',var2d%L/2.d0,']'
!$$$    'set xrange[',0.d0,':',var2d%L,']'
    write(2,'(A)')'set origin 0,0.5'
    write(2,'(A,I4.4,A)')"set title 'chrg. ig=",ig,"'"
    call print_normal(ig,'chrg')
    write(2,'(A)')'set origin 0.5,0.5'
    write(2,'(A,I4.4,A)')"set title 'ion loc. ig=",ig,"'"
    call print_normal(ig,'iloc')
    write(2,'(A)')'set origin 0,0'
    write(2,'(A,I4.4,A)')"set title 'hart. ig=",ig,"'"
    call print_normal(ig,'hart')
    write(2,'(A)')'set origin 0.5,0'
    write(2,'(A,I4.4,A)')"set title 'hart.+ion ig=",ig,"'"
    call print_normal(ig,'elec')
    write(2,'(A)')'set nomultiplot'
    write(2,'(A)')'pause -1'
    !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  enddo
  !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  write(2,'(A)')'quit'
  close(2)
  !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  contains
    subroutine print_normal(ig,plot_obj)
    implicit none
    integer,intent(in) ::ig
    character(4)       ::plot_obj
    character(17)      ::filename='plt_aaaa_bbbb.dat'
    write(filename(5:8),'(A)')plot_obj
    write(filename(10:13),'(I4.4)')ig
    write(2,30)filename
    30 format("plot '",A,"' w l")
    end subroutine print_normal
end subroutine m_2d_average_all_riffle

subroutine print_2d_average_riffle(plot_object,cginp,var2d,esm,v)
  use fft_1d
  implicit none
  type(esm_2d)  ::var2d
  type(cg2esm)  ::cginp
  type(esm_prm) ::esm
  integer       ::ig,k1,k2,ks1,ks2,igs,ige,err
  complex(kind=8),intent(in)  ::v(cginp%nrx2,cginp%nry2,cginp%nrz2)
  complex(kind=8),allocatable ::vr2(:)
  character(4),intent(in)     ::plot_object
  character(18) ::gpfile='plt_aaaa_riffle.gp'
  !!! FOR FFT !!!
  type(fft1_struct) ::fs
  real(kind=8),allocatable ::fftw1(:,:),fftw2(:,:)
  allocate(fftw1(var2d%n0,2),fftw2(var2d%n0,2),stat=err)
  call chkalloc('fft_pot','fftw1,fftw2',err)
  call fft1_init(var2d%n0,fs)
  !!!!!!!!!!!!!!!
  allocate(vr2(var2d%n0))
  !
  !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  write(gpfile(5:8),'(A)')plot_object
  open(2,file=gpfile)
  !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  igs=esm%gp_s
  if(esm%gp_e<0)then
  ige=var2d%nkf_2d
  else
    ige=esm%gp_e
  endif
  do ig=igs,ige
    !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    k1=var2d%kkg_2d(1,ig)
    k2=var2d%kkg_2d(2,ig)
    ks1=mod(var2d%n1+k1,var2d%n1)+1
    ks2=mod(var2d%n2+k2,var2d%n2)+1
    write(2,'(A)')'reset'
    write(2,'(A)')'set nokey'
    write(2,'(A,F8.4,A,F8.4,A)') &
    'set xrange[',-var2d%L/2.d0,':',var2d%L/2.d0,']'
!$$$    'set xrange[',0.d0,':',var2d%L,']'
    write(2,'(3A,I3.3,A)')"set title '",plot_object,', ig=',ig,"'"
if (esm%normal .eq. 'X') then
    vr2(:)=v(:,ks1,ks2)
else if (esm%normal .eq. 'Z') then
    vr2(:)=v(ks1,ks2,:)
endif
!    call wfft1d(var2d%n0,vr2,1)
    call fft1_bw(fs,vr2,fftw1,fftw2)
    call print_normal(ig)
    write(2,'(A)')'pause -1'
    !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  enddo
  !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  write(2,'(A)')'quit'
  close(2)
  !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!bug fix by Y. Suwa, 2006/09/08
  call fft1_done(fs)
! 
 deallocate(fftw1,fftw2,vr2)
  contains
    subroutine print_normal(ig)
    implicit none
    integer,intent(in) ::ig
    integer            ::ix,ixx
    real(kind=8)       ::x
    character(17)      ::filename='plt_aaaa_bbbb.dat'
    write(filename(5:8),'(A)')plot_object
    write(filename(10:13),'(I4.4)')ig
    write(2,30)filename
    30 format("plot '",A,"' w l")
    open(1,file=filename)
!$$$    do ix=1,var2d%n0
!$$$      x=(ix-1)*var2d%L/var2d%n0
!$$$      write(1,'(f8.4,2X,F22.16)') x,dble(vr2(ix))
!$$$    enddo
    do ix=-var2d%n0/2,var2d%n0/2-1
      x=(ix-1)*var2d%L/var2d%n0
      ixx=ix+var2d%n0
      if(ixx.gt.var2d%n0)ixx=ixx-var2d%n0
      write(1,'(f8.4,2X,F22.16)') x,dble(vr2(ixx))
    enddo
    close(1)
  end subroutine print_normal
end subroutine print_2d_average_riffle

subroutine ordg0_2d(nam,qb,kk,ak,vec2d,nk,ney,nez)
  implicit none
  integer,intent(in)      ::nam
  real(kind=8),intent(in) ::qb,vec2d(2,2)
  integer,intent(out)     ::nk
  integer                 ::kk(2,nam)
  real(kind=8)            ::ak(nam)
  !local
  integer                 ::icycle,neymax,nezmax,ierr,ney,nez
  integer,parameter       ::ncycle = 40
  real(kind=8),parameter  ::factor = 2.d0
  !
  ney=1;nez=1
  do icycle=1,ncycle
    call ykarr4_2d(nam,ney,nez,qb,nk,ierr,neymax,nezmax,kk,ak,vec2d)
    if(ierr.eq.0)then
      goto 1000
    else
      if(neymax.ge.ney)then
        ney = int(dble(ney) * factor)
      endif
      if(nezmax.ge.nez)then
        nez = int(dble(nez) * factor)
      endif
    endif
  enddo
  stop 'ERROR!! increase ncycle in ordg0_2d(ewald method)'
1000 continue
  !
  return
end subroutine ordg0_2d

subroutine ykarr4_2d(nam,ney,nez,qf,nkf,ierr,neymax,nezmax,kk,ak,vec2d)
  implicit none
  integer,intent(in)      ::nam
  real(kind=8),intent(in) ::qf,vec2d(2,2)
  integer,intent(out)     ::nkf
  integer                 ::kk(2,nam)
  real(kind=8)            ::ak(nam)
  !local
  real(kind=8)        ::qqf,cy,cz,s
  integer             ::i,iy,iz,ney,nez,ierr,neymax,nezmax
  integer,allocatable ::ind(:),kkg(:,:)
  !
  allocate(ind(nam),kkg(2,nam))
  qqf=qf*qf
  nkf=0
  do iy=-ney,ney
    do iz=-nez,nez
      cy=iy*vec2d(1,1)+iz*vec2d(1,2)
      cz=iy*vec2d(2,1)+iz*vec2d(2,2)
      s=cy*cy+cz*cz
      if(s.le.qqf)then
        if(nkf+1.ge.nam)then
          write(6,*)'Error!!! Increase gsafe?'
          stop
        endif
        nkf=nkf+1
        ak(nkf)=s
        kk(1,nkf)=iy
        kk(2,nkf)=iz
      endif
    enddo
  enddo
  neymax=0
  nezmax=0
  do i=1,nkf
    neymax=max(neymax,abs(kk(1,i)))
    nezmax=max(nezmax,abs(kk(2,i)))
  enddo
  if((neymax.ge.ney).or.(nezmax.ge.nez))then
    ierr=1
    deallocate(ind,kkg)
    return
  else
    ierr=0
    ney=neymax
    nez=nezmax
  endif
  !
  ! heap sort
  !
  call esm_hsort(nkf,ak,ind)
  do i=1,nkf
    kkg(:,i) = kk(:,i)
  enddo
  do i=1,nkf
    kk(:,i) = kkg(:,ind(i))
  enddo
  deallocate(ind,kkg)
  return
end subroutine ykarr4_2d
!
! hsort
!
subroutine esm_hsort(n,ra,ind)
  implicit none
  integer,intent(in)        ::n
  integer,dimension(n)      ::ind
  real(kind=8),dimension(n) ::ra
  integer                   ::i,j,l,ir,iind
  real(kind=8)              ::rra
  !
  l = n/2+1
  ir = n
  do i=1,n
    ind(i) = i
  enddo
  !
10  continue
  if(l.gt.1)then
    l    = l-1
    rra  = ra(l)
    iind = ind(l)
  else
    rra     = ra(ir)
    iind    = ind(ir)
    ra(ir)  = ra(1)
    ind(ir) = ind(1)
    ir=ir-1
    if(ir.eq.1)then
      ra(1)  = rra
      ind(1) = iind
      return
    endif
  endif
  !
  i=l
  j=l+1
20  if(j.le.ir)then
    if(j.lt.ir)then
      if(ra(j).lt.ra(j+1))j=j+1
    endif
    if(rra.lt.ra(j))then
      ra(i)  = ra(j)
      ind(i) = ind(j)
      i = j
      j = j + j
    else
      j=ir+1
    endif
    goto 20
  endif
  ra(i)  = rra
  ind(i) = iind
  goto 10
  !
  return
end subroutine esm_hsort

end module esm_util
