!
! Copyright (c) 2012-2013 Yoshihide Yoshimoto
!

#include "fftgrid.h"

module exchange_kern
  use fft_3d
  integer:: ifgam
  integer:: niw, njw, nwx, nwy, nwz, nwx2, nwy2, nwz2, nwxyz
  integer:: mwx, infwx
  integer:: nlx, nly, nlz, nlxyz
  type (fft3_struct):: fs
  real(kind=8),allocatable:: clkern(:,:,:),clkerns(:,:,:,:)
  real(kind=8),allocatable:: ztmp(:,:,:,:),ztmp2(:,:)
  real(kind=8),allocatable:: pxi(:,:,:,:)
  real(kind=8),allocatable:: vsum(:,:,:,:)
end module exchange_kern

subroutine get_exchange_nlxyz(nlxo, nlyo, nlzo)
  use exchange_kern
  implicit none
  integer,intent(out):: nlxo,nlyo,nlzo
  nlxo = nlx
  nlyo = nly
  nlzo = nlz
  return
end subroutine get_exchange_nlxyz

subroutine init_exchange(ifgami, niwi, njwi, nwxi, nwyi, nwzi)
  use exchange_kern
  use fft_3d
  implicit none
  integer:: ifgami, niwi, njwi, nwxi, nwyi, nwzi

  if (fft_3d_cap_rh) then
     ifgam = ifgami
  else
     ifgam = 0
     ifgami = 0
  end if

  niw = niwi
  njw = njwi
  nwx = nwxi
  nwy = nwyi
  nwz = nwzi
  nwx2 = nwx*2
  nwy2 = nwy*2
  nwz2 = nwz*2
  nwxyz = nwx2*nwy2*nwz2

  if (ifgam.eq.1) then
     infwx = 0
     mwx = nwx+1
     nlx = FFT_PAD_NX(nwx+1)
  else
     infwx = -nwx+1
     mwx = nwx2
     nlx = FFT_PAD_NX(nwx2)
  end if
  nly = FFT_PAD_NY(nwy2)
  nlz = FFT_PAD_NZ(nwz2)
  nlxyz = nlx*nly*nlz

  if (ifgam.eq.1) then
     call fft3_init_rh(nwx2,nwy2,nwz2,nlx,nly,nlz,fs)
  else
     call fft3_init(nwx2,nwy2,nwz2,nlx,nly,nlz,fs)
  end if
  allocate(clkern(nlx,nly,nlz),clkerns(nlx,nly,nlz,6))

  allocate(pxi(2,nlx,nly,nlz), &
       ztmp(2,nlx,nly,nlz),ztmp2(2,nlxyz), &
       vsum(2,nlx,nly,nlz))

  return
end subroutine init_exchange

subroutine setup_exchange_clkern(nkmesh, bb, q, rc, rd)
  use exchange_kern
  use util
  implicit none
  integer:: nkmesh
  real(kind=8):: bb(3,3),q(3)
  real(kind=8),intent(in):: rc,rd

  real(kind=8),parameter:: pi = 3.14159265358979323846d0

  real(kind=8):: vb,aqi(3),g(3),gsq,ga,dsq,epsilon
  integer:: ix,iy,iz,i1,i2,i3

  vb = cellvol(bb)
  dsq = ( pi/(rd-rc) )**2
  epsilon = 1.0d-6 * ( (vb/nkmesh)**(1.0d0/3.0d0) )

  clkern = 0.0d0
!$omp parallel do private(ix,iy,iz,i1,i2,i3,aqi,g,gsq,ga)
  do iz = -nwz+1,nwz-1
     do iy = -nwy+1,nwy-1
        do ix = infwx,nwx-1
           i1=mod(nwx2+ix,nwx2)+1
           i2=mod(nwy2+iy,nwy2)+1
           i3=mod(nwz2+iz,nwz2)+1
           aqi(1) = q(1)+ix
           aqi(2) = q(2)+iy
           aqi(3) = q(3)+iz

           g(1) = bb(1,1)*aqi(1)+bb(1,2)*aqi(2)+bb(1,3)*aqi(3)
           g(2) = bb(2,1)*aqi(1)+bb(2,2)*aqi(2)+bb(2,3)*aqi(3)
           g(3) = bb(3,1)*aqi(1)+bb(3,2)*aqi(2)+bb(3,3)*aqi(3)

           gsq = g(1)**2 + g(2)**2 + g(3)**2
           ga = sqrt(gsq)

           ! Ref: James Spencer and Ali Alavi, PRB 77, 193110 (2008)
           ! Ref: Yu-ichiro Matsushita and Kazuma Nakamura
           ! and Atsushi Oshiyama, PRB 84 075205 (2011)
           ! 
           ! if (ga .lt. epsilon) then
           !    clkern(i1,i2,i3) = 2*pi*rc**2
           ! else
           !    clkern(i1,i2,i3) = 4.0d0*pi/gsq * ( 1.0d0 - cos(ga*rc) )
           ! end if

           if (ga .lt. epsilon) then
              clkern(i1,i2,i3) = pi*(rc**2+rd**2) - 4.0d0/pi*(rd-rc)**2
           else
              clkern(i1,i2,i3) = 4.0d0*pi/gsq + &
                   2.0d0*pi*( -1.0d0/gsq + 1.0d0/(gsq-dsq) ) &
                   *( cos(ga*rc) + cos(ga*rd) )
           end if
        end do
     end do
  end do

  return
end subroutine setup_exchange_clkern

subroutine setup_exchange_clkerns(nkmesh, bb, q, rc, rd)
  use exchange_kern
  use util
  implicit none
  integer:: nkmesh
  real(kind=8):: bb(3,3),q(3)
  real(kind=8),intent(in):: rc,rd

  real(kind=8),parameter:: pi = 3.14159265358979323846d0

  real(kind=8):: vb,aqi(3),g(3),gsq,ga,dsq,epsilon,fac
  integer:: ix,iy,iz,i1,i2,i3

  vb = cellvol(bb)
  dsq = ( pi/(rd-rc) )**2
  epsilon = 1.0d-6 * ( (vb/nkmesh)**(1.0d0/3.0d0) )

  clkerns = 0.0d0
!$omp parallel do private(ix,iy,iz,i1,i2,i3,aqi,g,gsq,ga,fac)
  do iz = -nwz+1,nwz-1
     do iy = -nwy+1,nwy-1
        do ix = infwx,nwx-1
           i1=mod(nwx2+ix,nwx2)+1
           i2=mod(nwy2+iy,nwy2)+1
           i3=mod(nwz2+iz,nwz2)+1
           aqi(1) = q(1)+ix
           aqi(2) = q(2)+iy
           aqi(3) = q(3)+iz

           g(1) = bb(1,1)*aqi(1)+bb(1,2)*aqi(2)+bb(1,3)*aqi(3)
           g(2) = bb(2,1)*aqi(1)+bb(2,2)*aqi(2)+bb(2,3)*aqi(3)
           g(3) = bb(3,1)*aqi(1)+bb(3,2)*aqi(2)+bb(3,3)*aqi(3)

           gsq = g(1)**2 + g(2)**2 + g(3)**2
           ga = sqrt(gsq)

           ! Ref: James Spencer and Ali Alavi, PRB 77, 193110 (2008)
           ! Ref: Yu-ichiro Matsushita and Kazuma Nakamura
           ! and Atsushi Oshiyama, PRB 84 075205 (2011)
           ! 
           ! if (ga .lt. epsilon) then
           !    clkern(i1,i2,i3) = 2*pi*rc**2
           ! else
           !    clkern(i1,i2,i3) = 4.0d0*pi/gsq * ( 1.0d0 - cos(ga*rc) )
           ! end if

           if (ga .lt. epsilon) then
              clkerns(i1,i2,i3,1:6) = 0.0d0
           else
              fac = -8.0d0*pi/gsq + &
                   4.0d0*pi*( 1.0d0/gsq - gsq/(gsq-dsq)**2 ) &
                   *( cos(ga*rc) + cos(ga*rd) ) + &
                   2.0d0*pi*( 1.0d0/gsq - 1.0d0/(gsq-dsq) ) &
                   *( ga*rc*sin(ga*rc) + ga*rd*sin(ga*rd) )
              g(:) = g(:)/ga

              clkerns(i1,i2,i3,1) = fac*g(1)**2
              clkerns(i1,i2,i3,2) = fac*g(2)**2
              clkerns(i1,i2,i3,3) = fac*g(3)**2
              clkerns(i1,i2,i3,4) = fac*2.0d0*g(2)*g(3)
              clkerns(i1,i2,i3,5) = fac*2.0d0*g(3)*g(1)
              clkerns(i1,i2,i3,6) = fac*2.0d0*g(1)*g(2)
           end if
        end do
     end do
  end do

  return
end subroutine setup_exchange_clkerns

subroutine teardown_exchange()
  use exchange_kern
  use fft_3d
  implicit none

  deallocate(pxi,ztmp,ztmp2,vsum)
  deallocate(clkern,clkerns)
  call fft3_done(fs)
  return
end subroutine

subroutine deploy_pxj_kern(mjw,pxj)
  use exchange_kern
  use fft_3d
  implicit none
  integer,intent(in):: mjw
  real(kind=8):: FFT_A1M_DEFSZ(pxj,nlxyz,mjw)
  integer:: i,jb

  ztmp2 = 0.0d0
  do jb=1,mjw
     call fft3_bw(fs,FFT_A1M_RL(pxj,1,jb),ztmp2)
  end do

  return
end subroutine deploy_pxj_kern

subroutine compute_exchange(miw, mjw, wfi, pxj, vxp, wgt)
  use exchange_kern
  use fft_3d
  implicit none
  integer:: miw, mjw
  real(kind=8):: FFT_A3M_DEFSZ(wfi,nlx,nly,nlz,miw)
  real(kind=8):: FFT_A3M_DEFSZ(pxj,nlx,nly,nlz,mjw)
  real(kind=8):: FFT_A3M_DEFSZ(vxp,nlx,nly,nlz,miw)
  real(kind=8):: wgt(mjw)

  integer:: i1,i2,i3
  integer:: ib,jb

  if ( (miw .gt. niw) .or. (mjw .gt. njw) ) then
     write(6,*) 'invalid miw, mjw = ', miw, mjw
     stop
  end if

  ztmp = 0.0d0
  pxi = 0.0d0

  do ib=1,miw

!$omp parallel do private(i1,i2,i3)
     do i3 = 1,nwz2
        do i2 = 1,nwy2
           do i1 = 1,mwx
              FFT_A3D_RL(pxi,i1,i2,i3) = FFT_A3M_RL(wfi,i1,i2,i3,ib)
              FFT_A3D_IM(pxi,i1,i2,i3) = FFT_A3M_IM(wfi,i1,i2,i3,ib)
           end do
        end do
     end do

     call fft3_bw(fs,pxi,ztmp2)

!$omp parallel do private(i3)
     do i3=1,nlz
        vsum(:,:,:,i3) = 0.0d0
     end do
     do jb=1,mjw
        if (ifgam.eq.1) then
!$omp parallel do private(i1,i2,i3)
           do i3 = 1,nwz2
              do i2 = 1,nwy2
                 do i1 = 1,nwx
                    FFT_A3D_RL(ztmp,i1,i2,i3) = &
                         FFT_A3M_RL(pxj,i1,i2,i3,jb)*FFT_A3D_RL(pxi,i1,i2,i3)
                    FFT_A3D_IM(ztmp,i1,i2,i3) = &
                         FFT_A3M_IM(pxj,i1,i2,i3,jb)*FFT_A3D_IM(pxi,i1,i2,i3)
                 end do
              end do
           end do
        else
!$omp parallel do private(i1,i2,i3)
           do i3 = 1,nwz2
              do i2 = 1,nwy2
                 do i1 = 1,nwx2
                    FFT_A3D_RL(ztmp,i1,i2,i3) = &
                         FFT_A3M_RL(pxj,i1,i2,i3,jb)*FFT_A3D_RL(pxi,i1,i2,i3) &
                         + FFT_A3M_IM(pxj,i1,i2,i3,jb)*FFT_A3D_IM(pxi,i1,i2,i3)
                    FFT_A3D_IM(ztmp,i1,i2,i3) = &
                         FFT_A3M_RL(pxj,i1,i2,i3,jb)*FFT_A3D_IM(pxi,i1,i2,i3) &
                         - FFT_A3M_IM(pxj,i1,i2,i3,jb)*FFT_A3D_RL(pxi,i1,i2,i3)
                 end do
              end do
           end do
        end if
        call fft3_fw(fs,ztmp,ztmp2)

!$omp parallel do private(i1,i2,i3)
        do i3 = 1,nwz2
           do i2 = 1,nwy2
              do i1 = 1,mwx
                 FFT_A3D_RL(ztmp,i1,i2,i3) = clkern(i1,i2,i3)*FFT_A3D_RL(ztmp,i1,i2,i3)
                 FFT_A3D_IM(ztmp,i1,i2,i3) = clkern(i1,i2,i3)*FFT_A3D_IM(ztmp,i1,i2,i3)
              end do
           end do
        end do
        call fft3_bw(fs,ztmp,ztmp2)

        if (ifgam.eq.1) then
!$omp parallel do private(i1,i2,i3)
           do i3 = 1,nwz2
              do i2 = 1,nwy2
                 do i1 = 1,nwx
                    FFT_A3D_RL(vsum,i1,i2,i3) = FFT_A3D_RL(vsum,i1,i2,i3) &
                         + wgt(jb)*FFT_A3D_RL(ztmp,i1,i2,i3)*FFT_A3M_RL(pxj,i1,i2,i3,jb)
                    FFT_A3D_IM(vsum,i1,i2,i3) = FFT_A3D_IM(vsum,i1,i2,i3) &
                         + wgt(jb)*FFT_A3D_IM(ztmp,i1,i2,i3)*FFT_A3M_IM(pxj,i1,i2,i3,jb)
                 end do
              end do
           end do
        else
!$omp parallel do private(i1,i2,i3)
           do i3 = 1,nwz2
              do i2 = 1,nwy2
                 do i1 = 1,nwx2
                    FFT_A3D_RL(vsum,i1,i2,i3) = FFT_A3D_RL(vsum,i1,i2,i3) &
                         + wgt(jb)*( FFT_A3D_RL(ztmp,i1,i2,i3)*FFT_A3M_RL(pxj,i1,i2,i3,jb) &
                         - FFT_A3D_IM(ztmp,i1,i2,i3)*FFT_A3M_IM(pxj,i1,i2,i3,jb) )
                    FFT_A3D_IM(vsum,i1,i2,i3) = FFT_A3D_IM(vsum,i1,i2,i3) &
                         + wgt(jb)*( FFT_A3D_RL(ztmp,i1,i2,i3)*FFT_A3M_IM(pxj,i1,i2,i3,jb) &
                         + FFT_A3D_IM(ztmp,i1,i2,i3)*FFT_A3M_RL(pxj,i1,i2,i3,jb) )
                 end do
              end do
           end do
        end if
     end do
     call fft3_fw(fs,vsum,ztmp2)

!$omp parallel do private(i1,i2,i3)
     do i3 = 1,nwz2
        do i2 = 1,nwy2
           do i1 = 1,mwx
              FFT_A3M_RL(vxp,i1,i2,i3,ib) = FFT_A3M_RL(vxp,i1,i2,i3,ib) - FFT_A3D_RL(vsum,i1,i2,i3)
              FFT_A3M_IM(vxp,i1,i2,i3,ib) = FFT_A3M_IM(vxp,i1,i2,i3,ib) - FFT_A3D_IM(vsum,i1,i2,i3)
           end do
        end do
     end do
  end do

  return

end subroutine compute_exchange

subroutine compute_excheng(miw, mjw, wfi, pxj, exeng, exengs, wgti, wgtj)
  use exchange_kern
  use fft_3d
  implicit none
  integer:: miw, mjw
  real(kind=8):: FFT_A3M_DEFSZ(wfi,nlx,nly,nlz,miw)
  real(kind=8):: FFT_A3M_DEFSZ(pxj,nlx,nly,nlz,mjw)
  real(kind=8):: exeng,exengs(6)
  real(kind=8):: wgti(miw),wgtj(mjw)

  integer:: i1,i2,i3
  integer:: ib,jb,ii
  real(kind=8):: ar,ai,tmp

  if ( (miw .gt. niw) .or. (mjw .gt. njw) ) then
     write(6,*) 'invalid miw, mjw = ', miw, mjw
     stop
  end if

  ztmp = 0.0d0
  pxi = 0.0d0

  do ib=1,miw

!$omp parallel do private(i1,i2,i3)
     do i3 = 1,nwz2
        do i2 = 1,nwy2
           do i1 = 1,nwx2
              FFT_A3D_RL(pxi,i1,i2,i3) = FFT_A3M_RL(wfi,i1,i2,i3,ib)
              FFT_A3D_IM(pxi,i1,i2,i3) = FFT_A3M_IM(wfi,i1,i2,i3,ib)
           end do
        end do
     end do

     call fft3_bw(fs,pxi,ztmp2)

     do jb=1,mjw
!$omp parallel do private(i1,i2,i3)
        do i3 = 1,nwz2
           do i2 = 1,nwy2
              do i1 = 1,nwx2
                 FFT_A3D_RL(ztmp,i1,i2,i3) = &
                      FFT_A3M_RL(pxj,i1,i2,i3,jb)*FFT_A3D_RL(pxi,i1,i2,i3) &
                      + FFT_A3M_IM(pxj,i1,i2,i3,jb)*FFT_A3D_IM(pxi,i1,i2,i3)
                 FFT_A3D_IM(ztmp,i1,i2,i3) = &
                      FFT_A3M_RL(pxj,i1,i2,i3,jb)*FFT_A3D_IM(pxi,i1,i2,i3) &
                      - FFT_A3M_IM(pxj,i1,i2,i3,jb)*FFT_A3D_RL(pxi,i1,i2,i3)
              end do
           end do
        end do
        call fft3_fw(fs,ztmp,ztmp2)

!$omp parallel do private(i1,i2,i3)
        do i3 = 1,nwz2
           do i2 = 1,nwy2
              do i1 = 1,nwx2
                 FFT_A3D_RL(vsum,i1,i2,i3) = clkern(i1,i2,i3)*FFT_A3D_RL(ztmp,i1,i2,i3)
                 FFT_A3D_IM(vsum,i1,i2,i3) = clkern(i1,i2,i3)*FFT_A3D_IM(ztmp,i1,i2,i3)
              end do
           end do
        end do
        call fft3_bw(fs,vsum,ztmp2)

        tmp = 0.0d0
!$omp parallel do private(i1,i2,i3,ar,ai),reduction(+:tmp)
        do i3 = 1,nwz2
           do i2 = 1,nwy2
              do i1 = 1,nwx2
                 ar = wgtj(jb)*( FFT_A3D_RL(vsum,i1,i2,i3)*FFT_A3M_RL(pxj,i1,i2,i3,jb) &
                      - FFT_A3D_IM(vsum,i1,i2,i3)*FFT_A3M_IM(pxj,i1,i2,i3,jb) )
                 ai = wgtj(jb)*( FFT_A3D_RL(vsum,i1,i2,i3)*FFT_A3M_IM(pxj,i1,i2,i3,jb) &
                      + FFT_A3D_IM(vsum,i1,i2,i3)*FFT_A3M_RL(pxj,i1,i2,i3,jb) )

                 tmp = tmp - 0.5d0*wgti(ib)/nwxyz*( &
                      FFT_A3D_RL(pxi,i1,i2,i3)*ar + FFT_A3D_IM(pxi,i1,i2,i3)*ai )
              end do
           end do
        end do
        exeng = exeng + tmp

        do ii = 1,6
!$omp parallel do private(i1,i2,i3)
           do i3 = 1,nwz2
              do i2 = 1,nwy2
                 do i1 = 1,nwx2
                    FFT_A3D_RL(vsum,i1,i2,i3) = clkerns(i1,i2,i3,ii)*FFT_A3D_RL(ztmp,i1,i2,i3)
                    FFT_A3D_IM(vsum,i1,i2,i3) = clkerns(i1,i2,i3,ii)*FFT_A3D_IM(ztmp,i1,i2,i3)
                 end do
              end do
           end do
           call fft3_bw(fs,vsum,ztmp2)

           tmp = 0.0d0
!$omp parallel do private(i1,i2,i3,ar,ai),reduction(+:tmp)
           do i3 = 1,nwz2
              do i2 = 1,nwy2
                 do i1 = 1,nwx2
                    ar = wgtj(jb)*( FFT_A3D_RL(vsum,i1,i2,i3)*FFT_A3M_RL(pxj,i1,i2,i3,jb) &
                         - FFT_A3D_IM(vsum,i1,i2,i3)*FFT_A3M_IM(pxj,i1,i2,i3,jb) )
                    ai = wgtj(jb)*( FFT_A3D_RL(vsum,i1,i2,i3)*FFT_A3M_IM(pxj,i1,i2,i3,jb) &
                         + FFT_A3D_IM(vsum,i1,i2,i3)*FFT_A3M_RL(pxj,i1,i2,i3,jb) )

                    tmp = tmp - 0.5d0*wgti(ib)/nwxyz*( &
                         FFT_A3D_RL(pxi,i1,i2,i3)*ar + FFT_A3D_IM(pxi,i1,i2,i3)*ai )
                 end do
              end do
           end do
           exengs(ii) = exengs(ii) + tmp
        end do
     end do

  end do

  return

end subroutine compute_excheng
