!
! Copyright (c) 2012-2013 Yoshihide Yoshimoto
!

#define FFT3D_PROF
#define FFT3D_FFTW_EXPORT_WISDOM

module FFTW3
  use, intrinsic :: iso_c_binding
  logical,parameter:: fft_3d_cap_rh = .true.
  include 'fftw3.f03'
end module

module fft_3d
#ifdef _OPENMP
  use omp_lib
#endif
  use FFTW3
  !
  type fft3_struct
     integer:: nx,ny,nz,nxyz,lx,ly,lz,lxyz
     integer:: nmx
     type(C_PTR):: planf
     type(C_PTR):: planb
     logical:: ifsft
     type(C_PTR):: inout
     logical:: rh_trs
  end type fft3_struct

  logical,private:: fftw_init = .false.
  private:: cpyd2b, cpyb2d, cpyd2b_p, cpyb2d_p, cpyb2d_ps

contains

  subroutine cpyd2b_p(n,d,b)
    implicit none
    integer,intent(in):: n
    real(kind=8):: d(2*n)
    type(C_PTR):: b
    integer:: i
    real(kind=8),pointer:: bp(:)

    call c_f_pointer(b,bp,[2*n])
#ifdef __FUJITSU
!OCL XFILL
    do i=1,2*n
       bp(i) = d(i)
    end do
#else
!$omp parallel do private(i)
    do i=1,2*n
       bp(i) = d(i)
    end do
#endif
    return
  end subroutine cpyd2b_p

  subroutine cpyb2d_p(n,b,d)
    implicit none
    integer,intent(in):: n
    real(kind=8):: d(2*n)
    type(C_PTR):: b
    integer:: i
    real(kind=8),pointer:: bp(:)

    call c_f_pointer(b,bp,[2*n])
#ifdef __FUJITSU
!OCL XFILL
    do i=1,2*n
       d(i) = bp(i)
    end do
#else
!$omp parallel do private(i)
    do i=1,2*n
       d(i) = bp(i)
    end do
#endif
    return
  end subroutine cpyb2d_p

  subroutine cpyb2d_ps(n,b,d,scl)
    implicit none
    integer,intent(in):: n
    real(kind=8):: d(2*n),scl
    type(C_PTR):: b
    integer:: i
    real(kind=8),pointer:: bp(:)

    call c_f_pointer(b,bp,[2*n])
#ifdef __FUJITSU
!OCL XFILL
    do i=1,2*n
       d(i) = bp(i)*scl
    end do
#else
!$omp parallel do private(i)
    do i=1,2*n
       d(i) = bp(i)*scl
    end do
#endif
    return
  end subroutine cpyb2d_ps

  subroutine cpyd2b(d,b,nx,ny,nz,lx,ly,lz)
    implicit none
    integer,intent(in):: nx,ny,nz,lx,ly,lz
    real(kind=8):: d(2,lx,ly,lz)
    type(C_PTR):: b
    complex(kind=8),pointer:: bc(:,:,:)
    complex(kind=8):: t
    integer:: i,j,k

    if (nx.eq.lx .and. ny.eq.ly .and. nz.eq.lz) then
       call cpyd2b_p(nx*ny*nz,d,b)
    else
    call c_f_pointer(b,bc,[nx,ny,nz])
!$omp parallel do private(i,j,k,t)
       do k=1,nz
          do j=1,ny
!OCL XFILL
             do i=1,nx
                t = dcmplx(d(1,i,j,k),d(2,i,j,k))
                bc(i,j,k) = t
             end do
          end do
       end do
    end if
    return    
  end subroutine cpyd2b

  subroutine cpyb2d(b,d,nx,ny,nz,lx,ly,lz)
    implicit none
    integer,intent(in):: nx,ny,nz,lx,ly,lz
    real(kind=8):: d(2,lx,ly,lz)
    type(C_PTR):: b
    complex(kind=8),pointer:: bc(:,:,:)
    integer:: i,j,k

    if (nx.eq.lx .and. ny.eq.ly .and. nz.eq.lz) then
       call cpyb2d_p(nx*ny*nz,b,d)
    else
       call c_f_pointer(b,bc,[nx,ny,nz])
!$omp parallel do private(i,j,k)
       do k=1,nz
          do j=1,ny
!OCL XFILL
             do i=1,nx
                d(1,i,j,k) = dble(bc(i,j,k))
                d(2,i,j,k) = aimag(bc(i,j,k))
             end do
          end do
       end do
    end if
    return    
  end subroutine cpyb2d

  subroutine cpyb2ds(b,d,nx,ny,nz,lx,ly,lz,scl)
    implicit none
    integer,intent(in):: nx,ny,nz,lx,ly,lz
    real(kind=8):: d(2,lx,ly,lz),scl
    type(C_PTR):: b
    complex(kind=8),pointer:: bc(:,:,:)
    integer:: i,j,k

    if (nx.eq.lx .and. ny.eq.ly .and. nz.eq.lz) then
       call cpyb2d_ps(nx*ny*nz,b,d,scl)
    else
       call c_f_pointer(b,bc,[nx,ny,nz])
!$omp parallel do private(i,j,k)
       do k=1,nz
          do j=1,ny
!OCL XFILL
             do i=1,nx
                d(1,i,j,k) = dble(bc(i,j,k))*scl
                d(2,i,j,k) = aimag(bc(i,j,k))*scl
             end do
          end do
       end do
    end if
    return    
  end subroutine cpyb2ds
  !
  subroutine fft3_init(mx,my,mz,lmx,lmy,lmz,fs)
    implicit none
    integer,intent(in):: mx,my,mz,lmx,lmy,lmz
    type (fft3_struct):: fs
    integer:: ierr
    complex(kind=8),pointer:: inout(:,:,:)
    !
    fs%nx = mx
    fs%ny = my
    fs%nz = mz
    fs%nxyz = mx*my*mz
    fs%lx = lmx
    fs%ly = lmy
    fs%lz = lmz
    fs%lxyz = lmx*lmy*lmz
    fs%nmx = mx

    if (.not. fftw_init) then
#ifdef _OPENMP
       ierr = fftw_init_threads()
       call fftw_plan_with_nthreads( omp_get_max_threads() )
#endif
       ierr = fftw_import_wisdom_from_filename( &
            C_CHAR_'xtapp_fftw_wisdom.dat' // C_NULL_CHAR)
       if (ierr .eq. 0) then
          ierr = fftw_import_system_wisdom()
       end if
       fftw_init = .true.
    end if

    fs%inout = fftw_alloc_complex(int(mx*my*mz,C_SIZE_T))
    call c_f_pointer(fs%inout,inout,[mx,my,mz])

    fs%planf = fftw_plan_dft_3d(mz, my, mx, inout, inout, &
         FFTW_FORWARD, FFTW_PATIENT)
    fs%planb = fftw_plan_dft_3d(mz, my, mx, inout, inout, &
         FFTW_BACKWARD, FFTW_PATIENT)

    fs%ifsft = .false.

    fs%rh_trs = .false.
    return
  end subroutine fft3_init

  subroutine fft3_init_rh(mx,my,mz,lmx,lmy,lmz,fs)
    implicit none
    integer,intent(in):: mx,my,mz,lmx,lmy,lmz
    type (fft3_struct):: fs
    integer:: ierr
    complex(kind=8),pointer:: inouth(:,:,:)
    real(kind=8),pointer:: inoutr(:,:,:)
    !
    fs%nx = mx
    fs%ny = my
    fs%nz = mz
    fs%nxyz = mx*my*mz
    fs%lx = lmx
    fs%ly = lmy
    fs%lz = lmz
    fs%lxyz = lmx*lmy*lmz
    fs%nmx = mx/2+1

    if (.not. fftw_init) then
#ifdef _OPENMP
       ierr = fftw_init_threads()
       call fftw_plan_with_nthreads( omp_get_max_threads() )
#endif
       ierr = fftw_import_wisdom_from_filename( &
            C_CHAR_'xtapp_fftw_wisdom.dat' // C_NULL_CHAR)
       if (ierr .eq. 0) then
          ierr = fftw_import_system_wisdom()
       end if
       fftw_init = .true.
    end if

    fs%inout = fftw_alloc_complex(int((mx/2+1)*my*mz,C_SIZE_T))
    call c_f_pointer(fs%inout,inouth,[(mx/2+1),my,mz])
    call c_f_pointer(fs%inout,inoutr,[2*(mx/2+1),my,mz])

    fs%planf = fftw_plan_dft_r2c_3d(mz, my, mx, inoutr, inouth, &
         FFTW_PATIENT)
    fs%planb = fftw_plan_dft_c2r_3d(mz, my, mx, inouth, inoutr, &
         FFTW_PATIENT)

    fs%ifsft = .false.

    fs%rh_trs = .true.
    return
  end subroutine fft3_init_rh

  subroutine fft3_done(fs)
    implicit none
    type (fft3_struct):: fs
    integer:: ierr
    !
    call fftw_destroy_plan(fs%planf)
    call fftw_destroy_plan(fs%planb)
    call fftw_free(fs%inout)
#ifdef FFT3D_FFTW_EXPORT_WISDOM
    ierr = fftw_export_wisdom_to_filename( &
         C_CHAR_'xtapp_fftw_wisdom.new.dat' // C_NULL_CHAR)
#endif

    return
  end subroutine fft3_done

  subroutine fft3_fw(fs,dat,tmp)
#ifdef FFT3D_PROF
    use timeprof
#endif
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(2,fs%lx,fs%ly,fs%lz)
    real(kind=8):: tmp(2,fs%lx,fs%ly,fs%lz)
    real(kind=8):: scl
    complex(kind=8),pointer:: inout(:,:,:)
    real(kind=8),pointer:: inr(:,:,:)

#ifdef FFT3D_PROF
    call tpf_begin(tpf_ffts)
    call tpf_begin(tpf_fftsc)
#endif
    call c_f_pointer(fs%inout,inout,[fs%nmx,fs%ny,fs%nz])

    call cpyd2b(dat,fs%inout,fs%nmx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz)
    if (fs%rh_trs) then
       call c_f_pointer(fs%inout,inr,[2*fs%nmx,fs%ny,fs%nz])
       call fftw_execute_dft_r2c(fs%planf,inr,inout)
    else
       call fftw_execute_dft(fs%planf,inout,inout)
    end if
    scl = 1.0d0/fs%nxyz
    call cpyb2ds(fs%inout,dat,fs%nmx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz,scl)
#ifdef FFT3D_PROF
    call tpf_end(tpf_fftsc)
    call tpf_end(tpf_ffts)
#endif
    return
  end subroutine fft3_fw

  subroutine fft3_bw(fs,dat,tmp)
#ifdef FFT3D_PROF
    use timeprof
#endif
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(2,fs%lx,fs%ly,fs%lz)
    real(kind=8):: tmp(2,fs%lx,fs%ly,fs%lz)
    complex(kind=8),pointer:: inout(:,:,:)
    real(kind=8),pointer:: outr(:,:,:)

#ifdef FFT3D_PROF
    call tpf_begin(tpf_ffts)
    call tpf_begin(tpf_fftsc)
#endif
    call c_f_pointer(fs%inout,inout,[fs%nmx,fs%ny,fs%nz])

    call cpyd2b(dat,fs%inout,fs%nmx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz)
    if (fs%rh_trs) then
       call c_f_pointer(fs%inout,outr,[2*fs%nmx,fs%ny,fs%nz])
       call fftw_execute_dft_c2r(fs%planb,inout,outr)
    else
       call fftw_execute_dft(fs%planb,inout,inout)
    end if
    call cpyb2d(fs%inout,dat,fs%nmx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz)
#ifdef FFT3D_PROF
    call tpf_end(tpf_fftsc)
    call tpf_end(tpf_ffts)
#endif
    return
  end subroutine fft3_bw
  !
  subroutine fft3_fwsx(fs,dat,tmp,ix)
#ifdef FFT3D_PROF
    use timeprof
#endif
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(2,fs%lx,fs%ly,fs%lz)
    real(kind=8):: tmp(2,fs%lx,fs%ly,fs%lz)
    integer:: ix
    real(kind=8):: scl
    complex(kind=8),pointer:: inout(:,:,:)
    real(kind=8),pointer:: inr(:,:,:)

#ifdef FFT3D_PROF
    call tpf_begin(tpf_fftc)
#endif
    call c_f_pointer(fs%inout,inout,[fs%nmx,fs%ny,fs%nz])

    if (ix.eq.1) then
       call cpyd2b(dat,fs%inout,fs%nmx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz)
    else if (ix.eq.-1) then
       call cpyd2b(tmp,fs%inout,fs%nmx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz)
    else
       write(6,*) 'fft3_fwsx: z, abs(ix) != 1'
       stop
    end if

    if (fs%rh_trs) then
       call c_f_pointer(fs%inout,inr,[2*fs%nmx,fs%ny,fs%nz])
       call fftw_execute_dft_r2c(fs%planf,inr,inout)
    else
       call fftw_execute_dft(fs%planf,inout,inout)
    end if

    scl = 1.0d0/fs%nxyz
    if (ix.eq.1) then
       call cpyb2ds(fs%inout,dat,fs%nmx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz,scl)
    else if (ix.eq.-1) then
       call cpyb2ds(fs%inout,tmp,fs%nmx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz,scl)
    end if

#ifdef FFT3D_PROF
    call tpf_end(tpf_fftc)
#endif
    return
  end subroutine fft3_fwsx

  subroutine fft3_bwsx(fs,dat,tmp,ix)
#ifdef FFT3D_PROF
    use timeprof
#endif
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(2,fs%lx,fs%ly,fs%lz)
    real(kind=8):: tmp(2,fs%lx,fs%ly,fs%lz)
    integer:: ix
    complex(kind=8),pointer:: inout(:,:,:)
    real(kind=8),pointer:: outr(:,:,:)

#ifdef FFT3D_PROF
    call tpf_begin(tpf_fftc)
#endif
    call c_f_pointer(fs%inout,inout,[fs%nmx,fs%ny,fs%nz])

    if (ix.eq.1) then
       call cpyd2b(dat,fs%inout,fs%nmx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz)
    else if (ix.eq.-1) then
       call cpyd2b(tmp,fs%inout,fs%nmx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz)
    else
       write(6,*) 'fft3_bwsx: z, abs(ix) != 1'
       stop
    end if

    if (fs%rh_trs) then
       call c_f_pointer(fs%inout,outr,[2*fs%nmx,fs%ny,fs%nz])
       call fftw_execute_dft_c2r(fs%planb,inout,outr)
    else
       call fftw_execute_dft(fs%planb,inout,inout)
    end if

    if (ix.eq.1) then
       call cpyb2d(fs%inout,dat,fs%nmx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz)
    else if (ix.eq.-1) then
       call cpyb2d(fs%inout,tmp,fs%nmx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz)
    end if

#ifdef FFT3D_PROF
    call tpf_end(tpf_fftc)
#endif
    return
  end subroutine fft3_bwsx

end module fft_3d
