!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015,2016 Yoshihide Yoshimoto
!
module wfnc_init
contains
subroutine initwfn(ncomp,nspin,nw,aa,dp,mgs,mssg, &
     myrank,nproc,comm)
  use iounit
  use converter
  use types
  use meshgrid
  use util
  implicit none
  integer:: ncomp,nspin,nw,mssg
  type (meshgrid_struct):: mgs
  real(kind=8):: aa(3,3)

  type (wfn_for_k):: dp(mgs%nks,nspin)

  integer:: myrank,nproc,comm

  real(kind=8),allocatable:: akk(:)
  integer:: nhx,nhy,nhz,iseed,ierr
  integer:: iks,is,ig

  integer:: bkm,ekm

  real(kind=8):: bb(3,3)
  integer:: nks
!SOPTION CONTAREA(nkmary,kg,g_count,g_displ)
  type (irdkpt_struct),pointer:: irdkpt(:)
  integer,pointer:: nkmary(:),kg(:,:,:)
  integer,pointer:: g_count(:,:),g_displ(:,:)

  real(kind=8),parameter:: pi = 3.1415926535897932385d0

  call recp_uvec(aa,bb)
  nks = mgs%nks
  irdkpt => mgs%irdkpt
  nkmary => mgs%nkm
  kg => mgs%kg
  g_count => mgs%g_count
  g_displ => mgs%g_displ
  
  iseed = 456214531

  if (mssg.gt.-1) write(6,'(1x,a)') 'initwfn: iks,nhx,nhy,nhz ='
  do iks=1,nks
     allocate(akk(nkmary(iks)),stat=ierr)
     call chkalloc('initwfn','akk',ierr)

     call kg2akk(bb,irdkpt(iks)%sk,nkmary(iks),kg(:,:,iks),akk)

     nhx = abs(kg(1,1,iks))
     nhy = abs(kg(2,1,iks))
     nhz = abs(kg(3,1,iks))
     do ig=1,nkmary(iks)
        nhx = max(nhx,abs(kg(1,ig,iks)))
        nhy = max(nhy,abs(kg(2,ig,iks)))
        nhz = max(nhz,abs(kg(3,ig,iks)))
     end do
     if (mssg.gt.-1) write(6,'(3x,i4,3(1x,i4))') iks,nhx,nhy,nhz

     bkm = g_displ(myrank,iks)+1
     ekm = g_displ(myrank,iks)+g_count(myrank,iks)
     do is=1,nspin
        call initwfn1k(ncomp,nkmary(iks),nw,kg(:,:,iks),akk,nhx,nhy,nhz,iseed, &
             bkm,ekm,dp(iks,is),comm)
     end do

     deallocate(akk)
  end do
  return
end subroutine initwfn

subroutine initwfn1k(ncomp,nkm,nw,kg,akk,nhx,nhy,nhz,iseed, &
     bkm,ekm,dp,comm)
  use util
  use types
  use mpi
  implicit none
  integer:: ncomp,nw,nkm,bkm,ekm,comm
  real(kind=8):: akk(nkm)
  integer:: kg(3,nkm),nhx,nhy,nhz,iseed

  type (wfn_for_k):: dp

  real(kind=8):: winfac,cut,sum,psum
  complex(kind=8),allocatable:: wf(:,:,:,:)
  complex(kind=8),allocatable:: ctemp(:),pctemp(:)
  integer:: ierr
  integer:: k1,k2,k3,ig,ib,jb
!SOPTION CONTAREA(wfp)
  complex(kind=8),pointer:: wfp(:,:,:)
  real(kind=8):: rlp,aimp
  integer:: ic

  real(kind=8),parameter:: pi = 3.1415926535897932385d0

  !  
  ! from Research Institute for Mathematical Sciences, Kyoto University
  ! Ko^kyu^roku vol. 1032, 21-45 (1998)
  ! Satoru Tanaka, Shuya Kanagawa
  ! pseudo randoum number B using Schrage technique
  !
  integer,parameter:: a = 16807 ! = 7^5
  integer,parameter:: m = 2147483647 ! = 2^31 - 1
  integer,parameter:: r = 2836 ! = m/a
  integer,parameter:: q = 127773 ! = mod(m,a)
  integer:: tmp

  allocate(wf(ncomp,-nhx:nhx,-nhy:nhy,-nhz:nhz), &
       pctemp(nw),ctemp(nw),stat=ierr)
  call chkalloc('initwfn1k','wf',ierr)
  wfp => dp%wf
  wfp = 0.0d0

  ! 27 = 3**3
  cut = sqrt(maxval(akk))
  if (27*nw .lt. nkm) then
     cut = (dble(27*nw)/dble(nkm))**(1.0d0/3.0d0)*cut
  end if

  do ib=1,nw
     do k3=-nhz,nhz
        do k2=-nhy,nhy
           do k1=-nhx,nhx
              do ic=1,ncomp
                 tmp = iseed/q
                 iseed = a*(iseed - q*tmp) - r*(iseed/q)
                 if (iseed.lt.0) then
                    iseed = iseed + m
                 end if
                 rlp = dble(1.0d0)/dble(m)*iseed

                 tmp = iseed/q
                 iseed = a*(iseed - q*tmp) - r*(iseed/q)
                 if (iseed.lt.0) then
                    iseed = iseed + m
                 end if
                 aimp = dble(1.0d0)/dble(m)*iseed

                 wf(ic,k1,k2,k3) = dcmplx(rlp,aimp)
              end do
           end do
        end do
     end do
     do ig=bkm,ekm
        if (sqrt(akk(ig)) .lt. cut) then
           winfac = 0.5d0*( cos(pi/cut*sqrt(akk(ig))) + 1.0d0 )
        else
           winfac = 0.0d0
        end if
        k1=kg(1,ig)
        k2=kg(2,ig)
        k3=kg(3,ig)
        wfp(:,ig,ib) = winfac*wf(:,k1,k2,k3)
     end do
  end do

  ! modified gram-schmidt
  do ib=1,nw
     psum = 0.0
     do ig=bkm,ekm
        do ic=1,ncomp
           psum = psum &
                + dble(wfp(ic,ig,ib))**2 + aimag(wfp(ic,ig,ib))**2
        end do
     end do
     call MPI_ALLREDUCE(psum,sum,1, &
          MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
     sum=1.0d0/sqrt(sum)
     wfp(:,:,ib) = sum*wfp(:,:,ib)
     do jb=ib+1,nw
        pctemp(jb) = (0.0, 0.0)
        do ig=bkm,ekm
           do ic=1,ncomp
              pctemp(jb) = pctemp(jb) + conjg(wfp(ic,ig,ib))*wfp(ic,ig,jb)
           end do
        end do
     end do
     if (ib .lt. nw) then
        call MPI_ALLREDUCE(pctemp(ib+1),ctemp(ib+1),nw-ib, &
             MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
     end if
     do jb=ib+1,nw
        wfp(:,:,jb) =  wfp(:,:,jb) - ctemp(jb)*wfp(:,:,ib)
     end do
  end do

  !
  ! orthonormalization of wave funciton wrt S has not been completed.
  dp%ifothnm = .false.
  !
  ! wave function is randomly generated
  dp%ifrand = .true.

  deallocate(wf,pctemp,ctemp)
  return
end subroutine initwfn1k
end module wfnc_init
