!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015,2016 Yoshihide Yoshimoto
!
module uspp_scrdd
contains
  subroutine makedd(nspin,ndens,bkc,ekc,kqg,vpot,dd,kd,va,asi, &
       laprj,nsi,nkd,nvx,nvy,nvz,dqgs,bspots,comm)
    use bsplpot
    use types
    implicit none
    integer:: nspin,ndens
    integer:: bkc,ekc,kqg(3,bkc:ekc)
    integer:: laprj,nsi,nkd,nvx,nvy,nvz
    real(kind=8):: va,asi(3,nsi)
    complex(kind=8):: dd(laprj,laprj,nsi,nspin)
    integer:: kd(nsi)
    complex(kind=8):: vpot(-nvx:nvx,-nvy:nvy,-nvz:nvz,nspin*ndens)
    type (qg_for_atm):: dqgs(nkd,ndens)
    type (bspot_for_atm):: bspots(nkd)
    integer:: comm

    integer:: is

    if (nspin.eq.1) then
       call makedd_core(ndens,bkc,ekc,kqg,vpot, &
            dd,kd,va,asi,laprj,nsi,nkd,nvx,nvy,nvz, &
            dqgs,bspots,comm)
    else
       do is=1,nspin
          call makedd_core(1,bkc,ekc,kqg,vpot(-nvx,-nvy,-nvz,is), &
               dd(1,1,1,is), &
               kd,va,asi,laprj,nsi,nkd,nvx,nvy,nvz, &
               dqgs,bspots,comm)
       end do
    end if
    return
  end subroutine makedd

  subroutine makedd_core(ndens,bkc,ekc,kqg,vpot,dd,kd,va,asi, &
       laprj,nsi,nkd,nvx,nvy,nvz,dqgs,bspots,comm)
    use bsplpot
    use util
    use types
    use mpi
    use timeprof
#ifdef _OPENMP
    use omp_lib
#endif
    implicit none
    integer:: ndens
    integer:: bkc,ekc,kqg(3,bkc:ekc)
    integer:: laprj,nsi,nkd,nvx,nvy,nvz
    real(kind=8):: va,asi(3,nsi)
    complex(kind=8):: dd(laprj,laprj,nsi)
    integer:: kd(nsi)
    complex(kind=8):: vpot(-nvx:nvx,-nvy:nvy,-nvz:nvz,ndens)
    type (qg_for_atm):: dqgs(nkd,ndens)
    type (bspot_for_atm):: bspots(nkd)
    integer:: comm
    !
    real(kind=8),parameter:: pi2 = 2.0d0*3.14159265358979323846d0
    integer,parameter:: blkkc = 100, blkthr = 200
    !
    complex(kind=8):: cwrk(bkc:ekc)
!SOPTION CONTAREA(qg)
    complex(kind=8),pointer:: qg(:,:)
    complex(kind=8):: pdd(laprj,laprj,nsi)
    !
    integer:: ierr,ikd,it,i1,i2,ig,itt
    integer:: ip1,ip2
    integer:: k1max,k2max,k3max,k1,k2,k3,l1,m1,l2,m2
    real(kind=8):: t1,t2,t3,ph
    integer:: nsikd,itk,itk2it(nsi)
    integer:: bbkc,bekc
    logical:: ifblock
    integer:: idens

    call tpf_begin(tpf_mkdd)
    !
#ifdef _OPENMP
    if (ekc-bkc+1 .gt. omp_get_max_threads()*blkthr) then
       ifblock = .true.
    else
       ifblock = .false.
    end if
#else
    ifblock = .true.
#endif
    !
    ! check
    k1max=0
    k2max=0
    k3max=0
    do ig=bkc,ekc
       k1max=max(k1max,abs(kqg(1,ig)))
       k2max=max(k2max,abs(kqg(2,ig)))
       k3max=max(k3max,abs(kqg(3,ig)))
    end do
    !
    if ((k1max.gt.nvx).or.(k2max.gt.nvy).or.(k3max.gt.nvz)) then
       write(6,*)'********  ERROR  MAKEDD  *******'
       write(6,*)'K?MAX=',k1max,k2max,k3max
       write(6,*)'NV?=',nvx,nvy,nvz
       stop
    end if

    pdd=0.0d0
    !
    do idens=1,ndens
       dd = 0.0d0
       do itt=1,nkd
          if (bspots(itt)%ifncpp) then
             cycle
          end if
          qg => dqgs(itt,idens)%qg

          nsikd = 0
          do it=1, nsi
             if (kd(it).eq.itt) then
                nsikd = nsikd + 1
                itk2it(nsikd) = it
             end if
          end do

          if (ifblock) then
!$omp parallel do private (it,ig,k1,k2,k3,l1,i1,m1,l2,i2,m2,t1,t2,t3,ph), &
!$omp private(itk,bbkc,bekc,ip1,ip2),reduction(+:dd)
             do bbkc=bkc,ekc,blkkc
                bekc=min(bbkc+blkkc-1,ekc)
                do itk=1,nsikd
                   it=itk2it(itk)
                   t1=asi(1,it)
                   t2=asi(2,it)
                   t3=asi(3,it)
                   do ig=bbkc,bekc
                      k1=-kqg(1,ig)
                      k2=-kqg(2,ig)
                      k3=-kqg(3,ig)
                      ph=pi2*(t1*k1+t2*k2+t3*k3)
                      cwrk(ig)=dcmplx(cos(ph),sin(ph))*vpot(k1,k2,k3,idens)
                   end do
                   k1=0
                   do ip1 = 1, bspots(itt)%nprj
                      l1 = bspots(itt)%ip2l(ip1)
                      i1 = bspots(itt)%ip2r(ip1)
                      m1 = bspots(itt)%ip2m(ip1)
                      do ip2 = 1, bspots(itt)%l2pt(l1)
                         l2 = bspots(itt)%ip2l(ip2)
                         i2 = bspots(itt)%ip2r(ip2)
                         m2 = bspots(itt)%ip2m(ip2)
                         if (.not.( (l1.eq.l2).and. &
                              ((i2.gt.i1) .or. ((i1.eq.i2).and.(m2.gt.m1))))) then
                            k1=k1+1
                            do ig=bbkc,bekc
                               dd(ip1,ip2,it) = dd(ip1,ip2,it) &
                                    +qg(ig,k1)*cwrk(ig)*va
                            end do
                            !
                            ! (m2,i2) is conjugate of (m1,i1)
                            !
                            if (ip1.eq.ip2) then
                               dd(ip1,ip1,it) = dble(dd(ip1,ip1,it))
                            else
                               dd(ip2,ip1,it) = conjg(dd(ip1,ip2,it))
                            end if
                         end if
                      end do
                   end do
                end do
             end do
          else
!POPTION INDEP(dd)
!$omp parallel do private (it,ig,k1,k2,k3,l1,i1,m1,l2,i2,m2,t1,t2,t3,ph), &
!$omp private(itk,cwrk,ip1,ip2)
             do itk=1,nsikd
                it=itk2it(itk)
                t1=asi(1,it)
                t2=asi(2,it)
                t3=asi(3,it)
                do ig=bkc,ekc
                   k1=-kqg(1,ig)
                   k2=-kqg(2,ig)
                   k3=-kqg(3,ig)
                   ph=pi2*(t1*k1+t2*k2+t3*k3)
                   cwrk(ig)=dcmplx(cos(ph),sin(ph))*vpot(k1,k2,k3,idens)
                end do
                k1=0
                do ip1 = 1, bspots(itt)%nprj
                   l1 = bspots(itt)%ip2l(ip1)
                   i1 = bspots(itt)%ip2r(ip1)
                   m1 = bspots(itt)%ip2m(ip1)
                   do ip2 = 1, bspots(itt)%l2pt(l1)
                      l2 = bspots(itt)%ip2l(ip2)
                      i2 = bspots(itt)%ip2r(ip2)
                      m2 = bspots(itt)%ip2m(ip2)
                      if (.not.( (l1.eq.l2).and. &
                           ((i2.gt.i1) .or. ((i1.eq.i2).and.(m2.gt.m1))))) then
                         k1=k1+1
                         do ig=bkc,ekc
                            dd(ip1,ip2,it) = dd(ip1,ip2,it) &
                                 +qg(ig,k1)*cwrk(ig)*va
                         end do
                         !
                         ! (m2,i2) can be equal to (m1,i1)
                         !
                         if (ip1.eq.ip2) then
                            dd(ip1,ip1,it) = dble(dd(ip1,ip1,it))
                         else
                            dd(ip2,ip1,it) = conjg(dd(ip1,ip2,it))
                         end if
                      end if
                   end do
                end do
             end do
          end if
       end do
       pdd = pdd + dd
    end do
    call tpf_begin(tpf_mpi)
    call MPI_ALLREDUCE(pdd,dd,laprj**2*nsi, &
         MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
    call tpf_end(tpf_mpi)
    do it=1,nsi
       ikd=kd(it)
       do ip1=1,bspots(ikd)%nprj
          l1 = bspots(ikd)%ip2l(ip1)
          i1 = bspots(ikd)%ip2r(ip1)
          m1 = bspots(ikd)%ip2m(ip1)
          do ip2=bspots(ikd)%l2ph(l1),bspots(ikd)%l2pt(l1)
             i2 = bspots(ikd)%ip2r(ip2)
             m2 = bspots(ikd)%ip2m(ip2)
             if (m1.eq.m2) then
                dd(ip1,ip2,it)=dd(ip1,ip2,it)+bspots(ikd)%ddi(i1,i2,l1)
             end if
          end do
       end do
    end do
    !
    call tpf_end(tpf_mkdd)
    return
  end subroutine makedd_core

  subroutine mkinidd(ndens,bb,bkc,ekc,kqg,vpot,dd,kd,va,asi, &
       laprj,nsi,nkd,nvx,nvy,nvz,nspin,bspotfc,bspots,comm)
    use bsplpot
    use types
    use converter
    implicit none
    integer:: ndens
    real(kind=8):: bb(3,3)
    integer:: bkc,ekc,kqg(3,bkc:ekc)
    integer:: laprj,nsi,nkd,nvx,nvy,nvz,nspin
    real(kind=8):: va,asi(3,nsi)
    complex(kind=8):: dd(laprj,laprj,nsi,nspin)
    integer:: kd(nsi)
    complex(kind=8):: vpot(-nvx:nvx,-nvy:nvy,-nvz:nvz,nspin*ndens)
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    integer:: comm

    real(kind=8),allocatable:: qgk(:,:)
!SOPTION CONTAREA(qg)
    complex(kind=8),pointer:: qg(:,:)
    type (qg_for_atm),allocatable:: dqgs(:,:)

    integer:: ikd,nob,ierr
    integer:: idens

    allocate(dqgs(nkd,ndens),stat=ierr)
    call chkalloc('mkinidd','qgs',ierr)
    do idens=1,ndens
       do ikd=1,nkd
          nob=bspots(ikd)%nob
          allocate(dqgs(ikd,idens)%qg(bkc:ekc,nob*(nob+1)/2),stat=ierr)
          call chkalloc('mkinidd','qg',ierr)
       end do
    end do
    allocate(qgk(4,bkc:ekc),stat=ierr)
    call chkalloc('mkinidd','qgk',ierr)
    call kqg2qgk(bb,ekc-bkc+1,kqg,qgk)
    do idens=1,ndens
       do ikd=1,nkd
          qg => dqgs(ikd,idens)%qg
          nob=bspots(ikd)%nob
          call evqg(bspotfc,bspots(ikd),ekc-bkc+1,qgk,va,nob,idens,qg)
       end do
    end do
    deallocate(qgk)

    call makedd(nspin,ndens,bkc,ekc,kqg,vpot,dd,kd,va,asi, &
         laprj,nsi,nkd,nvx,nvy,nvz,dqgs,bspots,comm)

    do idens=1,ndens
       do ikd=1,nkd
          deallocate(dqgs(ikd,idens)%qg)
       end do
    end do
    deallocate(dqgs)
    
  end subroutine mkinidd
end module uspp_scrdd
