!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015-2017 Yoshihide Yoshimoto
!
module nonlocal_proj
  type nlprj_struct
     integer:: nprj
     integer,pointer:: ip2t(:),ip2l(:),ip2r(:),ip2m(:),ip2h(:)
     complex(kind=8),pointer:: prj(:,:)
     real(kind=8),pointer:: prj_r2(:,:,:)
     integer:: mode
  end type nlprj_struct
contains
  subroutine nlprj_init(lpnum,maxr,nkd,bspots,nsi,kd,asi,bkm,ekm, &
       bb,sk,kg,vpj,nlpjs,ifinvsym)
    use bsplpot
    use util
    use timeprof
    implicit none
    integer,intent(in):: lpnum,maxr,nkd,nsi,bkm,ekm
    real(kind=8),intent(in):: bb(3,3),sk(3),vpj(bkm:ekm,maxr,lpnum,nkd)
    integer,intent(in)::  kg(3,bkm:ekm)
    real(kind=8),intent(in):: asi(3,nsi)
    integer,intent(in):: kd(nsi)
    type(bspot_for_atm):: bspots(nkd)
    type(nlprj_struct):: nlpjs
    logical,optional,intent(in):: ifinvsym

    complex(kind=8),parameter:: ci = (0.0d0, 1.0d0)
    
!SOPTION CONTAREA(ip2t,ip2l,ip2r,ip2m,ip2h)
    integer,pointer:: ip2t(:),ip2l(:),ip2r(:),ip2m(:),ip2h(:)
!SOPTION CONTAREA(prj,prj_r2)
    complex(kind=8),pointer:: prj(:,:)
    real(kind=8),pointer:: prj_r2(:,:,:)
    complex(kind=8),allocatable:: ylm(:,:)
    complex(kind=8),allocatable:: strfac(:,:)
    complex(kind=8):: tmp
    complex(kind=8):: ctemp
    complex(kind=8):: ctemp2
    integer:: it,ikd,il,ir,im,nprj,ip,iph,ig,ipa
    integer:: mode
    integer:: lmax
    integer:: bkms,ekms,ncomp

    call tpf_begin(tpf_npj)

    mode = 1
    if (present(ifinvsym)) mode = 2

    ncomp = maxval(bspots(:)%ncomp)
    if (minval(bspots(:)%ncomp).ne.ncomp) then
       write(6,*) 'nlprj_init(): inconsistent bspots%ncomp'
       stop
    end if
    if (mode.eq.2 .and. ncomp.eq.2) then
       write(6,*) 'nlprj_init(): inversion symmetry for ncomp = 2 is not supported.'
       stop
    end if
    bkms = (bkm-1)*ncomp+1
    ekms = ekm*ncomp

    nprj = 0
    do it=1,nsi
       ikd=kd(it)
       nprj = nprj + bspots(ikd)%nprj
    end do
    nlpjs%nprj = nprj
    allocate(nlpjs%ip2t(nprj),nlpjs%ip2l(nprj),nlpjs%ip2r(nprj), &
         nlpjs%ip2m(nprj),nlpjs%ip2h(nprj))
    select case (mode)
    case (1)
       allocate(nlpjs%prj(bkms:ekms,nprj))
    case (2)
       allocate(nlpjs%prj_r2(bkms:ekms,nprj,2))
    case default
       write(6,*) 'invalid mode in nlprj_init()'
       stop
    end select
    ip2t => nlpjs%ip2t
    ip2l => nlpjs%ip2l
    ip2r => nlpjs%ip2r
    ip2m => nlpjs%ip2m
    ip2h => nlpjs%ip2h
    ip = 0
    do it=1,nsi
       ikd=kd(it)
       iph = ip
       do ipa=1,bspots(ikd)%nprj
          ip = ip + 1
          ip2t(ip) = it
          ip2l(ip) = bspots(ikd)%ip2l(ipa)
          ip2r(ip) = bspots(ikd)%ip2r(ipa)
          ip2m(ip) = bspots(ikd)%ip2m(ipa)
          ip2h(ip) = iph
       end do
    end do
    lmax = getlmax(nkd,bspots)
    allocate(ylm(bkms:ekms,ncomp*lmax**2),strfac(bkm:ekm,nsi))
    call getylb(ekm-bkm+1,sk,kg,bb,ylm,lmax,ncomp)
    call mkstrfac(nsi,asi,ekm-bkm+1,kg,strfac)

    select case (mode)
    case (1)
       nlpjs%mode = 1
       prj => nlpjs%prj
!$omp parallel do private(ip,it,ikd,il,ir,im,ctemp,ig,tmp)
       do ip=1,nprj
          it = ip2t(ip)
          ikd = kd(it)
          il = ip2l(ip)
          ir = ip2r(ip)
          im = ip2m(ip)
          ctemp = (-ci)**bspots(ikd)%l2lq(il)
          if (ncomp.eq.1) then
             do ig=bkm,ekm
                tmp = ylm(ig,im)*vpj(ig,ir,il,ikd)
                prj(ig,ip) = tmp*ctemp*strfac(ig,it)
             end do
          else
             do ig=bkm,ekm
                tmp = ylm(2*ig-1,im)*vpj(ig,ir,il,ikd)
                prj(2*ig-1,ip) = tmp*ctemp*strfac(ig,it)
                tmp = ylm(2*ig,im)*vpj(ig,ir,il,ikd)
                prj(2*ig,ip) = tmp*ctemp*strfac(ig,it)
             end do
          end if
       end do
       nullify(prj)
    case (2)
       nlpjs%mode = 2
       prj_r2 => nlpjs%prj_r2
!$omp parallel do private(ip,it,ikd,il,ir,im,ctemp,ig,tmp,ctemp2)
       do ip=1,nprj
          it = ip2t(ip)
          ikd = kd(it)
          il = ip2l(ip)
          ir = ip2r(ip)
          im = ip2m(ip)
          ctemp = (-ci)**bspots(ikd)%l2lq(il)
          do ig=bkm,ekm
             tmp = ylm(ig,im)*vpj(ig,ir,il,ikd)
             ctemp2 = tmp*ctemp*strfac(ig,it)
             prj_r2(ig,ip,1) = dble(ctemp2)
             prj_r2(ig,ip,2) = aimag(ctemp2)
          end do
       end do
       nullify(prj_r2)
    end select

    deallocate(strfac,ylm)
    nullify(ip2t,ip2l,ip2r,ip2m,ip2h)

    call tpf_end(tpf_npj)
    return
  end subroutine nlprj_init

  subroutine nlprj_apply(laprj,nkd,bspots,kd,nsi,bkm,ekm, &
       va,dd,nw,p,hp,sp,bprj,eprj,dcprj,nlpjs,comm)
    use xmpi
    use arch_nlprj
    use bsplpot
    use timeprof
    implicit none
    integer,intent(in):: laprj,nkd,nsi,bkm,ekm,nw,bprj,eprj
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: va
    complex(kind=8),intent(in):: dd(laprj,laprj,nsi)
    complex(kind=8),intent(in):: p(bkm:ekm,nw)
    complex(kind=8),intent(inout):: hp(bkm:ekm,nw),sp(bkm:ekm,nw)
    complex(kind=8):: dcprj(bprj:eprj,nw)
    type(bspot_for_atm):: bspots(nkd)
    type(nlprj_struct):: nlpjs
    integer:: comm

    complex(kind=8),parameter:: ci = (0.0d0, 1.0d0)
    
!SOPTION CONTAREA(ip2t,ip2l,ip2r,ip2m,ip2h)
    integer,pointer:: ip2t(:),ip2l(:),ip2r(:),ip2m(:),ip2h(:)
!SOPTION CONTAREA(prj)
    complex(kind=8),pointer:: prj(:,:)
    complex(kind=8),allocatable:: coefprj1(:,:), coefprj2(:,:)
    complex(kind=8),allocatable:: coefprj3(:,:)
    integer:: nprj
    real(kind=8):: cnst
    integer:: it,ikd,ib,ip1,ip2,ip1a,ip2a
    integer:: l1,ir1,m1,l2,ir2,m2,ierr

    call tpf_begin(tpf_npj)

    ip2t => nlpjs%ip2t
    ip2l => nlpjs%ip2l
    ip2r => nlpjs%ip2r
    ip2m => nlpjs%ip2m
    ip2h => nlpjs%ip2h
    prj => nlpjs%prj
    nprj = nlpjs%nprj

    cnst=1.0d0/sqrt(va)
    allocate(coefprj1(nprj,nw),coefprj2(nprj,nw), &
         coefprj3(nprj,nw),stat=ierr)
    call chkalloc('nlprj_apply','coef',ierr)
    call ovlmat(bkm,ekm,nprj,nw,prj,p,coefprj2)

    !call MPI_ALLREDUCE(coefprj2,coefprj3,nprj*nw, &
    !     MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_allzsum(coefprj2,coefprj3,nprj*nw,comm,ierr)

    coefprj3 = cnst*coefprj3
    dcprj(bprj:eprj,:) = coefprj3(bprj:eprj,:)

!$omp parallel do private(ib,ip1,it,ikd,l1,ir1,m1,ip1a,ip2a,ip2,l2,ir2,m2)
    do ib=1,nw
       do ip1=1,nprj
          it = ip2t(ip1)
          ikd = kd(it)
          l1 = ip2l(ip1)
          ir1 = ip2r(ip1)
          m1 = ip2m(ip1)
          ip1a = ip1 - ip2h(ip1)
          coefprj1(ip1,ib) = (0.0d0,0.0d0)
          coefprj2(ip1,ib) = (0.0d0,0.0d0)
          do ip2a = 1,bspots(ikd)%nprj
             ip2 = ip2h(ip1) + ip2a
             l2 = bspots(ikd)%ip2l(ip2a)
             ir2= bspots(ikd)%ip2r(ip2a)
             m2 = bspots(ikd)%ip2m(ip2a)
             coefprj1(ip1,ib) = coefprj1(ip1,ib) + &
                  dd(ip1a,ip2a,it)*coefprj3(ip2,ib)
             if (m1.eq.m2) then
                coefprj2(ip1,ib) = coefprj2(ip1,ib) + &
                     bspots(ikd)%qq(ir1,ir2,l1)*coefprj3(ip2,ib)
             end if
          end do
          coefprj1(ip1,ib) = cnst*coefprj1(ip1,ib)
          coefprj2(ip1,ib) = cnst*coefprj2(ip1,ib)
       end do
    end do
    call rotprjs(bkm,ekm,nprj,nw,coefprj1,prj,hp)
    call rotprjs(bkm,ekm,nprj,nw,coefprj2,prj,sp)
    
    deallocate(coefprj1,coefprj2,coefprj3)
    nullify(ip2t,ip2l,ip2r,ip2m,ip2h,prj)

    call tpf_end(tpf_npj)
    return
  end subroutine nlprj_apply

  subroutine nlprj_apyi(laprj,nkd,bspots,kd,nsi,bkm,ekm, &
       va,dd,nw,p,hp,sp,bprj,eprj,dcprj,nlpjs,comm)
    use xmpi
    use arch_nlprji
    use bsplpot
    use timeprof
    implicit none
    integer,intent(in):: laprj,nkd,nsi,bkm,ekm,nw,bprj,eprj
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: va
    complex(kind=8),intent(in):: dd(laprj,laprj,nsi)
    real(kind=8),intent(in):: p(bkm:ekm,nw)
    real(kind=8),intent(inout):: hp(bkm:ekm,nw),sp(bkm:ekm,nw)
    complex(kind=8):: dcprj(bprj:eprj,nw)
    type(bspot_for_atm):: bspots(nkd)
    type(nlprj_struct):: nlpjs
    integer:: comm

    complex(kind=8),parameter:: ci = (0.0d0, 1.0d0)
    
!SOPTION CONTAREA(ip2t,ip2l,ip2r,ip2m,ip2h)
    integer,pointer:: ip2t(:),ip2l(:),ip2r(:),ip2m(:),ip2h(:)
!SOPTION CONTAREA(prj_r2)
    real(kind=8),pointer:: prj_r2(:,:,:)
    complex(kind=8),allocatable:: coefprj1(:,:), coefprj2(:,:)
    complex(kind=8),allocatable:: coefprj3(:,:)
    integer:: nprj
    real(kind=8):: cnst
    integer:: it,ikd,ib,ip1,ip2,ip1a,ip2a
    integer:: l1,ir1,m1,l2,ir2,m2,ierr

    call tpf_begin(tpf_npj)

    ip2t => nlpjs%ip2t
    ip2l => nlpjs%ip2l
    ip2r => nlpjs%ip2r
    ip2m => nlpjs%ip2m
    ip2h => nlpjs%ip2h
    prj_r2 => nlpjs%prj_r2
    nprj = nlpjs%nprj

    cnst=1.0d0/sqrt(va)
    allocate(coefprj1(nprj,nw),coefprj2(nprj,nw),coefprj3(nprj,nw))
    call ovlmat(bkm,ekm,nprj,nw,prj_r2,p,coefprj2)

    !call MPI_ALLREDUCE(coefprj2,coefprj3,nprj*nw, &
    !     MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_allzsum(coefprj2,coefprj3,nprj*nw,comm,ierr)

    coefprj3 = cnst*coefprj3
    dcprj(bprj:eprj,:) = coefprj3(bprj:eprj,:)

!$omp parallel do private(it,ikd,ip1,ip2,l1,ir1,m1,l2,ir2,m2,ip1a,ip2a)
    do ib=1,nw
       do ip1=1,nprj
          it = ip2t(ip1)
          ikd = kd(it)
          l1 = ip2l(ip1)
          ir1 = ip2r(ip1)
          m1 = ip2m(ip1)
          ip1a = ip1 - ip2h(ip1)
          coefprj1(ip1,ib) = (0.0d0,0.0d0)
          coefprj2(ip1,ib) = (0.0d0,0.0d0)
          do ip2a=1,bspots(ikd)%nprj
             ip2 = ip2h(ip1) + ip2a
             l2 = bspots(ikd)%ip2l(ip2a)
             ir2 = bspots(ikd)%ip2r(ip2a)
             m2 = bspots(ikd)%ip2m(ip2a)
             coefprj1(ip1,ib) = coefprj1(ip1,ib) + &
                  dd(ip1a,ip2a,it)*coefprj3(ip2,ib)
             if (m1.eq.m2) then
                coefprj2(ip1,ib) = coefprj2(ip1,ib) + &
                     bspots(ikd)%qq(ir1,ir2,l1)*coefprj3(ip2,ib)
             end if
          end do
          coefprj1(ip1,ib) = cnst*coefprj1(ip1,ib)
          coefprj2(ip1,ib) = cnst*coefprj2(ip1,ib)
       end do
    end do
    call rotprjs(bkm,ekm,nprj,nw,coefprj1,prj_r2,hp)
    call rotprjs(bkm,ekm,nprj,nw,coefprj2,prj_r2,sp)
    
    deallocate(coefprj1,coefprj2,coefprj3)
    nullify(ip2t,ip2l,ip2r,ip2m,ip2h,prj_r2)

    call tpf_end(tpf_npj)
    return
  end subroutine nlprj_apyi

  subroutine nlprj_apyg(laprj,nkd,bspots,kd,nsi,bkm,ekm, &
       va,dd,nw,p,hp,sp,bprj,eprj,dcprj,nlpjs,comm)
    use xmpi
    use arch_nlprjg
    use bsplpot
    use timeprof
    implicit none
    integer,intent(in):: laprj,nkd,nsi,bkm,ekm,nw,bprj,eprj
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: va
    real(kind=8),intent(in):: dd(laprj,laprj,nsi)
    complex(kind=8),intent(in):: p(bkm:ekm,nw)
    complex(kind=8),intent(inout):: hp(bkm:ekm,nw),sp(bkm:ekm,nw)
    complex(kind=8):: dcprj(bprj:eprj,nw)
    type(bspot_for_atm):: bspots(nkd)
    type(nlprj_struct):: nlpjs
    integer:: comm

!SOPTION CONTAREA(ip2t,ip2l,ip2r,ip2m,ip2h)
    integer,pointer:: ip2t(:),ip2l(:),ip2r(:),ip2m(:),ip2h(:)
!SOPTION CONTAREA(prj)
    complex(kind=8),pointer:: prj(:,:)
    real(kind=8),allocatable:: coefprj1(:,:), coefprj2(:,:)
    real(kind=8),allocatable:: coefprj3(:,:)
    integer:: nprj
    real(kind=8):: cnst
    integer:: it,ikd,ib,ip1,ip2,ip1a,ip2a
    integer:: l1,ir1,m1,l2,ir2,m2,ierr

    call tpf_begin(tpf_npj)

    ip2t => nlpjs%ip2t
    ip2l => nlpjs%ip2l
    ip2r => nlpjs%ip2r
    ip2m => nlpjs%ip2m
    ip2h => nlpjs%ip2h
    prj => nlpjs%prj
    nprj = nlpjs%nprj

    cnst=1.0d0/sqrt(va)
    allocate(coefprj1(nprj,nw),coefprj2(nprj,nw), &
         coefprj3(nprj,nw),stat=ierr)
    call chkalloc('nlprj_apply','coef',ierr)
    call ovlmat(bkm,ekm,nprj,nw,prj,p,coefprj2)

    !call MPI_ALLREDUCE(coefprj2,coefprj3,nprj*nw, &
    !     MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_alldsum(coefprj2,coefprj3,nprj*nw,comm,ierr)

    coefprj3 = cnst*coefprj3
    dcprj(bprj:eprj,:) = coefprj3(bprj:eprj,:)

!$omp parallel do private(it,ikd,ip1,ip2,l1,ir1,m1,l2,ir2,m2,ip1a,ip2a)
    do ib=1,nw
       do ip1=1,nprj
          it = ip2t(ip1)
          ikd = kd(it)
          l1 = ip2l(ip1)
          ir1 = ip2r(ip1)
          m1 = ip2m(ip1)
          ip1a = ip1 - ip2h(ip1)
          coefprj1(ip1,ib) = 0.0d0
          coefprj2(ip1,ib) = 0.0d0
          do ip2a=1,bspots(ikd)%nprj
             ip2 = ip2h(ip1) + ip2a
             l2 = bspots(ikd)%ip2l(ip2a)
             ir2 = bspots(ikd)%ip2r(ip2a)
             m2 = bspots(ikd)%ip2m(ip2a)
             coefprj1(ip1,ib) = coefprj1(ip1,ib) + &
                  dd(ip1a,ip2a,it)*coefprj3(ip2,ib)
             if (m1.eq.m2) then
                coefprj2(ip1,ib) = coefprj2(ip1,ib) + &
                     bspots(ikd)%qq(ir1,ir2,l1)*coefprj3(ip2,ib)
             end if
          end do
          coefprj1(ip1,ib) = cnst*coefprj1(ip1,ib)
          coefprj2(ip1,ib) = cnst*coefprj2(ip1,ib)
       end do
    end do
    call rotprjs(bkm,ekm,nprj,nw,coefprj1,prj,hp)
    call rotprjs(bkm,ekm,nprj,nw,coefprj2,prj,sp)
    
    deallocate(coefprj1,coefprj2,coefprj3)
    nullify(ip2t,ip2l,ip2r,ip2m,ip2h,prj)

    call tpf_end(tpf_npj)
    return
  end subroutine nlprj_apyg

  subroutine nlprj_apy1s(nkd,bspots,kd,nsi,bkm,ekm, &
       va,p,sp,bprj,eprj,dcprj,nlpjs,comm)
    use xmpi
    use bsplpot
    use timeprof
    implicit none
    integer,intent(in):: nkd,nsi,bkm,ekm,bprj,eprj
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: va
    complex(kind=8),intent(in):: p(bkm:ekm)
    complex(kind=8),intent(inout):: sp(bkm:ekm)
    complex(kind=8):: dcprj(bprj:eprj)
    type(bspot_for_atm):: bspots(nkd)
    type(nlprj_struct):: nlpjs
    integer:: comm

    complex(kind=8),parameter:: ci = (0.0d0, 1.0d0)
    
!SOPTION CONTAREA(ip2t,ip2l,ip2r,ip2m,ip2h)
    integer,pointer:: ip2t(:),ip2l(:),ip2r(:),ip2m(:),ip2h(:)
!SOPTION CONTAREA(prj)
    complex(kind=8),pointer:: prj(:,:)
    complex(kind=8),allocatable:: coefprj(:), tcfpj(:)
    complex(kind=8),allocatable:: coefprj2(:)
    integer:: nprj
    real(kind=8):: cnst
    integer:: it,ikd,ip,ip2,ip2a
    integer:: l1,ir1,m1,l2,ir2,m2,ig,ierr

    call tpf_begin(tpf_npj)

    ip2t => nlpjs%ip2t
    ip2l => nlpjs%ip2l
    ip2r => nlpjs%ip2r
    ip2m => nlpjs%ip2m
    ip2h => nlpjs%ip2h
    prj => nlpjs%prj
    nprj = nlpjs%nprj

    cnst=1.0d0/sqrt(va)
    allocate(coefprj(nprj),coefprj2(nprj),tcfpj(nprj))
    tcfpj=(0.0d0, 0.0d0)
!$omp parallel do private(ig)
    do ip=1,nprj
!SOPTION NOLOOPINTERCHANGE
!CDIR NOLOOPCHG
       do ig=bkm,ekm
          tcfpj(ip) = tcfpj(ip) + conjg(prj(ig,ip))*p(ig)
       end do
    end do

    !call MPI_ALLREDUCE(tcfpj,coefprj2,nprj, &
    !     MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_allzsum(tcfpj,coefprj2,nprj,comm,ierr)

    coefprj2 = cnst*coefprj2
    dcprj(bprj:eprj) = coefprj2(bprj:eprj)

!$omp parallel do private(ip,it,ikd,l1,ir1,m1,ip2a,ip2,l2,ir2,m2)
    do ip=1,nprj
       it = ip2t(ip)
       ikd = kd(it)
       l1 = ip2l(ip)
       ir1 = ip2r(ip)
       m1 = ip2m(ip)
       coefprj(ip) = (0.0d0,0.0d0)
       do ip2a=1,bspots(ikd)%nprj
          ip2 = ip2h(ip) + ip2a
          l2 = bspots(ikd)%ip2l(ip2a)
          ir2 = bspots(ikd)%ip2r(ip2a)
          m2 = bspots(ikd)%ip2m(ip2a)
          if (m1.eq.m2) then
             coefprj(ip) = coefprj(ip) + &
                  bspots(ikd)%qq(ir1,ir2,l1)*coefprj2(ip2)
          end if
       end do
       coefprj(ip) = cnst*coefprj(ip)
    end do
!$omp parallel do private(ig),reduction(+:sp)
!POPTION PSUM(sp)
    do ip=1,nprj
!SOPTION NOLOOPINTERCHANGE
!CDIR NOLOOPCHG
       do ig=bkm,ekm
          sp(ig) = sp(ig) + coefprj(ip)*prj(ig,ip)
       end do
    end do
    
    deallocate(coefprj,coefprj2,tcfpj)
    nullify(ip2t,ip2l,ip2r,ip2m,ip2h,prj)

    call tpf_end(tpf_npj)
    return
  end subroutine nlprj_apy1s

  subroutine nlprj_apy1si(nkd,bspots,kd,nsi,bkm,ekm, &
       va,p,sp,bprj,eprj,dcprj,nlpjs,comm)
    use xmpi
    use bsplpot
    use timeprof
    implicit none
    integer,intent(in):: nkd,nsi,bkm,ekm,bprj,eprj
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: va
    real(kind=8),intent(in):: p(bkm:ekm)
    real(kind=8),intent(inout):: sp(bkm:ekm)
    complex(kind=8):: dcprj(bprj:eprj)
    type(bspot_for_atm):: bspots(nkd)
    type(nlprj_struct):: nlpjs
    integer:: comm

    complex(kind=8),parameter:: ci = (0.0d0, 1.0d0)
    
!SOPTION CONTAREA(ip2t,ip2l,ip2r,ip2m,ip2h)
    integer,pointer:: ip2t(:),ip2l(:),ip2r(:),ip2m(:),ip2h(:)
!SOPTION CONTAREA(prj)
    real(kind=8),pointer:: prj_r2(:,:,:)
    complex(kind=8),allocatable:: coefprj(:),tcfpj(:)
    complex(kind=8),allocatable:: coefprj2(:)
    integer:: nprj
    real(kind=8):: cnst
    integer:: it,ikd,ip,ip2,ip2a
    integer:: l1,ir1,m1,l2,ir2,m2,ig,ierr
    complex:: ctemp

    call tpf_begin(tpf_npj)

    ip2t => nlpjs%ip2t
    ip2l => nlpjs%ip2l
    ip2r => nlpjs%ip2r
    ip2m => nlpjs%ip2m
    ip2h => nlpjs%ip2h
    prj_r2 => nlpjs%prj_r2
    nprj = nlpjs%nprj

    cnst=1.0d0/sqrt(va)
    allocate(coefprj(nprj),coefprj2(nprj),tcfpj(nprj))
    tcfpj=(0.0d0, 0.0d0)
!$omp parallel do private(ig)
    do ip=1,nprj
!SOPTION NOLOOPINTERCHANGE
!CDIR NOLOOPCHG
       do ig=bkm,ekm
          ctemp = dcmplx(prj_r2(ig,ip,1),prj_r2(ig,ip,2))
          tcfpj(ip) = tcfpj(ip) + conjg(ctemp)*p(ig)
       end do
    end do

    !call MPI_ALLREDUCE(tcfpj,coefprj2,nprj, &
    !     MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_allzsum(tcfpj,coefprj2,nprj,comm,ierr)

    coefprj2 = cnst*coefprj2
    dcprj(bprj:eprj) = coefprj2(bprj:eprj)

!$omp parallel do private(it,ikd,ip2,l1,ir1,m1,l2,ir2,m2,ip2a)
    do ip=1,nprj
       it = ip2t(ip)
       ikd = kd(it)
       l1 = ip2l(ip)
       ir1 = ip2r(ip)
       m1 = ip2m(ip)
       coefprj(ip) = (0.0d0,0.0d0)
       do ip2a=1,bspots(ikd)%nprj
          l2 = bspots(ikd)%ip2l(ip2a)
          ir2 = bspots(ikd)%ip2r(ip2a)
          m2 = bspots(ikd)%ip2m(ip2a)
          ip2 = ip2h(ip) + ip2a
          if (m1.eq.m2) then
             coefprj(ip) = coefprj(ip) + &
                  bspots(ikd)%qq(ir1,ir2,l1)*coefprj2(ip2)
          end if
       end do
       coefprj(ip) = cnst*coefprj(ip)
    end do
!$omp parallel do private(ig),reduction(+:sp)
!POPTION PSUM(sp)
    do ip=1,nprj
!SOPTION NOLOOPINTERCHANGE
!CDIR NOLOOPCHG
       do ig=bkm,ekm
          ctemp = dcmplx(prj_r2(ig,ip,1),prj_r2(ig,ip,2))
          sp(ig) = sp(ig) + dble(coefprj(ip)*ctemp)
       end do
    end do
    
    deallocate(coefprj,coefprj2,tcfpj)
    nullify(ip2t,ip2l,ip2r,ip2m,ip2h,prj_r2)

    call tpf_end(tpf_npj)
    return
  end subroutine nlprj_apy1si

  subroutine nlprj_apy1sg(nkd,bspots,kd,nsi,bkm,ekm, &
       va,p,sp,bprj,eprj,dcprj,nlpjs,comm)
    use xmpi
    use bsplpot
    use timeprof
    implicit none
    integer,intent(in):: nkd,nsi,bkm,ekm,bprj,eprj
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: va
    complex(kind=8),intent(in):: p(bkm:ekm)
    complex(kind=8),intent(inout):: sp(bkm:ekm)
    complex(kind=8):: dcprj(bprj:eprj)
    type(bspot_for_atm):: bspots(nkd)
    type(nlprj_struct):: nlpjs
    integer:: comm

!SOPTION CONTAREA(ip2t,ip2l,ip2r,ip2m,ip2h)
    integer,pointer:: ip2t(:),ip2l(:),ip2r(:),ip2m(:),ip2h(:)
!SOPTION CONTAREA(prj)
    complex(kind=8),pointer:: prj(:,:)
    real(kind=8),allocatable:: coefprj(:), tcfpj(:)
    real(kind=8),allocatable:: coefprj2(:)
    integer:: nprj
    real(kind=8):: cnst
    integer:: it,ikd,ip,ip2,ip2a
    integer:: l1,ir1,m1,l2,ir2,m2,ig,ierr

    call tpf_begin(tpf_npj)

    ip2t => nlpjs%ip2t
    ip2l => nlpjs%ip2l
    ip2r => nlpjs%ip2r
    ip2m => nlpjs%ip2m
    ip2h => nlpjs%ip2h
    prj => nlpjs%prj
    nprj = nlpjs%nprj

    cnst=1.0d0/sqrt(va)
    allocate(coefprj(nprj),coefprj2(nprj),tcfpj(nprj))
    tcfpj=0.0d0
!$omp parallel do private(ig)
    do ip=1,nprj
!SOPTION NOLOOPINTERCHANGE
!CDIR NOLOOPCHG
       do ig=max(bkm,2),ekm
          tcfpj(ip) = tcfpj(ip) + dble(conjg(prj(ig,ip))*p(ig))
       end do
       tcfpj(ip) = 2.0d0*tcfpj(ip)
       if (bkm.eq.1) then
          tcfpj(ip) = tcfpj(ip) + dble(conjg(prj(1,ip))*p(1))
       end if
    end do

    !call MPI_ALLREDUCE(tcfpj,coefprj2,nprj, &
    !     MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_alldsum(tcfpj,coefprj2,nprj,comm,ierr)

    coefprj2 = cnst*coefprj2
    dcprj(bprj:eprj) = coefprj2(bprj:eprj)

!$omp parallel do private(it,ikd,ip2,l1,ir1,m1,l2,ir2,m2,ip2a)
    do ip=1,nprj
       it = ip2t(ip)
       ikd = kd(it)
       l1 = ip2l(ip)
       ir1 = ip2r(ip)
       m1 = ip2m(ip)
       coefprj(ip) = 0.0d0
       do ip2a=1,bspots(ikd)%nprj
          l2 = bspots(ikd)%ip2l(ip2a)
          ir2 = bspots(ikd)%ip2r(ip2a)
          m2 = bspots(ikd)%ip2m(ip2a)
          ip2 = ip2h(ip) + ip2a
          if (m1.eq.m2) then
             coefprj(ip) = coefprj(ip) + &
                  bspots(ikd)%qq(ir1,ir2,l1)*coefprj2(ip2)
          end if
       end do
       coefprj(ip) = cnst*coefprj(ip)
    end do
!$omp parallel do private(ig),reduction(+:sp)
!POPTION PSUM(sp)
    do ip=1,nprj
!SOPTION NOLOOPINTERCHANGE
!CDIR NOLOOPCHG
       do ig=bkm,ekm
          sp(ig) = sp(ig) + coefprj(ip)*prj(ig,ip)
       end do
    end do
    
    deallocate(coefprj,coefprj2,tcfpj)
    nullify(ip2t,ip2l,ip2r,ip2m,ip2h,prj)

    call tpf_end(tpf_npj)
    return
  end subroutine nlprj_apy1sg

  subroutine nlprj_apyas(nkd,bspots,kd,nsi,bkm,ekm, &
       va,nw,p,sp,bprj,eprj,dcprj,nlpjs,comm)
    use xmpi
    use arch_nlprj
    use bsplpot
    use timeprof
    implicit none
    integer,intent(in):: nkd,nsi,bkm,ekm,nw,bprj,eprj
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: va
    complex(kind=8),intent(in):: p(bkm:ekm,nw)
    complex(kind=8),intent(inout):: sp(bkm:ekm,nw)
    complex(kind=8):: dcprj(bprj:eprj,nw)
    type(bspot_for_atm):: bspots(nkd)
    type(nlprj_struct):: nlpjs
    integer:: comm

    complex(kind=8),parameter:: ci = (0.0d0, 1.0d0)
    
!SOPTION CONTAREA(ip2t,ip2l,ip2r,ip2m,ip2h)
    integer,pointer:: ip2t(:),ip2l(:),ip2r(:),ip2m(:),ip2h(:)
!SOPTION CONTAREA(prj)
    complex(kind=8),pointer:: prj(:,:)
    complex(kind=8),allocatable:: coefprj2(:,:)
    complex(kind=8),allocatable:: coefprj3(:,:)
    integer:: nprj
    real(kind=8):: cnst
    integer:: it,ikd,ib,ip1,ip2,ip1a,ip2a
    integer:: l1,ir1,m1,l2,ir2,m2,ierr

    call tpf_begin(tpf_npj)

    ip2t => nlpjs%ip2t
    ip2l => nlpjs%ip2l
    ip2r => nlpjs%ip2r
    ip2m => nlpjs%ip2m
    ip2h => nlpjs%ip2h
    prj => nlpjs%prj
    nprj = nlpjs%nprj

    cnst=1.0d0/sqrt(va)
    allocate(coefprj2(nprj,nw), &
         coefprj3(nprj,nw),stat=ierr)
    call chkalloc('nlprj_apyas','coef',ierr)
    call ovlmat(bkm,ekm,nprj,nw,prj,p,coefprj2)

    !call MPI_ALLREDUCE(coefprj2,coefprj3,nprj*nw, &
    !     MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_allzsum(coefprj2,coefprj3,nprj*nw,comm,ierr)

    coefprj3 = cnst*coefprj3
    dcprj(bprj:eprj,:) = coefprj3(bprj:eprj,:)

!$omp parallel do private(ib,ip1,it,ikd,l1,ir1,m1,ip1a,ip2a,ip2,l2,ir2,m2)
    do ib=1,nw
       do ip1=1,nprj
          it = ip2t(ip1)
          ikd = kd(it)
          l1 = ip2l(ip1)
          ir1 = ip2r(ip1)
          m1 = ip2m(ip1)
          ip1a = ip1 - ip2h(ip1)
          coefprj2(ip1,ib) = (0.0d0,0.0d0)
          do ip2a = 1,bspots(ikd)%nprj
             ip2 = ip2h(ip1) + ip2a
             l2 = bspots(ikd)%ip2l(ip2a)
             ir2= bspots(ikd)%ip2r(ip2a)
             m2 = bspots(ikd)%ip2m(ip2a)
             if (m1.eq.m2) then
                coefprj2(ip1,ib) = coefprj2(ip1,ib) + &
                     bspots(ikd)%qq(ir1,ir2,l1)*coefprj3(ip2,ib)
             end if
          end do
          coefprj2(ip1,ib) = cnst*coefprj2(ip1,ib)
       end do
    end do
    call rotprjs(bkm,ekm,nprj,nw,coefprj2,prj,sp)
    
    deallocate(coefprj2,coefprj3)
    nullify(ip2t,ip2l,ip2r,ip2m,ip2h,prj)

    call tpf_end(tpf_npj)
    return
  end subroutine nlprj_apyas

  subroutine nlprj_unify(bprj,eprj,nw,laprj,nsi, &
       dcprj,pby,nlpjs, &
       nproc,myrank,pj_count,pj_displ,comm)
    use mpi
    use timeprof
    implicit none
    integer,intent(in):: bprj,eprj,nw,laprj,nsi
    complex(kind=8):: dcprj(bprj:eprj,nw)
    complex(kind=8):: pby(laprj,nsi,nw)
    type(nlprj_struct):: nlpjs
    integer,intent(in):: nproc,myrank,comm
    integer,intent(in):: pj_count(0:nproc-1),pj_displ(0:nproc-1)

    complex(kind=8),allocatable:: dcprjtr(:,:),cprjtr(:,:)

!SOPTION CONTAREA(ip2t,ip2r,ip2m,ip2h)
    integer,pointer:: ip2t(:),ip2r(:),ip2m(:),ip2h(:)
    integer:: it,ir,im,ip,ib,nprj,ierr
    integer,allocatable:: count(:),displ(:)
    integer:: ipa

    call tpf_begin(tpf_npj)

    ip2t => nlpjs%ip2t
    ip2r => nlpjs%ip2r
    ip2m => nlpjs%ip2m
    ip2h => nlpjs%ip2h
    nprj = nlpjs%nprj
    allocate(dcprjtr(nw,bprj:eprj),cprjtr(nw,nprj), &
         count(0:nproc-1),displ(0:nproc-1),stat=ierr)
    call chkalloc('nlprj_unify','tmp',ierr)

    do ib=1,nw
       do ip=bprj,eprj
          dcprjtr(ib,ip) = dcprj(ip,ib)
       end do
    end do
    count = nw*pj_count
    displ = nw*pj_displ
    call tpf_begin(tpf_mpi)
    call MPI_ALLGATHERV(dcprjtr(1,bprj),count(myrank), &
         MPI_DOUBLE_COMPLEX, &
         cprjtr,count,displ,MPI_DOUBLE_COMPLEX,comm,ierr)
    call tpf_end(tpf_mpi)

    do ip=1,nprj
       do ib=1,nw
          it = ip2t(ip)
          ir = ip2r(ip)
          im = ip2m(ip)
          ipa = ip - ip2h(ip)
          pby(ipa,it,ib) = cprjtr(ib,ip)
       end do
    end do

    deallocate(dcprjtr,cprjtr,count,displ)
    nullify(ip2t,ip2r,ip2m,ip2h)

    call tpf_end(tpf_npj)
    return
  end subroutine nlprj_unify

  subroutine nlprj_done(nlpjs)
    implicit none
    type(nlprj_struct):: nlpjs

    deallocate(nlpjs%ip2t,nlpjs%ip2l,nlpjs%ip2r,nlpjs%ip2m,nlpjs%ip2h)
    select case (nlpjs%mode)
    case (1)
       deallocate(nlpjs%prj)
    case (2)
       deallocate(nlpjs%prj_r2)
    case default
       write(6,*) 'invalid mode in nlprj_done()'
       stop
    end select
    return
  end subroutine nlprj_done

end module nonlocal_proj
