/*
  Copyright (c) 2012-2013 Yoshihide Yoshimoto
 */

#include <emmintrin.h>

#define _MM_SHUFFLE2(x,y) (x<<1)|y

int fftrd2_(int *invk, int *inp, double *dw, double *dx, double *dy) {
	
  int nvk,np;
  int p,ivk;
  int iw,ix,iy;
  __m128d tw,x0,x1,t1,zr;
  int tsd0, tsd1, tsd2;

  __m128d *w,*x,*y;

  nvk = *invk;
  np = *inp;
  w = (__m128d *)dw;
  x = (__m128d *)dx;
  y = (__m128d *)dy;

  zr = _fjsp_setzero_v2r8();

  tsd0 = nvk;
  tsd1 = nvk*2;
  tsd2 = nvk*np;

  ix = 0;
  iy = 0;
  iw = 0;
  for (p = 0; p < np; p++) {
    tw = w[iw];

    ix = tsd0*p;
    iy = tsd1*p;
    for (ivk = 0; ivk < nvk; ivk++) {
      x0 = x[ix];
      x1 = x[ix+tsd2];

      y[iy] = _mm_add_pd(x0,x1);

      t1 = _mm_sub_pd(x0,x1);

      y[iy+tsd0] = _fjsp_nmsub_cp_neg_sr12_v2r8(tw,t1,_fjsp_madd_cp_v2r8(tw,t1,zr));

      ix = ix + 1;
      iy = iy + 1;
    }
    iw = iw + 1;
  }
  return 0;
}

int bftrd2_(int *invk, int *inp, double *dw, double *dx, double *dy) {
	
  int nvk,np;
  int p,ivk;
  int iw,ix,iy;
  __m128d tw,x0,x1,t1,zr;
  int tsd0, tsd1, tsd2;

  __m128d *w,*x,*y;

  nvk = *invk;
  np = *inp;
  w = (__m128d *)dw;
  x = (__m128d *)dx;
  y = (__m128d *)dy;

  zr = _fjsp_setzero_v2r8();

  tsd0 = nvk;
  tsd1 = nvk*2;
  tsd2 = nvk*np;

  ix = 0;
  iy = 0;
  iw = 0;
  for (p = 0; p < np; p++) {
    tw = w[iw];

    ix = tsd0*p;
    iy = tsd1*p;
    for (ivk = 0; ivk < nvk; ivk++) {
      x0 = x[ix];
      x1 = x[ix+tsd2];

      y[iy] = _mm_add_pd(x0,x1);

      t1 = _mm_sub_pd(x0,x1);

      y[iy+tsd0] = _fjsp_madd_cp_neg_sr12_v2r8(tw,t1,_fjsp_madd_cp_v2r8(tw,t1,zr));

      ix = ix + 1;
      iy = iy + 1;
    }
    iw = iw + 1;
  }
  return 0;
}

int fftrd3_(int *invk, int *inp, double *dw, double *dx, double *dy) {
  const double sp3 = 0.866025403784438647;

  int nvk,np;
  int p,ivk;
  int iw,ix,iy;
  __m128d tw1,tw2,x1,x2,x3,zr;
  __m128d t1,t2,t3,t4,t5,tsp3,thlf;
  int tsd0, tsd1, tsd2;
  int tsd2m2;
  int tsd0m2;

  __m128d *w,*x,*y;

  nvk = *invk;
  np = *inp;
  w = (__m128d *)dw;
  x = (__m128d *)dx;
  y = (__m128d *)dy;

  zr = _fjsp_setzero_v2r8();

  tsd0 = nvk;
  tsd1 = nvk*3;
  tsd2 = nvk*np;

  tsd2m2 = tsd2*2;

  tsd0m2 = tsd0*2;

  tsp3 = _mm_set_pd(-sp3,sp3);
  thlf = _fjsp_set_v2r8(0.5,0.5);

  ix = 0;
  iy = 0;
  iw = 0;
  for (p = 0; p < np; p++) {
    tw1 = w[iw];
    tw2 = w[iw+1];
    iw = iw + 2;

    ix = tsd0*p;
    iy = tsd1*p;
    for (ivk = 0; ivk < nvk; ivk++) {
      x1 = x[ix];
      x2 = x[ix+tsd2];
      x3 = x[ix+tsd2m2];

      t1 = _mm_add_pd(x2,x3);
      t2 = _fjsp_nmsub_v2r8(thlf,t1,x1);

      t3 = _mm_sub_pd(x2,x3);
      t3 = _fjsp_madd_sr2_v2r8(tsp3,t3,zr);

      y[iy] = _mm_add_pd(x1,t1);

      t4 = _mm_add_pd(t2,t3);
      y[iy+tsd0] = _fjsp_nmsub_cp_neg_sr12_v2r8(tw1,t4,_fjsp_madd_cp_v2r8(tw1,t4,zr));
      
      t5 = _mm_sub_pd(t2,t3);
      y[iy+tsd0m2] = _fjsp_nmsub_cp_neg_sr12_v2r8(tw2,t5,_fjsp_madd_cp_v2r8(tw2,t5,zr));

      ix = ix + 1;
      iy = iy + 1;
    }
  }
  return 0;
}

int bftrd3_(int *invk, int *inp, double *dw, double *dx, double *dy) {
  const double sp3 = 0.866025403784438647;

  int nvk,np;
  int p,ivk;
  int iw,ix,iy;
  __m128d tw1,tw2,x1,x2,x3,zr;
  __m128d t1,t2,t3,t4,t5,tsp3,thlf;
  int tsd0, tsd1, tsd2;
  int tsd2m2;
  int tsd0m2;

  __m128d *w,*x,*y;

  nvk = *invk;
  np = *inp;
  w = (__m128d *)dw;
  x = (__m128d *)dx;
  y = (__m128d *)dy;

  zr = _fjsp_setzero_v2r8();

  tsd0 = nvk;
  tsd1 = nvk*3;
  tsd2 = nvk*np;

  tsd2m2 = tsd2*2;

  tsd0m2 = tsd0*2;

  tsp3 = _mm_set_pd(sp3,-sp3);
  thlf = _fjsp_set_v2r8(0.5,0.5);

  ix = 0;
  iy = 0;
  iw = 0;
  for (p = 0; p < np; p++) {
    tw1 = w[iw];
    tw2 = w[iw+1];
    iw = iw + 2;

    ix = tsd0*p;
    iy = tsd1*p;
    for (ivk = 0; ivk < nvk; ivk++) {
      x1 = x[ix];
      x2 = x[ix+tsd2];
      x3 = x[ix+tsd2m2];

      t1 = _mm_add_pd(x2,x3);
      t2 =  _fjsp_nmsub_v2r8(thlf,t1,x1);

      t3 = _mm_sub_pd(x2,x3);
      t3 = _fjsp_madd_sr2_v2r8(tsp3,t3,zr);

      y[iy] = _mm_add_pd(x1,t1);

      t4 = _mm_add_pd(t2,t3);
      y[iy+tsd0] = _fjsp_madd_cp_neg_sr12_v2r8(tw1,t4,_fjsp_madd_cp_v2r8(tw1,t4,zr));

      t5 = _mm_sub_pd(t2,t3);
      y[iy+tsd0m2] = _fjsp_madd_cp_neg_sr12_v2r8(tw2,t5,_fjsp_madd_cp_v2r8(tw2,t5,zr));

      ix = ix + 1;
      iy = iy + 1;
    }
  }
  return 0;
}

int fftrd4_(int *invk, int *inp, double *dw, double *dx, double *dy) {
	
  int nvk,np;
  int p,ivk;
  int iw,ix,iy;
  __m128d tw1,tw2,tw3,x1,x2,x3,x4,zr,cf;
  __m128d t1,t2,t3,t4,t5,t6,t7;
  int tsd0, tsd1, tsd2;
  int tsd2m2,tsd2m3;
  int tsd0m2,tsd0m3;

  __m128d *w,*x,*y;

  nvk = *invk;
  np = *inp;
  w = (__m128d *)dw;
  x = (__m128d *)dx;
  y = (__m128d *)dy;

  zr = _fjsp_setzero_v2r8();
  cf = _fjsp_set_v2r8(-1.0,1.0);

  tsd0 = nvk;
  tsd1 = nvk*4;
  tsd2 = nvk*np;

  tsd2m2 = tsd2*2;
  tsd2m3 = tsd2*3;

  tsd0m2 = tsd0*2;
  tsd0m3 = tsd0*3;

  ix = 0;
  iy = 0;
  iw = 0;
  for (p = 0; p < np; p++) {
    tw1 = w[iw];
    tw2 = w[iw+1];
    tw3 = w[iw+2];
    iw = iw + 3;

    ix = tsd0*p;
    iy = tsd1*p;
    for (ivk = 0; ivk < nvk; ivk++) {
      x1 = x[ix];
      x2 = x[ix+tsd2];
      x3 = x[ix+tsd2m2];
      x4 = x[ix+tsd2m3];

      t1 = _mm_add_pd(x1,x3);
      t2 = _mm_sub_pd(x1,x3);
      t3 = _mm_add_pd(x2,x4);

      t4 = _mm_sub_pd(x2,x4);
      t4 = _fjsp_madd_sr2_v2r8(cf,t4,zr);

      y[iy] = _mm_add_pd(t1,t3);

      t5 = _mm_add_pd(t2,t4);
      y[iy+tsd0] = _fjsp_nmsub_cp_neg_sr12_v2r8(tw1,t5,_fjsp_madd_cp_v2r8(tw1,t5,zr));

      t6 = _mm_sub_pd(t1,t3);
      y[iy+tsd0m2] = _fjsp_nmsub_cp_neg_sr12_v2r8(tw2,t6,_fjsp_madd_cp_v2r8(tw2,t6,zr));

      t7 = _mm_sub_pd(t2,t4);
      y[iy+tsd0m3] = _fjsp_nmsub_cp_neg_sr12_v2r8(tw3,t7,_fjsp_madd_cp_v2r8(tw3,t7,zr));

      ix = ix + 1;
      iy = iy + 1;
    }
  }
  return 0;
}

int bftrd4_(int *invk, int *inp, double *dw, double *dx, double *dy) {
	
  int nvk,np;
  int p,ivk;
  int iw,ix,iy;
  __m128d tw1,tw2,tw3,x1,x2,x3,x4,zr,cf;
  __m128d t1,t2,t3,t4,t5,t6,t7;
  int tsd0, tsd1, tsd2;
  int tsd2m2,tsd2m3;
  int tsd0m2,tsd0m3;

  __m128d *w,*x,*y;

  nvk = *invk;
  np = *inp;
  w = (__m128d *)dw;
  x = (__m128d *)dx;
  y = (__m128d *)dy;

  zr = _fjsp_setzero_v2r8();
  cf = _fjsp_set_v2r8(-1.0,1.0);

  tsd0 = nvk;
  tsd1 = nvk*4;
  tsd2 = nvk*np;

  tsd2m2 = tsd2*2;
  tsd2m3 = tsd2*3;

  tsd0m2 = tsd0*2;
  tsd0m3 = tsd0*3;

  ix = 0;
  iy = 0;
  iw = 0;
  for (p = 0; p < np; p++) {
    tw1 = w[iw];
    tw2 = w[iw+1];
    tw3 = w[iw+2];
    iw = iw + 3;

    ix = tsd0*p;
    iy = tsd1*p;
    for (ivk = 0; ivk < nvk; ivk++) {
      x1 = x[ix];
      x2 = x[ix+tsd2];
      x3 = x[ix+tsd2m2];
      x4 = x[ix+tsd2m3];

      t1 = _mm_add_pd(x1,x3);
      t2 = _mm_sub_pd(x1,x3);
      t3 = _mm_add_pd(x2,x4);

      t4 = _mm_sub_pd(x4,x2);
      t4 = _fjsp_madd_sr2_v2r8(cf,t4,zr);

      y[iy] = _mm_add_pd(t1,t3);

      t5 = _mm_add_pd(t2,t4);
      y[iy+tsd0] = _fjsp_madd_cp_neg_sr12_v2r8(tw1,t5,_fjsp_madd_cp_v2r8(tw1,t5,zr));

      t6 = _mm_sub_pd(t1,t3);
      y[iy+tsd0m2] = _fjsp_madd_cp_neg_sr12_v2r8(tw2,t6,_fjsp_madd_cp_v2r8(tw2,t6,zr));

      t7 = _mm_sub_pd(t2,t4);
      y[iy+tsd0m3] = _fjsp_madd_cp_neg_sr12_v2r8(tw3,t7,_fjsp_madd_cp_v2r8(tw3,t7,zr));

      ix = ix + 1;
      iy = iy + 1;
    }
  }
  return 0;
}

int fftrd5_(int *invk, int *inp, double *dw, double *dx, double *dy) {
  const double sp25 = 0.951056516295153572;
  const double sq54 = 0.559016994374947424;
  const double ss = 0.618033988749894848;

  int nvk,np;
  int p,ivk;
  int iw,ix,iy;
  __m128d w1,w2,w3,w4,x1,x2,x3,x4,x5,zr,cf;
  __m128d t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15;
  int tsd0, tsd1, tsd2;
  int tsd2m2,tsd2m3,tsd2m4;
  int tsd0m2,tsd0m3,tsd0m4;

  __m128d *w,*x,*y;

  nvk = *invk;
  np = *inp;
  w = (__m128d *)dw;
  x = (__m128d *)dx;
  y = (__m128d *)dy;

  zr = _fjsp_setzero_v2r8();
  cf = _fjsp_set_v2r8(-1.0,1.0);

  tsd0 = nvk;
  tsd1 = nvk*5;
  tsd2 = nvk*np;

  tsd2m2 = tsd2*2;
  tsd2m3 = tsd2*3;
  tsd2m4 = tsd2*4;

  tsd0m2 = tsd0*2;
  tsd0m3 = tsd0*3;
  tsd0m4 = tsd0*4;

  ix = 0;
  iy = 0;
  for (ivk = 0; ivk < nvk; ivk++) {
    x1 = x[ix];
    x2 = x[ix+tsd2];
    x3 = x[ix+tsd2m2];
    x4 = x[ix+tsd2m3];
    x5 = x[ix+tsd2m4];

    t1 = _mm_add_pd(x2,x5);
    t2 = _mm_add_pd(x3,x4);

    t5 = _fjsp_set_v2r8(sp25,sp25);
    t3 = _mm_sub_pd(x2,x5);
    t4 = _mm_sub_pd(x3,x4);
    t3 = _mm_mul_pd(t5,t3);
    t4 = _mm_mul_pd(t5,t4);

    t5 = _mm_add_pd(t1,t2);
    t6 = _mm_sub_pd(t1,t2);
    t7 = _fjsp_set_v2r8(sq54,sq54);
    t6 = _mm_mul_pd(t7,t6);

    t7 = _fjsp_set_v2r8(0.25,0.25);
    t7 = _fjsp_nmsub_v2r8(t7,t5,x1);

    t8 = _mm_add_pd(t7,t6);
    t9 = _mm_sub_pd(t7,t6);

    t12 = _fjsp_madd_sr2_v2r8(cf,t3,zr);
    t13 = _fjsp_madd_sr2_v2r8(cf,t4,zr);
    t10 = _fjsp_set_v2r8(ss,ss);
    t10 = _mm_mul_pd(t13,t10);
    t10 = _mm_add_pd(t12,t10);
    t11 = _fjsp_set_v2r8(ss,ss);
    t11 = _mm_mul_pd(t12,t11);
    t11 = _mm_sub_pd(t11,t13);

    y[iy] = _mm_add_pd(x1,t5);

    y[iy+tsd0] = _mm_add_pd(t8,t10);

    y[iy+tsd0m2] = _mm_add_pd(t9,t11);

    y[iy+tsd0m3] = _mm_sub_pd(t9,t11);

    y[iy+tsd0m4] = _mm_sub_pd(t8,t10);

    ix = ix + 1;
    iy = iy + 1;
  }

  ix = 0;
  iy = 0;
  iw = 4;
  for (p = 1; p < np; p++) {
    w1 = w[iw];
    w2 = w[iw+1];
    w3 = w[iw+2];
    w4 = w[iw+3];
    iw = iw + 4;

    ix = tsd0*p;
    iy = tsd1*p;

    for (ivk = 0; ivk < nvk; ivk++) {
      x1 = x[ix];
      x2 = x[ix+tsd2];
      x3 = x[ix+tsd2m2];
      x4 = x[ix+tsd2m3];
      x5 = x[ix+tsd2m4];

      t1 = _mm_add_pd(x2,x5);
      t2 = _mm_add_pd(x3,x4);

      t5 = _fjsp_set_v2r8(sp25,sp25);
      t3 = _mm_sub_pd(x2,x5);
      t4 = _mm_sub_pd(x3,x4);
      t3 = _mm_mul_pd(t5,t3);
      t4 = _mm_mul_pd(t5,t4);

      t5 = _mm_add_pd(t1,t2);
      t6 = _mm_sub_pd(t1,t2);
      t7 = _fjsp_set_v2r8(sq54,sq54);
      t6 = _mm_mul_pd(t7,t6);

      t7 = _fjsp_set_v2r8(0.25,0.25);
      t7 = _fjsp_nmsub_v2r8(t7,t5,x1);

      t8 = _mm_add_pd(t7,t6);
      t9 = _mm_sub_pd(t7,t6);

      t12 = _fjsp_madd_sr2_v2r8(cf,t3,zr);
      t13 = _fjsp_madd_sr2_v2r8(cf,t4,zr);
      t10 = _fjsp_set_v2r8(ss,ss);
      t10 = _mm_mul_pd(t13,t10);
      t10 = _mm_add_pd(t12,t10);
      t11 = _fjsp_set_v2r8(ss,ss);
      t11 = _mm_mul_pd(t12,t11);
      t11 = _mm_sub_pd(t11,t13);

      y[iy] = _mm_add_pd(x1,t5);

      t1 = _mm_setzero_pd();

      t12 = _mm_add_pd(t8,t10);
      y[iy+tsd0] = _fjsp_nmsub_cp_neg_sr12_v2r8(w1,t12,_fjsp_madd_cp_v2r8(w1,t12,zr));

      t13 = _mm_add_pd(t9,t11);
      y[iy+tsd0m2] = _fjsp_nmsub_cp_neg_sr12_v2r8(w2,t13,_fjsp_madd_cp_v2r8(w2,t13,zr));

      t14 = _mm_sub_pd(t9,t11);
      y[iy+tsd0m3] = _fjsp_nmsub_cp_neg_sr12_v2r8(w3,t14,_fjsp_madd_cp_v2r8(w3,t14,zr));

      t15 = _mm_sub_pd(t8,t10);
      y[iy+tsd0m4] = _fjsp_nmsub_cp_neg_sr12_v2r8(w4,t15,_fjsp_madd_cp_v2r8(w4,t15,zr));

      ix = ix + 1;
      iy = iy + 1;
    }
  }
  return 0;
}

int bftrd5_(int *invk, int *inp, double *dw, double *dx, double *dy) {
  const double sp25 = 0.951056516295153572;
  const double sq54 = 0.559016994374947424;
  const double ss = 0.618033988749894848;

  int nvk,np;
  int p,ivk;
  int iw,ix,iy;
  __m128d w1,w2,w3,w4,x1,x2,x3,x4,x5,zr,cf;
  __m128d t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15;
  int tsd0, tsd1, tsd2;
  int tsd2m2,tsd2m3,tsd2m4;
  int tsd0m2,tsd0m3,tsd0m4;

  __m128d *w,*x,*y;

  nvk = *invk;
  np = *inp;
  w = (__m128d *)dw;
  x = (__m128d *)dx;
  y = (__m128d *)dy;

  zr = _fjsp_setzero_v2r8();
  cf = _fjsp_set_v2r8(-1.0,1.0);

  tsd0 = nvk;
  tsd1 = nvk*5;
  tsd2 = nvk*np;

  tsd2m2 = tsd2*2;
  tsd2m3 = tsd2*3;
  tsd2m4 = tsd2*4;

  tsd0m2 = tsd0*2;
  tsd0m3 = tsd0*3;
  tsd0m4 = tsd0*4;

  ix = 0;
  iy = 0;
  for (ivk = 0; ivk < nvk; ivk++) {
    x1 = x[ix];
    x2 = x[ix+tsd2];
    x3 = x[ix+tsd2m2];
    x4 = x[ix+tsd2m3];
    x5 = x[ix+tsd2m4];

    t1 = _mm_add_pd(x2,x5);
    t2 = _mm_add_pd(x3,x4);

    t5 = _fjsp_set_v2r8(sp25,sp25);
    t3 = _mm_sub_pd(x2,x5);
    t4 = _mm_sub_pd(x3,x4);
    t3 = _mm_mul_pd(t5,t3);
    t4 = _mm_mul_pd(t5,t4);

    t5 = _mm_add_pd(t1,t2);
    t6 = _mm_sub_pd(t1,t2);
    t7 = _fjsp_set_v2r8(sq54,sq54);
    t6 = _mm_mul_pd(t7,t6);

    t7 = _fjsp_set_v2r8(0.25,0.25);
    t7 = _fjsp_nmsub_v2r8(t7,t5,x1);

    t8 = _mm_add_pd(t7,t6);
    t9 = _mm_sub_pd(t7,t6);

    t12 = _fjsp_nmsub_sr2_v2r8(cf,t3,zr);
    t13 = _fjsp_nmsub_sr2_v2r8(cf,t4,zr);
    t10 = _fjsp_set_v2r8(ss,ss);
    t10 = _mm_mul_pd(t13,t10);
    t10 = _mm_add_pd(t12,t10);
    t11 = _fjsp_set_v2r8(ss,ss);
    t11 = _mm_mul_pd(t12,t11);
    t11 = _mm_sub_pd(t11,t13);

    y[iy] = _mm_add_pd(x1,t5);

    y[iy+tsd0] = _mm_add_pd(t8,t10);

    y[iy+tsd0m2] = _mm_add_pd(t9,t11);

    y[iy+tsd0m3] = _mm_sub_pd(t9,t11);

    y[iy+tsd0m4] = _mm_sub_pd(t8,t10);

    ix = ix + 1;
    iy = iy + 1;
  }

  ix = 0;
  iy = 0;
  iw = 4;
  for (p = 1; p < np; p++) {
    w1 = w[iw];
    w2 = w[iw+1];
    w3 = w[iw+2];
    w4 = w[iw+3];
    iw = iw + 4;

    ix = tsd0*p;
    iy = tsd1*p;

    for (ivk = 0; ivk < nvk; ivk++) {
      x1 = x[ix];
      x2 = x[ix+tsd2];
      x3 = x[ix+tsd2m2];
      x4 = x[ix+tsd2m3];
      x5 = x[ix+tsd2m4];

      t1 = _mm_add_pd(x2,x5);
      t2 = _mm_add_pd(x3,x4);

      t5 = _fjsp_set_v2r8(sp25,sp25);
      t3 = _mm_sub_pd(x2,x5);
      t4 = _mm_sub_pd(x3,x4);
      t3 = _mm_mul_pd(t5,t3);
      t4 = _mm_mul_pd(t5,t4);

      t5 = _mm_add_pd(t1,t2);
      t6 = _mm_sub_pd(t1,t2);
      t7 = _fjsp_set_v2r8(sq54,sq54);
      t6 = _mm_mul_pd(t7,t6);

      t7 = _fjsp_set_v2r8(0.25,0.25);
      t7 = _fjsp_nmsub_v2r8(t7,t5,x1);

      t8 = _mm_add_pd(t7,t6);
      t9 = _mm_sub_pd(t7,t6);

      t12 = _fjsp_nmsub_sr2_v2r8(cf,t3,zr);
      t13 = _fjsp_nmsub_sr2_v2r8(cf,t4,zr);
      t10 = _fjsp_set_v2r8(ss,ss);
      t10 = _mm_mul_pd(t13,t10);
      t10 = _mm_add_pd(t12,t10);
      t11 = _fjsp_set_v2r8(ss,ss);
      t11 = _mm_mul_pd(t12,t11);
      t11 = _mm_sub_pd(t11,t13);

      y[iy] = _mm_add_pd(x1,t5);

      t1 = _mm_setzero_pd();

      t12 = _mm_add_pd(t8,t10);
      y[iy+tsd0] = _fjsp_madd_cp_neg_sr12_v2r8(w1,t12,_fjsp_madd_cp_v2r8(w1,t12,zr));

      t13 = _mm_add_pd(t9,t11);
      y[iy+tsd0m2] = _fjsp_madd_cp_neg_sr12_v2r8(w2,t13,_fjsp_madd_cp_v2r8(w2,t13,zr));

      t14 = _mm_sub_pd(t9,t11);
      y[iy+tsd0m3] = _fjsp_madd_cp_neg_sr12_v2r8(w3,t14,_fjsp_madd_cp_v2r8(w3,t14,zr));

      t15 = _mm_sub_pd(t8,t10);
      y[iy+tsd0m4] = _fjsp_madd_cp_neg_sr12_v2r8(w4,t15,_fjsp_madd_cp_v2r8(w4,t15,zr));

      ix = ix + 1;
      iy = iy + 1;
    }
  }
  return 0;
}

int fftrd6_(int *invk, int *inp, double *dw, double *dx, double *dy) {
  const double sp3 = 0.866025403784438647;

  int nvk,np;
  int p,ivk;
  int iw,ix,iy;
  __m128d w1,w2,w3,w4,w5,x1,x2,x3,x4,x5,x6,zr,cf;
  __m128d t1,t2,t3,t4,t5,u1,u2,u3,v1,v2,v3;
  int tsd0, tsd1, tsd2;
  int tsd2m2,tsd2m3,tsd2m4,tsd2m5;
  int tsd0m2,tsd0m3,tsd0m4,tsd0m5;

  __m128d *w,*x,*y;

  nvk = *invk;
  np = *inp;
  w = (__m128d *)dw;
  x = (__m128d *)dx;
  y = (__m128d *)dy;

  zr = _fjsp_setzero_v2r8();
  cf = _fjsp_set_v2r8(-1.0,1.0);

  tsd0 = nvk;
  tsd1 = nvk*6;
  tsd2 = nvk*np;

  tsd2m2 = tsd2*2;
  tsd2m3 = tsd2*3;
  tsd2m4 = tsd2*4;
  tsd2m5 = tsd2*5;

  tsd0m2 = tsd0*2;
  tsd0m3 = tsd0*3;
  tsd0m4 = tsd0*4;
  tsd0m5 = tsd0*5;

  ix = 0;
  iy = 0;
  for (ivk = 0; ivk < nvk; ivk++) {
    x1 = x[ix];
    x2 = x[ix+tsd2];
    x3 = x[ix+tsd2m2];
    x4 = x[ix+tsd2m3];
    x5 = x[ix+tsd2m4];
    x6 = x[ix+tsd2m5];

    t1 = _mm_add_pd(x3,x5);
    t2 = _fjsp_set_v2r8(0.5,0.5);
    t2 = _mm_mul_pd(t1,t2);
    t2 = _mm_sub_pd(x1,t2);
    t3 = _mm_sub_pd(x3,x5);
    t4 = _fjsp_set_v2r8(-sp3,-sp3);
    t3 = _mm_mul_pd(t4,t3);
    u1 = _mm_add_pd(x1,t1);
    u2 = _fjsp_nmsub_sr2_v2r8(cf,t3,t2);
    u3 = _fjsp_madd_sr2_v2r8(cf,t3,t2);

    t1 = _mm_add_pd(x6,x2);
    t2 = _fjsp_set_v2r8(0.5,0.5);
    t2 = _mm_mul_pd(t1,t2);
    t2 = _mm_sub_pd(x4,t2);
    t4 = _fjsp_set_v2r8(-sp3,-sp3);
    t3 = _mm_sub_pd(x6,x2);
    t3 = _mm_mul_pd(t4,t3);
    v1 = _mm_add_pd(x4,t1);
    v2 = _fjsp_nmsub_sr2_v2r8(cf,t3,t2);
    v3 = _fjsp_madd_sr2_v2r8(cf,t3,t2);

    y[iy] = _mm_add_pd(u1,v1);

    y[iy+tsd0] = _mm_sub_pd(u2,v2);

    y[iy+tsd0m2] = _mm_add_pd(u3,v3);

    y[iy+tsd0m3] = _mm_sub_pd(u1,v1);

    y[iy+tsd0m4] = _mm_add_pd(u2,v2);

    y[iy+tsd0m5] = _mm_sub_pd(u3,v3);

    ix = ix + 1;
    iy = iy + 1;
  }

  ix = 0;
  iy = 0;
  iw = 5;
  for (p = 1; p < np; p++) {
    w1 = w[iw];
    w2 = w[iw+1];
    w3 = w[iw+2];
    w4 = w[iw+3];
    w5 = w[iw+4];
    iw = iw + 5;

    ix = tsd0*p;
    iy = tsd1*p;

    for (ivk = 0; ivk < nvk; ivk++) {
      x1 = x[ix];
      x2 = x[ix+tsd2];
      x3 = x[ix+tsd2m2];
      x4 = x[ix+tsd2m3];
      x5 = x[ix+tsd2m4];
      x6 = x[ix+tsd2m5];

      t1 = _mm_add_pd(x3,x5);
      t2 = _fjsp_set_v2r8(0.5,0.5);
      t2 = _mm_mul_pd(t1,t2);
      t2 = _mm_sub_pd(x1,t2);
      t3 = _mm_sub_pd(x3,x5);
      t4 = _fjsp_set_v2r8(-sp3,-sp3);
      t3 = _mm_mul_pd(t4,t3);
      u1 = _mm_add_pd(x1,t1);
      u2 = _fjsp_nmsub_sr2_v2r8(cf,t3,t2);
      u3 = _fjsp_madd_sr2_v2r8(cf,t3,t2);

      t1 = _mm_add_pd(x6,x2);
      t2 = _fjsp_set_v2r8(0.5,0.5);
      t2 = _mm_mul_pd(t1,t2);
      t2 = _mm_sub_pd(x4,t2);
      t4 = _fjsp_set_v2r8(-sp3,-sp3);
      t3 = _mm_sub_pd(x6,x2);
      t3 = _mm_mul_pd(t4,t3);
      v1 = _mm_add_pd(x4,t1);
      v2 = _fjsp_nmsub_sr2_v2r8(cf,t3,t2);
      v3 = _fjsp_madd_sr2_v2r8(cf,t3,t2);

      y[iy] = _mm_add_pd(u1,v1);

      t1 = _mm_sub_pd(u2,v2);
      y[iy+tsd0] = _fjsp_nmsub_cp_neg_sr12_v2r8(w1,t1,_fjsp_madd_cp_v2r8(w1,t1,zr));

      t2 = _mm_add_pd(u3,v3);
      y[iy+tsd0m2] = _fjsp_nmsub_cp_neg_sr12_v2r8(w2,t2,_fjsp_madd_cp_v2r8(w2,t2,zr));

      t3 = _mm_sub_pd(u1,v1);
      y[iy+tsd0m3] = _fjsp_nmsub_cp_neg_sr12_v2r8(w3,t3,_fjsp_madd_cp_v2r8(w3,t3,zr));

      t4 = _mm_add_pd(u2,v2);
      y[iy+tsd0m4] = _fjsp_nmsub_cp_neg_sr12_v2r8(w4,t4,_fjsp_madd_cp_v2r8(w4,t4,zr));

      t5 = _mm_sub_pd(u3,v3);
      y[iy+tsd0m5] = _fjsp_nmsub_cp_neg_sr12_v2r8(w5,t5,_fjsp_madd_cp_v2r8(w5,t5,zr));

      ix = ix + 1;
      iy = iy + 1;
    }
  }
  return 0;
}

int bftrd6_(int *invk, int *inp, double *dw, double *dx, double *dy) {
  const double sp3 = 0.866025403784438647;

  int nvk,np;
  int p,ivk;
  int iw,ix,iy;
  __m128d w1,w2,w3,w4,w5,x1,x2,x3,x4,x5,x6,zr,cf;
  __m128d t1,t2,t3,t4,t5,u1,u2,u3,v1,v2,v3;
  int tsd0, tsd1, tsd2;
  int tsd2m2,tsd2m3,tsd2m4,tsd2m5;
  int tsd0m2,tsd0m3,tsd0m4,tsd0m5;

  __m128d *w,*x,*y;

  nvk = *invk;
  np = *inp;
  w = (__m128d *)dw;
  x = (__m128d *)dx;
  y = (__m128d *)dy;

  zr = _fjsp_setzero_v2r8();
  cf = _fjsp_set_v2r8(-1.0,1.0);

  tsd0 = nvk;
  tsd1 = nvk*6;
  tsd2 = nvk*np;

  tsd2m2 = tsd2*2;
  tsd2m3 = tsd2*3;
  tsd2m4 = tsd2*4;
  tsd2m5 = tsd2*5;

  tsd0m2 = tsd0*2;
  tsd0m3 = tsd0*3;
  tsd0m4 = tsd0*4;
  tsd0m5 = tsd0*5;

  ix = 0;
  iy = 0;
  for (ivk = 0; ivk < nvk; ivk++) {
    x1 = x[ix];
    x2 = x[ix+tsd2];
    x3 = x[ix+tsd2m2];
    x4 = x[ix+tsd2m3];
    x5 = x[ix+tsd2m4];
    x6 = x[ix+tsd2m5];

    t1 = _mm_add_pd(x3,x5);
    t2 = _fjsp_set_v2r8(0.5,0.5);
    t2 = _mm_mul_pd(t1,t2);
    t2 = _mm_sub_pd(x1,t2);
    t3 = _mm_sub_pd(x3,x5);
    t4 = _fjsp_set_v2r8(sp3,sp3);
    t3 = _mm_mul_pd(t4,t3);
    u1 = _mm_add_pd(x1,t1);
    u2 = _fjsp_nmsub_sr2_v2r8(cf,t3,t2);
    u3 = _fjsp_madd_sr2_v2r8(cf,t3,t2);

    t1 = _mm_add_pd(x6,x2);
    t2 = _fjsp_set_v2r8(0.5,0.5);
    t2 = _mm_mul_pd(t1,t2);
    t2 = _mm_sub_pd(x4,t2);
    t4 = _fjsp_set_v2r8(sp3,sp3);
    t3 = _mm_sub_pd(x6,x2);
    t3 = _mm_mul_pd(t4,t3);
    v1 = _mm_add_pd(x4,t1);
    v2 = _fjsp_nmsub_sr2_v2r8(cf,t3,t2);
    v3 = _fjsp_madd_sr2_v2r8(cf,t3,t2);

    y[iy] = _mm_add_pd(u1,v1);

    y[iy+tsd0] = _mm_sub_pd(u2,v2);

    y[iy+tsd0m2] = _mm_add_pd(u3,v3);

    y[iy+tsd0m3] = _mm_sub_pd(u1,v1);

    y[iy+tsd0m4] = _mm_add_pd(u2,v2);

    y[iy+tsd0m5] = _mm_sub_pd(u3,v3);

    ix = ix + 1;
    iy = iy + 1;
  }

  ix = 0;
  iy = 0;
  iw = 5;
  for (p = 1; p < np; p++) {
    t1 = _mm_set_pd(-0.0,0.0);
    w1 = w[iw];
    w2 = w[iw+1];
    w3 = w[iw+2];
    w4 = w[iw+3];
    w5 = w[iw+4];
    iw = iw + 5;

    ix = tsd0*p;
    iy = tsd1*p;

    for (ivk = 0; ivk < nvk; ivk++) {
      x1 = x[ix];
      x2 = x[ix+tsd2];
      x3 = x[ix+tsd2m2];
      x4 = x[ix+tsd2m3];
      x5 = x[ix+tsd2m4];
      x6 = x[ix+tsd2m5];

      t1 = _mm_add_pd(x3,x5);
      t2 = _fjsp_set_v2r8(0.5,0.5);
      t2 = _mm_mul_pd(t1,t2);
      t2 = _mm_sub_pd(x1,t2);
      t3 = _mm_sub_pd(x3,x5);
      t4 = _fjsp_set_v2r8(sp3,sp3);
      t3 = _mm_mul_pd(t4,t3);
      u1 = _mm_add_pd(x1,t1);
      u2 = _fjsp_nmsub_sr2_v2r8(cf,t3,t2);
      u3 = _fjsp_madd_sr2_v2r8(cf,t3,t2);
      
      t1 = _mm_add_pd(x6,x2);
      t2 = _fjsp_set_v2r8(0.5,0.5);
      t2 = _mm_mul_pd(t1,t2);
      t2 = _mm_sub_pd(x4,t2);
      t4 = _fjsp_set_v2r8(sp3,sp3);
      t3 = _mm_sub_pd(x6,x2);
      t3 = _mm_mul_pd(t4,t3);
      v1 = _mm_add_pd(x4,t1);
      v2 = _fjsp_nmsub_sr2_v2r8(cf,t3,t2);
      v3 = _fjsp_madd_sr2_v2r8(cf,t3,t2);

      y[iy] = _mm_add_pd(u1,v1);

      t1 = _mm_sub_pd(u2,v2);
      y[iy+tsd0] = _fjsp_madd_cp_neg_sr12_v2r8(w1,t1,_fjsp_madd_cp_v2r8(w1,t1,zr));

      t2 = _mm_add_pd(u3,v3);
      y[iy+tsd0m2] = _fjsp_madd_cp_neg_sr12_v2r8(w2,t2,_fjsp_madd_cp_v2r8(w2,t2,zr));

      t3 = _mm_sub_pd(u1,v1);
      y[iy+tsd0m3] = _fjsp_madd_cp_neg_sr12_v2r8(w3,t3,_fjsp_madd_cp_v2r8(w3,t3,zr));

      t4 = _mm_add_pd(u2,v2);
      y[iy+tsd0m4] = _fjsp_madd_cp_neg_sr12_v2r8(w4,t4,_fjsp_madd_cp_v2r8(w4,t4,zr));

      t5 = _mm_sub_pd(u3,v3);
      y[iy+tsd0m5] = _fjsp_madd_cp_neg_sr12_v2r8(w5,t5,_fjsp_madd_cp_v2r8(w5,t5,zr));

      ix = ix + 1;
      iy = iy + 1;
    }
  }
  return 0;
}

int fftrd8_(int *invk, int *inp, double *dw, double *dx, double *dy) {
  const double sp2 = 0.70710678118654752440;

  int nvk,np;
  int p,ivk;
  int iw,ix,iy;
  __m128d w1,w2,w3,w4,w5,w6,w7,x1,x2,x3,x4,x5,x6,x7,x8,zr,cf;
  __m128d t1,t2,t3,t4,t5,u1,u2,u3,u4,v1,v2,v3,v4,s;
  int tsd0, tsd1, tsd2;
  int tsd2m2,tsd2m3,tsd2m4,tsd2m5,tsd2m6,tsd2m7;
  int tsd0m2,tsd0m3,tsd0m4,tsd0m5,tsd0m6,tsd0m7;

  __m128d *w,*x,*y;

  nvk = *invk;
  np = *inp;
  w = (__m128d *)dw;
  x = (__m128d *)dx;
  y = (__m128d *)dy;

  zr = _fjsp_setzero_v2r8();
  cf = _fjsp_set_v2r8(-1.0,1.0);

  tsd0 = nvk;
  tsd1 = nvk*8;
  tsd2 = nvk*np;

  tsd2m2 = tsd2*2;
  tsd2m3 = tsd2*3;
  tsd2m4 = tsd2*4;
  tsd2m5 = tsd2*5;
  tsd2m6 = tsd2*6;
  tsd2m7 = tsd2*7;

  tsd0m2 = tsd0*2;
  tsd0m3 = tsd0*3;
  tsd0m4 = tsd0*4;
  tsd0m5 = tsd0*5;
  tsd0m6 = tsd0*6;
  tsd0m7 = tsd0*7;

  ix = 0;
  iy = 0;
  for (ivk = 0; ivk < nvk; ivk++) {
    x1 = x[ix];
    x2 = x[ix+tsd2];
    x3 = x[ix+tsd2m2];
    x4 = x[ix+tsd2m3];
    x5 = x[ix+tsd2m4];
    x6 = x[ix+tsd2m5];
    x7 = x[ix+tsd2m6];
    x8 = x[ix+tsd2m7];

    t1 = _mm_add_pd(x1,x5);
    t2 = _mm_sub_pd(x1,x5);
    t3 = _mm_add_pd(x3,x7);
    t4 = _fjsp_nmsub_v2r8(cf,x3,zr);
    t5 = _fjsp_madd_v2r8(cf,x7,zr);
    t5 = _mm_add_pd(t4,t5);
    t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
    u1 = _mm_add_pd(t1,t3);
    u2 = _mm_add_pd(t2,t4);
    u3 = _mm_sub_pd(t1,t3);
    u4 = _mm_sub_pd(t2,t4);

    t1 = _mm_add_pd(x2,x6);
    t2 = _mm_sub_pd(x2,x6);
    t3 = _mm_add_pd(x4,x8);
    t4 = _fjsp_nmsub_v2r8(cf,x4,zr);
    t5 = _fjsp_madd_v2r8(cf,x8,zr);
    t5 = _mm_add_pd(t4,t5);
    t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
    v1 = _mm_add_pd(t1,t3);
    v2 = _mm_add_pd(t2,t4);
    v3 = _mm_sub_pd(t1,t3);
    v4 = _mm_sub_pd(t2,t4);

    y[iy] = _mm_add_pd(u1,v1);
    y[iy+tsd0m4] = _mm_sub_pd(u1,v1);

    t5 = _fjsp_set_v2r8(sp2,sp2);
    s = _fjsp_madd_sr2_v2r8(cf,v2,v2);
    s = _mm_mul_pd(t5,s);

    y[iy+tsd0] =_mm_add_pd(u2,s);

    y[iy+tsd0m5] = _mm_sub_pd(u2,s);

    y[iy+tsd0m2] = _fjsp_madd_sr2_v2r8(cf,v3,u3);

    y[iy+tsd0m6] = _fjsp_nmsub_sr2_v2r8(cf,v3,u3);

    t5 = _fjsp_set_v2r8(sp2,sp2);
    s = _fjsp_msub_sr2_v2r8(cf,v4,v4);
    s = _mm_mul_pd(t5,s);

    y[iy+tsd0m3] = _mm_add_pd(u4,s);

    y[iy+tsd0m7] = _mm_sub_pd(u4,s);

    ix = ix + 1;
    iy = iy + 1;
  }

  ix = 0;
  iy = 0;
  iw = 7;
  for (p = 1; p < np; p++) {
    w1 = w[iw];
    w2 = w[iw+1];
    w3 = w[iw+2];
    w4 = w[iw+3];
    w5 = w[iw+4];
    w6 = w[iw+5];
    w7 = w[iw+6];
    iw = iw + 7;

    ix = tsd0*p;
    iy = tsd1*p;

    for (ivk = 0; ivk < nvk; ivk++) {
      x1 = x[ix];
      x2 = x[ix+tsd2];
      x3 = x[ix+tsd2m2];
      x4 = x[ix+tsd2m3];
      x5 = x[ix+tsd2m4];
      x6 = x[ix+tsd2m5];
      x7 = x[ix+tsd2m6];
      x8 = x[ix+tsd2m7];

      t1 = _mm_add_pd(x1,x5);
      t2 = _mm_sub_pd(x1,x5);
      t3 = _mm_add_pd(x3,x7);
      t4 = _fjsp_nmsub_v2r8(cf,x3,zr);
      t5 = _fjsp_madd_v2r8(cf,x7,zr);
      t5 = _mm_add_pd(t4,t5);
      t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
      u1 = _mm_add_pd(t1,t3);
      u2 = _mm_add_pd(t2,t4);
      u3 = _mm_sub_pd(t1,t3);
      u4 = _mm_sub_pd(t2,t4);

      t1 = _mm_add_pd(x2,x6);
      t2 = _mm_sub_pd(x2,x6);
      t3 = _mm_add_pd(x4,x8);
      t4 = _fjsp_nmsub_v2r8(cf,x4,zr);
      t5 = _fjsp_madd_v2r8(cf,x8,zr);
      t5 = _mm_add_pd(t4,t5);
      t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
      v1 = _mm_add_pd(t1,t3);
      v2 = _mm_add_pd(t2,t4);
      v3 = _mm_sub_pd(t1,t3);
      v4 = _mm_sub_pd(t2,t4);

      y[iy] = _mm_add_pd(u1,v1);

      t3 = _mm_sub_pd(u1,v1);
      y[iy+tsd0m4] = _fjsp_nmsub_cp_neg_sr12_v2r8(w4,t3,_fjsp_madd_cp_v2r8(w4,t3,zr));

      t5 = _fjsp_set_v2r8(sp2,sp2);
      s = _fjsp_madd_sr2_v2r8(cf,v2,v2);
      s = _mm_mul_pd(t5,s);

      t3 = _mm_add_pd(u2,s);
      y[iy+tsd0] = _fjsp_nmsub_cp_neg_sr12_v2r8(w1,t3,_fjsp_madd_cp_v2r8(w1,t3,zr));

      t3 = _mm_sub_pd(u2,s);
      y[iy+tsd0m5] = _fjsp_nmsub_cp_neg_sr12_v2r8(w5,t3,_fjsp_madd_cp_v2r8(w5,t3,zr));

      t3 = _fjsp_madd_sr2_v2r8(cf,v3,u3);
      y[iy+tsd0m2] = _fjsp_nmsub_cp_neg_sr12_v2r8(w2,t3,_fjsp_madd_cp_v2r8(w2,t3,zr));

      t3 = _fjsp_nmsub_sr2_v2r8(cf,v3,u3);
      y[iy+tsd0m6] = _fjsp_nmsub_cp_neg_sr12_v2r8(w6,t3,_fjsp_madd_cp_v2r8(w6,t3,zr));

      t5 = _fjsp_set_v2r8(sp2,sp2);
      s = _fjsp_msub_sr2_v2r8(cf,v4,v4);
      s = _mm_mul_pd(t5,s);

      t3 = _mm_add_pd(u4,s);
      y[iy+tsd0m3] = _fjsp_nmsub_cp_neg_sr12_v2r8(w3,t3,_fjsp_madd_cp_v2r8(w3,t3,zr));

      t3 = _mm_sub_pd(u4,s);
      y[iy+tsd0m7] = _fjsp_nmsub_cp_neg_sr12_v2r8(w7,t3,_fjsp_madd_cp_v2r8(w7,t3,zr));

      ix = ix + 1;
      iy = iy + 1;
    }
  }
  return 0;
}

int bftrd8_(int *invk, int *inp, double *dw, double *dx, double *dy) {
  const double sp2 = 0.70710678118654752440;

  int nvk,np;
  int p,ivk;
  int iw,ix,iy;
  __m128d w1,w2,w3,w4,w5,w6,w7,x1,x2,x3,x4,x5,x6,x7,x8,zr,cf;
  __m128d t1,t2,t3,t4,t5,u1,u2,u3,u4,v1,v2,v3,v4,s;
  int tsd0, tsd1, tsd2;
  int tsd2m2,tsd2m3,tsd2m4,tsd2m5,tsd2m6,tsd2m7;
  int tsd0m2,tsd0m3,tsd0m4,tsd0m5,tsd0m6,tsd0m7;

  __m128d *w,*x,*y;

  nvk = *invk;
  np = *inp;
  w = (__m128d *)dw;
  x = (__m128d *)dx;
  y = (__m128d *)dy;

  zr = _fjsp_setzero_v2r8();
  cf = _fjsp_set_v2r8(-1.0,1.0);

  tsd0 = nvk;
  tsd1 = nvk*8;
  tsd2 = nvk*np;

  tsd2m2 = tsd2*2;
  tsd2m3 = tsd2*3;
  tsd2m4 = tsd2*4;
  tsd2m5 = tsd2*5;
  tsd2m6 = tsd2*6;
  tsd2m7 = tsd2*7;

  tsd0m2 = tsd0*2;
  tsd0m3 = tsd0*3;
  tsd0m4 = tsd0*4;
  tsd0m5 = tsd0*5;
  tsd0m6 = tsd0*6;
  tsd0m7 = tsd0*7;

  ix = 0;
  iy = 0;
  for (ivk = 0; ivk < nvk; ivk++) {
    x1 = x[ix];
    x2 = x[ix+tsd2];
    x3 = x[ix+tsd2m2];
    x4 = x[ix+tsd2m3];
    x5 = x[ix+tsd2m4];
    x6 = x[ix+tsd2m5];
    x7 = x[ix+tsd2m6];
    x8 = x[ix+tsd2m7];

    t1 = _mm_add_pd(x1,x5);
    t2 = _mm_sub_pd(x1,x5);
    t3 = _mm_add_pd(x3,x7);
    t4 = _fjsp_madd_v2r8(cf,x3,zr);
    t5 = _fjsp_nmsub_v2r8(cf,x7,zr);
    t5 = _mm_add_pd(t4,t5);
    t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
    u1 = _mm_add_pd(t1,t3);
    u2 = _mm_add_pd(t2,t4);
    u3 = _mm_sub_pd(t1,t3);
    u4 = _mm_sub_pd(t2,t4);

    t1 = _mm_add_pd(x2,x6);
    t2 = _mm_sub_pd(x2,x6);
    t3 = _mm_add_pd(x4,x8);
    t4 = _fjsp_madd_v2r8(cf,x4,zr);
    t5 = _fjsp_nmsub_v2r8(cf,x8,zr);
    t5 = _mm_add_pd(t4,t5);
    t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
    v1 = _mm_add_pd(t1,t3);
    v2 = _mm_add_pd(t2,t4);
    v3 = _mm_sub_pd(t1,t3);
    v4 = _mm_sub_pd(t2,t4);

    y[iy] = _mm_add_pd(u1,v1);
    y[iy+tsd0m4] = _mm_sub_pd(u1,v1);

    t5 = _fjsp_set_v2r8(sp2,sp2);
    s = _fjsp_nmsub_sr2_v2r8(cf,v2,v2);
    s = _mm_mul_pd(t5,s);

    y[iy+tsd0] = _mm_add_pd(u2,s);

    y[iy+tsd0m5] = _mm_sub_pd(u2,s);

    y[iy+tsd0m2] = _fjsp_nmsub_sr2_v2r8(cf,v3,u3);

    y[iy+tsd0m6] = _fjsp_madd_sr2_v2r8(cf,v3,u3);

    t5 = _fjsp_set_v2r8(sp2,sp2);
    s = _fjsp_nmadd_sr2_v2r8(cf,v4,v4);
    s = _mm_mul_pd(t5,s);

    y[iy+tsd0m3] = _mm_add_pd(u4,s);

    y[iy+tsd0m7] = _mm_sub_pd(u4,s);

    ix = ix + 1;
    iy = iy + 1;
  }

  ix = 0;
  iy = 0;
  iw = 7;
  for (p = 1; p < np; p++) {
    w1 = w[iw];
    w2 = w[iw+1];
    w3 = w[iw+2];
    w4 = w[iw+3];
    w5 = w[iw+4];
    w6 = w[iw+5];
    w7 = w[iw+6];
    iw = iw + 7;

    ix = tsd0*p;
    iy = tsd1*p;

    for (ivk = 0; ivk < nvk; ivk++) {
      x1 = x[ix];
      x2 = x[ix+tsd2];
      x3 = x[ix+tsd2m2];
      x4 = x[ix+tsd2m3];
      x5 = x[ix+tsd2m4];
      x6 = x[ix+tsd2m5];
      x7 = x[ix+tsd2m6];
      x8 = x[ix+tsd2m7];

      t1 = _mm_add_pd(x1,x5);
      t2 = _mm_sub_pd(x1,x5);
      t3 = _mm_add_pd(x3,x7);
      t4 = _fjsp_madd_v2r8(cf,x3,zr);
      t5 = _fjsp_nmsub_v2r8(cf,x7,zr);
      t5 = _mm_add_pd(t4,t5);
      t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
      u1 = _mm_add_pd(t1,t3);
      u2 = _mm_add_pd(t2,t4);
      u3 = _mm_sub_pd(t1,t3);
      u4 = _mm_sub_pd(t2,t4);

      t1 = _mm_add_pd(x2,x6);
      t2 = _mm_sub_pd(x2,x6);
      t3 = _mm_add_pd(x4,x8);
      t4 = _fjsp_madd_v2r8(cf,x4,zr);
      t5 = _fjsp_nmsub_v2r8(cf,x8,zr);
      t5 = _mm_add_pd(t4,t5);
      t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
      v1 = _mm_add_pd(t1,t3);
      v2 = _mm_add_pd(t2,t4);
      v3 = _mm_sub_pd(t1,t3);
      v4 = _mm_sub_pd(t2,t4);

      y[iy] = _mm_add_pd(u1,v1);

      t3 = _mm_sub_pd(u1,v1);
      y[iy+tsd0m4] = _fjsp_madd_cp_neg_sr12_v2r8(w3,t4,_fjsp_madd_cp_v2r8(w4,t3,zr));

      t5 = _fjsp_set_v2r8(sp2,sp2);
      s = _fjsp_nmsub_sr2_v2r8(cf,v2,v2);
      s = _mm_mul_pd(t5,s);

      t3 = _mm_add_pd(u2,s);
      y[iy+tsd0] = _fjsp_madd_cp_neg_sr12_v2r8(w1,t3,_fjsp_madd_cp_v2r8(w1,t3,zr));

      t3 = _mm_sub_pd(u2,s);
      y[iy+tsd0m5] = _fjsp_madd_cp_neg_sr12_v2r8(w5,t3,_fjsp_madd_cp_v2r8(w5,t3,zr));

      t3 = _fjsp_nmsub_sr2_v2r8(cf,v3,u3);
      y[iy+tsd0m2] = _fjsp_madd_cp_neg_sr12_v2r8(w2,t3,_fjsp_madd_cp_v2r8(w2,t3,zr));

      t3 = _fjsp_madd_sr2_v2r8(cf,v3,u3);
      y[iy+tsd0m6] = _fjsp_madd_cp_neg_sr12_v2r8(w6,t3,_fjsp_madd_cp_v2r8(w6,t3,zr));

      t5 = _fjsp_set_v2r8(sp2,sp2);
      s = _fjsp_nmadd_sr2_v2r8(cf,v4,v4);
      s = _mm_mul_pd(t5,s);

      t3 = _mm_add_pd(u4,s);
      y[iy+tsd0m3] = _fjsp_madd_cp_neg_sr12_v2r8(w3,t3,_fjsp_madd_cp_v2r8(w3,t3,zr));

      t3 = _mm_sub_pd(u4,s);
      y[iy+tsd0m7] = _fjsp_madd_cp_neg_sr12_v2r8(w7,t3,_fjsp_madd_cp_v2r8(w7,t3,zr));

      ix = ix + 1;
      iy = iy + 1;
    }
  }
  return 0;
}

int fft_tpb_(int *ilxy, int *ibxy, int *iexy, int *ilz,
	     double *dsrc, double *ddst) {
  int lxy,bxy,exy,lz;
  __m128d *src, *dst;

  int i,j,pxy;
  int tsd1, tsd2, tbs;

  lxy = *ilxy;
  bxy = *ibxy;
  exy = *iexy;
  lz = *ilz;

  src = (__m128d *)dsrc;
  dst = (__m128d *)ddst;

  pxy = exy-bxy+1;

  tsd1 = lxy;
  tsd2 = lz;
  tbs = bxy-1;

  for (j = 0; j < lz; j++) {
    src = (__m128d *)dsrc + tbs + tsd1*j;
    dst = (__m128d *)ddst + j;
    for (i = 0; i < pxy; i++) {
      dst[tsd2*i] = src[i];
    }
  }

  return 0;
}

int fft_tpbs_(int *ilxy, int *ibxy, int *iexy, int *ilz,
	     double *dsrc, double *ddst, double *scl) {
  int lxy,bxy,exy,lz;
  __m128d *src, *dst;

  int i,j,pxy;
  __m128d ts;
  int tsd1, tsd2, tbs;

  lxy = *ilxy;
  bxy = *ibxy;
  exy = *iexy;
  lz = *ilz;

  src = (__m128d *)dsrc;
  dst = (__m128d *)ddst;

  ts = _fjsp_set_v2r8(*scl,*scl);
  pxy = exy-bxy+1;

  tsd1 = lxy;
  tsd2 = lz;
  tbs = bxy-1;

  for (j = 0; j < lz; j++) {
    src = (__m128d *)dsrc + tbs + tsd1*j;
    dst = (__m128d *)ddst + j;
    for (i = 0; i < pxy; i++) {
      dst[tsd2*i] = _mm_mul_pd(ts,src[i]);
    }
  }

  return 0;
}
