!
! Copyright (c) 2012 TAPP Consortium
! Copyright (c) 2012-2017 Yoshihide Yoshimoto
!
! Initially written by K. Yoshizawa 2012.8 for xTAPP
!
module subr_inputparam
  character(len=*),parameter:: magic_inputparam = '# main data'
contains
  !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
  subroutine read_tappinput(iin,ax,aa,ncomp,nspin,ndens,ismode,qf,qm,qc,beta, &
       qbtp,nkd,nsi,nw,znext,dspin,nopt,iopt,eps,eeps,ekbt, &
       uptime,niter0,niter1,itsb,ndiag0,ndiag1, &
       qbsp,nbsf,fwqcom,xctp,qexat,thr_hybrid,osqnrm,axetp,qprcnd,thr_noncol, &
       spaxs,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    use xcf_types
    implicit none

    integer, intent(in) :: iin
    real(8), intent(out)  :: ax
    real(8), dimension(3,3), intent(out) :: aa
    integer, intent(out) :: ncomp, nspin, ndens, ismode
    real(8), intent(out) :: qf, qm, qc, beta, qbtp(3)
    integer, intent(out) :: nkd, nsi
    integer, intent(out) :: nw
    real(8), intent(out) :: znext, dspin
    integer,intent(in):: nopt
    integer, intent(out) :: iopt(nopt)
    real(8), intent(out) :: eps, eeps, ekbt
    real(8), intent(out) :: uptime
    integer, intent(out) :: niter0, niter1, itsb, ndiag0, ndiag1
    real(8), intent(out) :: qexat(2),thr_hybrid
    integer, intent(out) :: osqnrm
    integer, intent(out) :: axetp
    real(8), intent(out) :: qprcnd(6)
    real(8), intent(out) :: thr_noncol, spaxs(3)

    real(8) :: qbsp, fwqcom
    integer :: nbsf, xctp

    integer,intent(in) :: mssg
    integer :: info

    real(8) :: lattice_factor
    real(8) :: lattice_list(9)

    integer :: number_component, number_spin, spin_mode, number_density

    real(8) :: cutoff_local_potential, cutoff_wave_function
    real(8) :: cutoff_uspp_q, xtrap_beta
    real(8) :: cutoff_btp_A, cutoff_btp_gc, cutoff_btp_sigma

    integer :: number_element, number_atom
    integer :: number_band
    real(8) :: extra_charge, diff_spin

    integer :: number_xtrap_stage, initial_wfn, initial_lpt, store_wfn, &
         chain_calc

    real(8) :: scf_converge, scf_converge_energy, elec_kbt
    real(8) :: control_uptime
    integer :: scf_number_iter_1st, scf_number_iter
    integer :: davidson_flag
    integer :: davidson_number_diag_1st, davidson_number_diag

    real(8) :: cutoff_bspl_fit, control_q_han_win_fac
    integer :: control_nbsf
    character(len=64):: xc_type

    real(8) :: cutoff_exch_att_rc, cutoff_exch_att_rd
    real(8) :: scf_threshold_hybrid
    integer :: compat_old_squ_norm
    character(len=64):: aux_eng_type
    real(8) :: cutoff_precond(4)
    real(kind=8) :: precond_beta, precond_thr

    real(8) :: scf_threshold_noncol, noncol_spin_axis(3)

    namelist /tappinput/ lattice_factor, &
         lattice_list, &
         number_component, number_spin, spin_mode, number_density, &
         cutoff_local_potential, cutoff_wave_function, &
         cutoff_uspp_q, xtrap_beta, &
         cutoff_btp_A, cutoff_btp_gc, cutoff_btp_sigma, &
         number_element, number_atom, number_band, &
         extra_charge, diff_spin, elec_kbt, &
         number_xtrap_stage, initial_wfn, initial_lpt, &
         store_wfn, chain_calc, &
         scf_converge, scf_converge_energy, &
         scf_number_iter_1st, scf_number_iter, &
         davidson_flag, &
         davidson_number_diag_1st, davidson_number_diag, &
         control_uptime, &
         cutoff_bspl_fit, control_q_han_win_fac, control_nbsf, &
         xc_type, &
         cutoff_exch_att_rc, cutoff_exch_att_rd, scf_threshold_hybrid, &
         compat_old_squ_norm, &
         aux_eng_type, &
         precond_thr, precond_beta, cutoff_precond, &
         scf_threshold_noncol, noncol_spin_axis

    ! # initial parameter

    lattice_factor = huge(1.d0)

    lattice_list = huge(1.0d0)

    number_component = 1
    number_spin = 1
    spin_mode = -1
    number_density = 1

    cutoff_local_potential = -1.0d0
    cutoff_wave_function = 0.0d0
    cutoff_uspp_q = -1.0d0
    xtrap_beta = 0.2d0

    cutoff_btp_A = 0.0d0
    cutoff_btp_gc = -1.0d0
    cutoff_btp_sigma = -1.0d0

    number_element = 0
    number_atom = 0

    extra_charge = 0.0d0
    diff_spin = 0.0d0
    elec_kbt = 0.0d0

    number_xtrap_stage = 2
    initial_wfn = 0
    initial_lpt = 0
    store_wfn = 0
    chain_calc = 0

    scf_converge = 1.0d-10
    scf_converge_energy = tiny(1.0d0)

    control_uptime = -1.0d0

    scf_number_iter_1st = 80
    scf_number_iter = 40
    davidson_flag = -1
    davidson_number_diag_1st = -1
    davidson_number_diag = 4

    cutoff_bspl_fit = -1.0d0
    control_nbsf = 2000
    control_q_han_win_fac = -1.0d0

    xc_type = 'PBE'

    cutoff_exch_att_rc = -1.0d0
    cutoff_exch_att_rd = -1.0d0
    scf_threshold_hybrid = 1.0d-8

    compat_old_squ_norm = 0

    aux_eng_type = 'NONE'

    precond_beta = 0.0d0
    precond_thr = -1.0d0
    cutoff_precond = -1.0d0

    scf_threshold_noncol = 1.0d-2
    noncol_spin_axis = 0.0d0

    ! read namelist

    read(iin,tappinput,err=100)

    ! set to default

    if (spin_mode.lt.0) then
       if (number_spin.eq.1) then
          spin_mode = 1
       else
          spin_mode = 2
       end if
    end if
    if (number_component.ne.1 .and. number_component.ne.2) then
       write(6,*) 'number_component must be 1 or 2'
       info = 1
    end if
    if (number_density.ne.1 .and. number_density.ne.4) then
       write(6,*) 'number_density must be 1 or 4'
       info = 1
    end if
    if (number_spin.eq.2) then
       if (number_component.eq.2) then
          write(6,*) 'number_component = 2 and number_spin = 2 are exclusive'
          info = 1
       end if
       if (number_density.ne.1) then
          write(6,*) 'when number_spin = 2, number_density must be 1'
          info = 1
       end if
    end if
    if (number_density.eq.4 .and. number_component.eq.1) then
       write(6,*) 'when number_density = 4, number_component must be 2'
       info = 1
    end if

    if (cutoff_local_potential .lt. 0.0d0) then
       cutoff_local_potential = 2.0d0*cutoff_wave_function + 0.1d0
    end if
    if (cutoff_uspp_q .lt. 0.0d0) then
       cutoff_uspp_q = 2.0d0*cutoff_wave_function
    end if
    if (cutoff_btp_A .eq. 0.0d0) then
       cutoff_btp_gc = cutoff_wave_function
       cutoff_btp_sigma = cutoff_wave_function
    end if

    if (davidson_number_diag_1st .lt. 0) then
       if (initial_wfn.eq.1 .and. initial_lpt.eq.1) then
          davidson_number_diag_1st = davidson_number_diag
       else
          davidson_number_diag_1st = 20
       end if
    end if

    if (cutoff_bspl_fit .lt. 0.0d0) then
       cutoff_bspl_fit = 1.5d0*cutoff_local_potential
    end if

    ! input check

    if ( (number_spin .ne. 1) .and. (number_spin .ne. 2) ) then
       if (mssg.gt.-1) write(6,'(1x,a,i2)') &
            'invalid value in number_spin = ', number_spin
       info = 1
    end if
    if ( (spin_mode .ne. 1) .and. (spin_mode .ne. 2) ) then
       if (mssg.gt.-1) write(6,'(1x,a,i2)') &
            'invalid value in spin_mode = ', spin_mode
       info = 1
    end if
    if (cutoff_uspp_q .gt. cutoff_local_potential) then
       if (mssg.gt.-1) then
          write(6,*)' CAUTION !'
          write(6,*)' cutoff_uspp_q > cutoff_local_potential'
          write(6,*)' setting cutoff_uspp_q = cutoff_local_potential'
       end if
       cutoff_uspp_q = cutoff_local_potential
    end if
    if (cutoff_local_potential .lt. 2.0d0*cutoff_wave_function) then
       if (mssg.gt.-1) then
          write(6,*)' warning: cutoff_local_potential < 2*cutoff_wave_function'
          write(6,*)' V_{loc}|\Psi> will not be calculated correctly.'
       end if
    end if
    if (cutoff_btp_A.gt.0.0d0 .and. cutoff_btp_gc.gt.cutoff_wave_function) then
       if (mssg.gt.-1) write(6,'(1x,a)') 'cutoff_btp_gc must be < cutoff_wave_function.'
       info = 1
    end if

    if (cutoff_bspl_fit.lt.cutoff_local_potential) then
       if (mssg.gt.-1) &
            write(6,'(1x,a)') 'cutoff_qbsp must be > cutoff_local_potential'
       info = 1
    end if
    if (control_nbsf.le.0) then
       if (mssg.gt.-1) write(6,'(1x,a)') 'cutoff_nbsf must be > 0'
       info = 1
    end if
    if (control_q_han_win_fac.ge.0.0d0 &
         .and. control_q_han_win_fac.lt.2.0d0) then
       if (mssg.gt.-1) write(6,'(1x,a)') 'cutoff_q_han_win_fac must be >= 2.0'
       info = 1
    end if

    if (trim(aux_eng_type).ne.'NONE' .and.  &
         trim(aux_eng_type).ne.'DFTD2' .and. &
         trim(aux_eng_type).ne.'DFTD2_SG' .and. &
         trim(aux_eng_type).ne.'DFTD3_SG') then
       if (mssg.gt.-1) then
          write(6,'(1x,a,a)') 'unknown aux_eng_type = ', trim(aux_eng_type)
       end if
       info = 1
    end if

    if (store_wfn.lt.0 .or. store_wfn.gt.1) then
       if (mssg.gt.-1) then
          write(6,'(1x,a,a)') 'invalid value for store_wfn = ', store_wfn
       end if
       info = 1
    end if

    if (chain_calc.lt.0 .or. chain_calc.gt.3) then
       if (mssg.gt.-1) then
          write(6,'(1x,a,a)') 'invalid value for chain_calc = ', chain_calc
       end if
       info = 1
    end if

    if (precond_thr .gt. 0.0d0) then
       if (precond_thr .ge. 1000.0d0) then
          precond_thr = huge(1.0d0)
       end if
       if (precond_beta .lt. 0.0d0 .or. precond_beta .gt. 1.0d0) then
          if (mssg.gt.-1) then
             write(6,'(1x,a,a)') 'invalid range for precond_beta = ', precond_beta
          end if
          info = 1
       end if
       if (minval(qprcnd).lt.0.0d0) then
          if (mssg.gt.-1) then
             write(6,'(1x,a,a)') 'invalid range for cutoff_precond'
          end if
          info = 1
       end if
       if (elec_kbt.le.0.0d0) then
          if (mssg.gt.-1) then
             write(6,'(1x,a,a)') 'elec_kbt > 0' // &
                  ' when preconditioning of local potential is enabled'
          end if
          info = 1
       end if
    end if

    if (mssg.gt.-1) then
       write(6,'(a)') magic_inputparam
       write(6,tappinput)
    end if

    ! substitute vars

    ax = lattice_factor

    aa(1,1) = lattice_list(1)
    aa(2,1) = lattice_list(2)
    aa(3,1) = lattice_list(3)
    aa(1,2) = lattice_list(4)
    aa(2,2) = lattice_list(5)
    aa(3,2) = lattice_list(6)
    aa(1,3) = lattice_list(7)
    aa(2,3) = lattice_list(8)
    aa(3,3) = lattice_list(9)
    !  aa = ax*aa

    ncomp = number_component
    nspin = number_spin
    ndens = number_density
    ismode = spin_mode

    qm = cutoff_wave_function
    qf = cutoff_local_potential
    qc = cutoff_uspp_q
    beta = xtrap_beta

    qbtp(1) = cutoff_btp_A
    qbtp(2) = cutoff_btp_gc
    qbtp(3) = cutoff_btp_sigma

    nkd = number_element
    nsi = number_atom

    nw = number_band

    znext = extra_charge
    dspin = diff_spin

    iopt(1) = number_xtrap_stage
    iopt(2) = initial_wfn
    iopt(3) = initial_lpt
    iopt(4) = store_wfn
    iopt(5) = chain_calc

    uptime = control_uptime

    ekbt = elec_kbt

    eps = scf_converge
    eeps = scf_converge_energy
    niter0 = scf_number_iter_1st
    niter1 = scf_number_iter
    itsb = davidson_flag
    ndiag0 = davidson_number_diag_1st
    ndiag1 = davidson_number_diag

    qbsp = cutoff_bspl_fit
    nbsf = control_nbsf
    fwqcom = control_q_han_win_fac

    if (trim(xc_type).eq.xcf_pbe_name) then
       xctp = xcf_pbe
    else if (trim(xc_type).eq.xcf_pbesol_name) then
       xctp = xcf_pbesol
    else if (trim(xc_type).eq.xcf_capz_name) then
       xctp = xcf_capz
    else if (trim(xc_type).eq.xcf_pw91_name) then
       xctp = xcf_pw91
    else if (trim(xc_type).eq.xcf_ldapw92_name) then
       xctp = xcf_ldapw92
    else if (trim(xc_type).eq.xcf_hartree_fock_name) then
       xctp = xcf_hartree_fock
    else if (trim(xc_type).eq.xcf_pbe0_name) then
       xctp = xcf_pbe0
    else if (trim(xc_type).eq.xcf_b3lyp_name) then
       xctp = xcf_b3lyp
    end if

    qexat(1) = cutoff_exch_att_rc
    qexat(2) = cutoff_exch_att_rd
    thr_hybrid = scf_threshold_hybrid
    osqnrm = compat_old_squ_norm

    thr_noncol = scf_threshold_noncol
    spaxs = noncol_spin_axis
    if (maxval(abs(spaxs)).ne.0.0d0) then
       spaxs = spaxs/sqrt(spaxs(1)**2+spaxs(2)**2+spaxs(3)**2)
    end if

    if (trim(aux_eng_type).eq.'NONE') then
       axetp = 0
    else if (trim(aux_eng_type).eq.'DFTD2') then
       axetp = 1
    else if (trim(aux_eng_type).eq.'DFTD2_SG') then
       axetp = 2
    else if (trim(aux_eng_type).eq.'DFTD3_SG') then
       axetp = 3
    end if

    qprcnd(1) = cutoff_precond(1)
    qprcnd(2) = cutoff_precond(2)
    qprcnd(3) = cutoff_precond(3)
    qprcnd(4) = cutoff_precond(4)
    qprcnd(5) = precond_beta
    qprcnd(6) = precond_thr

    return

100 continue
    if (mssg.gt.-1) then
       write(6,'(A)')'error occurs @ reading namelist : tappinput'
       write(6,'(A)')'stop.'
    end if

  end subroutine read_tappinput
end module subr_inputparam

module subr_inputstropt
  character(len=*),parameter:: magic_tappstropt = '# struct_opt data'
contains
  subroutine read_tappstropt(iin,ax,nsi,eepsa,feps, &
       decr,okatom,ncycl,nrfr,most,syspext,stscl,cdm,optmode,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    use xcf_types
    implicit none

    integer, intent(in) :: iin, nsi
    real(8), intent(in) :: ax
    real(8), intent(out) :: eepsa, feps(2)
    real(8), intent(out) :: decr, okatom(2)
    integer, intent(out) :: ncycl, nrfr, most
    real(8), intent(out) :: syspext
    real(8), intent(out) :: stscl(6), cdm(6)
    integer, intent(out) :: optmode

    integer,intent(in) :: mssg
    integer :: info

    real(8) :: search_1d_fratio, displacement_max
    real(8) :: cell_deform_displ_max
    real(8) :: converge_energy, converge_force, converge_stress
    integer :: number_cycle, refresh_cycle
    integer :: search_1d_max_step

    real(8) :: extern_pressure
    real(8) :: stress_scale(6), cell_deform(6)

    character(len=24) :: mode

    namelist /struct_opt/ &
         converge_energy, converge_force, converge_stress, &
         search_1d_fratio, &
         displacement_max, &
         cell_deform_displ_max, &
         number_cycle, refresh_cycle, &
         search_1d_max_step, &
         extern_pressure, stress_scale, cell_deform, mode

    ! # initial parameter

    converge_energy = 1.0d-10
    converge_force = 1.0d-3
    converge_stress = 1.0d-3
    search_1d_fratio = 0.1d0
    displacement_max = 0.5d0
    cell_deform_displ_max = 0.05d0

    number_cycle = -1
    refresh_cycle = 8
    search_1d_max_step = 4

    extern_pressure = 0.0d0
    stress_scale(1) = 0.0d0
    stress_scale(2:6) = -1.0d0
    cell_deform = 0.0d0

    mode = 'NONE'

    ! read namelist

    read(iin,struct_opt,err=100)

    ! set to default

    if (number_cycle.lt.0) then
       number_cycle = 3*nsi
       if (maxval(abs(stress_scale)) .gt. 0.0d0) then
          number_cycle = number_cycle + 6
       end if
    end if

    if (maxval(stress_scale(2:6)).eq.-1.0d0 &
         .and. minval(stress_scale(2:6)).eq.-1.0d0) then
       stress_scale(2:6) = stress_scale(1)
    end if

    if (trim(mode).ne.'NONE' .and. &
         trim(mode).ne.'HYPER_PLANE' .and. &
         trim(mode).ne.'FORCE_INVERSION') then
       if (mssg.gt.-1) then
          write(6,*) 'error: unknown structure optimize mode'
       end if
       mode = 'NONE'
       info = 1
    end if

    if (trim(mode).eq.'HYPER_PLANE' .or. &
         trim(mode).eq.'FORCE_INVERSION' ) then
       if (maxval(abs(stress_scale)) .gt. 0.0d0) then
          if (mssg.gt.-1) then
             write(6,*) 'error: cell deformation is not allowed ' &
                  // 'in HYPER_PLANE and FORCE_INVERSION mode'
          end if
          stress_scale = 0.0d0
          info = 1
       end if
    end if

    if (mssg.gt.-1) then
       write(6,'(a)') magic_tappstropt
       write(6,struct_opt,err=100)
    end if

    ! substitute vars

    eepsa = converge_energy
    feps(1) = converge_force
    feps(2) = converge_stress
    decr = search_1d_fratio
    okatom(1) = displacement_max
    okatom(2) = cell_deform_displ_max

    ncycl = number_cycle
    nrfr = refresh_cycle
    most = search_1d_max_step

    syspext = extern_pressure
    stscl(:) = stress_scale
    cdm = cell_deform

    if (trim(mode).eq.'NONE') then
       optmode = 0
    else if (trim(mode).eq.'HYPER_PLANE') then
       optmode = 1
    else if (trim(mode).eq.'FORCE_INVERSION') then
       optmode = 2
    end if

    return

100 continue
    if (mssg.gt.-1) then
       write(6,'(A)')'error occurs @ reading namelist : tappstropt'
    end if
    info = 1
    return

  end subroutine read_tappstropt
end module subr_inputstropt

module subr_inputsym
  character(len=*),parameter:: magic_tappsym = '# symmetry data'
contains
  !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
  subroutine read_tappsym(iin,ng,nsv,nnp,rg,pg,srg,nsymq,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    implicit none

    integer,intent(in) :: iin, nsymq

    integer :: ng, nsv, nnp
    integer :: rg(3,3,nsymq), pg(3,nsymq), srg(nsymq)
    integer,intent(in) :: mssg
    integer :: info

    integer :: i, j
    integer :: ig, igh

    character(len=32):: symmetry_format
    integer :: number_sym_op, has_inversion, denom_trans
    integer :: input_mode

    namelist /symmetry/ symmetry_format, &
         number_sym_op, has_inversion, denom_trans

    symmetry_format = 'reciprocal'
    denom_trans = 1
    has_inversion = 0

    read(iin,symmetry)

    if (number_sym_op.gt.nsymq) then
       if (mssg.gt.-1) write(6,'(1x,a,i2)') &
            'number_sym_op is too large', number_sym_op
       info = 1
    end if
    if ( (has_inversion .ne. 0) .and. (has_inversion .ne. 1 ) ) then
       if (mssg.gt.-1) write(6,'(1x,a,i2)') &
            'invalid has_inversion = ', has_inversion
       info = 1
    end if
    if (trim(symmetry_format).eq.'reciprocal') then
       input_mode = 1
    else if (trim(symmetry_format).eq.'reciprocal_mag') then
       input_mode = 2
    else
       if (mssg.gt.-1) write(6,'(1x,a)') 'invalid symmetry format'
       info = 1
    end if
    if (denom_trans .le. 0) then
       if (mssg.gt.-1) write(6,'(1x,a)') 'denom_trans must be > 0'
       info = 1
    end if

    if (input_mode .eq. 1) then
       do igh=1,number_sym_op
          ig = 2*igh-1
          read(iin,*) ((rg(i,j,ig),i=1,3),j=1,3), (pg(i,ig),i=1,3)
          srg(ig) = 1
          
          ig = 2*igh
          rg(:,:,ig) = rg(:,:,ig-1)
          pg(:,ig) = pg(:,ig-1)
          srg(ig) = -1
       end do

       if (mssg.gt.-1) then
          write(6,'(a)') magic_tappsym
          write(6,symmetry)
          write(6,'(3(2x,3(1x,i2)),3x,3(1x,i2),a)') &
               ((rg(i,j,1),i=1,3),j=1,3),(pg(i,1),i=1,3), ' ! rg(3,3), pg(3)'
          do ig=3,2*number_sym_op,2
             write(6,'(3(2x,3(1x,i2)),3x,3(1x,i2))') &
                  ((rg(i,j,ig),i=1,3),j=1,3),(pg(i,ig),i=1,3)
          end do
       end if
       number_sym_op = number_sym_op*2
    else
       do ig=1,number_sym_op
          read(iin,*) ((rg(i,j,ig),i=1,3),j=1,3), (pg(i,ig),i=1,3), srg(ig)
       end do
       do ig=1,number_sym_op
          if (abs(srg(ig)).ne.1) then
             write(6,'(1x,a)') 'srg must be 1 or -1'
             info = 1
             exit
          end if
       end do

       if (mssg.gt.-1) then
          write(6,'(a)') magic_tappsym
          write(6,symmetry)
          write(6,'(3(2x,3(1x,i2)),3x,3(1x,i2),4x,i2,a)') &
               ((rg(i,j,1),i=1,3),j=1,3),(pg(i,1),i=1,3),srg(1), &
               ' ! rg(3,3), pg(3), srg'
          do ig=2,number_sym_op
             write(6,'(3(2x,3(1x,i2)),3x,3(1x,i2),4x,i2)') &
                  ((rg(i,j,ig),i=1,3),j=1,3),(pg(i,ig),i=1,3),srg(ig)
          end do
       end if
    end if
    
    ng = number_sym_op
    nsv = has_inversion
    nnp = denom_trans

  end subroutine read_tappsym
end module subr_inputsym

module subr_inputatom
  character(len=*),parameter:: magic_tappatom = '# atom data'
contains
  !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
  subroutine read_tappatom(iin,nkd,zoin,znin,nsi,kd,asi,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    implicit none

    integer, intent(in) :: iin, nkd
    integer, intent(in) :: nsi

    real(8), pointer :: zoin(:), znin(:)
    integer, dimension(:), pointer :: kd
    real(8), dimension(:,:), pointer :: asi

    integer, intent(in) :: mssg
    integer :: info

    integer :: ikd,it

    allocate(zoin(nkd),znin(nkd),kd(nsi),asi(3,nsi))

    do ikd = 1, nkd
       read(iin,*) zoin(ikd), znin(ikd)
       if (zoin(ikd).lt.0.0d0) then
          if (mssg.gt.-1) write(6,'(1x,a,i4)') 'zo < 0 for # ',ikd
          info = 1
       end if
       if (znin(ikd).lt.0.0d0) then
          if (mssg.gt.-1) write(6,'(1x,a,i4)') 'zn < 0 for # ',ikd
          info = 1
       end if
    end do

    do it = 1, nsi
       read(iin,*) kd(it), asi(1,it), asi(2,it), asi(3,it)
       if (kd(it).le.0) then
          if (mssg.gt.-1) write(6,'(1x,a,i4)') 'atom_kind <= 0 for # ',it
          info = 1
       end if
       if (kd(it).gt.nkd) then
          if (mssg.gt.-1) &
               write(6,'(1x,a,i4)') 'atom_kind > number_element for # ',it
          info = 1
       end if
    end do

    if (mssg.gt.-1) then
       write(6,'(a)') magic_tappatom
       write(6,'(1x,f13.9,1x,f13.9,a)') zoin(1),znin(1), ' ! zo, zn'
       do ikd=2,nkd
          write(6,'(1x,f13.9,1x,f13.9)') zoin(ikd),znin(ikd)
       end do
       write(6,'(1x,i5,3(1x,e18.11),a)') kd(1),asi(1,1),asi(2,1),asi(3,1), &
            ' ! atom_kind, pos_a, pos_b, pos_c (in lattice coordiate)'
       do it=2,nsi
          write(6,'(1x,i5,3(1x,e18.11))') kd(it),asi(1,it),asi(2,it),asi(3,it)
       end do
    end if

  end subroutine read_tappatom
end module subr_inputatom

module subr_inputkpoints
  character(len=*),parameter:: magic_tappkpoints = '# k-points data'
contains
  !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
  subroutine read_tappkpoints(iin,nw,ax,aa,nb1,nb2,ekbt, &
       rf,ndx,ndy,ndz,nk,nsk,mmm,idat,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    implicit none

    integer, intent(in) :: iin, nw
    real(kind=8),intent(in):: ax,aa(3,3),ekbt

    integer :: nb1, nb2
    real(kind=8):: rf
    integer :: ndx, ndy, ndz
    integer :: nk

    integer :: nsk
    integer, dimension(:,:,:), pointer :: mmm
    integer :: idat(3)

    integer, intent(in) :: mssg
    integer :: info

    integer :: i, j, k
    integer :: insk
    real(kind=8):: vtmp(3), tmp
    integer :: ll(3)

    character(len=48) :: dos_mode
    real(kind=8):: cutoff_dos_cos
    integer :: dos_band_lower, dos_band_upper
    integer :: dos_mesh(3)
    integer :: bz_mesh
    integer :: bz_number_tile
    integer :: rmesh_range(3), rmesh_number_shell

    namelist /smpl_kpt/ dos_mode, dos_band_lower, dos_band_upper, &
         dos_mesh, bz_mesh, bz_number_tile, cutoff_dos_cos, &
         rmesh_range, rmesh_number_shell

    dos_band_lower = 1
    dos_band_upper = nw
    cutoff_dos_cos = -1.0d0
    dos_mesh = 0
    rmesh_range = 0
    rmesh_number_shell = 0

    read(iin, smpl_kpt, err=100)

    if (trim(dos_mode) .eq. 'METHFESSEL_PAXTON') then
       dos_mesh = 0
       if (ekbt .le. 0.0d0) then
          if (mssg.gt.-1) then
             write(6,*) 'error: elec_kbt <= 0 for METHFESSEL_PAXTON mode'
          end if
          info = 1
       end if
    else if (trim(dos_mode) .eq. 'FERMI') then
       dos_mesh = -1
       if (ekbt .le. 0.0d0) then
          if (mssg.gt.-1) then
             write(6,*) 'error: elec_kbt <= 0 for FERMI mode'
          end if
          info = 1
       end if
    else if (trim(dos_mode) .eq. 'COS') then
       continue
    end if

    if ( (trim(dos_mode) .eq. 'COS') ) then
       if (maxval(dos_mesh).eq.0 .and. bz_number_tile.ge.2) then
          if (mssg.gt.-1) then
             write(6,*) 'error: dos_mesh <= 0 for manual COS fit mode'
          end if
          info = 1
       end if
       if (bz_number_tile.ge.2) then
          if ((rmesh_number_shell.le.0) .and. (maxval(rmesh_range).le.0)) then
             if (mssg.gt.-1) then
                write(6,*) 'error: rmesh_range and rmesh_number_shell <= 0 for manual COS fit mode'
             end if
             info = 1
          end if
          if ((rmesh_number_shell.gt.0) .and. (minval(rmesh_range).gt.0)) then
             if (mssg.gt.-1) then
                write(6,*) 'error: rmesh_range and rmesh_number_shell are exclusive'
             end if
             info = 1
          end if
       end if
    end if
    if (bz_number_tile.ge.2 .and. cutoff_dos_cos.le.0) then
       if (mssg.gt.-1) &
            write(6,*) 'error: set cutoff_dos_cos > 0 when nsk >= 2'
       info = 1
    end if
    if (dos_band_lower.gt.nw) then
       if (mssg.gt.-1) write(6,*) 'error: dos_band_lower > number_band'
       info = 1
    end if
    if (dos_band_lower.gt.dos_band_upper) then
       if (mssg.gt.-1) write(6,*) 'error: dos_band_lower > dos_band_upper'
       info = 1
    end if

    nb1 = dos_band_lower
    nb2 = dos_band_upper
    nk = bz_mesh
    nsk = bz_number_tile
    if (rmesh_number_shell.gt.0) then
       idat(1) = rmesh_number_shell
       idat(2) = -1
       idat(3) = -1
    else if (maxval(rmesh_range).gt.0) then
       idat = rmesh_range
    end if

    allocate(mmm(3,2,nsk))
    mmm = 1

    do insk = 1, nsk
       do j = 1, 2
          read(iin,*)(mmm(i,j,insk),i=1,3)
       end do
    end do

    if (trim(dos_mode) .eq. 'COS' .and. maxval(dos_mesh) .eq. 0) then
       do i=1,3
          dos_mesh(i) = 2*nk/mmm(i,2,1)
          if (mod(2*nk,mmm(i,2,1)).ne.0) then
             dos_mesh(i) = dos_mesh(i) + 1
          end if
       end do
    end if
    ndx = dos_mesh(1)
    ndy = dos_mesh(2)
    ndz = dos_mesh(3)

    if (trim(dos_mode) .eq. 'COS' .and. cutoff_dos_cos .lt. 0.0d0) then
       cutoff_dos_cos = 0.0d0

       ll = (nk/mmm(:,2,1)-1)/2
       do k=-1,1,2
          do j=-1,1,2
             do i=-1,1,2
                vtmp = i*ll(1)*aa(:,1)+j*ll(2)*aa(:,2)+k*ll(3)*aa(:,3)
                tmp = vtmp(1)*vtmp(1)+vtmp(2)*vtmp(2)+vtmp(3)*vtmp(3)
                tmp = sqrt(tmp)
                cutoff_dos_cos = max(cutoff_dos_cos,tmp)
             end do
          end do
       end do
       cutoff_dos_cos = 1.1d0*ax*cutoff_dos_cos
    end if
    rf = cutoff_dos_cos

    if (mssg.gt.-1) then
       write(6,'(a)') magic_tappkpoints
       write(6,smpl_kpt)

       do insk = 1, nsk
          do j = 1, 2
             write(6,'(3(1x,i5))')(mmm(i,j,insk),i=1,3)
          end do
       end do
    end if

    return

100 continue
    if (mssg.gt.-1) then
       write(6,'(A)')'error occurs @ reading namelist : smpl_kpt'
    end if
    info = 1
    return
  end subroutine read_tappkpoints
end module subr_inputkpoints

module subr_inputmoldyn
  character(len=*),parameter:: magic_tappmoldyn = '# md data'
contains
  subroutine read_tappmd(iin,mdmode,deltm,imdstp,emdstp,imtrw, &
       nkd,mass,nsi,cavl,syskbt,syspext,va0,sysvol, &
       mdparam,mdstat,msgcycl,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    use chain_moldyn
    implicit none

    integer, intent(in) :: iin, nkd, nsi
    real(8), intent(in) :: va0
    character(len=16):: mdmode
    integer, intent(out) :: imdstp,emdstp,imtrw,msgcycl
    real(8),pointer,intent(out) :: mass(:),cavl(:,:)
    real(8),intent(out) :: deltm,syskbt,syspext,sysvol
    real(8):: mdparam(num_mdparam),mdstat(num_mdstat)

    integer,intent(in) :: mssg
    integer :: info

    real(8):: delta_time
    integer:: init_md_step, end_md_step, mode_xtrap_rho_wfn
    integer:: message_cycle
    real(8):: system_volume, system_kbt, system_pressure
    real(8):: stern_omega_zeta, stern_omega_eta
    real(8):: stern_zeta, stern_eta, stern_sigma
    real(8):: brdsn_tauinv
    integer:: abinitio_cycle
    real(8):: wang_landau_update_factor
    integer:: wang_landau_weight_cycle, wang_landau_factor_cycle

    integer:: ikd, it

    namelist /mol_dyn/ &
         mdmode, delta_time, init_md_step, end_md_step, &
         mode_xtrap_rho_wfn, &
         message_cycle, &
         system_volume, system_kbt, system_pressure, &
         stern_omega_zeta, stern_omega_eta, &
         stern_zeta, stern_eta, stern_sigma, &
         brdsn_tauinv, &
         abinitio_cycle, &
         wang_landau_update_factor, &
         wang_landau_weight_cycle, &
         wang_landau_factor_cycle
         
    ! # initial parameter

    mdmode = 'NVE'
    delta_time = -1.0d0
    init_md_step = 0
    end_md_step = huge(1)
    mode_xtrap_rho_wfn = 2
    message_cycle = 1
    system_volume = va0
    system_kbt = -1.0d0
    system_pressure = 0.0d0
    stern_omega_zeta = -1.0d0
    stern_omega_eta = -1.0d0
    stern_zeta = 0.0d0
    stern_eta = 0.0d0
    stern_sigma = 0.0d0
    brdsn_tauinv = 0.0d0
    abinitio_cycle = 1
    wang_landau_update_factor = 0.0d0
    wang_landau_weight_cycle = -1
    wang_landau_factor_cycle = -1

    ! read namelist

    read(iin,mol_dyn,err=100)

    allocate(mass(nkd),cavl(3,nsi))
    if (mdmode .ne. 'SMPL') then
       do ikd=1,nkd
          read(iin,*) mass(ikd)
       end do
       do it=1,nsi
          read(iin,*) cavl(1,it),cavl(2,it),cavl(3,it)
       end do
    else
       mass = 0.0d0
       cavl = 0.0d0
    end if

    ! set to default

    ! input check

    if (mdmode .eq. 'NVE' .or. mdmode .eq. 'MBMT_ST' .or. &
       mdmode .eq. 'NPT_ST' .or. mdmode .eq. 'MT_ST_R' .or. &
       mdmode .eq. 'MBMT_ST_R' .or. mdmode .eq. 'NPT_ST_R' .or. &
       mdmode .eq. 'NVT_BRDSN' .or. &
       mdmode .eq. 'SMPL') then
       continue
    else
       if (mssg.gt.-1) write(6,'(1x,a)') 'unknown md_mode'
       info = 1
    end if

    if (delta_time .le. 0.0d0) then
       if (mssg.gt.-1) then
          write(6,'(1x,a,i3)') 'delta_time <= 0.0 or delta_time is unspecfied'
       end if
       info = 1
    end if

    if (mode_xtrap_rho_wfn.lt.0 .or. mode_xtrap_rho_wfn.gt.2) then
       if (mssg.gt.-1) then
          write(6,'(1x,a,i3)') 'unsupported mode_xtrap_rho_wfn = ', &
               mode_xtrap_rho_wfn
       end if
       info = 1
    end if
    if (system_kbt.lt.0.0d0) then
       if (mssg.gt.-1) write(6,'(1x,a)') 'system_kbt is not specified'
       info = 1
    end if

    
    if ( (mdmode .eq. 'NPT_ST') .or. (mdmode .eq. 'MBMT_ST') .or. &
         (mdmode .eq. 'MT_ST_R') .or. (mdmode .eq. 'MBMT_ST_R') ) then
       if (stern_omega_zeta.lt.0.0d0) then
          if (mssg.gt.-1) write(6,'(1x,a)') 'stern_omega_zeta < 0'
          info = 1
       end if
       if (stern_omega_eta.lt.0.0d0) then
          if (mssg.gt.-1) write(6,'(1x,a)') 'stern_omega_eta < 0'
          info = 1
       end if
    end if

    if ( mdmode .eq. 'NVT_BRDSN') then
       if (brdsn_tauinv.lt.0.0d0) then
          if (mssg.gt.-1) write(6,'(1x,a)') 'brdsn_tauinv < 0'
          info = 1
       end if
    end if

    if ( (mdmode .eq. 'MBMT_ST') .or. (mdmode .eq. 'MT_ST_R') .or. &
         (mdmode .eq. 'MBMT_ST_R') ) then
       if (wang_landau_update_factor.lt.0.0d0) then
          if (mssg.gt.-1) write(6,'(1x,a)') 'wang_landau_update_factor < 0'
          info = 1
       end if
       if (wang_landau_weight_cycle.lt.0) then
          if (mssg.gt.-1) write(6,'(1x,a)') 'wang_landau_weight_cycle < 0'
          info = 1
       end if
       if (wang_landau_factor_cycle.le.wang_landau_weight_cycle) then
          if (mssg.gt.-1) write(6,'(1x,a)') &
               'wang_landau_factor_cycle <= wang_landau_weight_cycle'
          info = 1
       end if
    end if

    ! print input

    if (mssg.gt.-1) then
       write(6,'(a)') magic_tappmoldyn
       write(6,mol_dyn)
       if (mdmode .ne. 'SMPL') then
          write(6,'(1x,e18.11,a)') mass(1), ' ! mass in a.m.u.'
          do ikd=2,nkd
             write(6,'(1x,e18.11)') mass(ikd)
          end do
          write(6,'(3(1x,e18.11),a)') cavl(1,1),cavl(2,1),cavl(3,1), ' ! velocity in Cartesian coordinate'
          do it=2,nsi
             write(6,'(3(1x,e18.11))') cavl(1,it),cavl(2,it),cavl(3,it)
          end do
       end if
    end if

    ! substitute vars

    deltm = delta_time
    imdstp = init_md_step
    emdstp = end_md_step
    imtrw = mode_xtrap_rho_wfn
    msgcycl = message_cycle
    sysvol = system_volume
    syskbt = system_kbt
    syspext = system_pressure

    mdparam = 0.0d0
    mdstat = 0.0d0
    if ( (mdmode .eq. 'NPT_ST') .or. (mdmode .eq. 'MBMT_ST') .or. &
         (mdmode .eq. 'MT_ST_R') .or. (mdmode .eq. 'MBMT_ST_R') ) then
       mdparam(1) = stern_omega_zeta
       mdparam(2) = stern_omega_eta
       mdstat(1) = stern_zeta
       mdstat(2) = stern_eta
       mdstat(3) = stern_sigma
    else if (mdmode .eq. 'NVT_BRDSN') then
       mdparam(1) = brdsn_tauinv
    end if
    mdparam(3) = abinitio_cycle
    mdstat(6) = wang_landau_update_factor
    mdparam(4) = wang_landau_weight_cycle
    mdparam(5) = wang_landau_factor_cycle

    return

100 continue
    if (mssg.gt.-1) then
       write(6,'(A)')'error occurs @ reading namelist : mol_dyn'
    end if
    info = 1
    return

  end subroutine read_tappmd
end module subr_inputmoldyn

module subr_inputpef
  character(len=*),parameter:: magic_tapppef = '# trace band data'
  integer,parameter:: nkptl = 6
contains
  !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
  subroutine read_tapppef(iin,nbk,nkpt,ak,nkfi,nw,nb2,itsb,ndiag,deps, &
          ifbunp,ifrho,nbba1,nbba2,iaxis,ndiv,rzero, &
          rmin,rmax,nsph,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    implicit none

    integer, intent(in) :: iin,mssg
    integer,intent(out) :: info

    integer,intent(out):: ifbunp,ifrho,nbk
    integer,pointer,intent(out):: nkfi(:)
    real(kind=8),pointer,intent(out):: ak(:,:)
    real(kind=8),intent(out):: deps
    character(len=nkptl),pointer:: nkpt(:)
    integer,pointer,intent(out):: ndiv(:),iaxis(:)
    real(kind=8),pointer:: rzero(:,:),rmin(:),rmax(:)
    integer:: nw,nb2
    integer,intent(out):: itsb,ndiag,nbba1,nbba2,nsph

    character(len=32):: distrib_mode
    integer:: output_charge, output_wave_function
    integer:: davidson_flag, davidson_number_diag
    real(kind=8):: diag_converge
    integer:: distrib_band_lower, distrib_band_upper
    integer:: distrib_number_sphere, distrib_number_plate
    integer:: number_band, number_band_traced, number_trace_block

    integer:: i,j,is

    namelist /trace_band/ distrib_mode, output_charge, output_wave_function, &
         davidson_flag, davidson_number_diag, diag_converge, &
         distrib_band_lower, distrib_band_upper, &
         distrib_number_sphere, distrib_number_plate, &
         number_band, number_band_traced, number_trace_block

    distrib_mode = 'none'
    output_charge = 0
    output_wave_function = 0
    davidson_flag = -1
    davidson_number_diag = 40
    diag_converge = 1.0d-5
    distrib_band_lower = -1
    distrib_band_upper = -1
    distrib_number_sphere = 0
    distrib_number_plate = 0
    number_band = nw
    number_band_traced = -1
    number_trace_block = -1

    read(iin, trace_band, err=100)

    if (distrib_band_lower .lt. 0) then
       distrib_band_lower = 1
    end if
    if (distrib_band_upper .lt. 0) then
       distrib_band_upper = number_band
    end if
    if (number_band_traced .lt. 0) then
       number_band_traced = number_band
    end if

    if (number_trace_block .le. 0) then
       write(6,*) 'number_trace_block is not specified'
       info = 1
       number_trace_block = 1
    end if
    allocate(nkpt(number_trace_block+1), ak(3,number_trace_block+1), &
         nkfi(number_trace_block))

    if (trim(distrib_mode) .eq. 'sphere') then
       continue
    else if (trim(distrib_mode) .eq. 'plate') then
       continue
    else if (trim(distrib_mode) .eq. 'none') then
       distrib_number_sphere = 1
       distrib_number_plate = 1
    else
       if (mssg.gt.-1) write(6,*) 'unknown distrib_mode'
       info = 1
    end if

    read(iin,*) (nkpt(i),i=1,number_trace_block+1)
    do i=1,3
       read(iin,*) (ak(i,j),j=1,number_trace_block+1)
    end do
    read(iin,*) (nkfi(i),i=1,number_trace_block)

    if (trim(distrib_mode) .eq. 'sphere') then
       allocate(rzero(3,distrib_number_sphere),rmin(distrib_number_sphere), &
            rmax(distrib_number_sphere),ndiv(distrib_number_sphere), &
            iaxis(1))
       do is=1,distrib_number_sphere
          read(iin,*) (rzero(i,is),i=1,3),rmin(is),rmax(is),ndiv(is)
       end do
    else if (trim(distrib_mode) .eq. 'plate') then
       allocate(iaxis(distrib_number_plate),rmin(distrib_number_plate), &
            rmax(distrib_number_plate),ndiv(distrib_number_plate), &
            rzero(3,1))
       do is=1,distrib_number_plate
          read(iin,*) iaxis(is),rmin(is),rmax(is),ndiv(is)
       end do
    else
       allocate(iaxis(1),rmin(1),rmax(1),ndiv(1),rzero(3,1))
    end if

    ! check input

    if (output_charge .eq. 1 .and. output_wave_function .eq. 1) then
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'output_charge and output_wave_function is exclusive'
       info = 1
    end if

    if (trim(distrib_mode).eq.'sphere') then
       if (distrib_number_sphere.le.0) then
          if (mssg.gt.-1) write(6,'(1x,a)') 'distrib_number_sphere must be > 0'
          info = 1
       end if
    else if (trim(distrib_mode).eq.'plate') then
       if (distrib_number_plate.le.0) then
          if (mssg.gt.-1) write(6,'(1x,a)') 'distrib_number_plate must be > 0'
          info = 1
       end if
    end if

    do i=1,number_trace_block
       if (nkfi(i).le.0) then
          if (mssg.gt.-1) write(6,'(1x,a,i3)') 'nkfi(i) <= 0 for i = ', i
          info = 1
       end if
    end do

    if (distrib_band_upper .gt. number_band) then
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'distrib_band_upper must be <= number_band'
       info = 1
    end if
    if (distrib_band_lower .gt. number_band) then
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'distrib_band_lower must be <= number_band'
       info = 1
    end if
    if (number_band_traced .gt. number_band) then
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'number_band_traced must be <= number_band'
       info = 1
    end if

    ! print input

    if (mssg.gt.-1) then
       write(6,'(a)') magic_tapppef
       write(6, trace_band)

       write(6,'(3x,a)',advance='no') trim(nkpt(1))
       do i=2,number_trace_block
          write(6,'(8x,a)',advance='no') trim(nkpt(i))
       end do
       write(6,'(8x,a)') trim(nkpt(number_trace_block+1))

       do i=1,3
          do j=1,number_trace_block
             write(6,'(1x,f8.5)',advance='no') ak(i,j)
          end do
          write(6,'(1x,f8.5,'' ! ak('',i1,'')'')') ak(i,number_trace_block+1),i
       end do

       write(6,'(2x)',advance='no')
       do i=1,number_trace_block-1
          write(6,'(6x,i3)',advance='no') nkfi(i)
       end do
       write(6,'(6x,i3)') nkfi(number_trace_block)

       if (trim(distrib_mode) .eq. 'sphere') then
          write(6,'(5(1x,e18.11),1x,i5,a)') &
               (rzero(i,1),i=1,3),rmin(1),rmax(1),ndiv(1), &
               ' ! rzero(3),rmin,rmax,ndiv'
          do is=2,distrib_number_sphere
             write(6,'(5(1x,e18.11),1x,i5,a)') &
                  (rzero(i,is),i=1,3),rmin(is),rmax(is),ndiv(is)
          end do
       else if (trim(distrib_mode) .eq. 'plate') then
          write(6,'(1x,i5,2(1x,e18.11),1x,i5,a)') &
               iaxis(1),rmin(1),rmax(1),ndiv(1), &
               ' ! iaxis,rmin,rmax,ndiv'
          do is=2,distrib_number_plate
             write(6,'(1x,i5,2(1x,e18.11),1x,i5)') &
                  iaxis(is),rmin(is),rmax(is),ndiv(is)
          end do
       end if
    end if

    ! substitute vars
    
    ifbunp = 0 
    if (trim(distrib_mode) .eq. 'none') then
       nsph = 1
       ifbunp = 0
    else if (trim(distrib_mode) .eq. 'plate') then
       ifbunp = 1
       nsph = distrib_number_plate
    else if (trim(distrib_mode) .eq. 'sphere') then
       ifbunp = 2
       nsph = distrib_number_sphere
    end if
    nbk = number_trace_block
    ifrho = 0
    if (output_charge .eq. 1) ifrho = 1
    if (output_wave_function .eq. 1) ifrho = 2
    itsb = davidson_flag
    ndiag = davidson_number_diag
    deps = diag_converge
    nbba1 = distrib_band_lower
    nbba2 = distrib_band_upper
    nw = number_band
    nb2 = number_band_traced

    return

100 continue
    if (mssg.gt.-1) then
       write(6,'(A)')'error occurs @ reading namelist : trace_band'
    end if
    info = 1
    return
  end subroutine read_tapppef
end module subr_inputpef

module subr_inputwfn2chg
  character(len=*),parameter:: magic_tappwfn2chg = '# inspect wfn data'
contains
  !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
  subroutine read_tappwfn2chg(iin,nw,nsi,ifbunp,ifrho,nbba1,nbba2, &
       iaxis,ndiv,rzero,rmin,rmax,nsph,itwav,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    implicit none

    integer,intent(in) :: iin,nw,nsi,mssg
    integer,intent(out) :: info

    integer,intent(out):: ifbunp,ifrho
    integer,pointer,intent(out):: ndiv(:),iaxis(:)
    real(kind=8),pointer:: rzero(:,:),rmin(:),rmax(:)
    integer,intent(out):: nbba1,nbba2,nsph,itwav

    character(len=32):: distrib_mode
    integer:: output_charge
    integer:: distrib_band_lower, distrib_band_upper
    integer:: distrib_number_sphere, distrib_number_plate
    integer:: pdos_target_atom

    integer:: i,is

    namelist /inspect_wfn/ distrib_mode, output_charge, &
         distrib_band_lower, distrib_band_upper, &
         distrib_number_sphere, distrib_number_plate, &
         pdos_target_atom

    distrib_mode = 'none'
    output_charge = 0
    distrib_band_lower = 1
    distrib_band_upper = nw
    distrib_number_sphere = 0
    distrib_number_plate = 0
    pdos_target_atom = -1

    read(iin, inspect_wfn, err=100)

    if (trim(distrib_mode) .eq. 'sphere') then
       continue
    else if (trim(distrib_mode) .eq. 'plate') then
       continue
    else if (trim(distrib_mode) .eq. 'pdos') then
       distrib_number_sphere = 1
       distrib_number_plate = 1
    else if (trim(distrib_mode) .eq. 'wannier') then
       distrib_number_sphere = 1
       distrib_number_plate = 1
    else if (trim(distrib_mode) .eq. 'kinet') then
       distrib_number_sphere = 1
       distrib_number_plate = 1
    else if (trim(distrib_mode) .eq. 'none') then
       distrib_number_sphere = 1
       distrib_number_plate = 1
    else
       if (mssg.gt.-1) write(6,*) 'unknown distrib_mode'
       info = 1
    end if

    if (trim(distrib_mode) .eq. 'sphere') then
       allocate(rzero(3,distrib_number_sphere),rmin(distrib_number_sphere), &
            rmax(distrib_number_sphere),ndiv(distrib_number_sphere), &
            iaxis(1))
       do is=1,distrib_number_sphere
          read(iin,*) (rzero(i,is),i=1,3),rmin(is),rmax(is),ndiv(is)
       end do
    else if (trim(distrib_mode) .eq. 'plate') then
       allocate(iaxis(distrib_number_plate),rmin(distrib_number_plate), &
            rmax(distrib_number_plate),ndiv(distrib_number_plate), &
            rzero(3,1))
       do is=1,distrib_number_plate
          read(iin,*) iaxis(is),rmin(is),rmax(is),ndiv(is)
       end do
    else ! none or pdos or wannier or kinet
       allocate(rzero(3,1),rmin(1),rmax(1),ndiv(1),iaxis(1))
       if (trim(distrib_mode).eq.'pdos' .and. pdos_target_atom.eq.0) then
          read(iin,*) rzero(1,1),rzero(2,1),rzero(3,1)
       end if
    end if

    ! check input

    if (trim(distrib_mode).eq.'sphere') then
       if (distrib_number_sphere.le.0) then
          if (mssg.gt.-1) write(6,'(1x,a)') 'distrib_number_sphere must be > 0'
          info = 1
       end if
    else if (trim(distrib_mode).eq.'plate') then
       if (distrib_number_plate.le.0) then
          if (mssg.gt.-1) write(6,'(1x,a)') 'distrib_number_plate must be > 0'
          info = 1
       end if
    else if (trim(distrib_mode).eq.'pdos') then
       if (pdos_target_atom.lt.0 .or. pdos_target_atom .gt. nsi) then
          if (mssg.gt.-1) write(6,'(1x,a)') &
               'pdos_target_atom must be 0 or in [1,nsi]'
          info = 1
       end if
    end if

    ! print input

    if (mssg.gt.-1) then
       write(6,'(a)') magic_tappwfn2chg
       write(6, inspect_wfn)

       if (trim(distrib_mode) .eq. 'sphere') then
          write(6,'(5(1x,e18.11),1x,i5,a)') &
               (rzero(i,1),i=1,3),rmin(1),rmax(1),ndiv(1), &
               ' ! rzero(3),rmin,rmax,ndiv'
          do is=2,distrib_number_sphere
             write(6,'(5(1x,e18.11),1x,i5,a)') &
                  (rzero(i,is),i=1,3),rmin(is),rmax(is),ndiv(is)
          end do
       else if (trim(distrib_mode) .eq. 'plate') then
          write(6,'(1x,i5,2(1x,e18.11),1x,i5,a)') &
               iaxis(1),rmin(1),rmax(1),ndiv(1), &
               ' ! iaxis,rmin,rmax,ndiv'
          do is=2,distrib_number_plate
             write(6,'(1x,i5,2(1x,e18.11),1x,i5)') &
                  iaxis(is),rmin(is),rmax(is),ndiv(is)
          end do
       else if (trim(distrib_mode).eq.'pdos' .and. pdos_target_atom.eq.0) then
          write(6,'(3(1x,e18.11))') rzero(1,1),rzero(2,1),rzero(3,1)
       end if
    end if

    ! substitute vars
    
    if (trim(distrib_mode) .eq. 'none') then
       ifbunp = 0
       nsph = 1
    else if (trim(distrib_mode) .eq. 'plate') then
       ifbunp = 1
       nsph = distrib_number_plate
    else if (trim(distrib_mode) .eq. 'sphere') then
       ifbunp = 2
       nsph = distrib_number_sphere
    else if (trim(distrib_mode) .eq. 'pdos') then
       ifbunp = 3
       if (pdos_target_atom .eq. 0) then
          nsph = 1
       end if
    else if (trim(distrib_mode) .eq. 'wannier') then
       ifbunp = 4
    else if (trim(distrib_mode) .eq. 'kinet') then
       ifbunp = 5
    end if
    if (output_charge .eq. 1) then
       ifrho = 1
    else
       ifrho = 0
    end if
    nbba1 = distrib_band_lower
    nbba2 = distrib_band_upper
    itwav = pdos_target_atom

    return

100 continue
    if (mssg.gt.-1) then
       write(6,'(A)')'error occurs @ reading namelist : inspect_wfn'
    end if
    info = 1
    return
  end subroutine read_tappwfn2chg
end module subr_inputwfn2chg

module subr_inputstm
  character(len=*),parameter:: magic_tappstm = '# stm data'
contains
  subroutine read_tappstm(iin,nbias,eferm,vs,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    use xcf_types
    implicit none

    integer, intent(in) :: iin
    integer, intent(out) :: nbias
    real(8), intent(out) :: eferm
    real(8), pointer, intent(out) :: vs(:)

    integer,intent(in) :: mssg
    integer :: info

    integer :: number_bias
    real(8) :: stm_fermi_energy

    integer:: i

    namelist /stminput/ number_bias, stm_fermi_energy

    ! # initial parameter

    number_bias = 0
    stm_fermi_energy = 0.0d0

    ! read namelist

    read(iin,stminput,err=100)
    allocate(vs(number_bias))
    do i=1,number_bias
       read(iin,*) vs(i)
    end do

    ! set to default

    ! print status

    if (mssg.gt.-1) then
       write(6,'(a)') magic_tappstm
       write(6,stminput)
       do i=1,number_bias
          write(6,'(1x,e18.11)') vs(i)
       end do
    end if

    ! substitute vars
    
    nbias = number_bias
    eferm = stm_fermi_energy

    return

100 continue
    if (mssg.gt.-1) then
       write(6,'(A)')'error occurs @ reading namelist : tappstm'
    end if
    info = 1
    return

  end subroutine read_tappstm
end module subr_inputstm

module subr_inputwannier
  character(len=*),parameter:: magic_tappwannier = '# wannier data'
contains
  !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
  subroutine read_tappwannier(iin,nspin,nw,aa,nkd,icell,wan_params &
       ,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    use subr_wannier
    implicit none

    integer,intent(in) :: iin,nspin,nw,nkd,mssg
    real(kind=8),intent(in) :: aa(3,3)
    integer:: icell
    type (wannier_params_struct):: wan_params
    integer,intent(out) :: info

    integer:: wannier_band_lower, wannier_band_upper, wannier_cell_type
    real(kind=8):: wannier_eng_lower, wannier_eng_upper
    real(kind=8):: wannier_eng_inner_lower, wannier_eng_inner_upper
    integer:: wannier_number_gaussian
    integer:: wannier_target_number
    logical:: wannier_ident_b_mat
    integer:: wannier_range_lattice_trans(3)
    integer:: wannier_number_basis
    integer:: wannier_target_spin
    real(kind=8):: wannier_threshold_spillage
    real(kind=8):: wannier_threshold_spread
    integer:: wannier_ini_basis_mode
    real(kind=8):: wannier_spread_mix
    real(kind=8):: wannier_sprd_ini_step_len_fac
    character(len=16):: program_mode

    integer:: i,j,ikd,ngau
    real(kind=8):: aalen(3),del,axdel
    character(len=256):: cbuf
    integer:: ierr

    real(kind=8),parameter:: pi = 3.14159265358979323846d0

    namelist /max_loc_wannier/ wannier_band_lower, wannier_band_upper, &
         wannier_cell_type, wannier_eng_lower, wannier_eng_upper, &
         wannier_eng_inner_lower, wannier_eng_inner_upper, &
         wannier_number_gaussian, &
         wannier_target_number, &
         wannier_ident_b_mat, &
         wannier_range_lattice_trans, &
         wannier_number_basis, &
         wannier_target_spin, &
         wannier_threshold_spillage, &
         wannier_threshold_spread, &
         wannier_sprd_ini_step_len_fac, &
         wannier_spread_mix, &
         wannier_ini_basis_mode, &
         program_mode

    wannier_band_lower = 1
    wannier_band_upper = nw
    wannier_cell_type = 9
    wannier_number_gaussian = 0
    wannier_ident_b_mat = .true.
    wannier_range_lattice_trans = (/3, 3, 3/)
    wannier_number_basis = -1
    wannier_target_spin = 1
    wannier_threshold_spillage = 1.0d-4
    wannier_threshold_spread = 1.0d-4
    wannier_spread_mix = 1.0d0
    wannier_sprd_ini_step_len_fac = 50.0d0
    wannier_ini_basis_mode = 1
    wannier_target_number = -1
    wannier_eng_inner_lower = huge(1.0d0)
    wannier_eng_inner_upper = -huge(1.0d0)
    program_mode = 'generate_rcp'

    read(iin, max_loc_wannier, err=100)

    if (wannier_ini_basis_mode .eq. 1) then
       if (wannier_number_basis .lt. 0 .or. &
            wannier_ident_b_mat ) then
          wannier_number_basis = wannier_number_gaussian
       end if
       wan_params%ngau = wannier_number_basis
       wan_params%nigs = wannier_number_gaussian
       allocate(wan_params%lgau(wan_params%nigs), &
            wan_params%mgau(wan_params%nigs), wan_params%alpg(wan_params%nigs), &
            wan_params%taug(3,wan_params%nigs), &
            wan_params%cgau(wan_params%nigs), &
            wan_params%qgau(3,wan_params%nigs), &
            wan_params%B_MAT(wan_params%nigs,wan_params%ngau))
       do i=1,wan_params%nigs
          read(iin,'(a)') cbuf
          read(cbuf,*,iostat=ierr) wan_params%lgau(i),wan_params%mgau(i), &
               wan_params%alpg(i),wan_params%taug(:,i),wan_params%cgau(i), &
               wan_params%qgau(1,i),wan_params%qgau(2,i),wan_params%qgau(3,i)
          if (ierr.ne.0) then
             wan_params%qgau(:,i) = 0.0d0
             read(cbuf,*,iostat=ierr) wan_params%lgau(i),wan_params%mgau(i), &
                  wan_params%alpg(i),wan_params%taug(:,i),wan_params%cgau(i)
             if (ierr.ne.0) then
                read(cbuf,*,iostat=ierr) wan_params%lgau(i),wan_params%mgau(i), &
                     wan_params%alpg(i),wan_params%taug(:,i)
                wan_params%cgau(i) = 1
             end if
          end if
       end do
       if (wannier_ident_b_mat) then
          wan_params%B_MAT = 0.0d0
          do i=1,wan_params%ngau
             wan_params%B_MAT(i,i) = 1.0d0
          end do
       else
          do j=1,wan_params%ngau
             read(iin,*) (wan_params%B_MAT(i,j),i=1,wan_params%nigs)
          end do
       end if
    else if (wannier_ini_basis_mode .eq. 2) then
       allocate(wan_params%gbkd(nkd))
       do ikd=1,nkd
          read(iin,*) ngau
          wan_params%gbkd(ikd)%ngau = ngau
          allocate(wan_params%gbkd(ikd)%lgau(ngau), &
               wan_params%gbkd(ikd)%mgau(ngau), &
               wan_params%gbkd(ikd)%alpg(ngau))
          do i=1,ngau
             read(iin,*) wan_params%gbkd(ikd)%lgau(i), &
                  wan_params%gbkd(ikd)%mgau(i), &
                  wan_params%gbkd(ikd)%alpg(i)
          end do
       end do
    else
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'wannier_ini_basis_mode is invalid'
       info = 1
    end if

    ! check input

    if (wannier_target_spin.lt.1 .or. wannier_target_spin.gt.nspin) then
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'wannier_target_spin is out of range'
       info = 1
    end if

    if (wannier_cell_type .lt. 1 .or. wannier_cell_type .gt. 9) then
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'invalid wannier_cell_type is specified'
       info = 1
    end if
    if (wannier_ini_basis_mode.eq.1 .and.  wannier_number_gaussian.lt.0) then
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'wannier_number_gaussian must be specified'
       info = 1
    end if

    del = epsilon(aa)
    axdel = sum(abs(aa))*del
    aalen(1) = sqrt(dot_product(aa(:,1),aa(:,1)))
    aalen(2) = sqrt(dot_product(aa(:,2),aa(:,2)))
    aalen(3) = sqrt(dot_product(aa(:,3),aa(:,3)))
    select case (wannier_cell_type)
    case (1)
       if (abs(aalen(1)-aalen(2)).gt.axdel .or. &
            abs(aalen(2)-aalen(3)).gt.axdel .or. &
            abs(aalen(3)-aalen(1)).gt.axdel .or. &
            abs(dot_product(aa(:,2),aa(:,3))/aalen(2)/aalen(3)).gt.del .or. &
            abs(dot_product(aa(:,3),aa(:,1))/aalen(3)/aalen(1)).gt.del .or. &
            abs(dot_product(aa(:,1),aa(:,2))/aalen(1)/aalen(2)).gt.del) then
          if (mssg.gt.-1) write(6,'(1x,a)') &
               'invalid cell form for wannier_cell_type = 1: SC'
          info = 1
       end if
    case(2)
       if (abs(aalen(1)-aalen(2)).gt.axdel .or. &
            abs(aalen(2)-aalen(3)).gt.axdel .or. &
            abs(aalen(3)-aalen(1)).gt.axdel .or. &
            abs(dot_product(aa(:,2),aa(:,3))/aalen(2)/aalen(3)-0.5d0) &
                .gt.del .or. &
            abs(dot_product(aa(:,3),aa(:,1))/aalen(3)/aalen(1)-0.5d0) &
                .gt.del .or. &
            abs(dot_product(aa(:,1),aa(:,2))/aalen(1)/aalen(2)-0.5d0) &
                .gt.del) then
          if (mssg.gt.-1) write(6,'(1x,a)') &
               'invalid cell form for wannier_cell_type = 2: FCC'
          info = 1
       end if
    case(3)
       if (abs(aalen(1)-aalen(2)).gt.axdel .or. &
            abs(aalen(2)-aalen(3)).gt.axdel .or. &
            abs(aalen(3)-aalen(1)).gt.axdel .or. &
            abs(dot_product(aa(:,2),aa(:,3))/aalen(2)/aalen(3)+1.0d0/3.0d0) &
                .gt.del .or. &
            abs(dot_product(aa(:,3),aa(:,1))/aalen(3)/aalen(1)+1.0d0/3.0d0) &
                .gt.del .or. &
            abs(dot_product(aa(:,1),aa(:,2))/aalen(1)/aalen(2)+1.0d0/3.0d0) &
                .gt.del) then
          if (mssg.gt.-1) write(6,'(1x,a)') &
               'invalid cell form for wannier_cell_type = 3: BCC'
          info = 1
       end if
    case(4)
       if (abs(aalen(1) - aalen(2)).gt.axdel .or. &
            abs(dot_product(aa(:,2),aa(:,3))/aalen(2)/aalen(3)).gt.del .or. &
            abs(dot_product(aa(:,3),aa(:,1))/aalen(3)/aalen(1)).gt.del .or. &
            abs(dot_product(aa(:,1),aa(:,2))/aalen(1)/aalen(2) + 0.5d0).gt.del) then
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'invalid cell form for wannier_cell_type = 4: HCP'
          info = 1
       end if
    case(5)
       if (abs(dot_product(aa(:,2),aa(:,3))/aalen(2)/aalen(3)).gt.del .or. &
            abs(dot_product(aa(:,3),aa(:,1))/aalen(3)/aalen(1)).gt.del .or. &
            abs(dot_product(aa(:,1),aa(:,2))/aalen(1)/aalen(2)).gt.del) then
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'invalid cell form for wannier_cell_type = 5: Ortho and Tetra'
          info = 1
       end if
    case(6)
       if (aa(2,1).ne.0.0d0 .or. aa(3,1).ne.0.0d0 .or. &
            aa(1,2).ne.0.0d0 .or. aa(3,2).ne.0.0d0) then
          if (mssg.gt.-1) write(6,'(1x,a)') &
               'invalid cell form for wannier_cell_type = 6: Monoclinic'
          info = 1
       end if
    case(7)
       if (aa(2,1).ne.0.0d0 .or. aa(3,1).ne.0.0d0 .or. &
            aa(3,2).ne.0.0d0) then
          if (mssg.gt.-1) write(6,'(1x,a)') &
               'invalid cell form for wannier_cell_type = 7: Triclinic'
          info = 1
       end if
    case(8)
       if (abs(aalen(1)-aalen(2)).gt.axdel .or. &
            abs(aalen(2)-aalen(3)).gt.axdel .or. &
            abs(aalen(3)-aalen(1)).gt.axdel .or. &
            abs(dot_product(aa(:,2),aa(:,3))-dot_product(aa(:,3),aa(:,1))).gt.del .or. &
            abs(dot_product(aa(:,3),aa(:,1))-dot_product(aa(:,1),aa(:,2))).gt.del .or. &
            abs(dot_product(aa(:,1),aa(:,2))-dot_product(aa(:,2),aa(:,3))).gt.del) then
          if (mssg.gt.-1) write(6,'(1x,a)') &
               'invalid cell form for wannier_cell_type = 8: Rhombohedral'
          info = 1
       end if
    end select

    if (trim(program_mode) .ne. 'generate_rcp' .and. &
         trim(program_mode) .ne. 'recp_to_real') then
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'invalid program_mode for wannier'
       info = 1
    end if

    if (trim(program_mode) .eq. 'recp_to_real' .and. &
         wannier_target_number .lt. 0) then
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'wannier_target_number is unspecified'
       info = 1
    end if

    if (mssg.gt.-1) then
       write(6,'(a)') magic_tappwannier
       write(6, max_loc_wannier)
       if (wannier_ini_basis_mode .eq. 1) then
          do i=1,wan_params%nigs
             write(6,'(1x,2i4,f8.3,1x,3e18.11,i4,3f8.2)') &
                  wan_params%lgau(i),wan_params%mgau(i), &
                  wan_params%alpg(i),wan_params%taug(:,i),wan_params%cgau(i), &
                  wan_params%qgau(1,i),wan_params%qgau(2,i),wan_params%qgau(3,i)
          end do
          if (.not. wannier_ident_b_mat) then
             do j=1,wan_params%ngau
                write(6,'(1x,''('',e18.11,'', '',e18.11,'')''$)') &
                     (wan_params%B_MAT(i,j),i=1,wan_params%nigs)
                write(6,*)
             end do
          end if
       else if (wannier_ini_basis_mode .eq. 2) then
          do ikd=1,nkd
             write(6,*) wan_params%gbkd(ikd)%ngau
             do i=1,wan_params%gbkd(ikd)%ngau
                write(6,*) wan_params%gbkd(ikd)%lgau(i), &
                     wan_params%gbkd(ikd)%mgau(i), &
                     wan_params%gbkd(ikd)%alpg(i)
             end do
          end do
       end if
    end if

    ! substitute vars

    wan_params%inibase_mode = wannier_ini_basis_mode
    wan_params%tspin = wannier_target_spin
    wan_params%nblw = wannier_band_lower
    wan_params%nbup = wannier_band_upper
    icell = wannier_cell_type
    wan_params%engl = wannier_eng_lower
    wan_params%engu = wannier_eng_upper
    wan_params%enginl = wannier_eng_inner_lower
    wan_params%enginu = wannier_eng_inner_upper
    wan_params%iwan = wannier_target_number
    wan_params%nocc = wannier_number_basis
    wan_params%Na1 = wannier_range_lattice_trans(1)
    wan_params%Na2 = wannier_range_lattice_trans(2)
    wan_params%Na3 = wannier_range_lattice_trans(3)
    wan_params%spll_thr = wannier_threshold_spillage
    wan_params%sprd_thr = wannier_threshold_spread
    wan_params%sprd_mix =  wannier_spread_mix
    wan_params%sprd_stpl_fac =  wannier_sprd_ini_step_len_fac
    wan_params%pgm_mode = program_mode

    if (wannier_ini_basis_mode .eq. 1) then
       !
       ! degree -> radian
       !
       wan_params%qgau(:,:) = (pi/180.0d0)*wan_params%qgau(:,:)
    end if

    return

100 continue
    if (mssg.gt.-1) then
       write(6,'(A)')'error occurs @ reading namelist : max_loc_wannier'
    end if
    info = 1
    return
  end subroutine read_tappwannier
end module subr_inputwannier

module subr_filemap
  character(len=*),parameter:: magic_filemap = '# file map data'
contains
  subroutine read_filemap(iin,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    use iounit
    implicit none

    integer, intent(in) :: iin

    integer,intent(in) :: mssg
    integer :: info

    integer:: number_PP_file, ipot
    character(len=256):: basename, cbuf

    namelist /filemap/ basename, number_PP_file

    ! # initial parameter

    basename = 'NONE'

    ! read namelist

    iou_fmap = .false.
    read(iin,filemap,iostat=info)
    if (info.ne.0) then
       if (mssg.gt.-1) then
          write(6,'(A)')' filemap namelist is unavailable'
          write(6,'(A)')' do not enable file mapping'
       end if
       info = 0
       return
    end if
    do ipot=1,number_PP_file
       read(iin,'(a)') cbuf
       read(cbuf,*,iostat=info) iou_potnam(ipot), iou_rhinam(ipot), &
            iou_ioncnam(ipot)
       if (info.ne.0) then
          read(cbuf,*,iostat=info) iou_potnam(ipot), iou_rhinam(ipot)
          if (info.ne.0) then
             read(cbuf,*,iostat=info) iou_potnam(ipot)
             iou_rhinam(ipot) = 'NONE'
          end if
          iou_ioncnam(ipot) = 'NONE'
       end if
       if (info.ne.0) then
          if (mssg.gt.-1) then
             write(6,'(A)')' file map data: error in reading PP filenames'
          end if
          return
       end if
    end do
    iou_fmap = .true.
    iou_bnam = basename

    ! print status

    if (mssg.gt.-1) then
       write(6,'(a)') magic_filemap
       write(6,filemap)
       do ipot=1,number_PP_file
          write(6,'(1x,a)') ''''//trim(iou_potnam(ipot))//''', ''' &
               //trim(iou_rhinam(ipot))//''', ''' &
               //trim(iou_ioncnam(ipot))//''''
       end do
    end if

    return

  end subroutine read_filemap
end module subr_filemap

module subr_inputinichr
  character(len=*),parameter:: magic_tappinichr = '# initial charge'
contains
  !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
  subroutine read_tappinichr(iin,nsi,udirchr,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    implicit none

    integer, intent(in) :: iin
    integer, intent(in) :: nsi

    real(kind=8),pointer :: udirchr(:,:)

    integer, intent(in) :: mssg
    integer :: info
    character(len=64) :: mode

    namelist /initial_charge/ mode

    integer :: it, ios
    real(kind=8):: adir
    integer:: flipchr

    allocate(udirchr(3,nsi))
    udirchr = 0.0d0

    read(iin, initial_charge, iostat=ios)
    if (ios.lt.0) then
       ! supress message here
       !
       ! if (mssg.gt.-1) then
       !   write(6,'(A)')' initial_charge namelist is unavailable'
       ! end if
       !
       ! save status
       ! info = 0
       return
    else if (ios.gt.0) then
       if (mssg.gt.-1) then
          write(6,'(A)')'error occurs @ reading namelist : initial_charge'
       end if
       info = 1
    end if

    if (trim(mode) .ne. 'flip_atomic_charge' .and. &
         trim(mode) .ne. 'direct_atomic_charge') then
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'invalid mode was specified'
       info = 1
       return
    end if

    if (trim(mode) .eq. 'flip_atomic_charge') then
       do it = 1, nsi
          read(iin,*) flipchr
          if (flipchr.ne.0 .and. flipchr.ne.1 .and. &
               mssg.gt.-1) then
             write(6,'(1x,a)') 'invalid flipchr was specfied @ it = ', it
             info = 1
          end if
          if (flipchr.eq.1) then
             udirchr(3,it) = -1.0d0
          else
             udirchr(3,it) = 1.0d0
          end if
       end do
    else if (trim(mode) .eq. 'direct_atomic_charge') then
       do it = 1, nsi
          read(iin,*) udirchr(1,it),udirchr(2,it),udirchr(3,it)
          adir = sqrt(udirchr(1,it)**2+udirchr(2,it)**2+udirchr(3,it)**2)
          if (adir.eq.0.0d0) then
             udirchr(:,it) = 0.0d0
          else
             udirchr(:,it) = udirchr(:,it)/adir
          end if
       end do
    else
       if (mssg.gt.-1) write(6,'(1x,a)') &
            'invalid mode was specified'
       info = 1
       return
    end if

    if (mssg.gt.-1) then
       write(6,'(a)') magic_tappinichr
       write(6,initial_charge)
       if (trim(mode) .eq. 'flip_atomic_charge') then
          do it=1,nsi
             if (udirchr(3,it).gt.0.0d0) then
                write(6,'(1x,i2)') 0
             else
                write(6,'(1x,i2)') 1
             end if
          end do
       else if (trim(mode) .eq. 'direct_atomic_charge') then
          do it=1,nsi
             write(6,'(3(1x,f8.4))') udirchr(1,it),udirchr(2,it),udirchr(3,it)
          end do
       end if
    end if

    return
  end subroutine read_tappinichr
end module subr_inputinichr

module subr_inputengpart
  character(len=*),parameter:: magic_tappengpart = '# energy_partition data'
contains
  subroutine read_tappengpart(iin,nkd,engparcs,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    use subr_engpart
    implicit none

    integer, intent(in) :: iin,nkd
    type (eng_part_struct):: engparcs

    integer,intent(in) :: mssg
    integer :: info

    real(8) :: cutoff_energy_partition
    character(len=64) :: mode_eng_part, mode_hart_part
    integer :: number_bin_weight
    integer :: ikd,ierr

    namelist /energy_part/ &
         cutoff_energy_partition, &
         mode_hart_part, &
         mode_eng_part, &
         number_bin_weight

    ! # initial parameter

    cutoff_energy_partition = -1.0d0
    mode_hart_part = 'rhophi'
    mode_eng_part = 'skip'
    number_bin_weight = 10

    ! read namelist

    read(iin,energy_part,err=100)

    ! set to default

    if (mssg.gt.-1) then
       write(6,'(a)') magic_tappengpart
       write(6,energy_part,err=100)
    end if

    ! substitute vars

    if (cutoff_energy_partition .le. 0.0d0) then
       if (mssg.gt.-1) then
          write(6,'(A)') 'cutoff_energy_partition < 0'
       end if
       info = 1
    else
       engparcs%qfengpar = cutoff_energy_partition
    end if
    
    if (trim(mode_hart_part).eq.'rhophi') then
       engparcs%mhtor = 1
    else if (trim(mode_hart_part).eq.'EE') then
       engparcs%mhtor = 2
    else
       engparcs%mhtor = -1
       if (mssg.gt.-1) then
          write(6,'(A)')'invalid mode_hart_part'
       end if
       info = 1
    end if

    if (trim(mode_eng_part).eq.'skip') then
       engparcs%mode = 0
    else if (trim(mode_eng_part).eq.'density') then
       engparcs%mode = 1
    else if (trim(mode_eng_part).eq.'voronoi') then
       engparcs%mode = 2
    else
       engparcs%mode = -1
       if (mssg.gt.-1) then
          write(6,'(A)')'invalid mode_eng_part'
       end if
       info = 1
    end if

    if (number_bin_weight .lt. 1) then
       if (mssg.gt.-1) then
          write(6,'(A)') 'number_bin_weight < 1'
       end if
       info = 1
    else
       engparcs%num_bin_wgt = number_bin_weight
    end if

    if (engparcs%mode .eq. 2) then
       allocate(engparcs%refsz(nkd))
       read(iin,*,iostat=ierr) engparcs%refsz(1)
       if (ierr.eq.0) then
          do ikd=2,nkd
             read(iin,*) engparcs%refsz(ikd)
          end do
       else
          engparcs%refsz = 1.0d0
       end if
       if (mssg.gt.-1) then
          do ikd=1,nkd
             write(6,'(1x,e18.11)') engparcs%refsz(ikd)
          end do
       end if
    end if

    return

100 continue
    if (mssg.gt.-1) then
       write(6,'(A)')'error occurs @ reading namelist : energy_part'
    end if
    info = 1
    return

  end subroutine read_tappengpart
end module subr_inputengpart
