!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015 Yoshihide Yoshimoto
!
module rotate_wavefun
contains
  subroutine rotwfn(bkm,ekm,nw,bprj,eprj,ee,p,hp,sp,dcprj, &
       ibc,nwc,cl,hcl,scl,dcprjcl,sunityq,initial,final,comm)
    use mpi
    use xmpi
    use eigsystm
    use arch_rotwfn
    use timeprof
    implicit none
    integer,intent(in):: bkm,ekm,nw,bprj,eprj,ibc,nwc
    real(kind=8),intent(inout):: ee(nwc)
    complex(kind=8),dimension(bkm:ekm,nw):: p,hp,sp
    complex(kind=8),dimension(bkm:ekm,nwc):: cl,hcl,scl
    complex(kind=8):: dcprj(bprj:eprj,nw)
    complex(kind=8):: dcprjcl(bprj:eprj,nwc)
    logical,intent(in):: sunityq,initial,final

    complex(kind=8),allocatable:: mat(:,:),phase(:),cpjtmp(:,:)
    complex(kind=8),allocatable:: smat(:,:), g(:,:), tmat(:,:)
    real(kind=8),allocatable:: eig(:)
    complex(kind=8),allocatable:: wftmp(:,:)

    integer:: ib,lnm,ig,ierr
    real(kind=8):: tmp
    integer:: comm,myrank

    call tpf_begin(tpf_rwf)

    if (nwc .ne. nw-ibc+1) then
       write(6,*)'rotwfn: nwc != nw - ibc + 1'
       stop
    end if

    lnm = getldm('Z',2*nwc)

    allocate(mat(lnm,2*nwc), &
         smat(lnm,2*nwc), g(nwc,nwc), &
         eig(2*nwc),wftmp(bkm:ekm,nwc),cpjtmp(bprj:eprj,nwc), &
         phase(nwc),tmat(lnm,2*nwc),stat=ierr)
    call chkalloc('rotwfn','tmp',ierr)
    !
    !     determine hamiltonian in sub-space
    !
    tmat = (0.0d0, 0.0d0)
    if (initial) then
       ! mat(1:nwc, 1:nwc) may have non-zero off-diagonal element
       ! in the first iteration.
       g = (0.0d0, 0.0d0)
       call ltrovlmat(bkm,ekm,nwc,p(bkm,ibc),hp(bkm,ibc),g)
       tmat(1:nwc, 1:nwc) = g
    end if

    g = (0.0d0, 0.0d0)
    call ovlmat(bkm,ekm,nwc,hcl,p(bkm,ibc),g)
    tmat(nwc+1:2*nwc, 1:nwc) = g

    g = (0.0d0, 0.0d0)
    call ltrovlmat(bkm,ekm,nwc,cl,hcl,g)
    tmat(nwc+1:2*nwc, nwc+1:2*nwc) = g

    !call MPI_ALLREDUCE(tmat,mat,lnm*(2*nwc), &
    !     MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_allzsum(tmat,mat,lnm*(2*nwc),comm,ierr)

    if (.not. initial) then
       ! mat(1:nwc, 1:nwc) do not have non-zero off-diagonal element
       ! after the first iteration.
       do ib=1,nwc
          mat(ib,ib) = ee(ib)
       end do
    end if

    tmat = (0.0d0, 0.0d0)

    g = (0.0d0, 0.0d0)
    if (sunityq) then
       call ovlmat(bkm,ekm,nwc,cl,p(bkm,ibc),g)
    else
       call ovlmat(bkm,ekm,nwc,scl,p(bkm,ibc),g)
    end if
    tmat(nwc+1:2*nwc, 1:nwc) = g

    g = (0.0d0, 0.0d0)
    if (sunityq) then
       call ltrovlmat(bkm,ekm,nwc,cl,cl,g)
    else
       call ltrovlmat(bkm,ekm,nwc,cl,scl,g)
    end if
    tmat(nwc+1:2*nwc, nwc+1:2*nwc) = g

    !call MPI_ALLREDUCE(tmat,smat,lnm*(2*nwc), &
    !     MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_allzsum(tmat,smat,lnm*(2*nwc),comm,ierr)

    do ib=1,nwc
       smat(ib,ib) = (1.0d0, 0.0d0)
    end do

    deallocate(tmat)

    call zhegv1vl(2*nwc,mat,smat,lnm,eig,8,comm,ierr)
    if (ierr.ne.0) then
       write(6,*)'zhegv1vl info =',ierr
       call MPI_FINALIZE(ierr)
       stop
    end if

    do ib=1,nwc
       ee(ib) = eig(ib)
    end do

    deallocate(eig,smat)

    !
    ! rotate p
    !      
    wftmp = 0.0d0 
    g = mat(1:nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,p(bkm,ibc),wftmp)
    g = mat(nwc+1:nwc+nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,cl,wftmp)

    call tpf_begin(tpf_rwfm)
    call zgemm('N','N',eprj-bprj+1,nwc,nwc,(1.0d0,0.0d0), &
         dcprj(bprj,ibc),eprj-bprj+1,mat(1,1),lnm, &
         (0.0d0,0.0d0),cpjtmp,eprj-bprj+1)
    call zgemm('N','N',eprj-bprj+1,nwc,nwc,(1.0d0,0.0d0), &
         dcprjcl,eprj-bprj+1,mat(nwc+1,1),lnm, &
         (1.0d0,0.0d0),cpjtmp,eprj-bprj+1)
    call tpf_end(tpf_rwfm)

    call MPI_COMM_RANK(comm,myrank,ierr)
    if (myrank .eq. 0) then
       do ib=1,nwc
          tmp = abs(wftmp(1,ib))
          if (tmp .gt. 1.0d-8) then
             phase(ib)=conjg(wftmp(1,ib))/tmp
          else
             ig = min(2,ekm)
             tmp = abs(wftmp(ig,ib))
             phase(ib)=conjg(wftmp(ig,ib))/tmp
          end if
       end do
    end if
    call MPI_BCAST(phase,nwc,MPI_DOUBLE_COMPLEX,0,comm,ierr)

!$omp parallel do
    do ib=1,nwc
       p(:,ib+ibc-1) = phase(ib)*wftmp(:,ib)
       dcprj(:,ib+ibc-1) = phase(ib)*cpjtmp(:,ib)
    end do

    if (final) then
       deallocate(g,mat,wftmp,cpjtmp,phase)
       call tpf_end(tpf_rwf)
       return
    end if

    !
    ! rotate hp
    !      
    wftmp = 0.0d0 
    g = mat(1:nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,hp(bkm,ibc),wftmp)
    g = mat(nwc+1:nwc+nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,hcl,wftmp)

!$omp parallel do
    do ib=1,nwc
       hp(:,ib+ibc-1) = phase(ib)*wftmp(:,ib)
    end do

    if (sunityq) then
       deallocate(g,mat,wftmp,cpjtmp,phase)
       call tpf_end(tpf_rwf)
       return
    end if
    !
    ! rotate sp
    !      
    wftmp = 0.0d0 
    g = mat(1:nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,sp(bkm,ibc),wftmp)
    g = mat(nwc+1:nwc+nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,scl,wftmp)

!$omp parallel do
    do ib=1,nwc
       sp(:,ib+ibc-1) = phase(ib)*wftmp(:,ib)
    end do

    deallocate(g,mat,wftmp,cpjtmp,phase)

    call tpf_end(tpf_rwf)
    return
  end subroutine rotwfn

  subroutine rotwfni(bkm,ekm,nw,bprj,eprj,ee,p,hp,sp,dcprj, &
       ibc,nwc,cl,hcl,scl,dcprjcl,initial,final,comm)
    use xmpi
    use eigsystm
    use arch_rotwfni
    use timeprof
    implicit none
    integer,intent(in):: bkm,ekm,nw,bprj,eprj,ibc,nwc
    real(kind=8),intent(inout):: ee(nwc)
    real(kind=8),dimension(bkm:ekm,nw):: p,hp,sp
    real(kind=8),dimension(bkm:ekm,nwc):: cl,hcl,scl
    complex(kind=8):: dcprj(bprj:eprj,nw)
    complex(kind=8):: dcprjcl(bprj:eprj,nwc)
    logical,intent(in):: initial,final

    real(kind=8),allocatable:: mat(:,:)
    complex(kind=8),allocatable:: cpjtmp(:,:)
    real(kind=8),allocatable:: smat(:,:),g(:,:),tmat(:,:)
    real(kind=8),allocatable:: eig(:),wftmp(:,:)
    real(kind=8),allocatable:: trprj(:,:),tiprj(:,:),cpjtr(:,:),cpjti(:,:)

    integer:: ib,lnm,ierr
    integer:: ip
    integer:: comm

    call tpf_begin(tpf_rwf)

    if (nwc .ne. nw-ibc+1) then
       write(6,*)'rotwfn: nwc != nw - ibc + 1'
       stop
    end if

    lnm = getldm('D',2*nwc)

    allocate(mat(lnm,2*nwc), &
         smat(lnm,2*nwc), g(nwc,nwc), &
         eig(2*nwc),wftmp(bkm:ekm,nwc),cpjtmp(bprj:eprj,nwc), &
         tmat(lnm,2*nwc),stat=ierr)
    call chkalloc('rotwfn','tmp',ierr)
    !
    !     determine hamiltonian in sub space
    !
    tmat = 0.0d0
    if (initial) then
       ! mat(1:nwc, 1:nwc) may have non-zero off-diagonal element
       ! in the first iteration.
       g = 0.0d0
       call ltrovlmat(bkm,ekm,nwc,p(bkm,ibc),hp(bkm,ibc),g)
       tmat(1:nwc, 1:nwc) = g
    end if

    g = 0.0d0
    call ovlmat(bkm,ekm,nwc,hcl,p(bkm,ibc),g)
    tmat(nwc+1:2*nwc, 1:nwc) = g

    g = 0.0d0
    call ltrovlmat(bkm,ekm,nwc,cl,hcl,g)
    tmat(nwc+1:2*nwc, nwc+1:2*nwc) = g

    !call MPI_ALLREDUCE(tmat,mat,lnm*(2*nwc), &
    !     MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_alldsum(tmat,mat,lnm*(2*nwc),comm,ierr)

    if (.not. initial) then
       ! mat(1:nwc, 1:nwc) do not have non-zero off-diagonal element
       ! after the first iteration.
       do ib=1,nwc
          mat(ib,ib) = ee(ib)
       end do
    end if

    tmat = 0.0d0

    g = 0.0d0
    call ovlmat(bkm,ekm,nwc,scl,p(bkm,ibc),g)
    tmat(nwc+1:2*nwc, 1:nwc) = g

    g = 0.0d0
    call ltrovlmat(bkm,ekm,nwc,cl,scl,g)
    tmat(nwc+1:2*nwc, nwc+1:2*nwc) = g

    !call MPI_ALLREDUCE(tmat,smat,lnm*(2*nwc), &
    !     MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_alldsum(tmat,smat,lnm*(2*nwc),comm,ierr)

    do ib=1,nwc
       smat(ib,ib) = 1.0d0
    end do

    deallocate(tmat)

    call dsygv1vl(2*nwc,mat,smat,lnm,eig,8,comm,ierr)
    if (ierr.ne.0) then
       write(6,*)'dsygv1vl info =',ierr
       call MPI_FINALIZE(ierr)
       stop
    end if

    do ib=1,nwc
       ee(ib) = eig(ib)
    end do

    deallocate(eig,smat)

    ! rotate p
    !
    wftmp = 0.0d0 
    g = mat(1:nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,p(bkm,ibc),wftmp)
    g = mat(nwc+1:nwc+nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,cl,wftmp)

    call tpf_begin(tpf_rwfm)
    allocate(trprj(bprj:eprj,nwc),tiprj(bprj:eprj,nwc), &
         cpjtr(bprj:eprj,nwc),cpjti(bprj:eprj,nwc))
    do ib=1,nwc
       do ip=bprj,eprj
          trprj(ip,ib) = dble(dcprj(ip,ib+ibc-1))
          tiprj(ip,ib) = aimag(dcprj(ip,ib+ibc-1))
       end do
    end do
    call dgemm('N','N',eprj-bprj+1,nwc,nwc,1.0d0, &
         trprj,eprj-bprj+1,mat(1,1),lnm,0.0d0,cpjtr,eprj-bprj+1)
    call dgemm('N','N',eprj-bprj+1,nwc,nwc,1.0d0, &
         tiprj,eprj-bprj+1,mat(1,1),lnm,0.0d0,cpjti,eprj-bprj+1)

    do ib=1,nwc
       do ip=bprj,eprj
          trprj(ip,ib) = dble(dcprjcl(ip,ib))
          tiprj(ip,ib) = aimag(dcprjcl(ip,ib))
       end do
    end do
    call dgemm('N','N',eprj-bprj+1,nwc,nwc,1.0d0, &
         trprj,eprj-bprj+1,mat(nwc+1,1),lnm,1.0d0,cpjtr,eprj-bprj+1)
    call dgemm('N','N',eprj-bprj+1,nwc,nwc,1.0d0, &
         tiprj,eprj-bprj+1,mat(nwc+1,1),lnm,1.0d0,cpjti,eprj-bprj+1)
    deallocate(trprj,tiprj)

!$omp parallel do private(ip)
    do ib=1,nwc
       p(:,ib+ibc-1) = wftmp(:,ib)
       do ip=bprj,eprj
          dcprj(ip,ib+ibc-1) = dcmplx(cpjtr(ip,ib),cpjti(ip,ib))
       end do
    end do
    deallocate(cpjtr,cpjti)
    call tpf_end(tpf_rwfm)

    if (final) then
       deallocate(g,mat,wftmp,cpjtmp)
       call tpf_end(tpf_rwf)
       return
    end if

    ! rotate hp
    !
    wftmp = 0.0d0 
    g = mat(1:nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,hp(bkm,ibc),wftmp)
    g = mat(nwc+1:nwc+nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,hcl,wftmp)

!$omp parallel do
    do ib=1,nwc
       hp(:,ib+ibc-1) = wftmp(:,ib)
    end do

    ! rotate sp
    !
    wftmp = 0.0d0 
    g = mat(1:nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,sp(bkm,ibc),wftmp)
    g = mat(nwc+1:nwc+nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,scl,wftmp)

!$omp parallel do
    do ib=1,nwc
       sp(:,ib+ibc-1) = wftmp(:,ib)
    end do

    deallocate(g,mat,wftmp,cpjtmp)

    call tpf_end(tpf_rwf)
    return
  end subroutine rotwfni

  subroutine rotwfng(bkm,ekm,nw,bprj,eprj,ee,p,hp,sp,dcprj, &
       ibc,nwc,cl,hcl,scl,dcprjcl,sunityq,initial,final,comm)
    use mpi
    use xmpi
    use eigsystm
    use arch_rotwfng
    use timeprof
    implicit none
    integer,intent(in):: bkm,ekm,nw,bprj,eprj,ibc,nwc
    real(kind=8),intent(inout):: ee(nwc)
    complex(kind=8),dimension(bkm:ekm,nw):: p,hp,sp
    complex(kind=8),dimension(bkm:ekm,nwc):: cl,hcl,scl
    complex(kind=8):: dcprj(bprj:eprj,nw)
    complex(kind=8):: dcprjcl(bprj:eprj,nwc)
    logical,intent(in):: sunityq,initial,final
    integer,intent(in):: comm

    real(kind=8),allocatable:: mat(:,:)
    complex(kind=8),allocatable:: cpjtmp(:,:)
    real(kind=8),allocatable:: smat(:,:), g(:,:), tmat(:,:)
    real(kind=8),allocatable:: eig(:)
    complex(kind=8),allocatable:: wftmp(:,:)
    real(kind=8),allocatable:: trprj(:,:),tiprj(:,:),cpjtr(:,:),cpjti(:,:)

    integer:: ib,lnm,ierr
    integer:: ip

    call tpf_begin(tpf_rwf)

    if (nwc .ne. nw-ibc+1) then
       write(6,*)'rotwfng: nwc != nw - ibc + 1'
       stop
    end if

    lnm = getldm('D',2*nwc)

    allocate(mat(lnm,2*nwc), &
         smat(lnm,2*nwc), g(nwc,nwc), &
         eig(2*nwc),wftmp(bkm:ekm,nwc),cpjtmp(bprj:eprj,nwc), &
         tmat(lnm,2*nwc),stat=ierr)
    call chkalloc('rotwfng','tmp',ierr)
    !
    !     determine hamiltonian in sub-space
    !
    tmat = 0.0d0
    if (initial) then
       ! mat(1:nwc, 1:nwc) may have non-zero off-diagonal element
       ! in the first iteration.
       g = 0.0d0
       call ltrovlmat(bkm,ekm,nwc,p(bkm,ibc),hp(bkm,ibc),g)
       tmat(1:nwc, 1:nwc) = g
    end if

    g = 0.0d0
    call ovlmat(bkm,ekm,nwc,hcl,p(bkm,ibc),g)
    tmat(nwc+1:2*nwc, 1:nwc) = g

    g = 0.0d0
    call ltrovlmat(bkm,ekm,nwc,cl,hcl,g)
    tmat(nwc+1:2*nwc, nwc+1:2*nwc) = g

    !call MPI_ALLREDUCE(tmat,mat,lnm*(2*nwc), &
    !     MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_alldsum(tmat,mat,lnm*(2*nwc),comm,ierr)

    if (.not. initial) then
       ! mat(1:nwc, 1:nwc) do not have non-zero off-diagonal element
       ! after the first iteration.
       do ib=1,nwc
          mat(ib,ib) = ee(ib)
       end do
    end if

    tmat = 0.0d0

    g = 0.0d0
    if (sunityq) then
       call ovlmat(bkm,ekm,nwc,cl,p(bkm,ibc),g)
    else
       call ovlmat(bkm,ekm,nwc,scl,p(bkm,ibc),g)
    end if
    tmat(nwc+1:2*nwc, 1:nwc) = g

    g = 0.0d0
    if (sunityq) then
       call ltrovlmat(bkm,ekm,nwc,cl,cl,g)
    else
       call ltrovlmat(bkm,ekm,nwc,cl,scl,g)
    end if
    tmat(nwc+1:2*nwc, nwc+1:2*nwc) = g

    !call MPI_ALLREDUCE(tmat,smat,lnm*(2*nwc), &
    !     MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_alldsum(tmat,smat,lnm*(2*nwc),comm,ierr)

    do ib=1,nwc
       smat(ib,ib) = 1.0d0
    end do

    deallocate(tmat)

    call dsygv1vl(2*nwc,mat,smat,lnm,eig,8,comm,ierr)
    if (ierr.ne.0) then
       write(6,*)'dsygv1vl info =',ierr
       call MPI_FINALIZE(ierr)
       stop
    end if

    do ib=1,nwc
       ee(ib) = eig(ib)
    end do

    deallocate(eig,smat)

    !
    ! rotate p
    !
    wftmp = 0.0d0
    g = mat(1:nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,p(bkm,ibc),wftmp)
    g = mat(nwc+1:nwc+nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,cl,wftmp)

    call tpf_begin(tpf_rwfm)
    allocate(trprj(bprj:eprj,nwc),tiprj(bprj:eprj,nwc), &
         cpjtr(bprj:eprj,nwc),cpjti(bprj:eprj,nwc))
    do ib=1,nwc
       do ip=bprj,eprj
          trprj(ip,ib) = dble(dcprj(ip,ib+ibc-1))
          tiprj(ip,ib) = aimag(dcprj(ip,ib+ibc-1))
       end do
    end do
    call dgemm('N','N',eprj-bprj+1,nwc,nwc,1.0d0, &
         trprj,eprj-bprj+1,mat(1,1),lnm,0.0d0,cpjtr,eprj-bprj+1)
    call dgemm('N','N',eprj-bprj+1,nwc,nwc,1.0d0, &
         tiprj,eprj-bprj+1,mat(1,1),lnm,0.0d0,cpjti,eprj-bprj+1)

    do ib=1,nwc
       do ip=bprj,eprj
          trprj(ip,ib) = dble(dcprjcl(ip,ib))
          tiprj(ip,ib) = aimag(dcprjcl(ip,ib))
       end do
    end do
    call dgemm('N','N',eprj-bprj+1,nwc,nwc,1.0d0, &
         trprj,eprj-bprj+1,mat(nwc+1,1),lnm,1.0d0,cpjtr,eprj-bprj+1)
    call dgemm('N','N',eprj-bprj+1,nwc,nwc,1.0d0, &
         tiprj,eprj-bprj+1,mat(nwc+1,1),lnm,1.0d0,cpjti,eprj-bprj+1)
    deallocate(trprj,tiprj)

!$omp parallel do private(ip)
    do ib=1,nwc
       p(:,ib+ibc-1) = wftmp(:,ib)
       do ip=bprj,eprj
          dcprj(ip,ib+ibc-1) = dcmplx(cpjtr(ip,ib),cpjti(ip,ib))
       end do
    end do
    deallocate(cpjtr,cpjti)
    call tpf_end(tpf_rwfm)

    if (final) then
       deallocate(g,mat,wftmp,cpjtmp)
       call tpf_end(tpf_rwf)
       return
    end if

    !
    ! rotate hp
    !
    wftmp = 0.0d0
    g = mat(1:nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,hp(bkm,ibc),wftmp)
    g = mat(nwc+1:nwc+nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,hcl,wftmp)

!$omp parallel do
    do ib=1,nwc
       hp(:,ib+ibc-1) = wftmp(:,ib)
    end do

    if (sunityq) then
       deallocate(g,mat,wftmp,cpjtmp)
       call tpf_end(tpf_rwf)
       return
    end if
    !
    ! rotate sp
    !
    wftmp = 0.0d0 
    g = mat(1:nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,sp(bkm,ibc),wftmp)
    g = mat(nwc+1:nwc+nwc,1:nwc)
    call rotvecs(bkm,ekm,nwc,g,scl,wftmp)

!$omp parallel do
    do ib=1,nwc
       sp(:,ib+ibc-1) = wftmp(:,ib)
    end do

    deallocate(g,mat,wftmp,cpjtmp)

    call tpf_end(tpf_rwf)
    return
  end subroutine rotwfng
end module rotate_wavefun
