!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2016,2017 Yoshihide Yoshimoto
!
module symchrg
contains
  subroutine ysymch(rho,nwk,idcwk,nlx,nly,nlz,nrxyz,ng,nrcell)
    implicit none
    integer:: nlx,nly,nlz,nrxyz,ng,nrcell
    real(kind=8):: rho(nlx,nly,nlz)
    integer:: nwk(nrxyz,3),idcwk(nrcell,2)
    !
    integer:: i,k
    real(kind=8):: s,fac
    !
    ! if the system has no symmetry (i.e. ng=1), do nothing.
    !
    if (ng.eq.1) then
       return
    endif
!$omp parallel do private(k,s,fac)
!POPTION INDEP(rho)
    do i=1,nrcell
       s=0.0d0
       fac=1.0d0/dble(idcwk(i,2))
!
! this loop is very short (idcwk(*,2) <= 24), so we don't vectorize it.
!
! for HITAC
!VOPTION NOVEC
! for VPP
!VOCL LOOP,SCALAR
       do k=idcwk(i,1),idcwk(i,1)+idcwk(i,2)-1
          s=s+rho(nwk(k,1),nwk(k,2),nwk(k,3))
       end do
       s=s*fac
!
! this loop is very short (idcwk(*,2) <= 24), so we don't vectorize it.
!
! for HITAC
!VOPTION NOVEC
! for VPP
!VOCL LOOP,SCALAR
       do k=idcwk(i,1),idcwk(i,1)+idcwk(i,2)-1
          rho(nwk(k,1),nwk(k,2),nwk(k,3))=s
       end do
    end do
    return
  end subroutine ysymch

  subroutine symmag(rho,nlx,nly,nlz,nrx2,nry2,nrz2,ng,nnp,rg,pg,srg,aa)
    implicit none
    integer,intent(in):: nlx,nly,nlz,nrx2,nry2,nrz2,ng,nnp
    real(kind=8):: rho(nlx,nly,nlz,3)
    integer,intent(in):: rg(3,3,ng),pg(3,ng),srg(ng)
    real(kind=8):: aa(3,3)
    !
    integer:: ig,ix,iy,iz,jx,jy,jz,kx,ky,kz,trg(3,3),tx,ty,tz,idet
    real(kind=8),allocatable:: rtmp(:,:,:,:)
    real(kind=8):: aai(3,3),tmat(3,3),det,s(3)
    !
    ! if the system has no symmetry (i.e. ng=1), do nothing.
    !
    if (ng.eq.1) then
       return
    endif

    allocate(rtmp(3,nrx2,nry2,nrz2))
    rtmp = 0.0d0

    det = aa(1,1)*( aa(2,2)*aa(3,3)-aa(2,3)*aa(3,2) ) &
         + aa(2,1)*( aa(3,2)*aa(1,3)-aa(3,3)*aa(1,2) ) &
         + aa(3,1)*( aa(1,2)*aa(2,3)-aa(1,3)*aa(2,2) )

    aai(1,1) = aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)
    aai(1,2) = aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)
    aai(1,3) = aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3)

    aai(2,1) = aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1)
    aai(2,2) = aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1)
    aai(2,3) = aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1)

    aai(3,1) = aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2)
    aai(3,2) = aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2)
    aai(3,3) = aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2)

    aai = aai/det

    do ig = 1,ng
       tx = pg(1,ig)*nrx2/nnp
       ty = pg(2,ig)*nry2/nnp
       tz = pg(3,ig)*nrz2/nnp

       idet = rg(1,1,ig)*( rg(2,2,ig)*rg(3,3,ig)-rg(2,3,ig)*rg(3,2,ig) ) &
            + rg(2,1,ig)*( rg(3,2,ig)*rg(1,3,ig)-rg(3,3,ig)*rg(1,2,ig) ) &
            + rg(3,1,ig)*( rg(1,2,ig)*rg(2,3,ig)-rg(1,3,ig)*rg(2,2,ig) )

       if (abs(idet).ne.1) then
          write(6,*) 'invalid operator in symmag(): ig = ', ig
          stop
       end if

       !
       ! trg = transpose(inverse(rg)) * idet
       !
       trg(1,1) = rg(2,2,ig)*rg(3,3,ig)-rg(3,2,ig)*rg(2,3,ig)
       trg(2,1) = rg(3,2,ig)*rg(1,3,ig)-rg(1,2,ig)*rg(3,3,ig)
       trg(3,1) = rg(1,2,ig)*rg(2,3,ig)-rg(2,2,ig)*rg(1,3,ig)
               
       trg(1,2) = rg(2,3,ig)*rg(3,1,ig)-rg(3,3,ig)*rg(2,1,ig)
       trg(2,2) = rg(3,3,ig)*rg(1,1,ig)-rg(1,3,ig)*rg(3,1,ig)
       trg(3,2) = rg(1,3,ig)*rg(2,1,ig)-rg(2,3,ig)*rg(1,1,ig)
               
       trg(1,3) = rg(2,1,ig)*rg(3,2,ig)-rg(3,1,ig)*rg(2,2,ig)
       trg(2,3) = rg(3,1,ig)*rg(1,2,ig)-rg(1,1,ig)*rg(3,2,ig)
       trg(3,3) = rg(1,1,ig)*rg(2,2,ig)-rg(2,1,ig)*rg(1,2,ig)

       !
       ! Spin is axial vector. therefore we must cancel out
       ! inversion component of the symmetry operation.
       ! This mean we have to use abs(idet) instead of idet.
       ! Because abs(idet) must be 1, this term does not exist
       ! in the formula.
       !
       tmat = matmul(aa,matmul(trg,aai))*srg(ig)

       ! The position vector of the spin density is polar vector.
       ! Therefore we have to care about the inversion.
       trg = trg/idet

!$omp parallel do private(ix,iy,iz,jx,jy,jz,kx,ky,kz,s)
       do iz = 1,nrz2
          do iy = 1,nry2
             do ix = 1,nrx2
                jx = (ix-1)
                jy = (iy-1)
                jz = (iz-1)

                kx = trg(1,1)*jx + trg(1,2)*jy + trg(1,3)*jz + tx
                ky = trg(2,1)*jx + trg(2,2)*jy + trg(2,3)*jz + ty
                kz = trg(3,1)*jx + trg(3,2)*jy + trg(3,3)*jz + tz
                kx = modulo(kx,nrx2)+1
                ky = modulo(ky,nry2)+1
                kz = modulo(kz,nrz2)+1

                s = matmul(tmat, rho(ix,iy,iz,1:3))
                rtmp(1:3,kx,ky,kz) = rtmp(1:3,kx,ky,kz) + s
             end do
          end do
       end do
    end do
    do iz=1,nrz2
       do iy=1,nry2
          do ix=1,nrx2
             rho(ix,iy,iz,:) = (1.0d0/ng)*rtmp(:,ix,iy,iz)
          end do
       end do
    end do
    deallocate(rtmp)

    return
  end subroutine symmag
  
  subroutine symstrsor(rho,nlx,nly,nlz,nrx2,nry2,nrz2,ng,nnp,rg,pg,aa)
    implicit none
    integer,intent(in):: nlx,nly,nlz,nrx2,nry2,nrz2,ng,nnp
    real(kind=8):: rho(6,nlx,nly,nlz)
    integer,intent(in):: rg(3,3,ng),pg(3,ng)
    real(kind=8):: aa(3,3)
    !
    integer:: ig,ix,iy,iz,jx,jy,jz,kx,ky,kz,trg(3,3),tx,ty,tz,idet
    real(kind=8),allocatable:: rtmp(:,:,:,:)
    real(kind=8):: aai(3,3),tmat(3,3),det,ss(3,3),tt(3,3)
    !
    ! if the system has no symmetry (i.e. ng=1), do nothing.
    !
    if (ng.eq.1) then
       return
    endif

    allocate(rtmp(6,nrx2,nry2,nrz2))
    rtmp = 0.0d0

    det = aa(1,1)*( aa(2,2)*aa(3,3)-aa(2,3)*aa(3,2) ) &
         + aa(2,1)*( aa(3,2)*aa(1,3)-aa(3,3)*aa(1,2) ) &
         + aa(3,1)*( aa(1,2)*aa(2,3)-aa(1,3)*aa(2,2) )

    aai(1,1) = aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)
    aai(1,2) = aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)
    aai(1,3) = aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3)

    aai(2,1) = aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1)
    aai(2,2) = aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1)
    aai(2,3) = aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1)

    aai(3,1) = aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2)
    aai(3,2) = aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2)
    aai(3,3) = aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2)

    aai = aai/det

    do ig = 1,ng
       tx = pg(1,ig)*nrx2/nnp
       ty = pg(2,ig)*nry2/nnp
       tz = pg(3,ig)*nrz2/nnp

       idet = rg(1,1,ig)*( rg(2,2,ig)*rg(3,3,ig)-rg(2,3,ig)*rg(3,2,ig) ) &
            + rg(2,1,ig)*( rg(3,2,ig)*rg(1,3,ig)-rg(3,3,ig)*rg(1,2,ig) ) &
            + rg(3,1,ig)*( rg(1,2,ig)*rg(2,3,ig)-rg(1,3,ig)*rg(2,2,ig) )

       if (abs(idet).ne.1) then
          write(6,*) 'invalid operator in symstrsor(): ig = ', ig
          stop
       end if

       !
       ! trg = transpose(inverse(rg)) * idet
       !
       trg(1,1) = rg(2,2,ig)*rg(3,3,ig)-rg(3,2,ig)*rg(2,3,ig)
       trg(2,1) = rg(3,2,ig)*rg(1,3,ig)-rg(1,2,ig)*rg(3,3,ig)
       trg(3,1) = rg(1,2,ig)*rg(2,3,ig)-rg(2,2,ig)*rg(1,3,ig)

       trg(1,2) = rg(2,3,ig)*rg(3,1,ig)-rg(3,3,ig)*rg(2,1,ig)
       trg(2,2) = rg(3,3,ig)*rg(1,1,ig)-rg(1,3,ig)*rg(3,1,ig)
       trg(3,2) = rg(1,3,ig)*rg(2,1,ig)-rg(2,3,ig)*rg(1,1,ig)

       trg(1,3) = rg(2,1,ig)*rg(3,2,ig)-rg(3,1,ig)*rg(2,2,ig)
       trg(2,3) = rg(3,1,ig)*rg(1,2,ig)-rg(1,1,ig)*rg(3,2,ig)
       trg(3,3) = rg(1,1,ig)*rg(2,2,ig)-rg(2,1,ig)*rg(1,2,ig)

       ! The position vector of the stress density is polar vector.
       ! Therefore we have to care about the inversion.
       trg = trg/idet

       tmat = matmul(aa,matmul(trg,aai))

!$omp parallel do private(ix,iy,iz,jx,jy,jz,kx,ky,kz,ss,tt)
       do iz = 1,nrz2
          do iy = 1,nry2
             do ix = 1,nrx2
                jx = (ix-1)
                jy = (iy-1)
                jz = (iz-1)

                kx = trg(1,1)*jx + trg(1,2)*jy + trg(1,3)*jz + tx
                ky = trg(2,1)*jx + trg(2,2)*jy + trg(2,3)*jz + ty
                kz = trg(3,1)*jx + trg(3,2)*jy + trg(3,3)*jz + tz
                kx = modulo(kx,nrx2)+1
                ky = modulo(ky,nry2)+1
                kz = modulo(kz,nrz2)+1

                ss(1,1) = rho(1,ix,iy,iz)
                ss(2,2) = rho(2,ix,iy,iz)
                ss(3,3) = rho(3,ix,iy,iz)
                ss(2,3) = rho(4,ix,iy,iz)
                ss(3,2) = rho(4,ix,iy,iz)
                ss(3,1) = rho(5,ix,iy,iz)
                ss(1,3) = rho(5,ix,iy,iz)
                ss(1,2) = rho(6,ix,iy,iz)
                ss(2,1) = rho(6,ix,iy,iz)
                tt = matmul(transpose(tmat),matmul(ss,tmat))
                rtmp(1,kx,ky,kz) = rtmp(1,kx,ky,kz) + tt(1,1)
                rtmp(2,kx,ky,kz) = rtmp(2,kx,ky,kz) + tt(2,2)
                rtmp(3,kx,ky,kz) = rtmp(3,kx,ky,kz) + tt(3,3)
                rtmp(4,kx,ky,kz) = rtmp(4,kx,ky,kz) + tt(2,3)
                rtmp(5,kx,ky,kz) = rtmp(5,kx,ky,kz) + tt(3,1)
                rtmp(6,kx,ky,kz) = rtmp(6,kx,ky,kz) + tt(1,2)
             end do
          end do
       end do
    end do
!$omp parallel do private(ix,iy,iz)
    do iz=1,nrz2
       do iy=1,nry2
          do ix=1,nrx2
             rho(:,ix,iy,iz) = (1.0d0/ng)*rtmp(:,ix,iy,iz)
          end do
       end do
    end do
    deallocate(rtmp)

    return
  end subroutine symstrsor
end module symchrg
