!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module vb_chrg_ovlp
contains
  subroutine vbchrg(dp,pby,nband,laprj,nsi,nkd &
       ,nspin,ncomp &
       ,nw,nbba1,nbba2 &
       ,va,aa,bb,asi,kd,ee,mgs,bspotfc,bspots,ifinvsym &
       ,ifbunp,ifrho,nfrok,nfwfk,nfout,nsph,iaxis,rmin,rmax,ndiv,rzero &
       ,ng,rg,nnp,pg,srg,itwav,ifoutkmsh,ndx,nfoutkmsh &
       ,nb1,nb2,wgtk &
       ,mssg,myrank,nproc,comm)
    use iounit, only: xrewind, NFVBCHRGWAV, NFRECIPWAN, NFIDXPAR
    use converter
    use types
    use bsplpot
    use distrib_wfn
    use mpi
    use meshgrid
    use subr_engpart
    implicit none
    integer:: nband,laprj,nsi,nkd
    integer:: nspin,ncomp
    integer:: nw,nbba1,nbba2
    integer:: ng,nnp
    integer:: mssg
    !
    type (meshgrid_struct):: mgs
    !
    type (wfn_for_k):: dp(mgs%nks,nspin)
    complex(kind=8):: pby(laprj,nsi,nw,mgs%nks,nspin)
    !
    real(kind=8):: va,aa(3,3),bb(3,3),asi(3,nsi)
    integer:: kd(nsi)
    real(kind=8):: ee(nband,mgs%nks,nspin)
    type (bspotfc_struct):: bspotfc
    type(bspot_for_atm):: bspots(nkd)
    logical:: ifinvsym
    integer:: rg(3,3,ng),pg(3,ng),srg(ng)
    !
    integer,intent(in):: ifbunp,ifrho,nfrok,nfwfk,nfout
    integer,intent(in):: nsph,ndiv(nsph),iaxis(nsph),itwav
    real(kind=8):: rmin(nsph),rmax(nsph),rzero(3,nsph)
    
    logical,intent(in):: ifoutkmsh
    integer,intent(in):: ndx(3),nfoutkmsh
    !
    integer,intent(in):: nb1,nb2
    real(kind=8),intent(in):: wgtk(nb1:nb2,mgs%nks,nspin)
    !
    integer:: myrank,nproc,comm

    integer:: nks
    integer:: nrx,nry,nrz
!SOPTION CONTAREA(nkm,kg,nsy,kz,nwk,idcwk,kf,kqg,g_count,g_displ)
    type (irdkpt_struct),pointer:: irdkpt(:)
    integer,pointer:: nkm(:),kg(:,:,:)
    integer,pointer:: nsy(:),kz(:,:,:),nwk(:,:),idcwk(:,:)
    integer,pointer:: kf(:,:),kqg(:,:)
    integer:: bkc,ekc
    integer,pointer:: g_count(:,:),g_displ(:,:)
    !
    real(kind=8),allocatable:: qgk(:,:)
!SOPTION CONTAREA(qg)
    complex(kind=8),pointer:: qg(:,:)
    type (qg_for_atm),allocatable:: dqgs(:)
    !
    integer:: ierr
    !
    integer:: iks,is,i,ikd,nob
    integer:: tifrho,isph
    !
    integer:: lpnum,maxr
    real(kind=8),allocatable:: r(:),wavr(:)
    real(kind=8),allocatable:: wavg(:)
    real(kind=8),allocatable:: akk(:),akg(:),dvpj(:,:,:,:)
    real(kind=8),allocatable:: tsum(:,:)
    integer:: wavil,nsmpl,bkm,ekm,im,ib
    real(kind=8):: tasi(3)
    external smpln
    real(kind=8):: smpln
    real(kind=8),allocatable:: tskmsh(:,:,:,:,:,:)
    integer:: wan_nk, wan_n_occ, wan_nkmesh

    real(kind=8),allocatable:: outchr(:,:,:)
    real(kind=8):: trssk(3)
    integer:: ikx(3),ic,irg,jrg,idiv

    integer,allocatable:: idxpar(:,:,:)
    real(kind=8),allocatable:: wgtpar(:,:,:,:)
    real(kind=8):: ekinp(nsi),ekinq(nsi),ekinr(nsi)
    integer:: inrx2,inry2,inrz2
    real(kind=8),allocatable:: awgtk(:)
    !
    nks = mgs%nks
    nrx = mgs%nrx
    nry = mgs%nry
    nrz = mgs%nrz

    irdkpt => mgs%irdkpt
    nkm => mgs%nkm
    kg => mgs%kg
    nsy => mgs%nsy
    kz => mgs%kz
    nwk => mgs%nwk
    idcwk => mgs%idcwk
    kf => mgs%kf
    kqg => mgs%kqg

    bkc = mgs%bkc
    ekc = mgs%ekc
    g_count => mgs%g_count
    g_displ => mgs%g_displ

    if (ifbunp .eq. 3) then
       if (myrank.eq.0) then
          call xrewind(NFVBCHRGWAV)
          read(NFVBCHRGWAV,*,iostat=ierr) nsmpl,wavil
          if (ierr.ne.0) then
             write(6,'(1x,a)') 'error in opening radial wfn for PDOS'
             call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
             stop
          end if

          if (itwav.eq.0) then
             write(nfout,'(2(1x,i5),3(1x,e18.11))') itwav,wavil,rzero(:,1)
          else
             write(nfout,'(2(1x,i5))') itwav,wavil
          end if
       end if
       call MPI_BCAST(nsmpl,1,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(wavil,1,MPI_INTEGER,0,comm,ierr)

       if (itwav.eq.0) then
          tasi = rzero(:,1)
       else
          tasi = asi(:,itwav)
       end if
       if (wavil.lt.1 .or. wavil.gt.3) then
          if (mssg.gt.-1) write(6,'(1x,a,i5)') 'invalid wavil = ',wavil
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
       allocate(wavr(nsmpl),r(nsmpl),stat=ierr)
       call chkalloc('vbchrg','wavr,r',ierr)
       if (myrank.eq.0) then
          do i=1,nsmpl
             read(NFVBCHRGWAV,*) r(i),wavr(i)
          end do
       end if
       call MPI_BCAST(r,nsmpl,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(wavr,nsmpl,MPI_DOUBLE_PRECISION,0,comm,ierr)

       if (ifoutkmsh) then
          allocate(tskmsh(0:2*wavil-1,ndx(1),ndx(2),ndx(3),nw,nspin))
          tskmsh = huge(1.0d0)
       end if
    end if

    if (ifbunp.eq.4) then

       if (myrank.eq.0) then
          call xrewind(NFRECIPWAN)
          read(NFRECIPWAN) wan_nk, wan_nkmesh, wan_n_occ
       end if
       call MPI_BCAST(wan_nk,1,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(wan_nkmesh,1,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(wan_n_occ,1,MPI_INTEGER,0,comm,ierr)

       if (mgs%nk.ne.wan_nk) then
          if (mssg.gt.-1) then
             write(6,'(1x,a,i4,1x,i4)') 'wan_nk != nk', &
                  wan_nk, mgs%nk
             call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
             stop
          end if
       end if
       if (mgs%nkmesh.ne.wan_nkmesh) then
          if (mssg.gt.-1) then
             write(6,'(1x,a,i4,1x,i4)') 'wan_nkmesh != nkmesh', &
                  wan_nkmesh, mgs%nkmesh
             call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
             stop
          end if
       end if
       call MPI_BARRIER(comm,ierr)

       if (ifoutkmsh) then
          allocate(tskmsh(0:wan_n_occ,ndx(1),ndx(2),ndx(3),nw,nspin))
          tskmsh = huge(1.0d0)
       end if
    end if

    if (ifbunp.eq.5) then
       allocate(idxpar(2*mgs%nrx,2*mgs%nry,2*mgs%nrz), &
            wgtpar(2*mgs%nrx,2*mgs%nry,2*mgs%nrz,nsi), &
            awgtk(nw),stat=ierr)
       call chkalloc('vbchrg','idxpar',ierr)
       if (myrank.eq.0) then
          call xrewind(NFIDXPAR)
          read(NFIDXPAR,*) inrx2,inry2,inrz2
          read(NFIDXPAR,*) idxpar
          if (inrx2.ne.mgs%nrx*2 .or. inry2.ne.mgs%nry*2 .or. &
               inrz2.ne.mgs%nrz*2) then
             write(6,*) 'invalid mesh in NFIDXPAR'
             stop
          end if
       end if
       call MPI_BCAST(idxpar,8*mgs%nrx*mgs%nry*mgs%nrz,MPI_DOUBLE_PRECISION,0,comm,ierr)
       ekinp = 0.0d0
       ekinq = 0.0d0
       ekinr = 0.0d0
       if (mssg.gt.-1) then
          write(6,*) 'itwav, rmax(1) = ', itwav, rmax(1)
       end if
       call setpartwgt(bb,mgs%nkfxa,mgs%kf,mgs%nrx*2,mgs%nry*2,mgs%nrz*2,idxpar, &
            nsi,rmax(1),wgtpar)
       ! if (mssg.gt.-1) then
       !   write(6,*) 'wgtpar = '
       !   write(6,*) wgtpar
       ! end if
    end if

    if (ifbunp.ge.0 .and. ifbunp.le.2) then
       allocate(dqgs(nkd),stat=ierr)
       call chkalloc('vbchrg','qgs',ierr)
       do ikd=1,nkd
          nob=bspots(ikd)%nob
          allocate(dqgs(ikd)%qg(bkc:ekc,nob*(nob+1)/2),stat=ierr)
          call chkalloc('vbchrg','qg',ierr)
       end do
       allocate(qgk(4,bkc:ekc),stat=ierr)
       call chkalloc('vbchrg','qgk',ierr)
       call kqg2qgk(bb,ekc-bkc+1,kqg,qgk)
       do ikd=1,nkd
          qg => dqgs(ikd)%qg
          nob=bspots(ikd)%nob
          call evqg(bspotfc,bspots(ikd),ekc-bkc+1,qgk,va,nob,1,qg)
       end do
       deallocate(qgk)

       if (ifoutkmsh) then
          allocate(tskmsh(0:(ndiv(1)+1)*ncomp**2,ndx(1),ndx(2),ndx(3),nbba1:nbba2,nspin))
          tskmsh = huge(1.0d0)
       end if
    end if

    do is=1,nspin
       do iks=1,nks
          if (mssg.gt.-1) then
             write(6,'(1x,a,i4,1x,i4)') 'iks,is = ',iks,is
          end if
          if (myrank.eq.0 .and. ifbunp.gt.0) then
             write(nfout,'(1x,i3,3(1x,e14.7),1x,i4)') &
                  iks,irdkpt(iks)%sk,irdkpt(iks)%ngir
          end if

          if (ifbunp .eq. 3 .or. ifbunp .eq. 4 .or. ifbunp .eq. 5) then

             bkm = g_displ(myrank,iks)+1
             ekm = g_displ(myrank,iks)+g_count(myrank,iks)

             if (ifbunp.eq.3) then
                allocate(akk(bkm:ekm),wavg(bkm:ekm),tsum(2*wavil-1,nw), &
                     stat=ierr)
                call chkalloc('vbchrg','akk',ierr)

                ! transformation by rg conserves akk
                call kg2akk(bb,irdkpt(iks)%sk,ekm-bkm+1,kg(:,bkm:ekm,iks),akk)
                call formwavg(akk,bkm,ekm,nsmpl,r,wavr,wavil,wavg)

                deallocate(akk)
             else if (ifbunp.eq.4) then
                allocate(tsum(wan_n_occ,nw), stat=ierr)
                call chkalloc('vbchrg','tsum',ierr)
             end if

             allocate(akg(bkm:ekm),stat=ierr)
             call chkalloc('vbchrg','akg',ierr)
             call kg2akg(bb,irdkpt(iks)%sk,g_count(myrank,iks), &
                  kg(:,bkm:ekm,iks),akg)
             lpnum = getlpnum(nkd,bspots)
             maxr = getmaxr(nkd,bspots)
             allocate(dvpj(bkm:ekm,maxr,lpnum,nkd),stat=ierr)
             call chkalloc('vbchrg','dvpj',ierr)
             do ikd=1,nkd
                call evbetg(bspotfc,bspots(ikd),g_count(myrank,iks),akg, &
                     g_count(myrank,iks),maxr,lpnum,dvpj(bkm,1,1,ikd))
             end do
             deallocate(akg)

             if (ifbunp.eq.3) then
                call ovlwforb(ncomp,aa,bb,va,irdkpt(iks),bkm,ekm, &
                     kg(:,bkm:ekm,iks),maxr,lpnum,nkd,dvpj,nsi,kd,asi, &
                     bspots,wavil,wavg,tasi,nw, &
                     dp(iks,is)%wf,ng,rg,nnp,pg,srg,tsum, &
                     ifoutkmsh,ndx,ee(:,iks,is),tskmsh(:,:,:,:,:,is),mssg,comm)

                if (myrank.eq.0) then
                   do ib=1,nw
                      write(nfout,'(8(1x,e18.11))') ee(ib,iks,is), &
                           (tsum(im,ib), im=1, 2*wavil-1)
                   end do
                end if
             else if (ifbunp.eq.4) then
                call ovlwannier(ncomp,aa,bb,va,irdkpt(iks),nkm(iks),bkm,ekm, &
                     kg(:,bkm:ekm,iks),maxr,lpnum,nkd,dvpj,nsi,kd,asi, &
                     bspots,wan_n_occ,nw,dp(iks,is)%wf,ng,rg,nnp,pg,srg,tsum, &
                     ifoutkmsh,ndx,ee(:,iks,is),tskmsh(:,:,:,:,:,is),mssg, &
                     myrank,nproc,g_count(:,iks),g_displ(:,iks),comm)

                if (myrank.eq.0) then
                   do ib=1,nw
                      write(nfout,'(8(1x,e18.11))') ee(ib,iks,is), &
                           (tsum(im,ib), im=1, wan_n_occ)
                   end do
                end if
             else if (ifbunp.eq.5) then
                do ib=1,nb1-1
                   if (nspin.eq.1) then
                      awgtk(ib) = 2.0d0/ncomp
                   else
                      awgtk(ib) = 1.0d0
                   end if
                end do
                do ib=nb1,nb2
                   awgtk(ib) = wgtk(ib,iks,is)/irdkpt(iks)%ngir
                end do
                do ib=nb2+1,nw
                   awgtk(ib) = 0.0d0
                end do
                call kinetpart(ncomp,aa,bb,va,irdkpt(iks),nkm(iks),bkm,ekm, &
                     kg(:,:,iks),nsi,kd,asi, &
                     nw,dp(iks,is)%wf,awgtk,ng,rg,nnp,pg,srg, &
                     mgs%nrx*2,mgs%nry*2,mgs%nrz*2,wgtpar,ekinp,ekinq,ekinr, &
                     mssg,myrank,nproc,g_count(:,iks),g_displ(:,iks),comm)
             end if

             if (ifbunp.eq.3) deallocate(wavg)
             if (ifbunp.eq.3 .or. ifbunp.eq.4) deallocate(tsum)

             deallocate(dvpj)

          else
             if (myrank.eq.0 .and. ifrho.eq.1) then
                write(nfrok) irdkpt(iks)%sk,dble(irdkpt(iks)%ngir)
             end if
             tifrho = ifrho
             do isph=1,nsph
                if (myrank.eq.0 .and. ((ifbunp.eq.1).or.(ifbunp.eq.2))) then
                   write(nfout,'(1x,i4)') isph
                end if
                allocate(outchr(0:ndiv(isph),ncomp**2,nbba1:nbba2))
                call bunpu(ifbunp,tifrho,nfout,nfrok,nfwfk &
                     ,ndiv(isph),rmin(isph),rmax(isph),iaxis(isph) &
                     ,rzero(1,isph) &
                     ,nw,nbba1,nbba2,kg(:,:,iks),nkm(iks) &
                     ,kd,asi,nsi,nkd,ee(1,iks,is),bb,va &
                     ,laprj,nrx,nry,nrz,dp(iks,is),pby(1,1,1,iks,is) &
                     ,bkc,ekc,kqg,dqgs,bspots,ncomp &
                     ,myrank,nproc,g_count(:,iks),g_displ(:,iks),comm &
                     ,outchr)
                tifrho = 0
                if (ifoutkmsh .and. isph.eq.1) then
                   do irg=1,irdkpt(iks)%ngir
                      jrg = irdkpt(iks)%irrg(irg)
                      trssk = matmul(rg(:,:,jrg),irdkpt(iks)%sk)
                      trssk = srg(jrg)*trssk
                      ikx = nint(trssk*ndx)
                      ikx = modulo(ikx,ndx)+1
                      do ib=nbba1,nbba2
                         tskmsh(0,ikx(1),ikx(2),ikx(3),ib,is) = ee(ib,iks,is)
                         do ic=1,ncomp**2
                            do idiv=0,ndiv(1)
                               im = 1 + idiv + (ndiv(1)+1)*(ic-1)
                               tskmsh(im,ikx(1),ikx(2),ikx(3),ib,is) &
                                    = outchr(idiv,ic,ib)
                            end do
                         end do
                      end do
                   end do
                end if
                deallocate(outchr)
             end do
          end if
       end do
    end do
    if (ifoutkmsh) then
       call xrewind(nfoutkmsh)
       if (ifbunp.eq.1 .or. ifbunp.eq.2) then
          write(nfoutkmsh) mgs%nkmesh,ndx,nbba2-nbba1+1,nspin, &
               nbba1,(ndiv(1)+1)*ncomp**2
       else if (ifbunp.eq.3) then
          write(nfoutkmsh) mgs%nkmesh,ndx,nw,nspin,itwav,2*wavil-1
       else if (ifbunp.eq.4) then
          write(nfoutkmsh) mgs%nkmesh,ndx,nw,nspin,0,wan_n_occ
       end if
       write(nfoutkmsh) tskmsh
       deallocate(tskmsh)
    end if

    if (ifbunp .eq. 5) then
       ekinp = 0.5d0*ekinp/dble(mgs%nkmesh)
       ekinq = 0.5d0*ekinq/dble(mgs%nkmesh)
       ekinr = 0.5d0*ekinr/dble(mgs%nkmesh)
       if (mssg.gt.-1) then
          write(6,'(a)') '# kinetic energy partitioned: '
          write(6,'(a)') '#  w|grad(phi)|^2 = '
          write(6,'(6x,e18.11)') ekinp
          write(6,'(a,e18.11)') '# Sum sym   ', sum(ekinp)
          write(6,'(a)') '# -w*conjg(phi)*lap(phi) = '
          write(6,'(6x,e18.11)') ekinq
          write(6,'(a,e18.11)') '# Sum asym  ', sum(ekinq)
          write(6,'(a)') '# -conjg(phi)*div(w*grad(phi)) = '
          write(6,'(6x,e18.11)') ekinr
          write(6,'(a,e18.11)') '# Sum masym ', sum(ekinr)
       end if
    end if
    if (ifbunp .eq. 3) then
       deallocate(wavr,r)
    else if (ifbunp .eq. 4) then
       continue
    else if (ifbunp .eq. 5) then
       deallocate(idxpar,wgtpar,awgtk)
    else
       do ikd=1,nkd
          deallocate(dqgs(ikd)%qg)
       end do
       deallocate(dqgs)
    end if

    return
  end subroutine vbchrg
  
  subroutine ovlwforb(ncomp,aa,bb,va,irdkpt,bkm,ekm,kg,maxr,lpnum,nkd,dvpj, &
       nsi,kd,asi,bspots,wavil,wavg,tasi,nw,wfp,ng,rg,nnp,pg,srg,tsum, &
       ifoutkmsh,ndx,ee,tskmsh,mssg,comm)
    use symwfn
    use util
    use nonlocal_proj
    use bsplpot
    use mpi
    use subr_kpoint
    use symutil
    implicit none
    integer,intent(in):: ncomp
    integer,intent(in):: bkm,ekm,maxr,lpnum,nkd,nsi,nw,ng
    integer,intent(in):: kg(3,bkm:ekm),rg(3,3,ng),nnp,pg(3,ng),srg(ng)
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: aa(3,3),bb(3,3),va
    type (irdkpt_struct):: irdkpt
    real(kind=8),intent(in):: dvpj(bkm:ekm,maxr,lpnum,nkd)
    integer,intent(in):: wavil
    real(kind=8),intent(in):: asi(3,nsi)
    real(kind=8),intent(in):: wavg(bkm:ekm),tasi(3)
!soption contarea(wfp)
    complex(kind=8),pointer:: wfp(:,:,:)
    real(kind=8):: tsum(2*wavil-1,nw)
    logical:: ifoutkmsh
    integer:: ndx(3)
    real(kind=8):: ee(nw), tskmsh(0:2*wavil-1,ndx(1),ndx(2),ndx(3),nw)
    type(bspot_for_atm):: bspots(nkd)
    integer,intent(in):: mssg,comm

    complex(kind=8),allocatable:: psum(:,:,:),asum(:,:,:)
    complex(kind=8),allocatable:: cprj(:)
    complex(kind=8),allocatable:: pwavg(:,:,:,:),swavg(:,:,:,:),strfac(:)
    complex(kind=8),allocatable:: ylm(:,:)

    type(nlprj_struct):: nlpjs
    real(kind=8):: trssk(3)
    integer,allocatable:: trskg(:,:)
    complex(kind=8),parameter:: ci = (0.0d0,1.0d0)
    real(kind=8),parameter:: pi = 3.1415926535897932385d0

    integer:: irg,jrg,im,ib,ig,nprj,ierr
    real(kind=8):: ph,tmp
    complex(kind=8):: ctemp,ctemp1(2),ctemp2(2)
    integer:: ikx(3)
    integer:: ic,jc,bkms,ekms

    complex(kind=8):: tsu2(2,2)

    allocate(pwavg(ncomp,bkm:ekm,2*wavil-1,ncomp), &
         swavg(ncomp,bkm:ekm,2*wavil-1,ncomp),asum(2*wavil-1,ncomp,nw), &
         psum(2*wavil-1,ncomp,nw),strfac(bkm:ekm), &
         trskg(3,bkm:ekm),ylm(bkm:ekm,wavil**2), &
         stat=ierr)
    call chkalloc('vbchrg','cprj,strfac',ierr)

    tsum = 0.0d0

    if (mssg.gt.-1) write(6,'(1x,a,i4)') 'ngir = ',irdkpt%ngir

    do irg=1,irdkpt%ngir
       jrg = irdkpt%irrg(irg)

       call transskkg(irdkpt%sk,ekm-bkm+1,kg, &
            rg(:,:,jrg),srg(jrg),trssk,trskg)

       if (mssg.gt.-1) write(6,'(1x,a,i3,1x,a,3(1x,f8.5))') &
            'irrg = ',irdkpt%irrg(irg),'trssk = ',trssk(1),trssk(2),trssk(3)

       call getylb_one(ekm-bkm+1,trssk,trskg,bb, &
            ylm,wavil)

       call nlprj_init(lpnum,maxr,nkd,bspots,nsi,kd,asi, &
            bkm,ekm,bb,trssk,trskg,dvpj,nlpjs)
       nprj = nlpjs%nprj

       allocate(cprj(nprj),stat=ierr)
       call chkalloc('ovlwforb','cprj',ierr)
       cprj = (0.0d0, 0.0d0)

       call mkstrfac(1,tasi,ekm-bkm+1,trskg,strfac)

       pwavg = 0.0d0
       ctemp = (-ci)**(wavil-1)
       bkms = (bkm-1)*ncomp+1
       ekms = ekm*ncomp
       do im=1,2*wavil-1
          do ig=bkm,ekm
             tmp = wavg(ig)*ylm(ig,(wavil-1)**2+im)
             do ic=1,ncomp
                pwavg(ic,ig,im,ic) = tmp*ctemp*strfac(ig)
             end do
          end do
       end do
       swavg = pwavg
       do ic=1,ncomp
          do im=1,2*wavil-1
             call nlprj_apy1s(nkd,bspots,kd,nsi,bkms,ekms, &
                  va,pwavg(1,bkm,im,ic),swavg(1,bkm,im,ic), &
                  1,nprj,cprj,nlpjs,comm)
          end do
       end do

       do im=1,2*wavil-1
          do jc=1,ncomp
             psum(im,jc,1) = (0.0d0, 0.0d0)
             do ig=bkm,ekm
                do ic=1,ncomp
                   psum(im,jc,1) = psum(im,jc,1) &
                        + conjg(swavg(ic,ig,im,jc))*pwavg(ic,ig,im,jc)
                end do
             end do
          end do
       end do
       call MPI_ALLREDUCE(psum,asum,(2*wavil-1)*ncomp, &
            MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
       asum(:,:,1)=asum(:,:,1)/va
       if (mssg.gt.-1) then
          do ic=1,ncomp
             do im=1,2*wavil-1
                write(6,'(1x,i3,1x,f8.5,1x,f8.5)') &
                     im,dble(asum(im,ic,1)),aimag(asum(im,ic,1))
             end do
          end do
       end if

       deallocate(cprj)
       call nlprj_done(nlpjs)

       call rg2tsu2(rg(:,:,jrg),aa,tsu2)
       psum = (0.0d0,0.0d0)
       if (srg(irdkpt%irrg(irg)).eq.1) then
          do ib=1,nw
             do im=1,2*wavil-1
                do ig=bkm,ekm
                   ph = dble(pg(1,jrg))*dble(trskg(1,ig)) &
                       +dble(pg(2,jrg))*dble(trskg(2,ig)) &
                       +dble(pg(3,jrg))*dble(trskg(3,ig))
                   ph = 2.0d0*pi/dble(nnp)*ph
                   if (ncomp.eq.1) then
                      ctemp1(1) = wfp(1,ig,ib)
                      ctemp1(1) = dcmplx(cos(ph),-sin(ph))*ctemp1(1)
                      ctemp2(1) = conjg(swavg(1,ig,im,1))
                      psum(im,1,ib) = psum(im,1,ib) + ctemp2(1)*ctemp1(1)
                   else
                      ctemp1 = matmul(tsu2,wfp(1:2,ig,ib))
                      ctemp1 = dcmplx(cos(ph),-sin(ph))*ctemp1
                      ctemp2 = conjg(swavg(1:2,ig,im,1))
                      psum(im,1,ib) = psum(im,1,ib) + ctemp2(1)*ctemp1(1) &
                           + ctemp2(2)*ctemp1(2)
                      ctemp2 = conjg(swavg(1:2,ig,im,2))
                      psum(im,2,ib) = psum(im,2,ib) + ctemp2(1)*ctemp1(1) &
                           + ctemp2(2)*ctemp1(2)
                   end if
                end do
             end do
          end do
       else
          do ib=1,nw
             do im=1,2*wavil-1
                do ig=bkm,ekm
                   ph = dble(pg(1,jrg))*dble(trskg(1,ig)) &
                       +dble(pg(2,jrg))*dble(trskg(2,ig)) &
                       +dble(pg(3,jrg))*dble(trskg(3,ig))
                   ph = 2.0d0*pi/dble(nnp)*ph
                   if (ncomp.eq.1) then
                      ctemp1(1) = wfp(1,ig,ib)
                      ctemp1(1) = dcmplx(cos(ph),-sin(ph))*conjg(ctemp1(1))
                      ctemp2(1) = conjg(swavg(1,ig,im,1))
                      psum(im,1,ib) = psum(im,1,ib) + ctemp2(1)*ctemp1(1)
                   else
                      ctemp1(1) = -conjg(wfp(2,ig,ib))
                      ctemp1(2) = conjg(wfp(1,ig,ib))
                      ctemp1 = dcmplx(cos(ph),-sin(ph))*matmul(tsu2,ctemp1)
                      ctemp2 = conjg(swavg(1:2,ig,im,1))
                      psum(im,1,ib) = psum(im,1,ib) + ctemp2(1)*ctemp1(1) &
                           + ctemp2(2)*ctemp1(2)
                      ctemp2 = conjg(swavg(1:2,ig,im,2))
                      psum(im,2,ib) = psum(im,2,ib) + ctemp2(1)*ctemp1(1) &
                           + ctemp2(2)*ctemp1(2)
                   end if
                end do
             end do
          end do
       end if

       call MPI_ALLREDUCE(psum,asum,(2*wavil-1)*ncomp*nw, &
            MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)

       do ib=1,nw
          do ic=1,ncomp
             do im=1,2*wavil-1
                tsum(im,ib) = tsum(im,ib) &
                     + dble(asum(im,ic,ib))**2 + aimag(asum(im,ic,ib))**2
             end do
          end do
       end do

       if (ifoutkmsh) then
          ikx = nint(trssk*ndx)
          ikx = modulo(ikx,ndx)+1
          do ib=1,nw
             tskmsh(0,ikx(1),ikx(2),ikx(3),ib) = ee(ib)
             do im=1,2*wavil-1
                if (ncomp.eq.1) then
                tskmsh(im,ikx(1),ikx(2),ikx(3),ib) = &
                     ( dble(asum(im,1,ib))**2 + aimag(asum(im,1,ib))**2 ) &
                     / va
                else
                   tskmsh(im,ikx(1),ikx(2),ikx(3),ib) = &
                        (  dble(asum(im,1,ib))**2 + aimag(asum(im,1,ib))**2 &
                         + dble(asum(im,2,ib))**2 + aimag(asum(im,2,ib))**2 ) &
                        / va * 0.5d0
                end if
             end do
          end do
       end if
    end do
    tsum = tsum/va

    deallocate(pwavg,swavg,psum,asum,strfac,trskg,ylm)

    return
  end subroutine ovlwforb

  subroutine formwavg(akk,bkm,ekm,nsmpl,r,wavr,il,wavg)
    implicit none
    integer:: bkm,ekm,nsmpl,il
    real(kind=8):: r(nsmpl),wavr(nsmpl)
    real(kind=8):: akk(bkm:ekm),wavg(bkm:ekm)
    !
    real(kind=8),parameter:: pi4 = 3.14159265358979323846d0*4.0d0
    real(kind=8),parameter:: eps = 1.0d-12

    real(kind=8):: work(nsmpl)
    real(kind=8):: q,z,sb
    integer:: i,ig,ist

    external smpln
    real(kind=8):: smpln
!
! TREATMENT FOR SK(3)=0
    if (bkm.eq.1 .and. akk(1).lt.eps) then
       ist=2
       if (il.eq.1) then
          do i=1,nsmpl
             work(i)=r(i)*wavr(i)
          end do
          wavg(1)=pi4*smpln(work,r,nsmpl)
       else
          wavg(1)=0.0d0
       end if
    else
       ist=bkm
    end if

    if (il.eq.1) then
!
!    START    L=1 (S)
!
!$omp parallel do private(q,z,sb,work,i)
!POPTION PARALLEL, TLOCAL(q,z,sb,work(nsmpl))
       do ig=ist,ekm
          q=sqrt(akk(ig))
          do i=1,nsmpl
             z=q*r(i)
             if (z.lt.1.d-1) then
                sb=-(1.d0/39916800.d0*z**10-1.d0/362880.d0*z**8 &
                     +1.d0/5040.d0*z**6-1.d0/120.d0*z**4 &
                     +1.d0/6.d0*z**2-1.d0)
             else
                sb=sin(z)/z
             end if
             work(i)=r(i)*sb*wavr(i)
          end do
          wavg(ig)=pi4*smpln(work,r,nsmpl)
       end do
    else if (il.eq.2) then
!
!    START    L=2 (P)
!
!$omp parallel do private(q,z,sb,work,i)
!POPTION PARALLEL, TLOCAL(q,z,sb,work(nsmpl))
       do ig=ist,ekm
          q=sqrt(akk(ig))
          do i=1,nsmpl
             z=q*r(i)
             if (z.lt.1.d-1) then
                sb=1.d0/3991680.d0*z**9-1.d0/45360.d0*z**7 &
                     +1.d0/840.d0*z**5-1.d0/30.d0*z**3+1.d0/3.d0*z
             else
                sb=sin(z)/z**2-cos(z)/z
             end if
             work(i)=r(i)*sb*wavr(i)
          end do
          wavg(ig)=pi4*smpln(work,r,nsmpl)
       end do
    else if (il.eq.3) then
!
!    START    L=3 (D)
!
!$omp parallel do private(q,z,sb,work,i)
!POPTION PARALLEL, TLOCAL(q,z,sb,work(nsmpl))
       do ig=ist,ekm
          q=sqrt(akk(ig))
          do i=1,nsmpl
             z=q*r(i)
             if (z.lt.1.d-1) then
                sb=1.d0/51891840.d0*z**10-1.d0/498960.d0*z**8 &
                     +1.d0/7560.d0*z**6-1.d0/210.d0*z**4+1.d0/15.d0*z**2
             else
                sb=(3.d0-1.d0*z**2)/z**3*sin(z)-3.d0/z**2*cos(z)
             end if
             work(i)=r(i)*sb*wavr(i)
          end do
          wavg(ig)=pi4*smpln(work,r,nsmpl)
       end do
    else
       write(6,*)'ERROR   IL=',il
       stop 'formwavg'
    end if

    return
  end subroutine formwavg


  subroutine ovlwannier(ncomp,aa,bb,va,irdkpt,nkm,bkm,ekm,kg,maxr,lpnum,nkd,dvpj, &
       nsi,kd,asi,bspots,wan_n_occ,nw,wfp,ng,rg,nnp,pg,srg,tsum, &
       ifoutkmsh,ndx,ee,tskmsh,mssg,myrank,nproc,g_count,g_displ,comm)
    use symwfn
    use util
    use nonlocal_proj
    use bsplpot
    use mpi
    use subr_kpoint
    use iounit
    use symutil
    implicit none
    integer,intent(in):: ncomp
    integer,intent(in):: nkm,bkm,ekm,maxr,lpnum,nkd,nsi,nw,ng
    integer,intent(in):: kg(3,bkm:ekm),rg(3,3,ng),nnp,pg(3,ng),srg(ng)
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: aa(3,3),bb(3,3),va
    type (irdkpt_struct):: irdkpt
    real(kind=8),intent(in):: dvpj(bkm:ekm,maxr,lpnum,nkd)
    integer,intent(in):: wan_n_occ
    real(kind=8),intent(in):: asi(3,nsi)
!soption contarea(wfp)
    complex(kind=8),pointer:: wfp(:,:,:)
    real(kind=8):: tsum(wan_n_occ,nw)
    logical:: ifoutkmsh
    integer:: ndx(3)
    real(kind=8):: ee(nw), tskmsh(0:wan_n_occ,ndx(1),ndx(2),ndx(3),nw)
    type(bspot_for_atm):: bspots(nkd)
    integer,intent(in):: myrank,nproc,g_count(0:nproc-1),g_displ(0:nproc-1)
    integer,intent(in):: mssg,comm

    complex(kind=8),allocatable:: psum(:,:),asum(:,:)
    complex(kind=8),allocatable:: cprj(:)
    complex(kind=8),allocatable:: pwavg(:,:,:),swavg(:,:,:)
    complex(kind=8),allocatable:: zwavg(:,:),pzwavg(:,:)

    type(nlprj_struct):: nlpjs
    real(kind=8):: trssk(3)
    integer,allocatable:: trskg(:,:)
    complex(kind=8),parameter:: ci = (0.0d0,1.0d0)
    real(kind=8),parameter:: pi = 3.1415926535897932385d0

    integer:: irg,jrg,ib,jb,ig,nprj,ierr
    real(kind=8):: ph
    complex(kind=8):: ctemp1(2),ctemp2(2)
    integer:: ikx(3)
    integer:: ic

    real(kind=8):: wan_sk(3)
    integer:: ik_map(5), wan_nkm

    integer:: mpi_typ_wfn

    complex(kind=8):: tsu2(2,2)

    allocate(pwavg(ncomp,bkm:ekm,wan_n_occ), &
         swavg(ncomp,bkm:ekm,wan_n_occ),asum(wan_n_occ,nw), &
         psum(wan_n_occ,nw),trskg(3,bkm:ekm), stat=ierr)
    call chkalloc('vbchrg','cprj,strfac',ierr)

    tsum = 0.0d0

    if (mssg.gt.-1) write(6,'(1x,a,i4)') 'ngir = ',irdkpt%ngir

    do irg=1,irdkpt%ngir
       jrg = irdkpt%irrg(irg)

       call transskkg(irdkpt%sk,ekm-bkm+1,kg, &
            rg(:,:,jrg),srg(jrg),trssk,trskg)

       if (mssg.gt.-1) write(6,'(1x,a,i3,1x,a,3(1x,f8.5))') &
            'irrg = ',irdkpt%irrg(irg),'trssk = ',trssk(1),trssk(2),trssk(3)

       call nlprj_init(lpnum,maxr,nkd,bspots,nsi,kd,asi, &
            bkm,ekm,bb,trssk,trskg,dvpj,nlpjs)
       nprj = nlpjs%nprj

       allocate(cprj(nprj),stat=ierr)
       call chkalloc('ovlwannier','cprj',ierr)
       cprj = (0.0d0, 0.0d0)

       ! read wannier as pwavg
       if (myrank.eq.0) then
          read(NFRECIPWAN) wan_sk, ik_map, wan_nkm
          if (maxval(abs(wan_sk-trssk)) .gt. 100.0d0*epsilon(1.0d0)) then
             write(6,*) 'wan_sk != trssk', wan_sk, trssk
          end if
          if (wan_nkm.ne.nkm) then
             write(6,*) 'wan_nkm != nkm', wan_nkm, nkm
          end if
       end if
       call MPI_BCAST(wan_sk,3,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(ik_map,5,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(wan_nkm,1,MPI_INTEGER,0,comm,ierr)

       call MPI_TYPE_CONTIGUOUS(ncomp,MPI_DOUBLE_COMPLEX,mpi_typ_wfn,ierr)
       call MPI_TYPE_COMMIT(mpi_typ_wfn,ierr)

       allocate(pzwavg(ncomp,bkm:ekm),zwavg(ncomp,wan_nkm))
       do jb=1,wan_n_occ
          if (myrank.eq.0) then
             read(NFRECIPWAN) ((zwavg(ic,ig),ic=1,ncomp),ig=1,wan_nkm)
          end if
          call MPI_SCATTERV(zwavg,g_count, &
               g_displ,mpi_typ_wfn, &
               pzwavg,g_count(myrank), &
               mpi_typ_wfn,0,comm,ierr)
          do ig=bkm,ekm
             pwavg(:,ig,jb) = pzwavg(:,ig)
          end do
       end do
       deallocate(pzwavg,zwavg)

       call MPI_TYPE_FREE(mpi_typ_wfn,ierr)

       swavg = pwavg
       do jb=1,wan_n_occ
          call nlprj_apy1s(nkd,bspots,kd,nsi,bkm,ekm, &
               va,pwavg(1,bkm,jb),swavg(1,bkm,jb), &
               1,nprj,cprj,nlpjs,comm)
       end do

       if (ncomp.eq.1) then
          do jb=1,wan_n_occ
             psum(jb,1) = (0.0d0, 0.0d0)
             do ig=bkm,ekm
                psum(jb,1) = psum(jb,1) + conjg(swavg(1,ig,jb))*pwavg(1,ig,jb)
             end do
          end do
       else
          do jb=1,wan_n_occ
             psum(jb,1) = (0.0d0, 0.0d0)
             do ig=bkm,ekm
                psum(jb,1) = psum(jb,1) &
                     + conjg(swavg(1,ig,jb))*pwavg(1,ig,jb) &
                     + conjg(swavg(2,ig,jb))*pwavg(2,ig,jb)
             end do
          end do
       end if

       call MPI_ALLREDUCE(psum(1,1),asum(1,1),wan_n_occ, &
            MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
       asum(:,1)=asum(:,1)
       if (mssg.gt.-1) then
          do jb=1,wan_n_occ
             write(6,'(1x,i3,1x,f8.5,1x,f8.5)') &
                  jb,dble(asum(jb,1)),aimag(asum(jb,1))
          end do
       end if

       deallocate(cprj)
       call nlprj_done(nlpjs)

       call rg2tsu2(rg(:,:,jrg),aa,tsu2)
       psum = (0.0d0,0.0d0)
       if (srg(irdkpt%irrg(irg)).eq.1) then
          do ib=1,nw
             do jb=1,wan_n_occ
                do ig=bkm,ekm
                   ph = dble(pg(1,jrg))*dble(trskg(1,ig)) &
                       +dble(pg(2,jrg))*dble(trskg(2,ig)) &
                       +dble(pg(3,jrg))*dble(trskg(3,ig))
                   ph = 2.0d0*pi/dble(nnp)*ph
                   if (ncomp.eq.1) then
                      ctemp1(1) = wfp(1,ig,ib)
                      ctemp1(1) = dcmplx(cos(ph),-sin(ph))*ctemp1(1)
                      ctemp2(1) = conjg(swavg(1,ig,jb))
                      psum(jb,ib) = psum(jb,ib) + ctemp2(1)*ctemp1(1)
                   else
                      ctemp1 = matmul(tsu2,wfp(1:2,ig,ib))
                      ctemp1 = dcmplx(cos(ph),-sin(ph))*ctemp1
                      ctemp2 = conjg(swavg(1:2,ig,jb))
                      psum(jb,ib) = psum(jb,ib) + ctemp2(1)*ctemp1(1) &
                           + ctemp2(2)*ctemp1(2)
                   end if
                end do
             end do
          end do
       else
          do ib=1,nw
             do jb=1,wan_n_occ
                do ig=bkm,ekm
                   ph = dble(pg(1,jrg))*dble(trskg(1,ig)) &
                       +dble(pg(2,jrg))*dble(trskg(2,ig)) &
                       +dble(pg(3,jrg))*dble(trskg(3,ig))
                   ph = 2.0d0*pi/dble(nnp)*ph
                   if (ncomp.eq.1) then
                      ctemp1(1) = wfp(1,ig,ib)
                      ctemp1(1) = dcmplx(cos(ph),-sin(ph))*conjg(ctemp1(1))
                      ctemp2(1) = conjg(swavg(1,ig,jb))
                      psum(jb,ib) = psum(jb,ib) + ctemp2(1)*ctemp1(1)
                   else
                      ctemp1(1) = -conjg(wfp(2,ig,ib))
                      ctemp1(2) = conjg(wfp(1,ig,ib))
                      ctemp1 = dcmplx(cos(ph),-sin(ph))*matmul(tsu2,ctemp1)
                      ctemp2 = conjg(swavg(1:2,ig,jb))
                      psum(jb,ib) = psum(jb,ib) + ctemp2(1)*ctemp1(1) &
                           + ctemp2(2)*ctemp1(2)
                   end if
                end do
             end do
          end do
       end if

       call MPI_ALLREDUCE(psum,asum,wan_n_occ*nw, &
            MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)

       do ib=1,nw
          do jb=1,wan_n_occ
             tsum(jb,ib) = tsum(jb,ib) &
                  + dble(asum(jb,ib))**2 + aimag(asum(jb,ib))**2
          end do
       end do
       if (ifoutkmsh) then
          ikx = nint(trssk*ndx)
          ikx = modulo(ikx,ndx)+1
          do ib=1,nw
             tskmsh(0,ikx(1),ikx(2),ikx(3),ib) = ee(ib)
             do jb=1,wan_n_occ
                tskmsh(jb,ikx(1),ikx(2),ikx(3),ib) = &
                     dble(asum(jb,ib))**2 + aimag(asum(jb,ib))**2
             end do
          end do
       end if
    end do
    tsum = tsum

    deallocate(pwavg,swavg,psum,asum,trskg)

    return
  end subroutine ovlwannier

  subroutine kinetpart(ncomp,aa,bb,va,irdkpt,nkm,bkm,ekm,kg, &
       nsi,kd,asi,nw,wfp,awgtk,ng,rg,nnp,pg,srg, &
       nx2,ny2,nz2,wgtpar,ekinp,ekinq,ekinr, &
       mssg,myrank,nproc,g_count,g_displ,comm)
    use fft_3d
    use symwfn
    use util
    use bsplpot
    use mpi
    use subr_kpoint
    use iounit
    use symutil
    implicit none
    integer,intent(in):: ncomp
    integer,intent(in):: nkm,bkm,ekm,nsi,nw,ng
    integer,intent(in):: kg(3,nkm),rg(3,3,ng),nnp,pg(3,ng),srg(ng)
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: aa(3,3),bb(3,3),va
    type (irdkpt_struct):: irdkpt
    real(kind=8),intent(in):: asi(3,nsi)
!soption contarea(wfp)
    complex(kind=8),pointer:: wfp(:,:,:)
    real(kind=8),intent(in):: awgtk(nw)
    integer,intent(in):: nx2,ny2,nz2
    real(kind=8),intent(in):: wgtpar(nx2,ny2,nz2,nsi)
    real(kind=8):: ekinp(nsi),ekinq(nsi),ekinr(nsi)
    integer,intent(in):: myrank,nproc,g_count(0:nproc-1),g_displ(0:nproc-1)
    integer,intent(in):: mssg,comm

    real(kind=8):: trssk(3)
    integer,allocatable:: trskg(:,:)
    complex(kind=8),parameter:: ci = (0.0d0,1.0d0)
    real(kind=8),parameter:: pi = 3.1415926535897932385d0

    integer:: irg,jrg,ib,ig,ierr
    real(kind=8):: ph
    complex(kind=8):: ctemp1(2)
    complex(kind=8),allocatable:: ptwfn(:,:),gtwfn(:,:)
    integer:: ic,iax,k1,k2,k3,j1,j2,j3
    real(kind=8):: fac(3),scl

    integer:: mpi_typ_wfn

    complex(kind=8):: tsu2(2,2)

    type (fft3_struct):: fs
    real(kind=8),allocatable:: ztmp(:,:,:,:,:),wtmp(:,:)
    real(kind=8),allocatable:: vtmp(:,:,:,:,:)
    real(kind=8):: tr,ti,tekp,tekq,tekr
    integer:: it

    allocate(trskg(3,nkm),ptwfn(ncomp,bkm:ekm), &
         gtwfn(ncomp,nkm),stat=ierr)
    call chkalloc('kinetpart','trskg,ptwfn,gtwfn',ierr)

    if (mssg.gt.-1) write(6,'(1x,a,i4)') 'ngir = ',irdkpt%ngir

    if (mssg.gt.-1) write(6,*) 'nx2,ny2,nz2 = ',nx2,ny2,nz2
    call fft3_init(nx2,ny2,nz2,nx2,ny2,nz2,fs)
    allocate(FFT_A3M_DEFSZ(ztmp,nx2,ny2,nz2,0:4), &
         FFT_A1D_DEFSZ(wtmp,nx2*ny2*nz2), &
         FFT_A3M_DEFSZ(vtmp,nx2,ny2,nz2,3))
    ztmp = 0.0d0
    wtmp = 0.0d0

    call MPI_TYPE_CONTIGUOUS(ncomp,MPI_DOUBLE_COMPLEX,mpi_typ_wfn,ierr)
    call MPI_TYPE_COMMIT(mpi_typ_wfn,ierr)
    do irg=1,irdkpt%ngir
       jrg = irdkpt%irrg(irg)

       call transskkg(irdkpt%sk,nkm,kg, &
            rg(:,:,jrg),srg(jrg),trssk,trskg)

       if (mssg.gt.-1) write(6,'(1x,a,i3,1x,a,3(1x,f8.5))') &
            'irrg = ',irdkpt%irrg(irg),'trssk = ',trssk(1),trssk(2),trssk(3)

       call rg2tsu2(rg(:,:,jrg),aa,tsu2)

       do ib=1,nw
          if (srg(irdkpt%irrg(irg)).eq.1) then
             do ig=bkm,ekm
                ph = dble(pg(1,jrg))*dble(trskg(1,ig)) &
                     +dble(pg(2,jrg))*dble(trskg(2,ig)) &
                     +dble(pg(3,jrg))*dble(trskg(3,ig))
                ph = 2.0d0*pi/dble(nnp)*ph
                if (ncomp.eq.1) then
                   ctemp1(1) = wfp(1,ig,ib)
                   ptwfn(1,ig) = dcmplx(cos(ph),-sin(ph))*ctemp1(1)
                else
                   ctemp1 = matmul(tsu2,wfp(1:2,ig,ib))
                   ptwfn(1:2,ig) = dcmplx(cos(ph),-sin(ph))*ctemp1
                end if
             end do
          else
             do ig=bkm,ekm
                ph = dble(pg(1,jrg))*dble(trskg(1,ig)) &
                     +dble(pg(2,jrg))*dble(trskg(2,ig)) &
                     +dble(pg(3,jrg))*dble(trskg(3,ig))
                ph = 2.0d0*pi/dble(nnp)*ph
                if (ncomp.eq.1) then
                   ctemp1(1) = wfp(1,ig,ib)
                   ptwfn(1,ig) = dcmplx(cos(ph),-sin(ph))*conjg(ctemp1(1))
                else
                   ctemp1(1) = -conjg(wfp(2,ig,ib))
                   ctemp1(2) = conjg(wfp(1,ig,ib))
                   ptwfn(1:2,ig) = dcmplx(cos(ph),-sin(ph))*matmul(tsu2,ctemp1)
                end if
             end do
          end if
          call MPI_ALLGATHERV(ptwfn,g_count(myrank), &
               mpi_typ_wfn,gtwfn,g_count,g_displ, &
               mpi_typ_wfn,comm,ierr)

          do ic=1,ncomp
             ztmp = 0.0d0
!$omp parallel do private(ig,k1,k2,k3,j1,j2,j3,fac,scl)
             do ig=1,nkm
                !
                ! BUG:
                ! this is not compatible with the modified kinetic energy
                ! formalism for cell deformation.
                !
                k1 = trskg(1,ig)
                k2 = trskg(2,ig)
                k3 = trskg(3,ig)
                fac(1) = bb(1,1)*(trssk(1)+k1) &
                     + bb(1,2)*(trssk(2)+k2) + bb(1,3)*(trssk(3)+k3)
                fac(2) = bb(2,1)*(trssk(1)+k1) &
                     + bb(2,2)*(trssk(2)+k2) + bb(2,3)*(trssk(3)+k3)
                fac(3) = bb(3,1)*(trssk(1)+k1) &
                     + bb(3,2)*(trssk(2)+k2) + bb(3,3)*(trssk(3)+k3)
                j1 = mod(nx2+k1,nx2) + 1
                j2 = mod(ny2+k2,ny2) + 1
                j3 = mod(nz2+k3,nz2) + 1
                FFT_A3M_RL(ztmp,j1,j2,j3,0) = dble(gtwfn(ic,ig))
                FFT_A3M_IM(ztmp,j1,j2,j3,0) = aimag(gtwfn(ic,ig))
                FFT_A3M_RL(ztmp,j1,j2,j3,1) = -fac(1)*aimag(gtwfn(ic,ig))
                FFT_A3M_IM(ztmp,j1,j2,j3,1) = fac(1)*dble(gtwfn(ic,ig))
                FFT_A3M_RL(ztmp,j1,j2,j3,2) = -fac(2)*aimag(gtwfn(ic,ig))
                FFT_A3M_IM(ztmp,j1,j2,j3,2) = fac(2)*dble(gtwfn(ic,ig))
                FFT_A3M_RL(ztmp,j1,j2,j3,3) = -fac(3)*aimag(gtwfn(ic,ig))
                FFT_A3M_IM(ztmp,j1,j2,j3,3) = fac(3)*dble(gtwfn(ic,ig))

                scl = fac(1)*fac(1)+fac(2)*fac(2)+fac(3)*fac(3)
                FFT_A3M_RL(ztmp,j1,j2,j3,4) = scl*dble(gtwfn(ic,ig))
                FFT_A3M_IM(ztmp,j1,j2,j3,4) = scl*aimag(gtwfn(ic,ig))
             end do
             do iax=0,4
                call fft3_bw(fs,FFT_A3M_RL(ztmp,1,1,1,iax),wtmp)
             end do

             scl = 1.0d0/dble(nx2*ny2*nz2)*awgtk(ib)
             do it=1,nsi
                tekp = 0.0d0
                tekq = 0.0d0
                vtmp = 0.0d0
!$omp parallel do private(j1,j2,j3,tr),reduction(+:tekp,tekq)
                do j3=1,nz2
                   do j2=1,ny2
                      do j1=1,nx2
                         tekp = tekp + scl*wgtpar(j1,j2,j3,it)*( &
                              FFT_A3M_RL(ztmp,j1,j2,j3,1)*FFT_A3M_RL(ztmp,j1,j2,j3,1) &
                              + FFT_A3M_IM(ztmp,j1,j2,j3,1)*FFT_A3M_IM(ztmp,j1,j2,j3,1) &
                              + FFT_A3M_RL(ztmp,j1,j2,j3,2)*FFT_A3M_RL(ztmp,j1,j2,j3,2) &
                              + FFT_A3M_IM(ztmp,j1,j2,j3,2)*FFT_A3M_IM(ztmp,j1,j2,j3,2) &
                              + FFT_A3M_RL(ztmp,j1,j2,j3,3)*FFT_A3M_RL(ztmp,j1,j2,j3,3) &
                              + FFT_A3M_IM(ztmp,j1,j2,j3,3)*FFT_A3M_IM(ztmp,j1,j2,j3,3) &
                              )
                         tekq = tekq + scl*wgtpar(j1,j2,j3,it)*( &
                              FFT_A3M_RL(ztmp,j1,j2,j3,0)*FFT_A3M_RL(ztmp,j1,j2,j3,4) &
                              + FFT_A3M_IM(ztmp,j1,j2,j3,0)*FFT_A3M_IM(ztmp,j1,j2,j3,4) &
                              + FFT_A3M_RL(ztmp,j1,j2,j3,0)*FFT_A3M_IM(ztmp,j1,j2,j3,4) &
                              - FFT_A3M_IM(ztmp,j1,j2,j3,0)*FFT_A3M_RL(ztmp,j1,j2,j3,4) &
                              )
                         tr = wgtpar(j1,j2,j3,it)
                         FFT_A3M_RL(vtmp,j1,j2,j3,1) = tr*FFT_A3M_RL(ztmp,j1,j2,j3,1)
                         FFT_A3M_IM(vtmp,j1,j2,j3,1) = tr*FFT_A3M_IM(ztmp,j1,j2,j3,1)
                         FFT_A3M_RL(vtmp,j1,j2,j3,2) = tr*FFT_A3M_RL(ztmp,j1,j2,j3,2)
                         FFT_A3M_IM(vtmp,j1,j2,j3,2) = tr*FFT_A3M_IM(ztmp,j1,j2,j3,2)
                         FFT_A3M_RL(vtmp,j1,j2,j3,3) = tr*FFT_A3M_RL(ztmp,j1,j2,j3,3)
                         FFT_A3M_IM(vtmp,j1,j2,j3,3) = tr*FFT_A3M_IM(ztmp,j1,j2,j3,3)
                      end do
                   end do
                end do
                ekinp(it) = ekinp(it) + tekp
                ekinq(it) = ekinq(it) + tekq
                do iax=1,3
                   call fft3_fw(fs,FFT_A3M_RL(vtmp,1,1,1,iax),wtmp)
                end do
                tekr = 0.0d0
!$omp parallel do private(ig,k1,k2,k3,j1,j2,j3,fac,tr,ti),reduction(+:tekr)
                do ig=1,nkm
                   !
                   ! BUG:
                   ! this is not compatible with the modified kinetic energy
                   ! formalism for cell deformation.
                   !
                   k1 = trskg(1,ig)
                   k2 = trskg(2,ig)
                   k3 = trskg(3,ig)
                   fac(1) = bb(1,1)*(trssk(1)+k1) &
                        + bb(1,2)*(trssk(2)+k2) + bb(1,3)*(trssk(3)+k3)
                   fac(2) = bb(2,1)*(trssk(1)+k1) &
                        + bb(2,2)*(trssk(2)+k2) + bb(2,3)*(trssk(3)+k3)
                   fac(3) = bb(3,1)*(trssk(1)+k1) &
                        + bb(3,2)*(trssk(2)+k2) + bb(3,3)*(trssk(3)+k3)
                   j1 = mod(nx2+k1,nx2) + 1
                   j2 = mod(ny2+k2,ny2) + 1
                   j3 = mod(nz2+k3,nz2) + 1
                   tr = -( fac(1)*FFT_A3M_IM(vtmp,j1,j2,j3,1) &
                        + fac(2)*FFT_A3M_IM(vtmp,j1,j2,j3,2) &
                        + fac(3)*FFT_A3M_IM(vtmp,j1,j2,j3,3) )
                   ti = fac(1)*FFT_A3M_RL(vtmp,j1,j2,j3,1) &
                        + fac(2)*FFT_A3M_RL(vtmp,j1,j2,j3,2) &
                        + fac(3)*FFT_A3M_RL(vtmp,j1,j2,j3,3)
                   tekr = tekr + awgtk(ib)*(dble(gtwfn(ic,ig))*tr + aimag(gtwfn(ic,ig))*ti)
                end do
                ekinr(it) = ekinr(it) - tekr
             end do
          end do
       end do
    end do
    call MPI_TYPE_FREE(mpi_typ_wfn,ierr)

    deallocate(trskg,ptwfn,gtwfn)

    deallocate(wtmp,ztmp,vtmp)
    call fft3_done(fs)

    return
  end subroutine kinetpart
end module vb_chrg_ovlp
