!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2017 Yoshihide Yoshimoto
!
program vbstm
  use mpi
#ifdef _OPENMP
  use omp_lib
#endif
  use chain_wfn
  use util
  use iounit
  use bsplpot
  use uspp_scrdd
  use stmitgdos
  use types
  use lpot_init
  use meshgrid
  use exchange
  use chkinput
  use subr_rddata
  use chain_atom_cg
#ifdef ESM
  use esm_manage
#endif
  implicit none
  !
  character(len=5),parameter:: prgnam = 'vbstm'
#include "config90.h"
#ifdef ESM
#include "esm_version.h"
#endif
  integer:: xctp
  real(kind=8):: bb(3,3),va
  integer:: nrxyz,nks,nspin,nkmesh
  integer:: ncomp,ndens
  real(kind=8):: spaxs(3)
  type (irdkpt_struct),pointer:: irdkpt(:)
  integer:: nrcell,nvx,nvy,nvz,nvxyz
  integer:: nwx,nwy,nwz

  integer:: nkr,nkfxa,nng
  integer:: laprj
  complex(kind=8),allocatable:: dd(:,:,:,:)
  !
  integer:: nkc
  !
  real(kind=8):: aa0(3,3),cdm(6),aa(3,3)
  integer:: nsi,nkd
  integer:: nband,nw,nb1,nb2,nk
  real(kind=8),pointer:: asi(:,:)
  real(kind=8):: znext,dspin
  integer,pointer:: kd(:)
  integer:: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq),ng,nsv,nnp
  integer:: iopt(nopt)
  real(kind=8):: qf,rf,qm,qc,qbtp(3),qexat(2)
  real(kind=8):: ekbt,uptime
  integer:: itsb,ndiag0
  integer:: ismode
  integer:: nsk
  integer,pointer:: mmm(:,:,:)
  integer:: idat(3)
  integer:: nrx,nry,nrz,ndx,ndy,ndz
  !
  real(kind=8):: qwf
  !
  complex(kind=8):: lpt(:,:), pby(:,:,:,:,:)
  allocatable pby,lpt
  !
  real(kind=8),allocatable:: ee(:,:,:)
  real(kind=8),allocatable:: wgtk(:,:,:)
  !
  type (bspotfc_struct):: bspotfc
  type (bspot_for_atm),allocatable:: bspots(:)
  !
  integer:: ik,is
  logical:: ifrgpbyee, ifinvsym
  real(kind=8):: clk
  integer:: mssg,ierr
  !
  type (wfn_for_k),allocatable:: dp(:,:)
  !
  integer:: myrank,nproc,comm
  integer:: bkm,ekm
  !
  type (meshgrid_struct):: mgs
!SOPTION CONTAREA(kqg)
  integer:: bkc,ekc
  integer,pointer:: kqg(:,:)

  real(kind=8):: eferm
  real(kind=8),pointer:: vs(:)
  integer:: nbias
  !
  real(kind=8),pointer:: udirchr(:,:)
  !
  call MPI_INIT(ierr)
  comm = MPI_COMM_WORLD
  call MPI_COMM_RANK(comm,myrank,ierr)
  call MPI_COMM_SIZE(comm,nproc,ierr)
  call MPI_BARRIER(comm,ierr)
  if (myrank.eq.0) then
     mssg=1
  else
     mssg=-1
  end if
  !
  clk=xetime(0)
  if (mssg.gt.-1) then
     write(6,*) prgnam
     write(6,'(1x,a,i4)') 'MPI: nproc = ',nproc
#ifdef _OPENMP
     write(6,'(1x,a,i4)') 'OpenMP: nthread = ', omp_get_max_threads()
#endif
  end if
#ifdef ESM
  call esm_probe(mssg)
#endif
  !
  call rdstmdata(aa0,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp,qexat &
       ,nkd,nsi,kd,asi &
       ,nband,ncomp,nspin,ndens,spaxs,znext,dspin &
       ,nrx,nry,nrz,nrxyz &
       ,qwf,nwx,nwy,nwz &
       ,ndx,ndy,ndz,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode,ekbt &
       ,itsb,ndiag0 &
       ,uptime,nbias,eferm,vs,NFINPUT,mssg)
  !
  if (iopt(5).ne.1) then
     allocate(udirchr(3,nsi))
     udirchr = 0.0d0
     call chkall(ng,rg,pg,srg,nnp,aa0,nsi,kd,asi,ndens,udirchr,comm)
     deallocate(udirchr)
  end if
  !
  call loadchkpots(nkd,nsi,aa0,qf,rf,qm,qc,ng,nnp,rg,pg,srg, &
       nk,nsk,mmm,idat,nvx,nvy,nvz,nvxyz,nks,nkmesh,irdkpt, &
       nng,nrcell,nkr,nkfxa,nkc,myrank,comm)
  !
  ifinvsym = .false.
  if (nsv.eq.1) then
     if (hasinvsym(ng,nnp,rg,pg,srg)) then
        ifinvsym = .true.
     else
        write(6,*) 'System does not have inversion symmetry'
        stop
     end if
  end if
  if (mssg.gt.-1 .and. ifinvsym) then
     write(6,*) 'Utilizing the inversion symmetry.'
  end if
  !
  allocate(ee(nband,nks,nspin),stat=ierr)
  call chkalloc(prgnam,'ee',ierr)
  !
  ! WARNING: ee must be cleared by 0 to notify
  !          ps2vcg() the SCF loop is initializing.
  !
  ee = 0.0d0
  !
  allocate(bspots(nkd),stat=ierr)
  call chkalloc(PRGNAM,'bspots',ierr)
  call loadbspot(NFBSPOT,nkd,xctp,bspotfc,bspots,myrank,comm)
  if (mssg.gt.-1) then
     call tellbspot(nkd,bspots)
  end if

  call read_mesh(ng,rg,nnp,pg,srg, &
       nk,nks,nkmesh,irdkpt,nrcell,nkr,nng,nkfxa,nkc, &
       nrx,nry,nrz,nrxyz,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz, &
       mgs,comm)

  kqg => mgs%kqg

  bkc = mgs%bkc
  ekc = mgs%ekc

  call potexch_setcutoff(xctp,qexat,aa0,nkmesh,mssg)

  if (iopt(5).eq.3) then
     if (mssg.gt.-1) write(6,*) 'loading atom position data by cgmrpt'
     call load_atom_pos_data(nsi,asi,cdm,myrank,comm)
     call cdm2aa(aa0,cdm,aa)
  else if (iopt(5).eq.0) then
     aa = aa0
  else
     if (mssg.gt.-1) write(6,*) 'invalid chain_calc mode'
     call MPI_FINALIZE(ierr)
     stop
  end if
  call recp_uvec(aa,bb)
  va = cellvol(aa)

#ifdef ESM  
  call esm_init_xTAPP(nsi,nkd,mgs,ng,aa,bb,kd,qf,va,bspots,mssg)
#endif
  if (iopt(3).ne.1) then
     write(6,'(1x,a)') 'initial_lpt must be 1'
     call MPI_FINALIZE(ierr)
     stop
  end if
  allocate(lpt(nvxyz,nspin*ndens),stat=ierr)
  call chkalloc(prgnam,'lpt',ierr)
  if (myrank.eq.0) then
     call loadlpt(nspin*ndens,lpt,nvx,nvy,nvz,spaxs,mssg)
  end if
  call MPI_BCAST(lpt,nvxyz*nspin*ndens,MPI_DOUBLE_COMPLEX,0,comm,ierr)
  call prntime('LOADLPT',mssg)
  !
  laprj = getlaprj(nkd,bspots)
  allocate(dd(laprj,laprj,nsi,nspin),stat=ierr)
  call chkalloc(prgnam,'dd',ierr)
  !
  call mkinidd(ndens,bb,bkc,ekc,kqg,lpt,dd,kd,va,asi, &
       laprj,nsi,nkd,nvx,nvy,nvz,nspin, &
       bspotfc,bspots,comm)
  call prntime('MKINIDD',mssg)
  !
  allocate(dp(nks,nspin),stat=ierr)
  call chkalloc(prgnam,'dp',ierr)
  do ik=1,nks
     bkm=mgs%g_displ(myrank,ik)+1
     ekm=mgs%g_displ(myrank,ik)+mgs%g_count(myrank,ik)
     do is=1,nspin
        allocate(dp(ik,is)%wf(ncomp,bkm:ekm,nw),stat=ierr)
        call chkalloc(prgnam,'dp%wf',ierr)
     end do
  end do

  allocate(pby(laprj,nsi,nw,nks,nspin),stat=ierr)
  call chkalloc(prgnam,'pby',ierr)
  !
  if (iopt(2).ne.1) then
     write(6,'(1x,a)') 'initial_wfn must be 1'
     call MPI_FINALIZE(ierr)
     stop
  end if
  allocate(wgtk(nb1:nb2,nks,nspin),stat=ierr)
  call chkalloc(prgnam,'wgtk',ierr)
  call load_wfn(ncomp,nspin,ndens,nw,aa,ifrgpbyee, &
       dp,mgs,nb1,nb2,wgtk,laprj,nsi,kd,pby, &
       nband,ee,nkd,bspots,mssg,myrank,nproc,comm)

  if (ifrgpbyee .and. ndiag0 .le. 0) then
     write(6,'(1x,a,a,a)') ' davidson_number_diag_1st must be >= 1, ', &
          'when wave functions on sk are mapped or ', &
          'the number of bands is expanded.'
     call MPI_FINALIZE(ierr)
     stop
  end if
  !
  ! In some architecture, two different I/O unit can not
  ! share the same file. Therefore NFWFI must be closed at this
  ! point for NFWFO.
  !
  close(NFWFI)
  !
  ! For charged state
  if (mssg.gt.-1) write(6,10) znext
10 format('CHARGE STATE CALCULATION:' &
       ,f8.3,' EXTRA ELECTRON IS ADDED')
  !
  clk=xetime(1)
  if (mssg.gt.-1) write(6,*) 'VBSTD START',clk

  call vbstd(nfstm,dp,lpt,pby &
       ,ndiag0,itsb,qbtp,ekbt,eferm,vs,nbias &
       ,nband,laprj,nsi,nkd,va,ng &
       ,ncomp,nspin,ndens,spaxs,ndx,ndy,ndz,nw,nb1,nb2 &
       ,aa,bb,asi,kd,ee,wgtk,dd &
       ,mgs,xctp,bspotfc,bspots,ifinvsym,mssg &
       ,myrank,nproc,comm)
  !
  deallocate(ee)
  deallocate(lpt,pby)
  deallocate(dd)
  deallocate(wgtk)
  do is=1,nspin
     do ik=1,nks
        deallocate(dp(ik,is)%wf)
     end do
  end do
  deallocate(dp)
  deallocate(kd,asi,mmm)
  deallocate(vs)
  clk=xetime(1)
  if (mssg.gt.-1) write(6,*) 'VBSTD is over. Time=', clk
  call MPI_FINALIZE(ierr)
  stop
end program vbstm
