!
! Copyright (c) 2003-2014 Kazuma Nakamura
! Copyright (c) 2014-2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

#define DEBUG_WANNIER_METAL

module wannier_metal

  type qgbl_for_atm
     complex(kind=8),pointer:: qg(:)
  end type qgbl_for_atm

contains
  subroutine wan_setup_bvec(aa,tbb,icell,nkb1,nkb2,nkb3, &
       NBmax,NB,b_LAT,VEC_b,WEIGHT_b,WEIGHT)
    implicit none
    real(kind=8),intent(in):: aa(3,3),tbb(3,3)
    integer,intent(in):: icell,nkb1,nkb2,nkb3,NBmax
    integer,intent(out):: NB
    real(kind=8),intent(out) :: b_LAT(3,NBmax)
    real(kind=8),intent(out) :: VEC_b(3,NBmax)
    real(kind=8),intent(out) :: WEIGHT_b(NBmax)
    real(kind=8),intent(out) :: WEIGHT

    real(kind=8),parameter :: tpi=2.0d0*3.14159265358979323846d0
    integer:: NBh
    real(kind=8):: alat
    real(kind=8) :: La,Lb,Lc
    real(kind=8) :: s,t,u,v
    !--- for triclinic Wannier
    real(kind=8) :: VEC_d(6)
    real(kind=8) :: bb(6,6)
    real(kind=8) :: w1,w2
    real(kind=8) :: flgact
    integer,parameter :: L_wdgetri = 6*10
    real(kind=8) :: wdgetri(L_wdgetri)
    integer :: info,ipiv(6)
    integer :: ab,ix,jx

    integer :: ib,ibvec
    integer :: IC,JC
    real(kind=8) :: SUM_REAL

    real(kind=8) :: a1(3),a2(3),a3(3)
    real(kind=8) :: b1(3),b2(3),b3(3)
    real(kind=8) :: del

    WRITE(6,*) ' '
    WRITE(6,*) '=========================================='
    WRITE(6,*) '=== WANNIER FUNCTION CALCULATION START ==='
    WRITE(6,*) '=========================================='

    alat = 0.0d0
    select case(icell) 
    case(1) !=== CASE(1) SIMPLE CUBIC CELL ===*
       alat=sqrt(dot_product(aa(:,1),aa(:,1)))
       NB=6
       NBh=3
    case(2) !=== CASE(2) FCC PRIMITIVE CELL ===*
       alat=sqrt(2.0d0*dot_product(aa(:,1),aa(:,1)))
       NB=8
       NBh=4
    case(3) !=== CASE(3) BCC PRIMITIVE CELL ===*
       alat=sqrt(4.0d0/3.0d0*dot_product(aa(:,1),aa(:,1)))
       NB=12
       Nbh=6
    case(4) !=== CASE(4) HCP PRIMITIVE CELL ===*
       NB=8
       NBh=4
    case(5) !=== CASE(5) ORTHORHOMBIC OR TETRAGONAL CELL ===*
       NB=6
       NBh=3
    case(6) !=== CASE(6) MONOCLINIC CELL ===*
       NB=8
       NBh=4
    case(7) !=== CASE(7) TRICLINIC CELL ===*
       NB=12
       NBh=6
    case(8) !=== CASE(8) RHOMBOHEDRAL CELL ===*
       NB=12
       NBh=6
    end select

    a1 = aa(:,1)
    a2 = aa(:,2)
    a3 = aa(:,3)
    b1 = tbb(:,1)
    b2 = tbb(:,2)
    b3 = tbb(:,3)

    write(6,*) ' '
    !===================================================*
    !=== NOTICE THAT THE b VECTORS ARE DEFINED AS    ===*
    !=== [vec(a) * vec(b)] = tpi IN THIS PROGRAM! ===*
    !===================================================*
    write(6,*) ' '
    write(6,*) '========================'
    write(6,*) 'CHECK OF (A_VEC * B_VEC)'
    write(6,*) '========================'
    write(6,*) ' '
    write(6,*) 'a1*b1=',b1(1)*a1(1)+b1(2)*a1(2)+b1(3)*a1(3)
    write(6,*) 'a2*b2=',b2(1)*a2(1)+b2(2)*a2(2)+b2(3)*a2(3)
    write(6,*) 'a3*b3=',b3(1)*a3(1)+b3(2)*a3(2)+b3(3)*a3(3)
    WRITE(6,*) ' '

    !================================*
    !=== SET B-VECTORS AND WEIGHT ===*
    !================================*
    select case(icell)
    case(1) !=== CASE(1) SIMPLE CUBIC CELL ===*
       b_LAT(:,:) = 0.0D0
       b_LAT(1,1) =  (1.0D0/DBLE(nkb1))
       b_LAT(2,2) =  (1.0D0/DBLE(nkb1))
       b_LAT(3,3) =  (1.0D0/DBLE(nkb1))
       b_LAT(1,4) = -(1.0D0/DBLE(nkb1))
       b_LAT(2,5) = -(1.0D0/DBLE(nkb1))
       b_LAT(3,6) = -(1.0D0/DBLE(nkb1))
       VEC_b(:,:) = 0.0D0
       VEC_b(:,1) =  (1.0D0/DBLE(nkb1)) * b1(:)
       VEC_b(:,2) =  (1.0D0/DBLE(nkb1)) * b2(:)
       VEC_b(:,3) =  (1.0D0/DBLE(nkb1)) * b3(:)
       VEC_b(:,4) = -(1.0D0/DBLE(nkb1)) * b1(:)
       VEC_b(:,5) = -(1.0D0/DBLE(nkb1)) * b2(:)
       VEC_b(:,6) = -(1.0D0/DBLE(nkb1)) * b3(:)
       WEIGHT_b(:) = 0.0D0
       WEIGHT_b(1) = (3.0D0/6.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(2) = (3.0D0/6.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(3) = (3.0D0/6.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(4) = (3.0D0/6.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(5) = (3.0D0/6.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(6) = (3.0D0/6.0D0) * (alat*nkb1/tpi)**2
    case(2) !=== CASE(2) FCC PRIMITIVE CELL ===*
       b_LAT(:,:) = 0.0D0
       b_LAT(1,1) =  (1.0D0/DBLE(nkb1))
       b_LAT(2,2) =  (1.0D0/DBLE(nkb1))
       b_LAT(3,3) =  (1.0D0/DBLE(nkb1))
       b_LAT(1,4) =  (1.0D0/DBLE(nkb1))
       b_LAT(2,4) =  (1.0D0/DBLE(nkb1))
       b_LAT(3,4) =  (1.0D0/DBLE(nkb1))
       b_LAT(1,5) = -(1.0D0/DBLE(nkb1))
       b_LAT(2,6) = -(1.0D0/DBLE(nkb1))
       b_LAT(3,7) = -(1.0D0/DBLE(nkb1))
       b_LAT(1,8) = -(1.0D0/DBLE(nkb1))
       b_LAT(2,8) = -(1.0D0/DBLE(nkb1))
       b_LAT(3,8) = -(1.0D0/DBLE(nkb1))
       VEC_b(:,:) = 0.0D0
       VEC_b(:,1) =  (1.0D0/DBLE(nkb1)) * b1(:)
       VEC_b(:,2) =  (1.0D0/DBLE(nkb1)) * b2(:)
       VEC_b(:,3) =  (1.0D0/DBLE(nkb1)) * b3(:)
       VEC_b(:,4) =  (1.0D0/DBLE(nkb1)) *(b1(:)+b2(:)+b3(:))
       VEC_b(:,5) = -(1.0D0/DBLE(nkb1)) * b1(:)
       VEC_b(:,6) = -(1.0D0/DBLE(nkb1)) * b2(:)
       VEC_b(:,7) = -(1.0D0/DBLE(nkb1)) * b3(:)
       VEC_b(:,8) = -(1.0D0/DBLE(nkb1)) *(b1(:)+b2(:)+b3(:))
       WEIGHT_b(:) = 0.0D0
       WEIGHT_b(1) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(2) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(3) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(4) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(5) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(6) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(7) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(8) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
    case(3) !=== CASE(3) BCC PRIMITIVE CELL ===*
       b_LAT(:,:) = 0.0D0
       b_LAT(1, 1) =  (1.0D0/DBLE(nkb1))
       b_LAT(2, 2) =  (1.0D0/DBLE(nkb1))
       b_LAT(3, 3) =  (1.0D0/DBLE(nkb1))
       b_LAT(1, 4) =  (1.0D0/DBLE(nkb1))
       b_LAT(2, 4) = -(1.0D0/DBLE(nkb1))
       b_LAT(2, 5) =  (1.0D0/DBLE(nkb1))
       b_LAT(3, 5) = -(1.0D0/DBLE(nkb1))
       b_LAT(3, 6) =  (1.0D0/DBLE(nkb1))
       b_LAT(1, 6) = -(1.0D0/DBLE(nkb1))
       b_LAT(1, 7) = -(1.0D0/DBLE(nkb1))
       b_LAT(2, 8) = -(1.0D0/DBLE(nkb1))
       b_LAT(3, 9) = -(1.0D0/DBLE(nkb1))
       b_LAT(1,10) = -(1.0D0/DBLE(nkb1))
       b_LAT(2,10) =  (1.0D0/DBLE(nkb1))
       b_LAT(2,11) = -(1.0D0/DBLE(nkb1))
       b_LAT(3,11) =  (1.0D0/DBLE(nkb1))
       b_LAT(3,12) = -(1.0D0/DBLE(nkb1))
       b_LAT(1,12) =  (1.0D0/DBLE(nkb1))
       VEC_b(:,:) = 0.0D0
       VEC_b(:, 1) =  (1.0D0/DBLE(nkb1)) * b1(:)
       VEC_b(:, 2) =  (1.0D0/DBLE(nkb1)) * b2(:)
       VEC_b(:, 3) =  (1.0D0/DBLE(nkb1)) * b3(:)
       VEC_b(:, 4) =  (1.0D0/DBLE(nkb1)) *(b1(:)-b2(:))
       VEC_b(:, 5) =  (1.0D0/DBLE(nkb1)) *(b2(:)-b3(:))
       VEC_b(:, 6) =  (1.0D0/DBLE(nkb1)) *(b3(:)-b1(:))
       VEC_b(:, 7) = -(1.0D0/DBLE(nkb1)) * b1(:)
       VEC_b(:, 8) = -(1.0D0/DBLE(nkb1)) * b2(:)
       VEC_b(:, 9) = -(1.0D0/DBLE(nkb1)) * b3(:)
       VEC_b(:,10) = -(1.0D0/DBLE(nkb1)) *(b1(:)-b2(:))
       VEC_b(:,11) = -(1.0D0/DBLE(nkb1)) *(b2(:)-b3(:))
       VEC_b(:,12) = -(1.0D0/DBLE(nkb1)) *(b3(:)-b1(:))
       WEIGHT_b( :) = 0.0D0
       WEIGHT_b( 1) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b( 2) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b( 3) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b( 4) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b( 5) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b( 6) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b( 7) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b( 8) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b( 9) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(10) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(11) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
       WEIGHT_b(12) = (1.0D0/8.0D0) * (alat*nkb1/tpi)**2
    case(4) !=== CASE(4) HCP PRIMITIVE CELL ===*
       b_LAT(:,:) = 0.0D0
       b_LAT(1,1) =  (1.0D0/DBLE(nkb1))
       b_LAT(2,2) =  (1.0D0/DBLE(nkb2))
       b_LAT(1,3) =  (1.0D0/DBLE(nkb1))
       b_LAT(2,3) = -(1.0D0/DBLE(nkb2))
       b_LAT(1,4) = -(1.0D0/DBLE(nkb1))
       b_LAT(2,5) = -(1.0D0/DBLE(nkb2))
       b_LAT(1,6) = -(1.0D0/DBLE(nkb1))
       b_LAT(2,6) =  (1.0D0/DBLE(nkb2))
       b_LAT(3,7) =  (1.0D0/DBLE(nkb3))
       b_LAT(3,8) = -(1.0D0/DBLE(nkb3))
       VEC_b(:,:) = 0.0D0
       VEC_b(:,1) =  (1.0D0/DBLE(nkb1)) * b1(:)
       VEC_b(:,2) =  (1.0D0/DBLE(nkb2)) * b2(:)
       VEC_b(:,3) =  (1.0D0/DBLE(nkb1)) * b1(:) &
            -  (1.0D0/DBLE(nkb2)) * b2(:)
       VEC_b(:,4) = -(1.0D0/DBLE(nkb1)) * b1(:)
       VEC_b(:,5) = -(1.0D0/DBLE(nkb2)) * b2(:)
       VEC_b(:,6) = -(1.0D0/DBLE(nkb1)) * b1(:) &
            +  (1.0D0/DBLE(nkb2)) * b2(:)
       VEC_b(:,7) =  (1.0D0/DBLE(nkb3)) * b3(:)
       VEC_b(:,8) = -(1.0D0/DBLE(nkb3)) * b3(:)
       La = DSQRT( a1(1)**2 + a1(2)**2 + a1(3)**2 )
       Lc = DSQRT( a3(1)**2 + a3(2)**2 + a3(3)**2 )
       WEIGHT_b(:) = 0.0D0
       WEIGHT_b(1) = (1.0D0/4.0D0) * (La*nkb1/tpi)**2
       WEIGHT_b(2) = (1.0D0/4.0D0) * (La*nkb1/tpi)**2
       WEIGHT_b(3) = (1.0D0/4.0D0) * (La*nkb1/tpi)**2
       WEIGHT_b(4) = (1.0D0/4.0D0) * (La*nkb1/tpi)**2
       WEIGHT_b(5) = (1.0D0/4.0D0) * (La*nkb1/tpi)**2
       WEIGHT_b(6) = (1.0D0/4.0D0) * (La*nkb1/tpi)**2
       WEIGHT_b(7) = (1.0D0/2.0D0) * (Lc*nkb3/tpi)**2
       WEIGHT_b(8) = (1.0D0/2.0D0) * (Lc*nkb3/tpi)**2
    case(5) !=== CASE(5) ORTHORHOMBIC OR TETRAGONAL CELL ===*
       b_LAT(:,:) = 0.0D0
       b_LAT(1,1) =  (1.0D0/DBLE(nkb1))
       b_LAT(2,2) =  (1.0D0/DBLE(nkb2))
       b_LAT(3,3) =  (1.0D0/DBLE(nkb3))
       b_LAT(1,4) = -(1.0D0/DBLE(nkb1))
       b_LAT(2,5) = -(1.0D0/DBLE(nkb2))
       b_LAT(3,6) = -(1.0D0/DBLE(nkb3))
       VEC_b(:,:) = 0.0D0
       VEC_b(:,1) =  (1.0D0/DBLE(nkb1)) * b1(:)
       VEC_b(:,2) =  (1.0D0/DBLE(nkb2)) * b2(:)
       VEC_b(:,3) =  (1.0D0/DBLE(nkb3)) * b3(:)
       VEC_b(:,4) = -(1.0D0/DBLE(nkb1)) * b1(:)
       VEC_b(:,5) = -(1.0D0/DBLE(nkb2)) * b2(:)
       VEC_b(:,6) = -(1.0D0/DBLE(nkb3)) * b3(:)
       La = DSQRT( a1(1)**2 + a1(2)**2 + a1(3)**2 )
       Lb = DSQRT( a2(1)**2 + a2(2)**2 + a2(3)**2 )
       Lc = DSQRT( a3(1)**2 + a3(2)**2 + a3(3)**2 )
       WEIGHT_b(:) = 0.0D0
       WEIGHT_b(1) = (1.0D0/2.0D0) * (La*nkb1/tpi)**2
       WEIGHT_b(2) = (1.0D0/2.0D0) * (Lb*nkb2/tpi)**2
       WEIGHT_b(3) = (1.0D0/2.0D0) * (Lc*nkb3/tpi)**2
       WEIGHT_b(4) = (1.0D0/2.0D0) * (La*nkb1/tpi)**2
       WEIGHT_b(5) = (1.0D0/2.0D0) * (Lb*nkb2/tpi)**2
       WEIGHT_b(6) = (1.0D0/2.0D0) * (Lc*nkb3/tpi)**2
    case(6) !=== CASE(6) MONOCLINIC CELL ===*
       b_LAT(:,:) = 0.0D0
       b_LAT(1,1) =  (1.0D0/DBLE(nkb1))
       b_LAT(3,2) =  (1.0D0/DBLE(nkb3))
       b_LAT(1,3) =  (1.0D0/DBLE(nkb1))
       b_LAT(3,3) = -(1.0D0/DBLE(nkb3))
       b_LAT(1,4) = -(1.0D0/DBLE(nkb1))
       b_LAT(3,5) = -(1.0D0/DBLE(nkb3))
       b_LAT(1,6) = -(1.0D0/DBLE(nkb1))
       b_LAT(3,6) =  (1.0D0/DBLE(nkb3))
       b_LAT(2,7) =  (1.0D0/DBLE(nkb2))
       b_LAT(2,8) = -(1.0D0/DBLE(nkb2))
       VEC_b(:,:) = 0.0D0
       VEC_b(:,1) = (1.0D0/DBLE(nkb1))*b1(:)
       VEC_b(:,2) = (1.0D0/DBLE(nkb3))*b3(:)
       VEC_b(:,3) = VEC_b(:,1)-VEC_b(:,2)
       VEC_b(:,4) =-VEC_b(:,1)
       VEC_b(:,5) =-VEC_b(:,2)
       VEC_b(:,6) =-VEC_b(:,3)
       VEC_b(:,7) = (1.0D0/DBLE(nkb2))*b2(:)
       VEC_b(:,8) =-VEC_b(:,7)
       s = VEC_b(1,1)
       t = VEC_b(3,1)
       u = VEC_b(3,2)
       v = VEC_b(2,7)
       WEIGHT_b(:) = 0.0D0
       WEIGHT_b(1) = (u-t)/(2.0d0*(s**2)*u)
       WEIGHT_b(2) = (s**2-t*(u-t))/(2.0d0*(s**2)*(u**2))
       WEIGHT_b(3) = t/(2.0d0*(s**2)*u)
       WEIGHT_b(4) = WEIGHT_b(1)
       WEIGHT_b(5) = WEIGHT_b(2)
       WEIGHT_b(6) = WEIGHT_b(3)
       WEIGHT_b(7) = 1.0d0/(2.0d0*(v**2))
       WEIGHT_b(8) = WEIGHT_b(7)
    case(7) !=== CASE(7) TRICLINIC CELL ===*
       b_LAT(:,:) = 0.0D0
       b_LAT(1,1) =  (1.0D0/DBLE(nkb1))
       b_LAT(2,2) =  (1.0D0/DBLE(nkb2))
       b_LAT(3,3) =  (1.0D0/DBLE(nkb3))
       b_LAT(1,4) =  (1.0D0/DBLE(nkb1))
       b_LAT(2,4) = -(1.0D0/DBLE(nkb2))
       b_LAT(2,5) =  (1.0D0/DBLE(nkb2))
       b_LAT(3,5) = -(1.0D0/DBLE(nkb3))
       b_LAT(3,6) =  (1.0D0/DBLE(nkb3))
       b_LAT(1,6) = -(1.0D0/DBLE(nkb1))
       b_LAT(:,7) = -b_LAT(:,1)
       b_LAT(:,8) = -b_LAT(:,2)
       b_LAT(:,9) = -b_LAT(:,3)
       b_LAT(:,10)= -b_LAT(:,4)
       b_LAT(:,11)= -b_LAT(:,5)
       b_LAT(:,12)= -b_LAT(:,6)
       VEC_b(:,:) = 0.0D0
       VEC_b(:,1) = (1.0D0/DBLE(nkb1))*b1(:)
       VEC_b(:,2) = (1.0D0/DBLE(nkb2))*b2(:)
       VEC_b(:,3) = (1.0D0/DBLE(nkb3))*b3(:)
       VEC_b(:,4) = VEC_b(:,1)-VEC_b(:,2)
       VEC_b(:,5) = VEC_b(:,2)-VEC_b(:,3)
       VEC_b(:,6) = VEC_b(:,3)-VEC_b(:,1)
       VEC_b(:,7) =-VEC_b(:,1)
       VEC_b(:,8) =-VEC_b(:,2)
       VEC_b(:,9) =-VEC_b(:,3)
       VEC_b(:,10)=-VEC_b(:,4)
       VEC_b(:,11)=-VEC_b(:,5)
       VEC_b(:,12)=-VEC_b(:,6)
       bb(:,:)=0.0d0
       do ib=1,Nbh
          ab=0
          do ix=1,3
             do jx=ix,3
                ab=ab+1
                bb(ib,ab)=VEC_b(ix,ib)*VEC_b(jx,ib)
             enddo
          enddo
       enddo
       info=0
       call dgetrf(Nbh,Nbh,bb,Nbh,ipiv,info)
       call dgetri(Nbh,bb,NBh,ipiv,wdgetri,L_wdgetri,info)
       VEC_d(1)=0.5d0 ! xx
       VEC_d(2)=0.0d0 ! xy
       VEC_d(3)=0.0d0 ! xz
       VEC_d(4)=0.5d0 ! yy
       VEC_d(5)=0.0d0 ! yz
       VEC_d(6)=0.5d0 ! zz
       WEIGHT_b(:) = 0.0D0
       do ib=1,Nbh
          ab=0
          SUM_REAL=0.0d0
          do ix=1,3
             do jx=ix,3
                ab=ab+1
                SUM_REAL=SUM_REAL+VEC_d(ab)*bb(ab,ib)
             enddo
          enddo
          WEIGHT_b(ib)=SUM_REAL
       enddo
       WEIGHT_b(7) = WEIGHT_b(1)
       WEIGHT_b(8) = WEIGHT_b(2)
       WEIGHT_b(9) = WEIGHT_b(3)
       WEIGHT_b(10)= WEIGHT_b(4)
       WEIGHT_b(11)= WEIGHT_b(5)
       WEIGHT_b(12)= WEIGHT_b(6)
    case(8) !=== CASE(8) rhombohedral cell ===*
       w1 = dot_product(b2(:)-b3(:),b2(:)-b3(:))
       w2 = dot_product(b2(:)+b3(:),b2(:)+b3(:))
       if (w1.lt.w2) then
          flgact = -1.0d0
          write(6,*) 'select b2 - b3 for the second shell'
       else
          flgact = 1.0d0
          write(6,*) 'select b2 + b3 for the second shell'
       end if

       b_LAT(:,:) = 0.0D0
       b_LAT(1,1) =  (1.0D0/DBLE(nkb1))

       b_LAT(2,2) =  (1.0D0/DBLE(nkb2))

       b_LAT(3,3) =  (1.0D0/DBLE(nkb3))

       b_LAT(2,4) =  (1.0D0/DBLE(nkb2))
       b_LAT(3,4) = flgact*(1.0D0/DBLE(nkb3))

       b_LAT(3,5) =  (1.0D0/DBLE(nkb3))
       b_LAT(1,5) = flgact*(1.0D0/DBLE(nkb1))

       b_LAT(1,6) =  (1.0D0/DBLE(nkb1))
       b_LAT(2,6) = flgact*(1.0D0/DBLE(nkb2))

       b_LAT(:,7) = -b_LAT(:,1)
       b_LAT(:,8) = -b_LAT(:,2)
       b_LAT(:,9) = -b_LAT(:,3)
       b_LAT(:,10) = -b_LAT(:,4)
       b_LAT(:,11) = -b_LAT(:,5)
       b_LAT(:,12) = -b_LAT(:,6)

       do ibvec=1,12
          VEC_b(:,ibvec) = b_LAT(1,ibvec)*b1 + b_LAT(2,ibvec)*b2 &
               + b_LAT(3,ibvec)*b3
       end do

       s = VEC_b(1,1)*VEC_b(1,1)+VEC_b(1,2)*VEC_b(1,2)+VEC_b(1,3)*VEC_b(1,3)
       t = VEC_b(1,4)*VEC_b(1,4)+VEC_b(1,5)*VEC_b(1,5)+VEC_b(1,6)*VEC_b(1,6)
       u = VEC_b(1,1)*VEC_b(2,1)+VEC_b(1,2)*VEC_b(2,2)+VEC_b(1,3)*VEC_b(2,3)
       v = VEC_b(1,4)*VEC_b(2,4)+VEC_b(1,5)*VEC_b(2,5)+VEC_b(1,6)*VEC_b(2,6)

       ! w1*s + w2*t = 0.5
       ! w1*u + w2*v = 0
       ! w2 = -w1*u/v
       ! w1*s - w1*u/v*t = 0.5
       ! (s-u*t/v)*w1 = 0.5
       ! w1 = 0.5/(s-u*t/v) = 0.5*v/(s*v-u*t)
       ! w2 = -0.5*v/(s*v-u*t)*u/v =  -0.5*u/(s*v-u*t)
       w1 = 0.5d0*v/(s*v-u*t)
       w2 = -0.5d0*u/(s*v-u*t)
       
       WEIGHT_b(:) = 0.0D0
       WEIGHT_b(1) = w1
       WEIGHT_b(2) = w1
       WEIGHT_b(3) = w1
       WEIGHT_b(4) = w2
       WEIGHT_b(5) = w2
       WEIGHT_b(6) = w2
       WEIGHT_b(7) = w1
       WEIGHT_b(8) = w1
       WEIGHT_b(9) = w1
       WEIGHT_b(10) = w2
       WEIGHT_b(11) = w2
       WEIGHT_b(12) = w2
    end select
    !!---
    Do ibvec=1,NB
       write(6,*) b_LAT(:,ibvec)
    ENDDO
    write(6,*) ' '

    SUM_REAL=0.0D0
    Do ibvec=1,NB
       SUM_REAL=SUM_REAL+WEIGHT_b(ibvec)
    ENDDO
    WEIGHT=SUM_REAL

    write(6,*) 'WEIGHT_b=',WEIGHT_b(1)
    write(6,*) 'WEIGHT  =',WEIGHT

    !======================*
    !=== CHECK SUM RULE ===*
    !======================*

    write(6,*) ' '
    write(6,*) 'SUM RULE CHECK OF B-VEC'
    write(6,*) ' '
    del = 1.0d2*epsilon(1.0d0)
    info = 0
    Do IC    = 1 , 3
       Do JC    = 1 , 3
          SUM_REAL = 0.0D0
          Do ibvec = 1 , NB
             SUM_REAL=SUM_REAL+WEIGHT_b(ibvec)*VEC_b(IC,ibvec)*VEC_b(JC,ibvec)
          ENDDO ! ibvec
          write(6,*) IC,JC,SUM_REAL
          if (IC.eq.JC) then
             if (abs(SUM_REAL-1.0d0).gt.del) info = 1
          else
             if (abs(SUM_REAL).gt.del) info = 1
          end if
       ENDDO ! JC
    ENDDO ! IC
    if (info.eq.1) then
       write(6,*) 'sum rule of b-vector is broken'
       stop
    end if
    return
  end subroutine wan_setup_bvec

!
! USPP support by
! A. Ferretti, A. Calzolari, B. Bonferroni and R. Di Felice,
! J. Phys: Condens. Matter 19, 036215 (2007)
!
  subroutine wan_calc_ism(dp,ncomp,aa,bb,mgs &
       ,ng,rg,nnp,pg,srg,NTK,NTG,NB_start,NB_end,NTB &
       ,N_BAND,N_BAND_BOTTOM &
       ,NB,b_LAT,mapkpt,SK0,OVERLAP,KPT &
       ,bspotfc,nkd,bspots,nsi,kd,asi,qgbl)
    use meshgrid
    use types
    use bsplpot
    use converter
    use util
    use symutil
    implicit none

    integer:: ncomp
    integer:: ng,nnp

    type (meshgrid_struct):: mgs
    type (wfn_for_k):: dp(mgs%nks)

    real(kind=8):: aa(3,3),bb(3,3)
    integer:: rg(3,3,ng),pg(3,ng),srg(ng)

    !=== ARRAYS FOR INTERSTATE MATRIX ===*
    real(kind=8) :: b_LAT(3,NB)

    integer:: NTK,NTG
    integer:: NB_start,NB_end,NTB
    integer:: NB  ! TOTAL NUMBER OF B-VEC
    integer:: mapkpt(5,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk)

    !--- 2011 1 18 ---*
    complex(kind=8),intent(out):: OVERLAP(NTB,NTB,NTK,NB)
    integer,intent(out) :: KPT(NTK,NB)

    integer:: nkd,nsi,kd(nsi)
    real(kind=8):: asi(3,nsi)
    type (bspotfc_struct):: bspotfc
    type(bspot_for_atm):: bspots(nkd)
    type (qgbl_for_atm):: qgbl(nkd,NB)

    integer:: NG0_KET, NG0_BRA
    integer:: KG0_KET(3,NTG), KG0_BRA(3,NTG)
    integer:: ik_map(5),ik_ib_map(5)
    integer:: ik,jk,ig,jrg,iph
    complex(Kind=8):: pf(NTG)
    real(kind=8):: ph
    integer:: ib,jb
    complex(kind=8):: ctemp2,cwrk(NTG)

    real(kind=8),parameter :: tpi=2.0d0*3.14159265358979323846d0
    complex(kind=8),parameter :: ci=(0.0D0,1.0D0)

    !=== ARRAYS FOR SORTING SAMPLE K-POINTS ===*
    real(kind=8) :: SK0(3,NTK)
    !=== ARRAYS FOR WAVE FUNCTIONS ===*
    complex(kind=8) :: C0_BRA(ncomp,0:NTG,NTB)
    complex(kind=8) :: C0_KET(ncomp,0:NTG,NTB)
    integer,allocatable :: map_bra(:,:,:)
    !=== ARRAYS FOR 3D FFT GRID ===*
    integer :: nwx,nwy,nwz
    !=== ARRAYS FOR K+G VECTORS ===*
    integer,pointer :: KG0(:,:,:)
    integer,pointer :: NG0(:)
    !=== ARRAYS FOR INTERSTATE MATRIX ===*
    real(kind=8) :: SL(3)
    integer :: ik_ib,ibvec
    integer :: N_BAND(NTK)
    integer :: N_BAND_BOTTOM(NTK)
    complex(kind=8),allocatable:: gmbra(:,:), gmket(:,:), zt(:,:,:)
    complex(kind=8),allocatable:: ylm(:,:,:)
    real(kind=8),allocatable:: vpj(:,:,:,:),akg(:)
    !--- for triclinic Wannier
    !=== FOR CALCULATIONS OF THE DETERMINANT ===*
    complex(kind=8) :: DET
    integer :: ILL
    !=== ARRAYS FOR BASIC VARIABLES

    integer:: ip1,ip2,ipa1,ipa2,l1,l2,m1,m2,i1,i2,k1
    integer:: it,ikd,nprj,ic,ib1,ib2
    integer:: it2h(nsi+1),lmax,maxr,lpnum
    real(kind=8):: va

    complex(kind=8):: tsu2(2,2), tspr(2)

    NG0 => mgs%nkm
    KG0 => mgs%kg

    va = cellvol(aa)

    nwx = mgs%nwx
    nwy = mgs%nwy
    nwz = mgs%nwz
    !
    ! widen nw[xyz] to consider mapping between different k-points
    !
    nwx = nwx + 1
    nwy = nwy + 1
    nwz = nwz + 1

    !---
    write(6,*) '====================='
    write(6,*) 'WRITE SAMPLE K-POINTS'
    write(6,*) '====================='
    DO IK=1,NTK
       WRITE(6,'(I5,3F15.10)') IK,SK0(:,IK)
    ENDDO

    !=============================*
    !=== SET INTERSTATE MATRIX ===*
    !=============================*

    WRITE(6,*) ' '
    WRITE(6,*) '====================================='
    WRITE(6,*) 'INTERSTATE MATRIX; LOG{det[M(k,k+b)]}'
    WRITE(6,*) '====================================='
    WRITE(6,*) ' '

    OVERLAP(:,:,:,:) = 0.0D0

    ip1 = 0
    do it=1,nsi
       ikd = kd(it)
       it2h(it) = ip1
       ip1 = ip1 + bspots(ikd)%nprj
    end do
    it2h(nsi+1) = ip1
    nprj = ip1
    lmax = getlmax(nkd,bspots)
    maxr = getmaxr(nkd,bspots)
    lpnum = getlpnum(nkd,bspots)
    allocate(gmbra(nprj,NTB),gmket(nprj,NTB),zt(ncomp,NTG,nprj), &
         vpj(NTG,maxr,lpnum,nkd),akg(NTG))

    !---
    !      write(6,*) ' '
    !      write(6,*) 'FOR GAMMA POINT'
    !      write(6,*) ' '
    !      Do i_band=1,NTB
    !      Do j_band=1,NTB
    !      SUM_CMPX=0.0D0
    !      Do ig=1,NG0(1) ! NG0(365)
    !          SUM_CMPX
    !      + = SUM_CMPX
    !      + + CONJG(C0(:,ig,i_band,365))
    !      + *       C0(:,ig,j_band,365)
    !         SUM_CMPX
    !     + = SUM_CMPX
    !     + + CONJG(C0(:,ig,i_band,1))
    !     + *       C0(:,ig,j_band,1)
    !      Enddo
    !      write(6,*) i_band,j_band,SUM_CMPX
    !      Enddo
    !      Enddo
    !      write(6,*) ' '
    !---
    ALLOCATE(map_bra(-nwx:nwx,-nwy:nwy,-nwz:nwz))
    Do ik=1,NTK
       ik_map = mapkpt(:,nint(mgs%nk*SK0(1,ik)), &
            nint(mgs%nk*SK0(2,ik)), &
            nint(mgs%nk*SK0(3,ik)))
       !=== SET C0_BRA ===*

       C0_BRA(:,0,:) = 0.0D0
       NG0_BRA = NG0(ik_map(1))
       jrg = ik_map(2)
!$omp parallel do private(ig,iph,ph)
       Do ig = 1 , NG0_BRA
          KG0_BRA(:,ig) = matmul(rg(:,:,jrg),KG0(:,ig,ik_map(1)))
          KG0_BRA(:,ig) = srg(jrg)*KG0_BRA(:,ig)
          iph = pg(1,jrg)*KG0_BRA(1,ig)+pg(2,jrg)*KG0_BRA(2,ig) &
               +pg(3,jrg)*KG0_BRA(3,ig)
          ph = tpi*dble(iph)/nnp
          pf(ig) = dcmplx(cos(ph),-sin(ph))
          KG0_BRA(1:3,ig) = KG0_BRA(1:3,ig) - ik_map(3:5)
       end do
       call rg2tsu2(rg(:,:,jrg),aa,tsu2)
       if (srg(jrg).eq.1) then
          if (ncomp.eq.1) then
!$omp parallel do private(ib,jb,ig)
             do ib=NB_start,NB_end
                jb = ib-NB_start+1
                do ig=1,NG0_BRA
                   C0_BRA(1,ig,jb) = dp(ik_map(1))%wf(1,ig,ib)*pf(ig)
                end do
             end do
          else
!$omp parallel do private(ib,jb,ig,tspr)
             do ib=NB_start,NB_end
                jb = ib-NB_start+1
                do ig=1,NG0_BRA
                   tspr = matmul(tsu2,dp(ik_map(1))%wf(1:2,ig,ib))
                   C0_BRA(1:2,ig,jb) = tspr*pf(ig)
                end do
             end do
          end if
       else
          if (ncomp.eq.1) then
!$omp parallel do private(ib,jb,ig)
             do ib=NB_start,NB_end
                jb = ib-NB_start+1
                do ig=1,NG0_BRA
                   C0_BRA(1,ig,jb) = conjg(dp(ik_map(1))%wf(1,ig,ib))*pf(ig)
                end do
             end do
          else
!$omp parallel do private(ib,jb,ig,tspr)
             do ib=NB_start,NB_end
                jb = ib-NB_start+1
                do ig=1,NG0_BRA
                   tspr(1) = -conjg(dp(ik_map(1))%wf(2,ig,ib))
                   tspr(2) = conjg(dp(ik_map(1))%wf(1,ig,ib))
                   C0_BRA(1:2,ig,jb) = matmul(tsu2,tspr)*pf(ig)
                end do
             end do
          end if
       end if

       !=== CONSTRUCT MAP FUNCTION ===*
       map_bra(:,:,:) = 0
       Do ig = 1 , NG0_BRA
          map_bra(KG0_BRA(1,ig),KG0_BRA(2,ig),KG0_BRA(3,ig)) = ig
       Enddo

       allocate(ylm(ncomp,NG0_BRA,ncomp*lmax**2))
       call getylb(NG0_BRA,SK0(:,ik),KG0_BRA,bb,ylm,lmax,ncomp)
       call kg2akg(bb,SK0(:,ik),NG0_BRA,KG0_BRA,akg)
       do ikd=1,nkd
          call evbetg(bspotfc,bspots(ikd),NG0_BRA,akg, &
               NTG,maxr,lpnum,vpj(:,:,:,ikd))
       end do
       do it=1,nsi
          ikd = kd(it)
!$omp parallel do private(ig,ph)
          do ig=1,NG0_BRA
             ph = tpi*(asi(1,it)*KG0_BRA(1,ig)+asi(2,it)*KG0_BRA(2,ig) &
                  +asi(3,it)*KG0_BRA(3,ig))
             cwrk(ig) = dcmplx(cos(ph),-sin(ph))
          end do
          do ipa1 = 1, bspots(ikd)%nprj
             l1 = bspots(ikd)%ip2l(ipa1)
             i1 = bspots(ikd)%ip2r(ipa1)
             m1 = bspots(ikd)%ip2m(ipa1)
             ctemp2 = (-ci)**bspots(ikd)%l2lq(l1)
             ip1 = it2h(it) + ipa1
             do ig = 1,NG0_BRA
                do ic=1,ncomp
                   zt(ic,ig,ip1) = &
                        cwrk(ig)*ylm(ic,ig,m1)*vpj(ig,i1,l1,ikd)*ctemp2
                end do
             end do
          end do
       end do
       deallocate(ylm)

       call zgemm('C','N',nprj,N_BAND(ik),ncomp*NG0_BRA,(1.0d0,0.0d0),zt,ncomp*NTG, &
            C0_BRA(1,1,N_BAND_BOTTOM(ik)+1),ncomp*(NTG+1),(0.0d0,0.0d0),gmbra,nprj)

       Do ibvec = 1 , NB
          SL(:) = SK0(:,ik) + b_LAT(:,ibvec)

          !=== SET IK_IB ===*

          Do jk = 1 , NTK
             IF( maxval(modulo( &
                  nint(mgs%nk*(SK0(:,jk)-SL)), mgs%nk)) .eq. 0 ) THEN
                ik_ib = jk
                EXIT
             ENDIF
          ENDDO ! jk
          KPT(ik,ibvec) = ik_ib
          !      write(6,*) 'ik=',ik,'ik_ib=',ik_ib
          ik_ib_map = mapkpt(:,nint(mgs%nk*SL(1)), &
               nint(mgs%nk*SL(2)), &
               nint(mgs%nk*SL(3)))

          !=== SET C0_KET ===*

          C0_KET(:,0,:)=0.0D0
          NG0_KET = NG0(ik_ib_map(1))
          jrg = ik_ib_map(2)
!$omp parallel do private(ig,iph,ph)
          do ig=1,NG0_KET
             KG0_KET(:,ig) = matmul(rg(:,:,jrg),KG0(:,ig,ik_ib_map(1)))
             KG0_KET(:,ig) = srg(jrg)*KG0_KET(:,ig)
             iph = pg(1,jrg)*KG0_KET(1,ig)+pg(2,jrg)*KG0_KET(2,ig) &
                  +pg(3,jrg)*KG0_KET(3,ig)
             ph = tpi*dble(iph)/nnp
             pf(ig) = dcmplx(cos(ph),-sin(ph))
             KG0_KET(1:3,ig) = KG0_KET(1:3,ig) - ik_ib_map(3:5)
          end do
          call rg2tsu2(rg(:,:,jrg),aa,tsu2)
          if (srg(jrg).eq.1) then
             if (ncomp.eq.1) then
!$omp parallel do private(ib,jb,ig)
                do ib=NB_start,NB_end
                   jb = ib-NB_start+1
                   do ig=1,NG0_KET
                      C0_KET(1,ig,jb)=dp(ik_ib_map(1))%wf(1,ig,ib)*pf(ig)
                   end do
                end do
             else
!$omp parallel do private(ib,jb,ig,tspr)
                do ib=NB_start,NB_end
                   jb = ib-NB_start+1
                   do ig=1,NG0_KET
                      tspr = dp(ik_ib_map(1))%wf(1:2,ig,ib)
                      C0_KET(1:2,ig,jb) = matmul(tsu2,tspr)*pf(ig)
                   end do
                end do
             end if
          else
             if (ncomp.eq.1) then
!$omp parallel do private(ib,jb,ig)
                do ib=NB_start,NB_end
                   jb = ib-NB_start+1
                   do ig=1,NG0_KET
                      C0_KET(1,ig,jb)=conjg(dp(ik_ib_map(1))%wf(1,ig,ib))*pf(ig)
                   end do
                end do
             else
!$omp parallel do private(ib,jb,ig,tspr)
                do ib=NB_start,NB_end
                   jb = ib-NB_start+1
                   do ig=1,NG0_KET
                      tspr(1) = -conjg(dp(ik_ib_map(1))%wf(2,ig,ib))
                      tspr(2) = conjg(dp(ik_ib_map(1))%wf(1,ig,ib))
                      C0_KET(1:2,ig,jb) = matmul(tsu2,tspr)*pf(ig)
                   end do
                end do
             end if
          endif

          allocate(ylm(ncomp,NG0_KET,ncomp*lmax**2))
          call getylb(NG0_KET,SL,KG0_KET,bb,ylm,lmax,ncomp)
          call kg2akg(bb,SL,NG0_KET,KG0_KET,akg)
          do ikd=1,nkd
             call evbetg(bspotfc,bspots(ikd),NG0_KET,akg, &
                  NTG,maxr,lpnum,vpj(:,:,:,ikd))
          end do
          do it=1,nsi
             ikd = kd(it)
!$omp parallel do private(ig,ph)
             do ig=1,NG0_KET
                ph = tpi*(asi(1,it)*KG0_KET(1,ig)+asi(2,it)*KG0_KET(2,ig) &
                     +asi(3,it)*KG0_KET(3,ig))
                cwrk(ig) = dcmplx(cos(ph),-sin(ph))
             end do
             do ipa2 = 1, bspots(ikd)%nprj
                l2 = bspots(ikd)%ip2l(ipa2)
                i2 = bspots(ikd)%ip2r(ipa2)
                m2 = bspots(ikd)%ip2m(ipa2)
                ctemp2 = (-ci)**bspots(ikd)%l2lq(l2)
                ip2 = it2h(it) + ipa2
                do ig = 1,NG0_KET
                   do ic=1,ncomp
                      zt(ic,ig,ip2) = &
                           cwrk(ig)*ylm(ic,ig,m2)*vpj(ig,i2,l2,ikd)*ctemp2
                   end do
                end do
             end do
          end do
          deallocate(ylm)

          call zgemm('C','N',nprj,N_BAND(ik_ib),ncomp*NG0_KET,(1.0d0,0.0d0), &
               zt,ncomp*NTG,C0_KET(1,1,N_BAND_BOTTOM(ik_ib)+1),ncomp*(NTG+1), &
               (0.0d0,0.0d0),gmket,nprj)

          CALL CALC_OVERLAP(NTK,NTB,ncomp,NTG,nwx,nwy,nwz, &
               ik,ik_ib,map_bra,NG0_KET,KG0_KET,N_BAND, &
               C0_BRA,C0_KET,OVERLAP(:,:,ik,ibvec), &
               N_BAND_BOTTOM)

          do ib2 = 1,N_BAND(ik_ib)
             do ib1 = 1,N_BAND(ik)
                do it = 1,nsi
                   ikd = kd(it)
                   k1 = 0
                   do ipa1 = 1,bspots(ikd)%nprj
                      l1 = bspots(ikd)%ip2l(ipa1)
                      i1 = bspots(ikd)%ip2r(ipa1)
                      m1 = bspots(ikd)%ip2m(ipa1)
                      ip1 = it2h(it) + ipa1
                      do ipa2 = 1,bspots(ikd)%l2pt(l1)
                         l2 = bspots(ikd)%ip2l(ipa2)
                         i2 = bspots(ikd)%ip2r(ipa2)
                         m2 = bspots(ikd)%ip2m(ipa2)
                         ip2 = it2h(it) + ipa2
                         if (.not.( (l1.eq.l2).and. &
                              ((i2.gt.i1) .or. ((i1.eq.i2).and.(m2.gt.m1))))) then
                            k1=k1+1
                            OVERLAP(ib1,ib2,ik,ibvec) = &
                                 OVERLAP(ib1,ib2,ik,ibvec) + &
                                 1.0d0/va*conjg(gmbra(ip1,ib1))* &
                                 qgbl(ikd,ibvec)%qg(k1)*gmket(ip2,ib2)

                            if (ipa1.ne.ipa2) then
                               OVERLAP(ib1,ib2,ik,ibvec) = &
                                    OVERLAP(ib1,ib2,ik,ibvec) + &
                                    1.0d0/va*conjg(gmbra(ip2,ib1))* &
                                    conjg(qgbl(ikd,ibvec)%qg(k1))*gmket(ip1,ib2)
                            end if
                         end if
                      end do
                   end do
                end do
             end do
          end do

          IF(N_BAND(ik)==N_BAND(ik_ib)) THEN

             !=== CALCULATE THE DETERMINANT. ===*

             call calc_det(N_BAND(ik),OVERLAP(:,:,ik,ibvec),NTB,DET,ILL)

             IF(ILL/=0) write(6,*) 'ILL=',ILL
             !      write(6,'(I5,A8,I5,A8,2F15.10)') 'IK=',ik,'IK_IB=',ik_ib,
             !     +                                 'LOG=',LOG(DET)
          ENDIF
       ENDDO ! ibvec
       ! WRITE(6,*) ' '
    ENDDO ! ik

    DEALLOCATE( map_bra )
    deallocate(gmbra,gmket,zt,vpj,akg)
    return
  end subroutine wan_calc_ism

  subroutine wan_init_amat(dsp,ncomp,nw,VOLUME,aa,bb,mgs &
       ,ng,rg,nnp,pg,srg,NTK,NTG,NB_start,NB_end,NTB &
       ,N_BAND,N_BAND_BOTTOM &
       ,NGAUSS,nigs,LGAUSS,MGAUSS,CGAUSS,TAU_GAUSS,ALPHA_GAUSS,qgauss,B_MAT &
       ,mapkpt,SK0,A_MAT)
    use types
    use meshgrid
    use util
    use symutil
    implicit none

    integer:: ncomp,nw
    integer:: ng,nnp

    type (meshgrid_struct):: mgs
    type (wfn_for_k):: dsp(mgs%nks)

    real(kind=8):: VOLUME,aa(3,3),bb(3,3)
    integer:: rg(3,3,ng),pg(3,ng),srg(ng)

    integer:: NTK,NTG
    integer:: NB_start,NB_end,NTB
    integer:: NGAUSS ! TOTAL NUMBER OF INITIAL GUESS GAUSSIAN
    integer:: nigs
    integer:: LGAUSS(nigs)
    integer:: MGAUSS(nigs)
    integer:: CGAUSS(nigs)
    real(kind=8):: TAU_GAUSS(3,nigs)
    real(kind=8):: ALPHA_GAUSS(nigs)
    real(kind=8):: qgauss(3,nigs)
    complex(kind=8) :: B_MAT(nigs,NGAUSS)
    integer:: mapkpt(5,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk)

    integer:: NG0_BRA
    integer:: KG0_BRA(3,NTG)
    integer:: ik_map(5)
    integer:: ik,ig,jrg,iph
    complex(Kind=8):: pf(NTG)

    real(kind=8),parameter :: pi=dacos(-1.0d0)
    real(kind=8),parameter :: tpi=2.0d0*pi
    complex(kind=8),parameter :: ci=(0.0D0,1.0D0)

    !=== ARRAYS FOR SORTING SAMPLE K-POINTS ===*
    real(kind=8) :: SK0(3,NTK)
    !=== ARRAYS FOR WAVE FUNCTIONS ===*
    complex(kind=8) :: C0_BRA(ncomp,0:NTG,NTB)
    !=== ARRAYS FOR K+G VECTORS ===*
    integer,pointer :: KG0(:,:,:)
    integer,pointer :: NG0(:)
    integer:: NB0(NTK)
    !=== ARRAYS FOR INITIAL GUESS ===*
    real(kind=8) :: NORM_GAUSS(nigs)
    integer :: i_gauss,j_gauss
    integer :: iL,jL,jM,jC
    real(kind=8) :: VEC_KG(3,NTG)
    real(kind=8) :: KG2
    real(kind=8) :: PHASE
    complex(kind=8) :: PF1(NTG)
    real(kind=8) :: PF2(NTG)
    real(kind=8) :: ZETA1,ZETA2
    real(kind=8) :: YLM
    complex(kind=8) :: A_MAT(NTB,NGAUSS,NTK)
    complex(kind=8) :: AA_MAT(NTB,nigs,NTK)
    complex(kind=8) :: AA_TMP(NTB,nigs)
    !=== ARRAYS FOR INTERSTATE MATRIX ===*
    integer :: i_band
    integer :: N_BAND(NTK)
    integer :: N_BAND_BOTTOM(NTK)
    !=== ARRAYS FOR BASIC VARIABLES
    real(kind=8) :: tmp(3)
    complex(kind=8) :: SUM_CMPX
    real(kind=8) :: a1(3),a2(3),a3(3)
    real(kind=8) :: b1(3),b2(3),b3(3)

    complex(kind=8):: tsu2(2,2),tspr(2)
    integer:: ib,jb
    real(kind=8):: eulm(3,3)

    NG0 => mgs%nkm
    KG0 => mgs%kg
    NB0 = nw

    a1 = aa(:,1)
    a2 = aa(:,2)
    a3 = aa(:,3)
    b1 = bb(:,1)
    b2 = bb(:,2)
    b3 = bb(:,3)

    !======================================================*
    !=== PREPARE INITIAL GUESS FOR OMEGA_I MINIMIZATION ===*
    !======================================================*

    !=== NORMALIZED CONST FOR GAUSSIAN ===*

    Do i_gauss = 1 , nigs

       iL = LGAUSS(i_gauss)

       IF(iL == 1) THEN

          NORM_GAUSS(i_gauss) &
               = ( 8.0D0 * ALPHA_GAUSS(i_gauss) / pi )**0.25D0 &
               * ( 4.0D0 * ALPHA_GAUSS(i_gauss) / 1.0D0  )**0.50D0

       ENDIF

       IF(iL == 2) THEN

          NORM_GAUSS(i_gauss) &
               = ( 8.0D0 * ALPHA_GAUSS(i_gauss) / pi )**0.25D0 &
               * ( 4.0D0 * ALPHA_GAUSS(i_gauss) / 1.0D0  )**0.50D0 &
               * ( 4.0D0 * ALPHA_GAUSS(i_gauss) / 3.0D0  )**0.50D0

       ENDIF

       IF(iL == 3) THEN

          NORM_GAUSS(i_gauss) &
               = ( 8.0D0 * ALPHA_GAUSS(i_gauss) / pi )**0.25D0 &
               * ( 4.0D0 * ALPHA_GAUSS(i_gauss) / 1.0D0  )**0.50D0 &
               * ( 4.0D0 * ALPHA_GAUSS(i_gauss) / 3.0D0  )**0.50D0 &
               * ( 4.0D0 * ALPHA_GAUSS(i_gauss) / 5.0D0  )**0.50D0

       ENDIF

    ENDDO ! OF i_gauss

    !      Do i_gauss = 1 , NGAUSS
    !       write(6,*) i_gauss,NORM_GAUSS(i_gauss)
    !      ENDDO

    !=== GAUSSIAN CENTER IN CARTESIAN COORD ===*

    Do i_gauss = 1 , nigs

       tmp(:) = TAU_GAUSS(:,i_gauss)

       TAU_GAUSS(:,i_gauss) &
            = tmp(1) * a1(:) &
            + tmp(2) * a2(:) &
            + tmp(3) * a3(:)

       write(6,1234) TAU_GAUSS(:,i_gauss)

    ENDDO
1234 FORMAT(3F20.12)

    !=======================*
    !=== CONSTRUCT A_MAT ===*
    !=======================*

    AA_MAT(:,:,:) = 0.0D0
    Do ik      = 1 , NTK
       ik_map = mapkpt(:,nint(mgs%nk*SK0(1,ik)), &
            nint(mgs%nk*SK0(2,ik)), &
            nint(mgs%nk*SK0(3,ik)))
       NG0_BRA = NG0(ik_map(1))
       C0_BRA(:,0,:) = 0.0D0
       jrg = ik_map(2)
       call rg2tsu2(rg(:,:,jrg),aa,tsu2)
       do ig = 1 , NG0_BRA
          KG0_BRA(:,ig) = matmul(rg(:,:,jrg),KG0(:,ig,ik_map(1)))
          KG0_BRA(:,ig) = srg(jrg)*KG0_BRA(:,ig)
          iph = pg(1,jrg)*KG0_BRA(1,ig)+pg(2,jrg)*KG0_BRA(2,ig) &
               +pg(3,jrg)*KG0_BRA(3,ig)
          pf(ig) = exp(-ci*tpi*dble(iph)/nnp)
          KG0_BRA(1:3,ig) = KG0_BRA(1:3,ig) - ik_map(3:5)
       end do
       if (srg(jrg).eq.1) then
          if (ncomp.eq.1) then
             do ib = NB_start, NB_end
                jb = ib - NB_start + 1
                do ig = 1, NG0_BRA
                   C0_BRA(1,ig,jb) = dsp(ik_map(1))%wf(1,ig,ib)*pf(ig)
                end do
             end do
          else
             do ib = NB_start, NB_end
                jb = ib - NB_start + 1
                do ig = 1, NG0_BRA
                   tspr = matmul(tsu2,dsp(ik_map(1))%wf(1:2,ig,ib))
                   C0_BRA(1:2,ig,jb) = tspr*pf(ig)
                end do
             end do
          end if
       else
          if (ncomp.eq.1) then
             do ib = NB_start, NB_end
                jb = ib - NB_start + 1
                do ig = 1, NG0_BRA
                   C0_BRA(1,ig,jb) = conjg(dsp(ik_map(1))%wf(1,ig,ib))*pf(ig)
                end do
             end do
          else
             do ib = NB_start, NB_end
                jb = ib - NB_start + 1
                do ig = 1, NG0_BRA
                   tspr(1) = -conjg(dsp(ik_map(1))%wf(2,ig,ib))
                   tspr(2) = conjg(dsp(ik_map(1))%wf(1,ig,ib))
                   C0_BRA(1:2,ig,jb) = matmul(tsu2,tspr)*pf(ig)
                end do
             end do
          end if
       endif

       Do j_gauss = 1 , nigs

          VEC_KG(:,:) = 0.0D0
          PF1(:) = 0.0D0
          PF2(:) = 0.0D0

          call geteuler(qgauss(1,j_gauss),qgauss(2,j_gauss),qgauss(3,j_gauss),eulm)
          eulm = transpose(eulm)
          
          Do ig = 1 , NG0_BRA

             VEC_KG(:,ig) = ( SK0(1,ik) + DBLE(KG0_BRA(1,ig)) ) * b1(:) &
                  + ( SK0(2,ik) + DBLE(KG0_BRA(2,ig)) ) * b2(:) &
                  + ( SK0(3,ik) + DBLE(KG0_BRA(3,ig)) ) * b3(:)
             PHASE = VEC_KG(1,ig) * TAU_GAUSS(1,j_gauss) &
                  + VEC_KG(2,ig) * TAU_GAUSS(2,j_gauss) &
                  + VEC_KG(3,ig) * TAU_GAUSS(3,j_gauss)
             PF1(ig) = EXP(-ci*PHASE )

             VEC_KG(:,ig) = matmul(eulm,VEC_KG(:,ig))

             KG2   = VEC_KG(1,ig)**2 + VEC_KG(2,ig)**2 &
                  + VEC_KG(3,ig)**2
             PF2(ig) = EXP(-KG2/4.0D0/ALPHA_GAUSS(j_gauss))

          ENDDO

          jL = LGAUSS(j_gauss)
          jM = MGAUSS(j_gauss)
          jC = CGAUSS(j_gauss)

          Do i_band  = 1 , N_BAND(ik)

             SUM_CMPX = 0.0D0

             !--- <Psi|s> ---*

             IF(jL==1) THEN
                Do ig = 1 , NG0(ik_map(1))

                   ZETA1 = ( pi / ALPHA_GAUSS(j_gauss) )**1.5D0
                   YLM   =  DSQRT( 1.0D0 / 4.0D0 / pi )

                   SUM_CMPX &
                        = SUM_CMPX &
                        + CONJG( C0_BRA(jC,ig,i_band+N_BAND_BOTTOM(ik)) ) &
                        * PF1(ig) &
                        * YLM &
                        * NORM_GAUSS(j_gauss) &
                        * ZETA1 * PF2(ig)

                ENDDO ! OF ig
             ENDIF

             IF(jL==2.AND.jM==1) THEN
                Do ig = 1 , NG0(ik_map(1))

                   ZETA1 = ( pi / ALPHA_GAUSS(j_gauss) )**1.5D0
                   ZETA2 = 1.0D0 / 2.0D0 / ALPHA_GAUSS(j_gauss)
                   YLM   = - DSQRT( 3.0D0 / 4.0D0 / pi ) * VEC_KG(2,ig)

                   SUM_CMPX &
                        = SUM_CMPX &
                        + CONJG( C0_BRA(jC,ig,i_band+N_BAND_BOTTOM(ik)) ) &
                        * PF1(ig) &
                        * (-ci)*YLM &
                        * NORM_GAUSS(j_gauss) &
                        * ZETA1 * ZETA2 * PF2(ig)

                ENDDO ! OF ig
             ENDIF

             !--- <Psi|p2> ---*

             IF(jL==2.AND.jM==2) THEN
                Do ig = 1 , NG0(ik_map(1))

                   ZETA1 = ( pi / ALPHA_GAUSS(j_gauss) )**1.5D0
                   ZETA2 = 1.0D0 / 2.0D0 / ALPHA_GAUSS(j_gauss)
                   YLM   =  DSQRT( 3.0D0 / 4.0D0 / pi ) * VEC_KG(3,ig)

                   SUM_CMPX &
                        = SUM_CMPX &
                        + CONJG( C0_BRA(jC,ig,i_band+N_BAND_BOTTOM(ik)) ) &
                        * PF1(ig) &
                        * (-ci)*YLM &
                        * NORM_GAUSS(j_gauss) &
                        * ZETA1 * ZETA2 * PF2(ig)

                ENDDO ! OF ig
             ENDIF ! OF (jL==2.AND.jM==2)

             !--- <Psi|p3> ---*

             IF(jL==2.AND.jM==3) THEN
                Do ig = 1 , NG0(ik_map(1))

                   ZETA1 = ( pi / ALPHA_GAUSS(j_gauss) )**1.5D0
                   ZETA2 = 1.0D0 / 2.0D0 / ALPHA_GAUSS(j_gauss)
                   YLM   = - DSQRT( 3.0D0 / 4.0D0 / pi ) * VEC_KG(1,ig)

                   SUM_CMPX &
                        = SUM_CMPX &
                        + CONJG( C0_BRA(jC,ig,i_band+N_BAND_BOTTOM(ik)) ) &
                        * PF1(ig) &
                        * (-ci)*YLM &
                        * NORM_GAUSS(j_gauss) &
                        * ZETA1 * ZETA2 * PF2(ig)

                ENDDO ! OF ig
             ENDIF

             !--- <Psi|d1> ---*

             IF(jL==3.AND.jM==1) THEN
                Do ig = 1 , NG0(ik_map(1))

                   ZETA1 = ( pi / ALPHA_GAUSS(j_gauss) )**1.5D0
                   ZETA2 = ( 1.0D0 / 2.0D0 / ALPHA_GAUSS(j_gauss) )**2.0D0
                   YLM   = 0.5D0 * DSQRT( 15.0D0 / pi ) &
                        * VEC_KG(1,ig) * VEC_KG(2,ig)

                   SUM_CMPX &
                        = SUM_CMPX &
                        + CONJG( C0_BRA(jC,ig,i_band+N_BAND_BOTTOM(ik)) ) &
                        * PF1(ig) &
                        * ( - YLM ) &
                        * NORM_GAUSS(j_gauss) &
                        * ZETA1 * ZETA2 * PF2(ig)

                ENDDO ! OF ig
             ENDIF

             !--- <Psi|d2> ---*

             IF(jL==3.AND.jM==2) THEN
                Do ig = 1 , NG0(ik_map(1))

                   ZETA1 = ( pi / ALPHA_GAUSS(j_gauss) )**1.5D0
                   ZETA2 = ( 1.0D0 / 2.0D0 / ALPHA_GAUSS(j_gauss) )**2.0D0
                   YLM   = (-0.5D0) * DSQRT( 15.0D0 / pi ) &
                        * VEC_KG(2,ig) * VEC_KG(3,ig)

                   SUM_CMPX &
                        = SUM_CMPX &
                        + CONJG( C0_BRA(jC,ig,i_band+N_BAND_BOTTOM(ik)) ) &
                        * PF1(ig) &
                        * ( - YLM ) &
                        * NORM_GAUSS(j_gauss) &
                        * ZETA1 * ZETA2 * PF2(ig)

                ENDDO ! OF ig
             ENDIF

             !--- <Psi|d3> ---*

             IF(jL==3.AND.jM==3) THEN
                Do ig = 1 , NG0(ik_map(1))

                   ZETA1 = ( pi / ALPHA_GAUSS(j_gauss) )**1.5D0
                   ZETA2 = ( 1.0D0 / 2.0D0 / ALPHA_GAUSS(j_gauss) )**2.0D0
                   KG2   = VEC_KG(1,ig)**2 &
                        + VEC_KG(2,ig)**2 &
                        + VEC_KG(3,ig)**2
                   YLM   = (0.25D0) * DSQRT( 5.0D0 / pi ) &
                        * ( 3.0D0 * VEC_KG(3,ig)**2 - KG2 )

                   SUM_CMPX &
                        = SUM_CMPX &
                        + CONJG( C0_BRA(jC,ig,i_band+N_BAND_BOTTOM(ik)) ) &
                        * PF1(ig) &
                        * ( - YLM ) &
                        * NORM_GAUSS(j_gauss) &
                        * ZETA1 * ZETA2 * PF2(ig)

                ENDDO ! OF ig
             ENDIF

             !--- <Psi|d4> ---*

             IF(jL==3.AND.jM==4) THEN
                Do ig = 1 , NG0(ik_map(1))

                   ZETA1 = ( pi / ALPHA_GAUSS(j_gauss) )**1.5D0
                   ZETA2 = ( 1.0D0 / 2.0D0 / ALPHA_GAUSS(j_gauss) )**2.0D0
                   YLM   = (-0.5D0) * DSQRT( 15.0D0 / pi ) &
                        * VEC_KG(3,ig) * VEC_KG(1,ig)

                   SUM_CMPX &
                        = SUM_CMPX &
                        + CONJG( C0_BRA(jC,ig,i_band+N_BAND_BOTTOM(ik)) ) &
                        * PF1(ig) &
                        * ( - YLM ) &
                        * NORM_GAUSS(j_gauss) &
                        * ZETA1 * ZETA2 * PF2(ig)

                ENDDO ! OF ig
             ENDIF

             !--- <Psi|d5> ---*

             IF(jL==3.AND.jM==5) THEN
                Do ig = 1 , NG0(ik_map(1))

                   ZETA1 = ( pi / ALPHA_GAUSS(j_gauss) )**1.5D0
                   ZETA2 = ( 1.0D0 / 2.0D0 / ALPHA_GAUSS(j_gauss) )**2.0D0
                   YLM   = (0.25D0) * DSQRT( 15.0D0 / pi ) &
                        * ( VEC_KG(1,ig)**2 - VEC_KG(2,ig)**2 )

                   SUM_CMPX &
                        = SUM_CMPX &
                        + CONJG( C0_BRA(jC,ig,i_band+N_BAND_BOTTOM(ik)) ) &
                        * PF1(ig) &
                        * ( - YLM ) &
                        * NORM_GAUSS(j_gauss) &
                        * ZETA1 * ZETA2 * PF2(ig)

                ENDDO ! OF ig
             ENDIF

             ! AA_MAT(i_band,j_gauss,ik) = SUM_CMPX
             AA_MAT(i_band,j_gauss,ik) = SUM_CMPX/DSQRT(VOLUME)

          ENDDO ! OF i_band
       ENDDO ! OF j_gauss
    ENDDO ! OF ik

    WRITE(6,*) ' '
    WRITE(6,*) 'AA_MAT'
    WRITE(6,*) ' '
    Do ik=1,NTK
       WRITE(6,*) 'ik=',ik
       Do i_band=1,N_BAND(ik)
          Do j_gauss=1,NGAUSS
             write(6,6001) i_band,j_gauss,AA_MAT(i_band,j_gauss,ik)
          ENDDO
       ENDDO
       WRITE(6,*) ' '
    ENDDO
    WRITE(6,*) ' '
6001 FORMAT(2I5,2F15.10)

    Do ik=1,NTK
       AA_TMP(:,:)=AA_MAT(:,:,ik)
       Do i_band=1,NTB
          Do i_gauss=1,NGAUSS
             SUM_CMPX=0.0D0
             Do j_gauss=1,nigs
                SUM_CMPX &
                     =SUM_CMPX &
                     +AA_TMP(i_band ,j_gauss) &
                     *B_MAT(j_gauss,i_gauss)
             ENDDO
             A_MAT(i_band,i_gauss,ik)=SUM_CMPX
          ENDDO ! OF i_gauss
       ENDDO ! OF i_band
    ENDDO ! OF ik

    WRITE(6,*) '========================='
    WRITE(6,*) 'FINISH CONSTRUCTING A_MAT'
    WRITE(6,*) '========================='

    return
  end subroutine wan_init_amat

  subroutine wannier_kazuma(OVERLAP,KPT,aa,bb,E_EIG,mgs &
       ,NTK,spll_thr,sprd_thr,NB_start,NB_end,NTB &
       ,N_BAND,N_BAND_BOTTOM &
       ,in_win_band &
       ,NGAUSS,A_MAT,DAMP,sprd_stpl_fac &
       ,nkb,n_occ &
       ,NB,VEC_b,WEIGHT_b,WEIGHT,mapkpt,SK0 &
       ,WF_POSITION,H_MAT_R,A_MAT_O)
    use types
    use meshgrid
    use util
    implicit none

    type (meshgrid_struct):: mgs

    real(kind=8):: aa(3,3),bb(3,3)

    !=== ARRAYS FOR INTERSTATE MATRIX ===*
    real(kind=8) :: VEC_b(3,NB)
    real(kind=8) :: WEIGHT_b(NB)
    real(kind=8) :: WEIGHT

    integer:: NTK
    integer:: NB_start,NB_end,NTB
    real(kind=8):: spll_thr,sprd_thr
    integer:: NGAUSS ! TOTAL NUMBER OF INITIAL GUESS GAUSSIAN
    integer:: n_occ
    integer:: NB  ! TOTAL NUMBER OF B-VEC
    real(kind=8) :: DAMP, sprd_stpl_fac
    integer :: nkb(3) ! k vector range is 0:nkb-1
    integer:: mapkpt(5,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk)

    complex(kind=8):: OVERLAP(NTB,NTB,NTK,NB)
    integer,intent(in):: KPT(NTK,NB)
    real(kind=8):: E_EIG(NTB,mgs%nks)

    integer:: ik_map(5)
    integer:: ik

    !real(kind=8),parameter :: au=27.21151d0
    !
    ! CODATA 2014
    real(kind=8),parameter :: har=27.21138602d0
    real(kind=8),parameter :: pi=dacos(-1.0d0)
    real(kind=8),parameter :: tpi=2.0d0*pi
    complex(kind=8),parameter :: ci=(0.0D0,1.0D0)
    complex(kind=8),parameter :: tci=(0.0D0,2.0D0)

    complex(kind=8) :: A_MAT_O(NTB,n_occ,NTK)
    !=== ARRAYS FOR SORTING SAMPLE K-POINTS ===*
    real(kind=8) :: SK0(3,NTK)
    !=== BIG ARRAYS ===*
    complex(kind=8),allocatable :: M_MAT(:,:,:,:)
    complex(kind=8) :: H_MAT_R(n_occ,n_occ,0:nkb(1)-1,0:nkb(2)-1,0:nkb(3)-1)
    !=== ARRAYS FOR INITIAL GUESS ===*
    integer :: i_gauss,j_gauss,k_gauss,l_gauss
    complex(kind=8) :: work_zheev(10*NTB)
    real(kind=8) :: rwork_zheev(3*NTB)
    complex(kind=8) :: A_MAT(NTB,NGAUSS,NTK)
    complex(kind=8) :: A_TMP(NTB,NGAUSS)
    !=== ARRAYS FOR INTERSTATE MATRIX ===*
    integer :: ik_ib,ibvec
    integer :: i_band,j_band,k_band,l_band
    integer :: N_BAND(NTK)
    integer :: N_BAND_BOTTOM(NTK)
    complex(kind=8) :: detcnt(NTK,NB)
    real(kind=8) :: wf_pos_sum(3), optwf_pos_sum(3)
    !--- for triclinic Wannier
    !=== FOR CALCULATIONS OF THE DETERMINANT (USING NUMPAC LIBLARY) ===*
    complex(kind=8) :: DET
    integer :: ILL
    !=== ARRAYS FOR SPILLAGE FUNCTIONAL MINIMIZATION ===*
    complex(kind=8) :: Z_MAT(NTB,NTB,NTK)
    complex(kind=8) :: Z_TMP(NTB,NTB)
    real(kind=8) :: Z_EIG(NTB)
    complex(kind=8) :: C_MAT(NTB,NGAUSS,NTK)
    complex(kind=8) :: C_TMP(NTB,NTB)
    real(kind=8) :: OMEGA_I_NEW,OMEGA_I_OLD,DEL_OMEGA_I
    integer :: I_SCF
    !=== inner energy window ===*
    integer,intent(in):: in_win_band(2,NTK)
    real(kind=8):: zm_diag
    !=== ARRAYS FOR PSEUDO EIGENVALUE PROBLEM ===*
    complex(kind=8) :: P_MAT_IN  (NGAUSS,NGAUSS)
    complex(kind=8) :: P_MAT_OUT (NGAUSS,NGAUSS)
    complex(kind=8) :: PSEUDO_MAT(NGAUSS,NGAUSS,NTK)
    real(kind=8) :: P_EIG(NGAUSS)
    real(kind=8) :: PSEUDO_EIG(NGAUSS,NTK)
    !=== ARRAYS FOR GRADIENT MATRIX ===*
    real(kind=8) :: WF_CHARGE(n_occ,NTK,NB)
    real(kind=8) :: WF_POSITION(3,n_occ)
    complex(kind=8) :: GRADIENT(n_occ,n_occ,NTK)
    complex(kind=8) :: DIRECTION(n_occ,n_occ,NTK)
    complex(kind=8) :: DIRECTION2(n_occ,n_occ,NTK)
    real(kind=8) :: OMEGA_I,OMEGA_OD,OMEGA_D
    real(kind=8) :: DELTA(4),OMEGA(4)
    real(kind=8) :: TR_GRAD(NTK)  ! TR_GRAD_OLD(NTK)
    !=== ARRAYS FOR UNITARY MATRIX ===*
    complex(kind=8) :: UNITARY(n_occ,n_occ,NTK)
    complex(kind=8) :: U_ORG  (n_occ,n_occ,NTK)
    complex(kind=8) :: U_OLD  (n_occ,n_occ,NTK)
    real(kind=8) :: SPREAD,SPREAD_OLD,DEL_SPREAD
    real(kind=8) :: STEP_LENGTH
    integer :: ILS
    integer :: I_STEP
    !=== ARRAYS FOR INTERPOLATED BAND DISPERSION ===*
    real(kind=8) :: PHASE
    complex(kind=8) :: PHASE_FACTOR
    complex(kind=8) :: H_MAT_K(n_occ,n_occ,NTK)
    integer :: ia1,ia2,ia3
    !=== ARRAYS FOR BASIC VARIABLES
    real(kind=8) :: SUM_REAL
    complex(kind=8) :: SUM_CMPX
    !=== arrays for the singular value decomposition
    complex(kind=8),allocatable:: umat(:,:),vhmat(:,:)
    integer:: lwork_zgesvd
    complex(kind=8),allocatable:: work_zgesvd(:)
    real(kind=8),allocatable:: sgv(:)
    real(kind=8),allocatable:: rwork_zgesvd(:)


    !--- 2011 1 18 ---*
    allocate(M_MAT(NTB,NTB,NTK,NB))

    ! for the singular value decomposition: zgesvd()
    ! 
    lwork_zgesvd = 50*NTB
    allocate(umat(NTB,NGAUSS),vhmat(NGAUSS,NGAUSS), &
         work_zgesvd(lwork_zgesvd),sgv(NGAUSS),rwork_zgesvd(5*NTB))

    WRITE(6,*) ' '
    WRITE(6,*) '========================================='
    WRITE(6,*) '=== WE NOW CALCULATE WANNIER FUNCTION ==='
    WRITE(6,*) '========================================='
    WRITE(6,*) ' '

    !=======================*
    !=== CONSTRUCT C_MAT ===*
    !=======================*
#ifdef DEBUG_WANNIER_METAL
    WRITE(6,*) ' '
    WRITE(6,*) '==========================='
    WRITE(6,*) 'CHECK OF UNITALITY OF C_MAT'
    WRITE(6,*) '==========================='
    WRITE(6,*) ' '
#endif
    !
    ! calculate the polar decomposition of A_MAT.
    ! A_MAT = U * P, where U is a unitary matrix and
    ! P is a Hermite matrix.
    ! by SVD, A_MAT = U_s * S * V^H.
    ! A_MAT = U_s * V^H * V * S * V^H
    !       =  (U_s * V^H) * (V * S * V^H)
    !       =  U * P
    ! U is stored into C_MAT.
    !
    C_MAT = 0.0d0
    do ik = 1, NTK
       A_TMP = A_MAT(:,:,ik) 
       call zgesvd('S','S',N_BAND(ik),NGAUSS,A_TMP,NTB,sgv, &
            umat,NTB,vhmat,NGAUSS, &
            work_zgesvd,lwork_zgesvd,rwork_zgesvd,ILL)
       if (ILL .ne. 0) then
          write(6,*) 'Error: zgesvd returns ', ILL
       end if
       call zgemm('N','N',N_BAND(ik),NGAUSS,NGAUSS,(1.0d0,0.0d0), &
            umat,NTB,vhmat,NGAUSS,(0.0d0,0.0d0),C_MAT(1,1,ik),NTB)

#ifdef DEBUG_WANNIER_METAL
       ILL = 0
       Do i_gauss = 1 , NGAUSS
          Do j_gauss = 1 , NGAUSS

             SUM_CMPX = 0.0D0

             Do i_band = 1 , N_BAND(ik)

                SUM_CMPX = SUM_CMPX &
                     + CONJG( C_MAT(i_band,i_gauss,ik) ) &
                     *        C_MAT(i_band,j_gauss,ik)

             ENDDO

             if (i_gauss.eq.j_gauss) then
                SUM_CMPX = SUM_CMPX - 1.0d0
             end if
             if (abs(SUM_CMPX) .gt. 1.0d4*epsilon(1.0d0)) then
                ILL = 1
                write(6,*) ik,i_gauss,j_gauss,SUM_CMPX
             end if

          ENDDO ! OF j_gauss
       ENDDO ! OF i_gauss
       if (ILL .ne. 0) then
          write(6,*) 'Error: C_MAT is not unitary, ik = ', ik
       end if
#endif
    end do
   
    do ik = 1 , NTK
       do i_band = 1 ,  N_BAND(ik)
          do j_band = 1 ,  N_BAND(ik)
             SUM_CMPX = 0.0D0
             do i_gauss = 1 , NGAUSS
                SUM_CMPX = SUM_CMPX &
                     +       C_MAT(i_band,i_gauss,ik) &
                     * conjg(C_MAT(j_band,i_gauss,ik))
             end do
             Z_TMP(i_band,j_band) = SUM_CMPX
          end do
       end do
       call wan_mat_clr_in_win(N_BAND_BOTTOM(ik), &
            in_win_band(:,ik),Z_TMP,NTB,1.0d0)
       call zheev('V','U',N_BAND(ik),Z_TMP,NTB,Z_EIG, &
            work_zheev,10*NTB,rwork_zheev,ILL)
       do i_gauss = 0, NGAUSS-1
          do i_band = 1, N_BAND(ik)
             C_MAT(i_band, i_gauss+1, ik) = Z_TMP(i_band, N_BAND(ik)-i_gauss)
          end do
       end do
    end do

    WRITE(6,*) ' '
    WRITE(6,*) '==========================================='
    WRITE(6,*) '=== MINIMIZATION OF SPILLAGE FUNCTIONAL ==='
    WRITE(6,*) '==========================================='
    WRITE(6,*) ' '

    I_SCF        = 0
    Z_MAT(:,:,:) = 0.0D0
    DEL_OMEGA_I  = 1.0D0
    OMEGA_I_OLD  = 0.0D0

    Do WHILE( DEL_OMEGA_I > spll_thr)

       !=== CONSTRUCT M_MAT ===*

       M_MAT(:,:,:,:) = 0.0D0

       Do ik    = 1 , NTK
          Do ibvec = 1 , NB

             ik_ib = KPT(ik,ibvec)

             Do i_band  = 1 , N_BAND(ik)
                Do j_gauss = 1 , NGAUSS

                   SUM_CMPX = 0.0D0

                   Do k_band = 1 , N_BAND(ik_ib)

                      SUM_CMPX &
                           = SUM_CMPX &
                           + OVERLAP(i_band,k_band ,ik,ibvec) &
                           * C_MAT  (k_band,j_gauss,ik_ib   )

                   ENDDO

                   M_MAT(i_band,j_gauss,ik,ibvec) = SUM_CMPX

                ENDDO ! j_gauss
             ENDDO ! i_band

          ENDDO ! ibvec
       ENDDO ! ik

       !=== CONSTRUCT Z_TMP ===*

       SUM_REAL     = 0.0D0
       C_MAT(:,:,:) = 0.0D0

       zm_diag = 0.0d0
       do ibvec = 1, NB
          zm_diag = zm_diag + WEIGHT_b(ibvec)
       end do

       Do ik = 1 , NTK

          Z_TMP(:,:) = 0.0D0

          Do i_band = 1 , N_BAND(ik)
             Do j_band = 1 , N_BAND(ik)

                SUM_CMPX = 0.0D0

                Do ibvec   = 1 , NB
                   Do k_gauss = 1 , NGAUSS

                      SUM_CMPX &
                           = SUM_CMPX &
                           + WEIGHT_b(ibvec) &
                           *        M_MAT(i_band,k_gauss,ik,ibvec) &
                           * CONJG( M_MAT(j_band,k_gauss,ik,ibvec) )

                   ENDDO
                ENDDO

                Z_TMP(i_band,j_band) = SUM_CMPX

             ENDDO ! j_band
          ENDDO ! i_band

          !=== DAMPING FOR Z_TMP ===*

          IF( I_SCF == 0 ) THEN

             Z_MAT(:,:,ik) = Z_TMP(:,:)

          ELSE

             Z_MAT(:,:,ik)=DAMP*Z_TMP(:,:)+(1.0D0-DAMP)*Z_MAT(:,:,ik)

          ENDIF

          !=== inner energy window ===*
          Z_TMP = Z_MAT(:,:,ik)
          call wan_mat_clr_in_win(N_BAND_BOTTOM(ik), &
               in_win_band(:,ik),Z_TMP,NTB,zm_diag)

          !=== DIAGONALIZE Z_TMP ===*

          Z_EIG = 0.0d0
          call zheev('V','U',N_BAND(ik),Z_TMP,NTB,Z_EIG, &
               work_zheev,10*NTB,rwork_zheev,ILL)
          rwork_zheev(1:N_BAND(ik)) = Z_EIG
          !
          !     to descending order
          !
          do i_band = 1, N_BAND(ik)
             Z_EIG(i_band) = rwork_zheev(N_BAND(ik)-i_band+1)
             C_TMP(:,i_band) = Z_TMP(:,N_BAND(ik)-i_band+1)
          end do

          IF(ILL/=0) write(6,*) 'ERROR HAPPENED IN DIAG Z_MAT'

          !=== UPDATE C_MAT ===*

          Do j_gauss = 1 , NGAUSS

             C_MAT(:,j_gauss,ik) = C_TMP(:,j_gauss)

          ENDDO

          !=== EVALUATE SPILLAGE FUNCTIONAL ===*

          SUM_CMPX = 0.0d0
          do i_gauss = 1 , NGAUSS
             do j_band = 1, N_BAND(ik)
                do i_band = 1, N_BAND(ik)
                   SUM_CMPX = SUM_CMPX &
                        + conjg(C_TMP(i_band,i_gauss))* &
                        Z_MAT(i_band,j_band,ik)*C_TMP(j_band,i_gauss)
                end do
             end do
          end do
          SUM_REAL = SUM_REAL + dble(SUM_CMPX)

       ENDDO ! OF IK

       OMEGA_I_NEW=DBLE(NGAUSS)*WEIGHT-SUM_REAL/DBLE(NTK)

       DEL_OMEGA_I=DABS( OMEGA_I_NEW - OMEGA_I_OLD )

       OMEGA_I_OLD=OMEGA_I_NEW

       I_SCF = I_SCF + 1

       write(6,6004) 'I_SCF='      ,I_SCF, &
            'DEL_OMEGA_I=',DEL_OMEGA_I, &
            'OMEGA_I'     ,OMEGA_I_NEW
6004   FORMAT(A8,I5,A15,F20.10,A10,F20.10)

    ENDDO ! OF DO-WHILE ##### OMEGA_I MINIMIZATION #####

    !===============================================*
    !=== PSEUDO EIGENSTATE AND PSEUDO EIGENVALUE ===*
    !===============================================*

    PSEUDO_MAT(:,:,:) = 0.0D0
    PSEUDO_EIG(:,:)   = 0.0D0

    Do ik = 1 , NTK

       !=== CONSTRUCT PSEUDO-KS HAMILTONIAN ===*

       P_MAT_IN(:,:) = 0.0D0

       Do i_gauss = 1 , NGAUSS
          Do j_gauss = 1 , NGAUSS

             SUM_CMPX = 0.0D0

             Do k_band = 1 , N_BAND(ik)
                ik_map = mapkpt(:,nint(mgs%nk*SK0(1,ik)), &
                     nint(mgs%nk*SK0(2,ik)), &
                     nint(mgs%nk*SK0(3,ik)))

                SUM_CMPX &
                     = SUM_CMPX &
                     + CONJG( C_MAT(k_band,i_gauss,ik) ) &
                     *        E_EIG(k_band+N_BAND_BOTTOM(ik),ik_map(1)) &
                     *        C_MAT(k_band,j_gauss,ik)

             ENDDO

             P_MAT_IN(i_gauss,j_gauss) = SUM_CMPX

          ENDDO ! OF j_gauss
       ENDDO ! OF i_gauss

       !=== DIAGONALIZE PSEUDO-KS HAMILTONIAN ===*

       call zheev('V','U',NGAUSS,P_MAT_IN,NGAUSS,P_EIG, &
            work_zheev,10*NTB,rwork_zheev,ILL)
       P_MAT_OUT = P_MAT_IN

       IF(ILL/=0) write(6,*) 'ERROR HAPPENED IN DIAG S_MAT'

       PSEUDO_MAT(:,:,ik) = P_MAT_OUT(:,:)
       PSEUDO_EIG(:,  ik) = P_EIG    (:  )

    ENDDO ! OF ik

    !      Do ik = 1 , NTK
    !       write(6,6005) PSEUDO_EIG(:,ik)*27.21151D0
    !      ENDDO
    !6005  FORMAT(30F15.10)
    !
    !====================================================*
    !=== PREPARE INITIAL GUESS FOR OMEGA MINIMIZATION ===*
    !====================================================*

    Z_MAT(:,:,:)=0.0D0 ! INITIAL GUESS <---> CELL-PERIODIC

    Do ik = 1 , NTK

       !=== CONSTRUCT A_TMP ===*

       A_TMP(:,:) = 0.0D0

       Do i_band  = 1 , n_occ
          Do j_gauss = 1 , NGAUSS

             SUM_CMPX = 0.0D0

             Do k_band  = 1 , N_BAND(ik)
                Do l_gauss = 1 , NGAUSS

                   SUM_CMPX &
                        = SUM_CMPX &
                        + CONJG( PSEUDO_MAT( l_gauss , i_band  , ik ) ) &
                        * CONJG( C_MAT     ( k_band  , l_gauss , ik ) ) &
                        *        A_MAT     ( k_band  , j_gauss , ik )

                ENDDO
             ENDDO

             A_TMP(i_band,j_gauss) = SUM_CMPX

          ENDDO ! j_gauss
       ENDDO ! i_band

       !
       ! calculate the polar decomposition of A_TMP.
       ! A_MAT is over written by the unitary matrix.
       !
       call zgesvd('S','S',n_occ,NGAUSS,A_TMP,NTB,sgv, &
            umat,NTB,vhmat,NGAUSS, &
            work_zgesvd,lwork_zgesvd,rwork_zgesvd,ILL)
       if (ILL .ne. 0) then
          write(6,*) 'Error: zgesvd returns ', ILL
       end if
       call zgemm('N','N',n_occ,NGAUSS,NGAUSS,(1.0d0,0.0d0), &
            umat,NTB,vhmat,NGAUSS,(0.0d0,0.0d0),A_MAT(1,1,ik),NTB)

       !=== CONSTRUCT Z_MAT; INITIAL GUESS <---> CELL-PERIODIC ===*

       Do i_band  = 1 , N_BAND(ik)
          Do j_gauss = 1 , NGAUSS

             SUM_CMPX = 0.0D0

             Do k_gauss = 1 , NGAUSS
                Do l_band  = 1 , n_occ

                   SUM_CMPX &
                        = SUM_CMPX &
                        + C_MAT     ( i_band  , k_gauss , ik ) &
                        * PSEUDO_MAT( k_gauss , l_band  , ik ) &
                        * A_MAT     ( l_band  , j_gauss , ik )

                ENDDO
             ENDDO

             Z_MAT(i_band , j_gauss , ik ) = SUM_CMPX

          ENDDO ! j_gauss
       ENDDO ! i_band

    ENDDO ! OF IK ##### INITIAL GUESS FOR OMEGA MINIMIZATION #####

#ifdef DEBUG_WANNIER_METAL
    WRITE(6,*) ' '
    WRITE(6,*) '==========================='
    WRITE(6,*) 'CHECK OF UNITALITY OF Z_MAT'
    WRITE(6,*) '==========================='
    WRITE(6,*) ' '

    ILL = 0
    Do ik = 1 , NTK
       Do i_gauss = 1 , NGAUSS
          Do j_gauss = 1 , NGAUSS

             SUM_CMPX = 0.0D0

             Do i_band = 1 , N_BAND(ik)

                SUM_CMPX = SUM_CMPX &
                     + CONJG( Z_MAT(i_band,i_gauss,ik) ) &
                     *        Z_MAT(i_band,j_gauss,ik)

             ENDDO
             if (i_gauss.eq.j_gauss) then
                SUM_CMPX = SUM_CMPX - 1.0d0
             end if
             if (abs(SUM_CMPX) .gt. 1.0d4*epsilon(1.0d0)) then
                ILL = 1
                write(6,*) i_gauss,j_gauss,SUM_CMPX
             end if

          ENDDO ! OF j_gauss
       ENDDO ! OF i_gauss
    ENDDO ! OF ik
    if (ILL .ne. 0) then
       write(6,*) 'Error: Z_MAT is not unitary'
    end if
#endif

    !
    !=========================================*
    !=== MINIMIZATION OF SPREAD FUNCTIONAL ===*
    !=========================================*

    !===========================================*
    !=== CONSTRUCT INITIAL INTERSTATE MATRIX ===*
    !===========================================*

    WRITE(6,*) ' '
    WRITE(6,*) '====================================='
    WRITE(6,*) 'INTERSTATE MATRIX; LOG{det[M(k,k+b)]}'
    WRITE(6,*) '====================================='
    WRITE(6,*) ' '

    M_MAT  (:,:,:,:) = OVERLAP(:,:,:,:)
    OVERLAP(:,:,:,:) = 0.0D0
    Do ik    = 1 , NTK
       Do ibvec = 1 , NB
          ik_ib = KPT( ik , ibvec )
          Do i_band = 1 , n_occ ! NOTICE n_occ = NGAUSS !
             Do j_band = 1 , n_occ ! NOTICE n_occ = NGAUSS !
                SUM_CMPX = 0.0D0
                Do k_band = 1 , N_BAND(ik)
                   Do l_band = 1 , N_BAND(ik_ib)
                      SUM_CMPX &
                           = SUM_CMPX &
                           + CONJG( Z_MAT(k_band,i_band,ik      ) ) &
                           *        M_MAT(k_band,l_band,ik,ibvec) &
                           *        Z_MAT(l_band,j_band,ik_ib   )
                   ENDDO
                ENDDO
                OVERLAP(i_band,j_band,ik,ibvec) = SUM_CMPX
             ENDDO ! OF j_band
          ENDDO ! OF i_band

          !=== CALCULATE THE DETERMINANT. ===*
          call calc_det(n_occ,OVERLAP(:,:,ik,ibvec),NTB,DET,ILL)

          IF(ILL /= 0) write(6,*) 'ILL=',ILL
          write(6,6007) 'IK=',ik,'IK_IB=',ik_ib,'LOG=',LOG(DET)
6007      FORMAT(A8,I5,A8,I5,A8,2F15.10)

          detcnt(ik,ibvec) = DET
       ENDDO ! ibvec
       WRITE(6,*) ' '
    ENDDO ! ik

    call calc_wf_pos_sum(mgs%nk,NTK,NB,SK0,KPT,VEC_b,WEIGHT_b, &
         detcnt, wf_pos_sum)
    write(6,*)
    write(6,'(1x,a)') 'Sum of the Wannier center: (invariant, Cartesian) ='
    write(6,'(1x,3(1x,e18.11))') wf_pos_sum
    write(6,*)

    !=================================*
    !=== INITIAL SPREAD FUNCTIONAL ===*
    !=================================*

    ILS = 1
    CALL CALC_OMEGA(NTK,NTB,n_occ,NB,VEC_b,ILS, &
         WEIGHT_b,OVERLAP,WF_CHARGE,WF_POSITION, &
         OMEGA_I,OMEGA_OD,OMEGA_D)

    write(6,*) 'OMEGA_I =',OMEGA_I
    write(6,*) 'OMEGA_OD=',OMEGA_OD
    write(6,*) 'OMEGA_D =',OMEGA_D

    WRITE(6,*) ' '
    WRITE(6,*) '========================================='
    WRITE(6,*) '=== MINIMIZATION OF SPREAD FUNCTIONAL ==='
    WRITE(6,*) '========================================='
    WRITE(6,*) ' '

    M_MAT(:,:,:,:) = OVERLAP(:,:,:,:)
    DEL_SPREAD = 1.0D0
    SPREAD_OLD = 0.0D0
    I_STEP = 1
    SPREAD = OMEGA_I+OMEGA_OD+OMEGA_D

    STEP_LENGTH = sprd_stpl_fac / 4.0D0 / WEIGHT

    Do WHILE(DEL_SPREAD > sprd_thr)

       !--- CALCULATE GRADIENT ---*

       CALL CALC_GRADIENT(NTK,NTB,n_occ,NB, &
            WF_CHARGE,WEIGHT_b,OVERLAP, &
            tci,GRADIENT,TR_GRAD)

       !--- SET CONJUGATE GRADIENT DIRECTION ---*

       !      IF(I_STEP == 1 ) THEN
       DIRECTION(:,:,:)=GRADIENT(:,:,:)
       !      ELSE
       !       Do ik = 1 , NTK
       !         DIRECTION(:,:,ik)
       !     + = GRADIENT (:,:,ik)
       !     + + ( TR_GRAD(ik) / TR_GRAD_OLD(ik) )
       !     + * DIRECTION(:,:,ik)
       !       ENDDO
       !       ENDIF
       !      TR_GRAD_OLD(:) = TR_GRAD(:)
       !
       !--- SET DIRECTION^2 ---*

       DIRECTION2(:,:,:) = 0.0D0

       Do ik     = 1 , NTK
          Do i_band = 1 , n_occ
             Do j_band = 1 , n_occ
                SUM_CMPX=0.0D0
                Do k_band = 1 , n_occ
                   SUM_CMPX &
                        = SUM_CMPX &
                        + DIRECTION(i_band,k_band,ik) &
                        * DIRECTION(k_band,j_band,ik)
                ENDDO ! k_band
                DIRECTION2(i_band,j_band,ik) = SUM_CMPX
             ENDDO ! j_band
          ENDDO ! i_band
       ENDDO ! ik

       !=== ONE-DIMENSIONAL SEARCH ===*

       ILS = 0
       DELTA(1) = 0.0D0
       OMEGA(1) = SPREAD
       write(6,*) 'ILS=',ILS,DELTA(1),OMEGA(1)

       !
       ! adjust step length to satisfy OMEGA(2) < OMEGA(1)
       !
       ILS = 1
       OMEGA(2) = huge(1.0d0)
       do while( OMEGA(2) .gt. OMEGA(1) )
          DELTA(2) = STEP_LENGTH

          CALL UNITARY_GEN(NTK,NTB,n_occ,ILS, &
               DIRECTION,DIRECTION2, &
               STEP_LENGTH,UNITARY,U_ORG,U_OLD)

          OVERLAP(:,:,:,:) = M_MAT(:,:,:,:)
          CALL UPDATE_OVERLAP(NTK,NTB,n_occ,NB, &
               KPT,UNITARY,OVERLAP)

          CALL CALC_OMEGA(NTK,NTB,n_occ,NB,VEC_b,ILS, &
               WEIGHT_b,OVERLAP,WF_CHARGE,WF_POSITION, &
               OMEGA_I,OMEGA_OD,OMEGA_D)

          OMEGA(2) = OMEGA_I+OMEGA_D+OMEGA_OD
          if ( OMEGA(2) .gt. OMEGA(1) ) then
             STEP_LENGTH = 0.5d0*STEP_LENGTH
          end if
       end do
       write(6,*) 'STEP_LENGTH=',STEP_LENGTH
       write(6,*) 'ILS=',ILS,DELTA(2),OMEGA(2)

       !===
       DO WHILE( OMEGA(1) > OMEGA(2) )

          ILS = ILS + 1
          DELTA(3) = DELTA(2) + STEP_LENGTH

          CALL UNITARY_GEN(NTK,NTB,n_occ,ILS, &
               DIRECTION,DIRECTION2, &
               STEP_LENGTH,UNITARY,U_ORG,U_OLD)

          OVERLAP(:,:,:,:) = M_MAT(:,:,:,:)
          CALL UPDATE_OVERLAP(NTK,NTB,n_occ,NB, &
               KPT,UNITARY,OVERLAP)

          CALL CALC_OMEGA(NTK,NTB,n_occ,NB,VEC_b,ILS, &
               WEIGHT_b,OVERLAP,WF_CHARGE,WF_POSITION, &
               OMEGA_I,OMEGA_OD,OMEGA_D)
          OMEGA(3) = OMEGA_I+OMEGA_D+OMEGA_OD
          write(6,*) 'ILS=',ILS,DELTA(3),OMEGA(3)

          IF(OMEGA(2) < OMEGA(3)) EXIT

          DELTA(1) = DELTA(2)
          OMEGA(1) = OMEGA(2)

          DELTA(2) = DELTA(3)
          OMEGA(2) = OMEGA(3)

       ENDDO ! DO WHILE
       !===
       DELTA(4) = 0.5D0 * ( OMEGA(1)*(DELTA(2)**2-DELTA(3)**2) &
            + OMEGA(2)*(DELTA(3)**2-DELTA(1)**2) &
            + OMEGA(3)*(DELTA(1)**2-DELTA(2)**2) ) &
            / ( OMEGA(1)*(DELTA(2)   -DELTA(3)   ) &
            + OMEGA(2)*(DELTA(3)   -DELTA(1)   ) &
            + OMEGA(3)*(DELTA(1)   -DELTA(2)   ) )

       ILS = 1

       CALL UNITARY_GEN(NTK,NTB,n_occ,ILS, &
            DIRECTION,DIRECTION2, &
            DELTA(4),UNITARY,U_ORG,U_OLD)

       OVERLAP(:,:,:,:) = M_MAT(:,:,:,:)
       CALL UPDATE_OVERLAP(NTK,NTB,n_occ,NB, &
            KPT,UNITARY,OVERLAP)

       CALL CALC_OMEGA(NTK,NTB,n_occ,NB,VEC_b,ILS, &
            WEIGHT_b,OVERLAP,WF_CHARGE,WF_POSITION, &
            OMEGA_I,OMEGA_OD,OMEGA_D)

       OMEGA(4) = OMEGA_I+OMEGA_D+OMEGA_OD

       write(6,*) 'DELTA_OPT=',DELTA(4), &
            'OMEGA_OPT=',OMEGA(4)
       write(6,*) 'OMEGA_I =',OMEGA_I
       write(6,*) 'OMEGA_OD=',OMEGA_OD
       write(6,*) 'OMEGA_D =',OMEGA_D

       !=== M_MAT UNDATE

       M_MAT(:,:,:,:) = OVERLAP(:,:,:,:)

       !=== A_MAT UPDATE ===*

       CALL A_MAT_UPDATE(NTK,NTB,NGAUSS,n_occ, &
            UNITARY,A_MAT)

       !=== CONVERGENCE CHECK ===*

       SPREAD = OMEGA(4)
       DEL_SPREAD = ABS( SPREAD - SPREAD_OLD )
       SPREAD_OLD = OMEGA(4)

       write(6,*) 'I_STEP=',I_STEP,SPREAD,DEL_SPREAD
       write(6,*) ' '
       I_STEP = I_STEP + 1
       !
       ! update STEP_LENGTH
       !
       STEP_LENGTH = DELTA(4)*0.5d0

    ENDDO ! OF ##### SPREAD-MINIMIZATION LOOP #####

#ifdef DEBUG_WANNIER_METAL
    WRITE(6,*) ' '
    WRITE(6,*) '========================'
    WRITE(6,*) 'UNITALITY CHECK OF A_MAT'
    WRITE(6,*) '========================'
    WRITE(6,*) ' '

    ILL = 0
    Do ik = 1 , NTK
       Do i_band = 1 , n_occ
          Do j_band = 1 , n_occ

             SUM_CMPX = 0.0D0

             Do k_band = 1 , n_occ

                SUM_CMPX = SUM_CMPX &
                     + CONJG( A_MAT(k_band,i_band,ik) ) &
                     *        A_MAT(k_band,j_band,ik)

             ENDDO

             if (i_gauss.eq.j_gauss) then
                !
                ! diagonals cal be both +1 and -1.
                !
                SUM_CMPX = SUM_CMPX - sign(dble(SUM_CMPX),1.0d0)
             end if
             if (abs(SUM_CMPX) .gt. 1.0d4*epsilon(1.0d0)) then
                ILL = 1
                write(6,*) ik,i_gauss,j_gauss,SUM_CMPX
             end if
          ENDDO
       ENDDO
    ENDDO ! OF ik
    if (ILL .ne. 0) then
       write(6,*) 'Error: A_MAT is not unitary'
    end if
#endif

    !======================*
    !=== Wannier center ===*
    !======================*
    write(6,*)
    write(6,'(1x,a)') 'Sum of the Wannier center (optimized, Cartesian) ='
    optwf_pos_sum(1) = sum(WF_POSITION(1,:))
    optwf_pos_sum(2) = sum(WF_POSITION(2,:))
    optwf_pos_sum(3) = sum(WF_POSITION(3,:))
    write(6,'(1x,3(1x,e18.11))') optwf_pos_sum
    write(6,'(1x,a)') 'Diff of sum of the Wannier center (invariant - optimized, lattice)'
    write(6,'(1x,3(1x,f18.10))') matmul(wf_pos_sum - optwf_pos_sum,bb)/tpi
    write(6,'(1x,a)') 'Sum of Wannier center (invariant, aligned, Cartesian) ='
    write(6,'(1x,3(1x,e18.11))') wf_pos_sum &
         - matmul(aa,nint(matmul(wf_pos_sum - optwf_pos_sum,bb)/tpi))
    write(6,*)

    !====================================*
    !=== INTERPOLATED BAND DISPERSION ===*
    !====================================*

    !--- (1) SAMPLE k-POINTS FOR INTERPOLATED BAND DISPERSION ---*

    !--- (2) H_MAT(k) IN WANNIER BASIS. ---*

    H_MAT_K(:,:,:) = 0.0D0

    Do ik     = 1 , NTK
       Do i_band = 1 , n_occ
          Do j_band = 1 , n_occ

             SUM_CMPX = 0.0D0

             Do k_band = 1 , n_occ

                SUM_CMPX &
                     = SUM_CMPX &
                     + CONJG( A_MAT(k_band,i_band,ik) ) &
                     *   PSEUDO_EIG(k_band       ,ik) &
                     *        A_MAT(k_band,j_band,ik)

             ENDDO

             H_MAT_K(i_band,j_band,ik) = SUM_CMPX

          ENDDO ! OF j_band
       ENDDO ! OF i_band
    ENDDO ! OF ik

    !--- (3) H_MAT(R) IN WANNIER BASIS. ---*

    H_MAT_R(:,:,:,:,:) = 0.0D0

    do ia3 = 0 , nkb(3)-1
       do ia2 = 0 , nkb(2)-1
          do ia1 = 0 , nkb(1)-1
             do j_band = 1 , n_occ
                do i_band = 1 , n_occ
                   SUM_CMPX = 0.0D0
                   Do ik = 1 , NTK
                      PHASE = tpi*(SK0(1,ik)*DBLE(ia1) &
                           +SK0(2,ik)*DBLE(ia2) &
                           +SK0(3,ik)*DBLE(ia3) )
                      PHASE_FACTOR = EXP(-ci*PHASE)
                      SUM_CMPX &
                           = SUM_CMPX &
                           + H_MAT_K(i_band,j_band,ik) &
                           * PHASE_FACTOR
                   ENDDO
                   H_MAT_R(i_band,j_band,ia1,ia2,ia3) = SUM_CMPX / DBLE(NTK)
                end do
             end do
          end do
       end do
    end do
     
    A_MAT_O = 0.0d0 
    Do ik = 1 , NTK
       ik_map = mapkpt(:,nint(mgs%nk*SK0(1,ik)), &
            nint(mgs%nk*SK0(2,ik)), &
            nint(mgs%nk*SK0(3,ik)))
       !===
       A_TMP(:,:) = 0.0D0
       Do i_band = 1 , N_BAND(ik)
          Do j_band = 1 , n_occ
             SUM_CMPX = 0.0D0
             Do k_band = 1 , n_occ
                Do l_band = 1 , n_occ
                   SUM_CMPX &
                        = SUM_CMPX &
                        + C_MAT     ( i_band , k_band , ik ) &
                        * PSEUDO_MAT( k_band , l_band , ik ) &
                        * A_MAT     ( l_band , j_band , ik )
                ENDDO ! OF l_band
             ENDDO ! OF k_band
             A_MAT_O(i_band,j_band,ik) = SUM_CMPX
          ENDDO ! j_band
       ENDDO ! i_band
    END DO

    deallocate(umat,vhmat,work_zgesvd,sgv,rwork_zgesvd)
    deallocate(M_MAT)

    return
  END subroutine wannier_kazuma

  !================================================================*

  SUBROUTINE CALC_GRADIENT(NTK,NTB,n_occ,NB, &
       WF_CHARGE,WEIGHT_b,OVERLAP, &
       tci,GRADIENT,TR_GRAD)

    implicit none
    !=== INPUT ===*
    integer         :: NTK
    integer         :: NTB
    integer         :: n_occ
    integer         :: NB
    real(kind=8)    :: WF_CHARGE(n_occ,NTK,NB)
    real(kind=8)    :: WEIGHT_b(NB)
    complex(kind=8) :: OVERLAP(NTB,NTB,NTK,NB)
    complex(kind=8) :: tci
    !=== OUTPUT ===*
    complex(kind=8) :: GRADIENT(n_occ,n_occ,NTK)
    real(kind=8)    :: TR_GRAD(NTK)
    !=== LOCAL ===*
    complex(kind=8) :: WF_R(n_occ,n_occ,NB)
    complex(kind=8) :: WF_T(n_occ,n_occ,NB)
    integer         :: ik,ibvec,i_band,j_band
    complex(kind=8) :: SUM_CMPX
    real(kind=8)    :: SUM_REAL

    GRADIENT(:,:,:) = 0.0D0

    Do ik     = 1 , NTK
       Do i_band = 1 , n_occ
          Do j_band = 1 , n_occ
             Do ibvec  = 1 , NB

                WF_R          (i_band,j_band,   ibvec) &
                     =        OVERLAP(i_band,j_band,ik,ibvec) &
                     * CONJG( OVERLAP(j_band,j_band,ik,ibvec) ) ! modified 2007 6 26

                WF_T     (i_band,j_band,   ibvec) &
                     = OVERLAP  (i_band,j_band,ik,ibvec) &
                     / OVERLAP  (j_band,j_band,ik,ibvec) & ! modified 2007 6 26
                * WF_CHARGE(j_band       ,ik,ibvec)  ! modified 2007 6 26

             ENDDO ! ibvec
          ENDDO ! j_band
       ENDDO ! i_band

       Do i_band = 1 , n_occ
          Do j_band = 1 , n_occ

             SUM_CMPX=0.0D0

             Do ibvec  = 1 , NB

                SUM_CMPX &
                     = SUM_CMPX &
                     + 4.0D0 &
                     * WEIGHT_b(ibvec) &
                     * ( ( (      WF_R(i_band,j_band,ibvec) &
                     - CONJG( WF_R(j_band,i_band,ibvec) ) ) / 2.0D0 ) &
                     - ( (      WF_T(i_band,j_band,ibvec) &
                     + CONJG( WF_T(j_band,i_band,ibvec) ) ) / tci ) )

             ENDDO ! ibvec

             GRADIENT(i_band,j_band,ik) = SUM_CMPX / CMPLX(NTK)

          ENDDO ! j_band
       ENDDO ! i_band
    ENDDO ! ik

    !--- Tr(G^2) ---*

    TR_GRAD(:) = 0.0D0
    SUM_REAL   = 0.0D0

    Do ik = 1 , NTK

       SUM_CMPX=0.0D0

       Do i_band = 1 , n_occ
          Do j_band = 1 , n_occ

             SUM_CMPX &
                  = SUM_CMPX &
                  + GRADIENT(i_band,j_band,ik) &
                  * GRADIENT(j_band,i_band,ik)

          ENDDO ! j_band
       ENDDO ! i_band

       TR_GRAD(ik) = DBLE( SUM_CMPX )
       SUM_REAL = SUM_REAL + TR_GRAD(ik)

    ENDDO ! ik

    write(6,*) 'Tr(G^2)=',SUM_REAL

    RETURN
  END SUBROUTINE CALC_GRADIENT

  !================================================================*

  SUBROUTINE CALC_OMEGA(NTK,NTB,n_occ,NB,VEC_b,ILS, &
       WEIGHT_b,OVERLAP,WF_CHARGE,WF_POSITION, &
       OMEGA_I,OMEGA_OD,OMEGA_D)
    implicit none
    !=== INPUT ===*
    integer         :: NTK
    integer         :: NTB
    integer         :: n_occ
    integer         :: NB
    real(kind=8)    :: VEC_b(3,NB)
    integer         :: ILS
    real(kind=8)    :: WEIGHT_b(NB)
    complex(kind=8) :: OVERLAP(NTB,NTB,NTK,NB)
    !=== OUTPUT ===*
    real(kind=8)    :: WF_CHARGE(n_occ,NTK,NB)
    real(kind=8)    :: WF_POSITION(3,n_occ)
    real(kind=8)    :: OMEGA_I,OMEGA_OD,OMEGA_D
    !=== LOCAL ===*
    integer         :: ik,ibvec,i_band,j_band
    real(kind=8)    :: VEC_SUM_REAL(3)
    real(kind=8)    :: SUM_REAL,SUM_REAL1,SUM_REAL2
    complex(kind=8) :: LOG_OVERLAP
    !--- (1) WANNIER POSITION ---*

    WF_POSITION(:,:) = 0.0D0

    Do i_band = 1 , n_occ

       VEC_SUM_REAL(:) = 0.0D0

       Do ik     = 1 , NTK
          Do ibvec  = 1 , NB

             LOG_OVERLAP=LOG(OVERLAP(i_band,i_band,ik,ibvec))

             VEC_SUM_REAL(:) &
                  = VEC_SUM_REAL(:) &
                  + WEIGHT_b(ibvec) &
                  * VEC_b (:,ibvec) &
                  * AIMAG( LOG_OVERLAP )

          ENDDO ! ibvec
       ENDDO ! ik

       WF_POSITION(:,i_band)=-VEC_SUM_REAL(:)/DBLE(NTK)

    ENDDO ! i_band

    !--- (2) WANNIER CHARGE ---*

    WF_CHARGE(:,:,:) = 0.0D0

    Do ik     = 1 , NTK
       Do ibvec  = 1 , NB
          Do i_band = 1 , n_occ

             SUM_REAL = VEC_b(1,ibvec)*WF_POSITION(1,i_band) &
                  + VEC_b(2,ibvec)*WF_POSITION(2,i_band) &
                  + VEC_b(3,ibvec)*WF_POSITION(3,i_band)

             LOG_OVERLAP = LOG ( OVERLAP(i_band,i_band,ik,ibvec) )

             WF_CHARGE(i_band,ik,ibvec) &
                  = AIMAG( LOG_OVERLAP ) &
                  + SUM_REAL

          ENDDO ! i_band
       ENDDO ! ibvec
    ENDDO ! ik

    !--- (3-a) OMEGA_I ---*

    SUM_REAL1 = 0.0D0

    Do ik     = 1 , NTK
       Do ibvec  = 1 , NB

          SUM_REAL2 = 0.0D0

          Do i_band = 1 , n_occ
             Do j_band = 1 , n_occ

                SUM_REAL2 &
                     = SUM_REAL2 &
                     + ( ABS( OVERLAP(i_band,j_band,ik,ibvec) ) )**2

             ENDDO ! j_band
          ENDDO ! i_band

          SUM_REAL1 &
               = SUM_REAL1 &
               + WEIGHT_b(ibvec) &
               * ( DBLE(n_occ) - SUM_REAL2 )

       ENDDO ! ibvec
    ENDDO ! ik

    OMEGA_I = SUM_REAL1 / DBLE(NTK)

    !--- (3-b) OMEGA_OD ---*

    SUM_REAL1 = 0.0D0

    Do ik     = 1 , NTK
       Do ibvec  = 1 , NB

          SUM_REAL2 = 0.0D0

          Do i_band = 1 , n_occ
             Do j_band = 1 , n_occ

                IF(i_band /= j_band) THEN

                   SUM_REAL2 &
                        = SUM_REAL2 &
                        + ( ABS( OVERLAP(i_band,j_band,ik,ibvec) ) )**2

                ENDIF

             ENDDO ! j_band
          ENDDO ! i_band

          SUM_REAL1 &
               = SUM_REAL1 &
               + WEIGHT_b(ibvec) &
               * SUM_REAL2

       ENDDO ! ibvec
    ENDDO ! ik

    OMEGA_OD = SUM_REAL1 / DBLE(NTK)

    !--- (3-c) OMEGA_D ---*

    SUM_REAL1 = 0.0D0

    Do ik     = 1 , NTK
       Do ibvec  = 1 , NB

          SUM_REAL2 = 0.0D0

          Do i_band = 1 , n_occ

             SUM_REAL2 &
                  = SUM_REAL2 &
                  + ( WF_CHARGE(i_band,ik,ibvec) )**2

          ENDDO ! i_band

          SUM_REAL1 &
               = SUM_REAL1 &
               + WEIGHT_b(ibvec) &
               * SUM_REAL2

       ENDDO ! ibvec
    ENDDO ! ik

    OMEGA_D = SUM_REAL1 / DBLE(NTK)

    !--- (4) OMEGA_I,OD AND OMEGA_D OF EACH BAND ---*

    IF(ILS == 1) THEN

       Do i_band = 1 , n_occ
          SUM_REAL1 = 0.0D0
          SUM_REAL2 = 0.0D0
          Do ik     = 1 , NTK
             Do ibvec  = 1 , NB
                SUM_REAL1 &
                     = SUM_REAL1 &
                     + WEIGHT_b(ibvec) &
                     * ( 1.0D0 - (ABS( OVERLAP(i_band,i_band,ik,ibvec) ) )**2 )
                SUM_REAL2 &
                     = SUM_REAL2 &
                     + WEIGHT_b(ibvec) &
                     * ( WF_CHARGE(i_band,ik,ibvec) )**2
             ENDDO ! ibvec
          ENDDO ! ik
          SUM_REAL1 =  SUM_REAL1 / DBLE(NTK)
          SUM_REAL2 =  SUM_REAL2 / DBLE(NTK)
          write(6,*) i_band,SUM_REAL1,SUM_REAL2
       ENDDO ! i_band

    ENDIF

    RETURN
  END SUBROUTINE CALC_OMEGA

  !================================================================*
  SUBROUTINE UNITARY_GEN(NTK,NTB,n_occ,ILS, &
       DIRECTION,DIRECTION2, &
       STEP_LENGTH,UNITARY,U_ORG,U_OLD)
    implicit none
    !=== INPUT ===*
    integer         :: NTK
    integer         :: NTB
    integer         :: n_occ
    real(kind=8)    :: STEP_LENGTH
    integer         :: ILS
    complex(kind=8), INTENT(IN) :: DIRECTION (n_occ,n_occ,NTK)
    complex(kind=8), INTENT(IN) :: DIRECTION2(n_occ,n_occ,NTK)
    !=== INPUT & OUTPUT ===*
    complex(kind=8) :: UNITARY(n_occ,n_occ,NTK)
    complex(kind=8) :: U_ORG  (n_occ,n_occ,NTK)
    complex(kind=8) :: U_OLD  (n_occ,n_occ,NTK)
    !=== LOCAL ===*
    complex(kind=8) :: D_MAT(n_occ,n_occ)
    real(kind=8)    :: D_EIG(n_occ)
    complex(kind=8) :: R_MAT(n_occ,n_occ)
    real(kind=8)    :: D_INV(n_occ)
    real(kind=8)    :: D_COS(n_occ)
    real(kind=8)    :: D_SIN(n_occ)
    complex(kind=8) :: RC_MAT(n_occ,n_occ)
    complex(kind=8) :: RS_MAT(n_occ,n_occ)
    complex(kind=8) :: G_MAT(n_occ,n_occ)
    complex(kind=8) :: work_zheev(10*NTB)
    real(kind=8) :: rwork_zheev(3*NTB)
    integer         :: ILL
    integer         :: ik,i_band,j_band,k_band
    complex(kind=8) :: SUM_CMPX

    UNITARY(:,:,:) = 0.0D0
    IF(ILS == 1) THEN
       Do ik = 1 , NTK
          !--- STEP (1) DIAGONALIZATION ---*

          D_MAT(:,:) = DIRECTION2(:,:,ik)
          call zheev('V','U',n_occ,D_MAT,n_occ,D_EIG,work_zheev,10*NTB, &
               rwork_zheev,ILL)
          R_MAT = D_MAT

          IF(ILL /= 0) write(6,*) 'IK=',ik,'ILL=',ILL

          !--- STEP (2) CONSTRUCT DIAGONAL MATRICES ---*

          D_INV(:) = 0.0D0
          D_COS(:) = 0.0D0
          D_SIN(:) = 0.0D0

          Do i_band = 1 , n_occ

             IF( D_EIG(i_band) == 0.0D0 ) THEN

                D_EIG(i_band) = 1.0D-20

             ENDIF

             D_EIG(i_band) = DSQRT( ABS( D_EIG(i_band) ) )
             D_INV(i_band) = 1.0D0 / D_EIG(i_band)
             D_COS(i_band) = DCOS(   D_EIG(i_band) * STEP_LENGTH )
             D_SIN(i_band) = DSIN(   D_EIG(i_band) * STEP_LENGTH )

          ENDDO

          !--- STEP (3) CONSTRUCT RC & RS MATRIX ---*

          RC_MAT(:,:) = 0.0D0
          RS_MAT(:,:) = 0.0D0

          Do i_band = 1 , n_occ
             Do j_band = 1 , n_occ

                RC_MAT(i_band,j_band) &
                     = R_MAT (i_band,j_band) &
                     * D_COS (       j_band)

                RS_MAT(i_band,j_band) &
                     = R_MAT (i_band,j_band) &
                     * D_SIN (       j_band) &
                     * D_INV (       j_band)

             ENDDO
          ENDDO

          !--- STEP (4) CONSTRUCT RC MATRIX ---*

          G_MAT(:,:) = DIRECTION(:,:,ik)

          Do i_band = 1 , n_occ
             Do j_band = 1 , n_occ

                SUM_CMPX = 0.0D0

                Do k_band = 1 , n_occ

                   SUM_CMPX &
                        = SUM_CMPX &
                        + G_MAT (i_band,k_band) &
                        * RS_MAT(k_band,j_band)

                ENDDO

                RC_MAT(i_band,j_band) &
                     = RC_MAT(i_band,j_band) + SUM_CMPX

             ENDDO
          ENDDO

          !--- STEP (5) CONSTRUCT U MATRIX ---*

          Do i_band = 1 , n_occ
             Do j_band = 1 , n_occ

                SUM_CMPX = 0.0D0

                Do k_band = 1 , n_occ

                   SUM_CMPX &
                        = SUM_CMPX &
                        +        RC_MAT(i_band,k_band) &
                        * CONJG( R_MAT (j_band,k_band) )

                ENDDO

                UNITARY(i_band,j_band,ik) = SUM_CMPX

             ENDDO
          ENDDO

       ENDDO ! OF ik === UNITARY MATRIX GENERATION LOOP ===

       U_ORG(:,:,:) = UNITARY(:,:,:)
       U_OLD(:,:,:) = UNITARY(:,:,:)

    ELSE

       Do ik = 1 , NTK
          DO i_band = 1 , n_occ
             DO j_band = 1 , n_occ

                SUM_CMPX = 0.0D0

                Do k_band = 1 , n_occ

                   SUM_CMPX &
                        = SUM_CMPX &
                        + U_OLD(i_band,k_band,ik) &
                        * U_ORG(k_band,j_band,ik)

                ENDDO

                UNITARY(i_band,j_band,ik) = SUM_CMPX

             ENDDO ! OF j_band
          ENDDO ! OF i_band
       ENDDO ! OF ik

       U_OLD(:,:,:) = UNITARY(:,:,:)

    ENDIF

    RETURN
  END SUBROUTINE UNITARY_GEN

  !================================================================*
  SUBROUTINE UPDATE_OVERLAP(NTK,NTB,n_occ,NB, &
       KPT,UNITARY,OVERLAP)
    implicit none
    !=== INPUT ===*
    integer         :: NTK
    integer         :: NTB
    integer         :: n_occ
    integer         :: NB
    integer         :: KPT(NTK,NB)
    complex(kind=8) :: UNITARY(n_occ,n_occ,NTK)
    !=== INPUT & OUTPUT ===*
    complex(kind=8) :: OVERLAP(NTB,NTB,NTK,NB)
    !=== LOCAL ===*
    integer         :: ik,ik_ib,ibvec
    integer         :: i_band,j_band,k_band
    complex(kind=8) :: SUM_CMPX
    complex(kind=8) :: U_MAT(n_occ,n_occ)

    Do ik    = 1 , NTK
       Do ibvec = 1 , NB

          ik_ib = KPT(ik,ibvec)

          Do i_band = 1 , n_occ
             Do j_band = 1 , n_occ

                SUM_CMPX = 0.0D0

                Do k_band = 1 , n_occ

                   SUM_CMPX &
                        = SUM_CMPX &
                        + OVERLAP( i_band , k_band , ik , ibvec ) &
                        * UNITARY( k_band , j_band , ik_ib )

                ENDDO

                U_MAT(i_band,j_band) = SUM_CMPX

             ENDDO
          ENDDO

          Do i_band = 1 , n_occ
             Do j_band = 1 , n_occ

                SUM_CMPX = 0.0D0

                Do k_band = 1 , n_occ

                   SUM_CMPX &
                        = SUM_CMPX &
                        + CONJG( UNITARY( k_band , i_band , ik ) ) &
                        *        U_MAT  ( k_band , j_band )

                ENDDO

                OVERLAP(i_band,j_band,ik,ibvec) = SUM_CMPX

             ENDDO
          ENDDO

       ENDDO ! OF ibvec
    ENDDO ! OF ik

    RETURN
  END SUBROUTINE UPDATE_OVERLAP

  !================================================================*
  SUBROUTINE A_MAT_UPDATE(NTK,NTB,NGAUSS,n_occ, &
       UNITARY,A_MAT)
    implicit none
    !=== INPUT ===*
    integer         :: NTK
    integer         :: NTB
    integer         :: NGAUSS
    integer         :: n_occ
    complex(kind=8) :: UNITARY(n_occ,n_occ,NTK)
    !=== INPUT & OUTPUT ===*
    complex(kind=8) :: A_MAT(NTB,NGAUSS,NTK)
    !=== LOCAL ===*
    integer         :: ik
    integer         :: i_band,j_band,k_band
    complex(kind=8) :: SUM_CMPX
    complex(kind=8) :: A_TMP(NTB,NGAUSS)

    Do ik = 1 , NTK

       A_TMP(:,:) = 0.0D0
       A_TMP(:,:) = A_MAT(:,:,ik)

       Do i_band = 1 , n_occ
          Do j_band = 1 , n_occ

             SUM_CMPX = 0.0D0

             Do k_band = 1 , n_occ

                SUM_CMPX &
                     = SUM_CMPX &
                     + A_TMP  (i_band,k_band   ) &
                     * UNITARY(k_band,j_band,ik)

             ENDDO

             A_MAT(i_band,j_band,ik) = SUM_CMPX

          ENDDO ! OF j_band
       ENDDO ! OF i_band
    ENDDO ! OF ik

    RETURN
  END SUBROUTINE A_MAT_UPDATE

  !================================================================*
  SUBROUTINE CALC_OVERLAP(NTK,NTB,ncomp,NTG,nwx,nwy,nwz, &
       ik,ik_ib,map_bra,NG0_KET,KG0_KET,N_BAND, &
       C0_BRA,C0_KET,OVERLAP, &
       N_BAND_BOTTOM)
    implicit none
    !=== INPUT ===*
    integer :: NTK
    integer :: NTB
    integer :: ncomp,NTG
    integer :: nwx,nwy,nwz
    integer :: ik,ik_ib
    integer,intent(in) :: map_bra(-nwx:nwx,-nwy:nwy,-nwz:nwz)
    integer :: NG0_KET
    integer :: KG0_KET(3,NTG)
    integer :: N_BAND(NTK)
    complex(kind=8) :: C0_BRA(ncomp,0:NTG,NTB)
    complex(kind=8) :: C0_KET(ncomp,0:NTG,NTB)
    integer         :: N_BAND_BOTTOM(NTK)
    !=== OUTPUT ===*
    complex(kind=8) :: OVERLAP(NTB,NTB)
    !=== LOCAL ===*
    integer :: PULLBACK(NTG)
    integer :: ig,ig1,ig2,i_band
    complex(kind=8),allocatable :: C0_BRA_PB(:,:,:)

    allocate(C0_BRA_PB(ncomp,0:NTG,NTB))
    !=== GENERATE PULLBACK FUNCTION ===*
    PULLBACK(:) = 0
    Do ig = 1 , NG0_KET
       PULLBACK(ig) &
            = map_bra( KG0_KET(1,ig) &
            , KG0_KET(2,ig) &
            , KG0_KET(3,ig) )
    Enddo
    !=== INTERSTATE MATRIX ===*
    if (ncomp.eq.1) then
       do i_band=1,N_BAND(ik)
          C0_BRA_PB(1,0,i_band+N_BAND_BOTTOM(ik)) = 0.0d0
          do ig2=1,NG0_KET
             ig1 = PULLBACK(ig2)
             C0_BRA_PB(1,ig2,i_band) = C0_BRA(1,ig1,i_band+N_BAND_BOTTOM(ik))
          end do
       end do
    else if (ncomp.eq.2) then
       do i_band=1,N_BAND(ik)
          C0_BRA_PB(1,0,i_band+N_BAND_BOTTOM(ik)) = 0.0d0
          C0_BRA_PB(2,0,i_band+N_BAND_BOTTOM(ik)) = 0.0d0
          do ig2=1,NG0_KET
             ig1 = PULLBACK(ig2)
             C0_BRA_PB(1,ig2,i_band) = C0_BRA(1,ig1,i_band+N_BAND_BOTTOM(ik))
             C0_BRA_PB(2,ig2,i_band) = C0_BRA(2,ig1,i_band+N_BAND_BOTTOM(ik))
          end do
       end do
    else
       write(6,*) 'calc_overlap: invalid ncomp =', ncomp
       stop
    end if
    call zgemm('C','N',N_BAND(ik),N_BAND(ik_ib),ncomp*NG0_KET, &
         (1.0d0,0.0d0),C0_BRA_PB(1,1,1),ncomp*(NTG+1), &
         C0_KET(1,1,N_BAND_BOTTOM(ik_ib)+1),ncomp*(NTG+1), &
         (0.0d0,0.0d0),OVERLAP,NTB)
    return
    deallocate(C0_BRA_PB)
    RETURN
  END SUBROUTINE CALC_OVERLAP

  subroutine calc_det(n,a,lda,det,istat)
    implicit none
    integer,intent(in):: n,lda
    complex(kind=8),intent(in):: a(lda,n)
    complex(kind=8):: det
    integer:: ipiv(n)
    integer:: istat

    integer:: i
    complex(kind=8):: atmp(lda,n)

    atmp = a

    call zgetrf(n,n,atmp,lda,ipiv,istat)
    if (istat.ne.0) then
       det = (0.0d0, 0.0d0)
       return
    end if
    det = (1.0d0, 0.0d0)
    istat = 0
    do i = 1, n
       if (abs(a(i,i)).le.epsilon(1.0d0)) then
          istat = i
          det = (0.0d0, 0.0d0)
          exit
       end if
       if (ipiv(i).eq.i) then
          det = det*atmp(i,i)
       else
          det = -det*atmp(i,i)
       end if
    end do
    return
  end subroutine calc_det

  !
  ! calculate electronic part of the polarization by using the Berry phase
  ! formula. The scheme of the separation of determinants
  ! by the log() is important. This relates to the branch cut of the log().
  !
  subroutine calc_wf_pos_sum(nk,NTK,NB,SK0,KPT,VEC_b,WEIGHT_b, &
       detcnt,wf_pos_sum)
    implicit none
    integer:: nk, NTK, NB
    real(kind=8):: SK0(3,NTK)
    integer:: KPT(NTK,NB)
    real(kind=8):: VEC_b(3,NB),WEIGHT_b(NB)
    complex(kind=8):: detcnt(NTK,NB)
    real(kind=8):: wf_pos_sum(3)

    integer:: ibvec,ik,jk,ik0,ik_ib
    complex(kind=8):: det
    integer:: markdet(NTK)

    do ibvec=1, NB
       markdet = 0
       jk = 0
       do while (jk .lt. NTK)
          ik = 1
          do while (markdet(ik).eq.1)
             ik = ik + 1
          end do
          ik0 = ik
          det = 1.0d0
          do while (.true.)
             jk = jk + 1
             markdet(ik) = 1
             det = det*detcnt(ik,ibvec)
             ik_ib = KPT( ik , ibvec )
             if( maxval(modulo( &
                  nint(nk*(SK0(:,ik0)-SK0(:,ik_ib))), nk)) .eq. 0 ) then
                exit
             else
                ik = ik_ib
             end if
          end do
          wf_pos_sum = wf_pos_sum &
               + WEIGHT_b(ibvec)*VEC_b(:,ibvec)*aimag(log(DET))
       end do
       if ( sum(markdet(:)) .ne. NTK ) then
          write(6,*) 'Error: invalid markdet for ibvec = ', ibvec
          stop
       end if
    end do
    wf_pos_sum = - wf_pos_sum/NTK
  end subroutine calc_wf_pos_sum

  !======================================*
  !=== WANNIER FUNCTION IN REAL SPACE ===*
  !======================================*

  subroutine wan_recp2real(ncomp,VOLUME,aa,mgs &
       ,ng,rg,nnp,pg,srg,NTK,NTG,NTB &
       ,I_WANNIER,n_occ,mapkpt,SK0)
    use types
    use meshgrid
    use fft_3d
    use util
    use iounit
    implicit none

    integer:: ncomp,ng,nnp

    type (meshgrid_struct):: mgs

    real(kind=8):: VOLUME,aa(3,3)
    integer:: rg(3,3,ng),pg(3,ng),srg(ng)

    integer:: NTK,NTG
    integer:: NTB
    integer:: I_WANNIER
    integer:: n_occ
    integer:: mapkpt(5,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk)

    integer:: KG0_TMP(3)
    integer:: ik_map(5)
    integer:: ik,ib,ig,jrg

    real(kind=8),parameter :: pi=dacos(-1.0d0)
    real(kind=8),parameter :: tpi=2.0d0*pi
    complex(kind=8),parameter :: ci=(0.0D0,1.0D0)

    !=== ARRAYS FOR SORTING SAMPLE K-POINTS ===*
    real(kind=8) :: SK0(3,NTK)
    !=== ARRAYS FOR WAVE FUNCTIONS ===*
    complex(kind=8) :: C0_TMP_1(ncomp,NTG,NTB)
    !=== ARRAYS FOR 3D FFT GRID ===*
    integer :: Lx,Ly,Lz
    type (fft3_struct):: fs
    !=== ARRAYS FOR K+G VECTORS ===*
    integer,pointer :: KG0(:,:,:)
    integer,pointer :: NG0(:)
    integer :: igb1,igb2,igb3
    !=== ARRAYS FOR VISUALIZING WANNIER FUNCTION ===*
    complex(kind=8), ALLOCATABLE :: WANNIER_REALSPACE(:,:,:,:)
    real(kind=8), ALLOCATABLE :: PSI_FFT3DB(:,:,:,:)
    real(kind=8), ALLOCATABLE :: WORK_FFT3DB(:)
    integer :: L_FFT3DB(3)
    integer :: N_FFT3DB(3)
    integer :: L123
    integer :: ir1,ir2,ir3
    integer :: jr1,jr2,jr3
    integer :: na_grid,nb_grid,nc_grid,i1,i2,i3
    !=== ARRAYS FOR BASIC VARIABLES
    real(kind=8) :: PHASE
    complex(kind=8) :: PHASE_FACTOR
    !== for check
    integer:: nk_in,ntk_in,n_occ_in
    integer:: ik_map_in(5),ng0_in
    real(kind=8):: sk0_in(3)
    integer:: ic

    write(6,*) 'translating wannier function '// &
         'from recpirocal space to real space'
    write(6,*) 'target_number = ', I_WANNIER

    NG0 => mgs%nkm
    KG0 => mgs%kg
    Lx = mgs%nwx + 1
    Ly = mgs%nwy + 1
    Lz = mgs%nwz + 1
    write(6,*) ' '
    write(6,'(A20,3I10)') 'FFT GRID L(x,y,z)=',Lx,Ly,Lz

    !=== DEFINE FFT GRID ===*

    N_FFT3DB(1) = algn235(Lx,1)
    N_FFT3DB(2) = algn235(Ly,1)
    N_FFT3DB(3) = algn235(Lz,1)
    L_FFT3DB(1) = FFT_PAD_NX(N_FFT3DB(1))
    L_FFT3DB(2) = FFT_PAD_NY(N_FFT3DB(2))
    L_FFT3DB(3) = FFT_PAD_NZ(N_FFT3DB(3))

    L123 = L_FFT3DB(1) * L_FFT3DB(2) * L_FFT3DB(3)

    !=== ALLOCATE ARRAYS DEFINITION ===*

    ALLOCATE(FFT_A3D_DEFSZ(PSI_FFT3DB,L_FFT3DB(1),L_FFT3DB(2),L_FFT3DB(3)))

    ALLOCATE( WORK_FFT3DB( 2*L123 ) )

    ALLOCATE(WANNIER_REALSPACE((-1)*N_FFT3DB(1):(1)*N_FFT3DB(1)-1 &
         ,(-1)*N_FFT3DB(2):(1)*N_FFT3DB(2)-1 &
         ,(-1)*N_FFT3DB(3):(1)*N_FFT3DB(3)-1,ncomp))

    call fft3_init(N_FFT3DB(1),N_FFT3DB(2),N_FFT3DB(3), &
         L_FFT3DB(1),L_FFT3DB(2),L_FFT3DB(3),fs)

    write(6,*) 'Lx,Ly,Lz=',Lx,Ly,Lz
    write(6,*) 'N_FFT3DB=',N_FFT3DB(:)
    write(6,*) 'L_FFT3DB=',L_FFT3DB(:)

    !=== CALCULATE WANNIER FUNCTION IN REALSPACE ===*

    WANNIER_REALSPACE(:,:,:,:) = 0.0D0
    call xrewind(NFRECIPWAN)
    read(NFRECIPWAN) nk_in,ntk_in,n_occ_in
    if (ntk_in.ne.NTK .or. n_occ_in.ne.n_occ) then
       write(6,*) 'inconsistent input wannier function ' // &
            'for recprocal->real conversion'
       write(6,'(1x,a,4(1x,i5))') 'ntk_in, ntk, n_occ_in, n_occ = ', &
            ntk_in, ntk, n_occ_in, n_occ
       stop
    end if
    Do ik     = 1 , NTK
       ik_map = mapkpt(:,nint(mgs%nk*SK0(1,ik)), &
            nint(mgs%nk*SK0(2,ik)), &
            nint(mgs%nk*SK0(3,ik)))
       !===
       C0_TMP_1(:,:,:) = 0.0D0
       read(NFRECIPWAN) sk0_in,ik_map_in,ng0_in
       if (maxval(abs(SK0(:,ik)-sk0_in)).gt.100.0d0*epsilon(1.0d0).or. &
            maxval(ik_map-ik_map_in).gt.0 .or. NG0(ik_map(1)).ne.ng0_in) then
          write(6,*) 'inconsistent input wannier functions ' // &
               'for recprocal->real conversion'
          stop
       end if
       do ib=1,n_occ
          READ(NFRECIPWAN) ((C0_TMP_1(ic,IG,IB),ic=1,ncomp),IG=1,NG0(ik_map(1)))
       end do

       do ic=1,ncomp
          !=== SET `PSI_FFT3DB' ===*
          
          PSI_FFT3DB = 0.0D0
          DO ig=1 , NG0(ik_map(1))
             jrg = ik_map(2)
             KG0_TMP(:) = matmul(rg(:,:,jrg),KG0(:,ig,ik_map(1)))
             KG0_TMP(:) = srg(jrg)*KG0_TMP(:)
             KG0_TMP(1:3) = KG0_TMP(1:3) - ik_map(3:5)
             igb1 = KG0_TMP(1)
             igb2 = KG0_TMP(2)
             igb3 = KG0_TMP(3)

             igb1 = MOD( N_FFT3DB(1)+igb1 , N_FFT3DB(1) ) + 1
             igb2 = MOD( N_FFT3DB(2)+igb2 , N_FFT3DB(2) ) + 1
             igb3 = MOD( N_FFT3DB(3)+igb3 , N_FFT3DB(3) ) + 1

             FFT_A3D_RL(PSI_FFT3DB,igb1,igb2,igb3) = &
                  DBLE(C0_TMP_1(ic,ig,I_WANNIER))
             FFT_A3D_IM(PSI_FFT3DB,igb1,igb2,igb3) = &
                  AIMAG(C0_TMP_1(ic,ig,I_WANNIER))
          ENDDO

          !=== FFT FOR `PSI_FFT3DB' ===*

          WORK_FFT3DB = 0.0D0
          call fft3_bw(fs, PSI_FFT3DB, WORK_FFT3DB)

          Do ir1 = (-1)*N_FFT3DB(1),(1)*N_FFT3DB(1)-1
             Do ir2 = (-1)*N_FFT3DB(2),(1)*N_FFT3DB(2)-1
                Do ir3 = (-1)*N_FFT3DB(3),(1)*N_FFT3DB(3)-1

                   !=== CALCULATE PHASE_FACTOR

                   PHASE = tpi*(DBLE(ir1)*SK0(1,ik)/DBLE(N_FFT3DB(1)) &
                        +DBLE(ir2)*SK0(2,ik)/DBLE(N_FFT3DB(2)) &
                        +DBLE(ir3)*SK0(3,ik)/DBLE(N_FFT3DB(3)))
                   PHASE_FACTOR = EXP(ci*PHASE)

                   !=== (ir1,ir2,ir3) ===> (jr1,jr2,jr3) ===*

                   jr1 = modulo( ir1, N_FFT3DB(1) ) + 1
                   jr2 = modulo( ir2, N_FFT3DB(2) ) + 1
                   jr3 = modulo( ir3, N_FFT3DB(3) ) + 1

                   !=== CALCULATE MAGNITUDE OF WFs OF GRID (ir1,ir2,ir3) ===*

                   WANNIER_REALSPACE(ir1,ir2,ir3,ic) &
                        = WANNIER_REALSPACE(ir1,ir2,ir3,ic) &
                        + dcmplx(FFT_A3D_RL(PSI_FFT3DB,jr1,jr2,jr3), &
                        FFT_A3D_IM(PSI_FFT3DB,jr1,jr2,jr3)) &
                        * PHASE_FACTOR

                ENDDO ! OF ir3
             ENDDO ! OF ir2
          ENDDO ! OF ir1
       end do

    ENDDO ! OF ik

    WANNIER_REALSPACE(:,:,:,:) &
         = WANNIER_REALSPACE(:,:,:,:) &
         /DBLE(NTK)/DSQRT(VOLUME)

    !=== WRITE MAGNITUDE OF WFs OF GRID (ir1,ir2,ir3) ===*

    na_grid=N_FFT3DB(1)
    nb_grid=N_FFT3DB(2)
    nc_grid=N_FFT3DB(3)
    call xrewind(NFREALWAN)
    write(NFREALWAN) 'TAPP:rho'
    write(NFREALWAN) 2*aa,ncomp
    write(NFREALWAN) 2*na_grid,2*nb_grid,2*nc_grid
    write(NFREALWAN) ((((dble(WANNIER_REALSPACE(i1,i2,i3,ic)) &
         ,i1=-na_grid,na_grid-1) &
         ,i2=-nb_grid,nb_grid-1) &
         ,i3=-nc_grid,nc_grid-1),ic=1,ncomp)
    close(NFREALWAN)

    deallocate(PSI_FFT3DB,WORK_FFT3DB,WANNIER_REALSPACE)

    return
  END subroutine wan_recp2real

  !=======================================================*
  !=== WRITE WANNIER FUNCTION DATA FOR CONSTRAINED LDA ===*
  !=======================================================*

  !=== TRANSFORM CELL-PERIODIC TO ROTATED FUNCTION

  subroutine wan_write_recpwan(ncomp,NTG,NTK, &
       NB_start,NB_end,NTB,SK0,n_occ,mgs,mapkpt,dp, &
       N_BAND,N_BAND_BOTTOM,A_MAT_O,ng,rg,nnp,pg,srg,aa)
    use meshgrid
    use iounit
    use types
    use symutil
    implicit none

    integer:: ncomp,NTG,NTK,NB_start,NB_end,NTB,n_occ
    real(kind=8) :: SK0(3,NTK)
    type (meshgrid_struct):: mgs
    integer:: mapkpt(5,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk)
    type (wfn_for_k):: dp(mgs%nks)
    integer:: N_BAND(NTK), N_BAND_BOTTOM(NTK)
    complex(kind=8),intent(in):: A_MAT_O(NTB,n_occ,NTK)
    integer:: ng,nnp
    integer:: rg(3,3,ng),pg(3,ng),srg(ng)
    real(kind=8):: aa(3,3)

    integer:: ik,ik_map(5),j_band,k_band
    integer:: ig,ib,jb,jrg,iph,ic
    complex(kind=8):: pf(NTG)
    integer:: KG0_TMP(3,NTG)
    complex(kind=8) :: C0_TMP_1(ncomp,NTG,NTB)
    complex(kind=8) :: C0_TMP_2(ncomp,NTG,NTB)
    complex(kind=8) :: SUM_CMPX(ncomp)

    integer,pointer :: KG0(:,:,:)
    integer,pointer :: NG0(:)
    
    real(kind=8),parameter :: pi=dacos(-1.0d0)
    real(kind=8),parameter :: tpi=2.0d0*pi
    complex(kind=8),parameter :: ci=(0.0D0,1.0D0)

    complex(kind=8):: tsu2(2,2),tspr(2)

    NG0 => mgs%nkm
    KG0 => mgs%kg

    call xrewind(NFRECIPWAN)
    write(NFRECIPWAN) mgs%nk, NTK, n_occ
    !---
    Do ik = 1 , NTK
       ik_map = mapkpt(:,nint(mgs%nk*SK0(1,ik)), &
            nint(mgs%nk*SK0(2,ik)), &
            nint(mgs%nk*SK0(3,ik)))
       !===
       jrg = ik_map(2)
       call rg2tsu2(rg(:,:,jrg),aa,tsu2)
       do ig=1, NG0(ik_map(1))
          KG0_TMP(:,ig) = matmul(rg(:,:,jrg),KG0(:,ig,ik_map(1)))
          KG0_TMP(:,ig) = srg(jrg)*KG0_TMP(:,ig)
          iph = pg(1,jrg)*KG0_TMP(1,ig)+pg(2,jrg)*KG0_TMP(2,ig) &
               +pg(3,jrg)*KG0_TMP(3,ig)
          pf(ig) = exp(-ci*tpi*dble(iph)/nnp)
          KG0_TMP(1:3,ig) = KG0_TMP(1:3,ig) - ik_map(3:5)
       end do
       if (ncomp.eq.1) then
          if (srg(jrg).eq.1) then
             do ib = NB_start, NB_end
                jb = ib - NB_start + 1
                do ig=1, NG0(ik_map(1))
                   C0_TMP_1(1,ig,jb) = dp(ik_map(1))%wf(1,ig,ib)*pf(ig)
                end do
             end do
          else
             do ib = NB_start, NB_end
                jb = ib - NB_start + 1
                do ig=1, NG0(ik_map(1))
                   C0_TMP_1(1,ig,jb) = conjg(dp(ik_map(1))%wf(1,ig,ib))*pf(ig)
                end do
             end do
          end if
       else
          if (srg(jrg).eq.1) then
             do ib = NB_start, NB_end
                jb = ib - NB_start + 1
                do ig=1, NG0(ik_map(1))
                   tspr = matmul(tsu2,dp(ik_map(1))%wf(1:2,ig,ib))*pf(ig)
                   C0_TMP_1(1:2,ig,jb) = tspr
                end do
             end do
          else
             do ib = NB_start, NB_end
                jb = ib - NB_start + 1
                do ig=1, NG0(ik_map(1))
                   tspr(1) = -conjg(dp(ik_map(1))%wf(2,ig,ib))
                   tspr(2) = conjg(dp(ik_map(1))%wf(1,ig,ib))
                   C0_TMP_1(1:2,ig,jb) = matmul(tsu2,tspr)*pf(ig)
                end do
             end do
          end if
       end if
       !===
       C0_TMP_2(:,:,:) = 0.0D0
       Do ig     = 1 , NG0(ik_map(1))
          Do j_band = 1 , n_occ
             SUM_CMPX = 0.0D0
             Do k_band = 1 , N_BAND(ik)
                SUM_CMPX &
                     = SUM_CMPX &
                     + C0_TMP_1( :, ig, k_band+N_BAND_BOTTOM(ik) ) &
                     * A_MAT_O(k_band, j_band, ik)
             ENDDO ! OF k_band
             C0_TMP_2(:, ig,j_band) = SUM_CMPX
          ENDDO ! OF j_band
       ENDDO ! OF ig
       !===
       write(NFRECIPWAN) SK0(:,ik), ik_map, NG0(ik_map(1))
       do ib=1,n_occ
          WRITE(NFRECIPWAN) ((C0_TMP_2(ic,IG,IB),ic=1,ncomp), &
               IG=1,NG0(ik_map(1)))
       end do
       !===
    ENDDO ! OF ik
    close(NFRECIPWAN)
    return
  end subroutine wan_write_recpwan

  !=======================================*
  !=== CALCULATE THE NUMBER OF BAND AT ===*
  !=== K-POINT BELOW THE ENERGY EPS_F  ===*
  !=======================================*

  subroutine wan_target_band(mgs,NTK,NTB,E_EIG,EPS_LOWER,EPS_UPPER, &
       mapkpt,SK0,N_BAND,N_BAND_BOTTOM)
    use meshgrid
    implicit none
    type (meshgrid_struct):: mgs
    integer,intent(in):: NTK,NTB
    real(kind=8):: E_EIG(NTB,mgs%nks)
    real(kind=8):: EPS_UPPER, EPS_LOWER
    integer:: mapkpt(5,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk)
    real(kind=8) :: SK0(3,NTK)
    integer:: N_BAND(NTK),N_BAND_BOTTOM(NTK)

    integer:: ik,ik_map(5),i_band
    integer:: SUM_INT

    N_BAND       (:) = 0
    N_BAND_BOTTOM(:) = 0
    Do ik=1,NTK
       ik_map = mapkpt(:,nint(mgs%nk*SK0(1,ik)), &
            nint(mgs%nk*SK0(2,ik)), &
            nint(mgs%nk*SK0(3,ik)))
       SUM_INT=0
       Do i_band=1,NTB
          IF( EPS_LOWER <= E_EIG(i_band,ik_map(1)) .AND. &
               EPS_UPPER >= E_EIG(i_band,ik_map(1))) THEN
             SUM_INT=SUM_INT+1
             IF(SUM_INT==1) THEN
                N_BAND_BOTTOM(ik)=i_band-1
             ENDIF
          ENDIF
       ENDDO ! OF i_band
       N_BAND(ik)=SUM_INT
    ENDDO ! OF ik

    ! OPEN(149,W,FILE='Ns_and_Nb')
    ! rewind(149)
    WRITE(6,*) ' '
    Do ik=1,NTK
       write(6,*) ik,N_BAND_BOTTOM(ik),N_BAND(ik)
       ! write(149,*) N_BAND_BOTTOM(ik)+NB_start-1,N_BAND(ik)
    Enddo
    WRITE(6,*) ' '

  end subroutine wan_target_band

  subroutine wan_set_inner_win(mgs,nkmesh,NTB,E_EIG, &
       eng_in_low,eng_in_upper,mapkpt,sklist,in_win_band)
    use meshgrid
    implicit none
    type (meshgrid_struct):: mgs
    integer,intent(in):: nkmesh,NTB
    real(kind=8):: E_EIG(NTB,mgs%nks)
    real(kind=8):: eng_in_low,eng_in_upper
    integer:: mapkpt(5,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk)
    real(kind=8):: sklist(3,nkmesh)
    integer:: in_win_band(2,nkmesh)

    integer:: ik,ik_map(5),ib,max_band

    do ik=1,nkmesh
       ik_map = mapkpt(:,nint(mgs%nk*sklist(1,ik)), &
            nint(mgs%nk*sklist(2,ik)), &
            nint(mgs%nk*sklist(3,ik)))
       in_win_band(1,ik) = 1
       max_band = 0
       do ib=1,NTB
          if (eng_in_low .le. E_EIG(ib,ik_map(1)) .and. &
               E_EIG(ib,ik_map(1)) .le. eng_in_upper) then
             if(max_band .eq. 0) then
                in_win_band(1,ik) = ib
             end if
             max_band = ib
          end if
       end do
       in_win_band(2,ik) = max_band
    end do

    write(6,*)
    write(6,*) '====================================='
    write(6,*) 'inner energy window'
    write(6,*) '====================================='
    write(6,*)
    do ik=1,nkmesh
       write(6,*) ik,in_win_band(1,ik),in_win_band(2,ik)
    end do
    return
  end subroutine wan_set_inner_win

  subroutine wan_mat_clr_in_win(N_BAND_BOTTOM, &
       in_win_band, zmat, NTB, diag)
    implicit none
    integer,intent(in):: N_BAND_BOTTOM, in_win_band(2), NTB
    real(kind=8),intent(in):: diag
    complex(kind=8):: zmat(NTB,NTB)
    integer:: ib, ib_start, ib_end

    ib_start = in_win_band(1) - N_BAND_BOTTOM
    ib_end = in_win_band(2) - N_BAND_BOTTOM
    do ib = ib_start, ib_end
       zmat(:,ib) = 0.0d0
       zmat(ib,:) = 0.0d0
       zmat(ib,ib) = diag
    end do
    return
  end subroutine wan_mat_clr_in_win

end module wannier_metal
