!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
module pw91_xcgga
  private
  public pw91_exchcor,pw91_exchcorsp
contains
  function Gcor(np,rs,a,alph1,bet1,bet2,bet3,bet4,p)
    implicit none
    integer:: np
    real(kind=8),pointer,dimension(:):: Gcor
    real(kind=8):: rs(np),a,alph1,bet1,bet2,bet3,bet4,p
    integer:: i
    !
    allocate(Gcor(np))
!$omp parallel do
    do i=1,np
       Gcor(i) = -2.0d0*a*(1.0d0 + alph1*rs(i))*Log(1.0d0 + 1.0d0/ &
            (2.0d0*a*(bet1*Sqrt(rs(i)) + bet2*rs(i) &
            + bet3*rs(i)**(3.0d0/2.0d0) + bet4*rs(i)**(1.0d0 + p))))
    end do
    !
    return
  end function Gcor
  !
  function DGcor(np,rs,a,alph1,bet1,bet2,bet3,bet4,p)
    implicit none
    integer:: np
    real(kind=8),pointer,dimension(:):: DGcor
    real(kind=8):: rs(np),a,alph1,bet1,bet2,bet3,bet4,p
    !
    real(kind=8):: p1,p2
    integer:: i
    !
    allocate(DGcor(np))
!$omp parallel do private(p1,p2)
    do i=1,np
       p1 = bet1*Sqrt(rs(i)) + bet2*rs(i) + bet3*rs(i)**(3.0d0/2.0d0) + &
            bet4*rs(i)**(1.0d0 + p)
       p2 = bet2 + bet1/(2.0d0*Sqrt(rs(i))) + 3.0d0*bet3*Sqrt(rs(i))/2.0d0 + &
            bet4*(1.0d0 + p)*rs(i)**p
       DGcor(i) = (1.0d0 + alph1*rs(i))*p2 &
            /(p1**2*(1.0d0 + 1.0d0/(2.0d0*a*p1))) - &
            2.0d0*a*alph1*Log(1.0d0 + 1.0d0/(2.0d0*a*p1))
    end do

    return
  end function DGcor
  !
  subroutine Ecpw91(np,rs,zeta,ec,ecrs,eczet)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: rs, zeta, ec, ecrs, eczet
    !
    real(kind=8):: fzz
    ! parameter (fzz = 8.0d0/(9.0d0*(-2.0d0+2.0d0**(4.0d0/3.0d0))))
    parameter (fzz = 1.70992093416137d0)
    real(kind=8),pointer,dimension(:):: eu,eurs,ep,eprs,alf,alfrs
    real(kind=8),allocatable:: f(:),fz(:)
    integer:: i
    !
    allocate(f(np),fz(np))
!$omp parallel do
    do i=1,np
       f(i) = (-2.0d0 + (1.0d0-zeta(i))**(4.0d0/3.0d0) &
            + (1.0d0+zeta(i))**(4.0d0/3.0d0))/ &
            (-2.0d0 + 2.0d0**(4.0d0/3.0d0))
       fz(i) = (4.0d0/3.0d0)*((1.0d0+zeta(i))**(1.0d0/3.0d0) &
            - (1.0d0-zeta(i))**(1.0d0/3.0d0))/ &
            (-2.0d0 + 2.0d0**(4.0d0/3.0d0))
    end do

    eu => Gcor(np, rs, 0.0310907d0, 0.21370d0, 7.5957d0, 3.5876d0, &
         1.6382d0, 0.49294d0, 1.0d0)
    ep => Gcor(np, rs, 0.01554535d0, 0.20548d0, 14.1189d0, 6.1977d0, &
         3.3662d0, 0.62517d0, 1.0d0)
    alf => Gcor(np, rs, 0.0168869d0, 0.11125d0, 10.357d0, 3.6231d0, &
         0.88026d0, 0.49671d0, 1.0d0)
!$omp parallel do
    do i=1,np
       alf(i) = - alf(i)
       ec(i) = eu(i) + (ep(i) - eu(i))*f(i)*zeta(i)**4 + &
            alf(i)*f(i)*(1.0d0 - zeta(i)**4)/fzz
       eczet(i) = 4.0d0*(ep(i)-eu(i)-alf(i)/fzz)*f(i)*zeta(i)**3 &
            + fz(i)*((ep(i)-eu(i))*zeta(i)**4 + &
         alf(i)*(1.0d0-zeta(i)**4)/fzz)
    end do
    deallocate(eu,ep,alf)
    !
    eurs => DGcor(np, rs, 0.0310907d0, 0.21370d0, 7.5957d0, 3.5876d0, &
         1.6382d0, 0.49294d0, 1.0d0)
    eprs => DGcor(np, rs, 0.01554535d0, 0.20548d0, 14.1189d0, 6.1977d0, &
         3.3662d0, 0.62517d0, 1.0d0)
    alfrs => DGcor(np, rs, 0.0168869d0, 0.11125d0, 10.357d0, 3.6231d0, &
         0.88026d0, 0.49671d0, 1.0d0)
!$omp parallel do
    do i=1,np
       alfrs(i) = - alfrs(i)
       ecrs(i) = eurs(i) + (eprs(i) - eurs(i))*f(i)*zeta(i)**4 &
            + alfrs(i)*f(i)*(1.0d0 - zeta(i)**4)/fzz
    end do
    deallocate(eurs,eprs,alfrs,f)
    !
    return
  end subroutine Ecpw91
  !
  subroutine Expw91(np, rs, s, ex, exrs, exs)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: rs, s, ex, exrs, exs
    !
    real(kind=8):: q1, q3, a, a1, a2, a3, a4, b1
    ! real(kind=8),parameter:: Pi=3.14159265358979323846d0
    ! parameter(q1=(9.0d0*Pi/4.0d0)**(1.0d0/3.0d0))
    parameter(q1=1.91915829267751d0)
    ! parameter(q3=-3.0d0/(4.0d0*Pi))
    parameter(q3=-0.238732414637843d0)
    parameter(a=7.7956d0, a1=0.19645d0, a2=0.27430d0, a3=0.15084d0)
    parameter(a4=-100d0, b1=0.004d0)
    !
    real(kind=8):: p1,p2
    integer:: i
    !
!$omp parallel do private(p1,p2)
    do i=1,np
       p1 = log(a*s(i) + sqrt(1.0d0+(a*s(i))**2))
       p2 = exp(a4*s(i)**2)
       ex(i) = q1*q3*(1.0d0 + (a2 - a3*p2)*s(i)**2 + a1*s(i)*p1)/ &
            (rs(i)*(1.0d0 + b1*s(i)**4 + a1*s(i)*p1))
       exrs(i) = -(q1*q3*(1.0d0 + (a2 - a3*p2)*s(i)**2 + a1*s(i)*p1)/ &
            (rs(i)**2*(1.0d0 + b1*s(i)**4 + a1*s(i)*p1)))
       exs(i) = -(q1*q3*(4*b1*s(i)**3 + a*a1*s(i)/ &
            Sqrt(1.0d0+(a*s(i))**2) + a1*p1)* &
            (1.0d0 + (a2 - a3*p2)*s(i)**2 + a1*s(i)*p1)/ &
            (rs(i)*(1.0d0 + b1*s(i)**4 + a1*s(i)*p1)**2)) + &
            q1*q3*(2.0d0*(a2 - a3*p2)*s(i) - 2.0d0*a3*a4*p2*s(i)**3 + &
            a*a1*s(i)/Sqrt(1.0d0 + (a*s(i))**2) + a1*p1)/ &
            (rs(i)*(1.0d0 + b1*s(i)**4 + a1*s(i)*p1))
    end do
    return
  end subroutine Expw91

  subroutine H0pw91(np, A, t, g, h0, h0a, h0t, h0g)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: A, t, g, h0, h0a, h0t, h0g
    !
    real(kind=8):: nu, cc0, alph, bet
    ! real(kind=8),parameter:: Pi=3.14159265358979323846d0
    ! parameter(nu=(16.0d0/Pi)*(3*Pi**2)**(1.0d0/3.0d0))
    parameter(nu=15.75592034948314d0)
    parameter(cc0=0.004235d0)
    ! real(kind=8),parameter:: cx=-0.001667212d0
    parameter(alph=0.09d0,bet=nu*cc0)
    !
    real(kind=8):: p1,p2,p3
    integer:: i
    !
!$omp parallel do private(p1,p2,p3)
    do i=1,np
       p1 = 1.0d0 + A(i)*t(i)**2 + A(i)**2*t(i)**4
       p2 = t(i)**2 + A(i)*t(i)**4
       p3 = Log(1.0d0 + 2.0d0*alph*p2/(bet*p1))
       h0(i) = bet**2*g(i)**3*p3/(2.0d0*alph)
       h0a(i) = bet**2*g(i)**3* &
            (-2*alph*p2*(t(i)**2 + 2*A(i)*t(i)**4)/(bet*p1**2) + &
            2.0d0*alph*t(i)**4/(bet*p1))/ &
            (2.0d0*alph*(1.0d0 + 2.0d0*alph*p2/(bet*p1)))
       h0t(i) = bet**2*g(i)**3* &
            (-2.0d0*alph* &
            (2.0d0*A(i)*t(i) + 4.0d0*A(i)**2*t(i)**3)*p2/(bet*p1**2) + &
            2.0d0*alph*(2.0d0*t(i) + 4.0d0*A(i)*t(i)**3)/(bet*p1))/ &
            (2.0d0*alph*(1.0d0 + 2.0d0*alph*p2/(bet*p1)))
       h0g(i) = 3.0d0*bet**2*g(i)**2*p3/(2.0d0*alph)
    end do
    return
  end subroutine H0pw91

  subroutine Apw91(np,Ecl,g,a,aecl,ag)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: Ecl,g,a,aecl,ag
    !
    real(kind=8):: nu, cc0, alph, bet
    ! real(kind=8),parameter:: Pi=3.14159265358979323846d0
    ! parameter(nu=(16.0d0/Pi)*(3*Pi**2)**(1.0d0/3.0d0))
    parameter(nu=15.75592034948314d0)
    parameter(cc0=0.004235d0)
    parameter(alph=0.09d0,bet=nu*cc0)
    !
    real(kind=8):: p1
    integer:: i
    !
!$omp parallel do private(p1)
    do i=1,np
       p1 = exp(-2.0d0*alph*Ecl(i)/(bet**2*g(i)**3))
       a(i) = 2.0d0*alph/(bet*(-1.0d0 + p1))
       aecl(i) = 4.0d0*alph**2/(bet**3*(-1.0d0 + p1)**2*g(i)**3)*p1
       ag(i) = -12.0d0*alph**2/(bet**3*(-1.0d0 + p1)**2*g(i)**4)*p1*ecl(i)
    end do
    return
  end subroutine Apw91

  subroutine H1pw91(np,rs,t,g,h1,h1rs,h1t,h1g)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: rs, t, g
    !
    real(kind=8),dimension(np):: h1,h1rs,h1t,h1g
    !
    real(kind=8):: q1, q2, nu, cc0, cx
    real(kind=8):: c1,c2,c3,c4,c5,c6,d1
    ! real(kind=8),parameter:: Pi=3.14159265358979323846d0
    ! parameter(q1=(9.0d0*Pi/4.0d0)**(1.0d0/3.0d0))
    parameter(q1=1.91915829267751d0)
    ! parameter(q2=2.0d0*(q1/Pi)**(1.0d0/2.0d0))
    parameter(q2=1.56318528359354d0)
    ! parameter(nu=(16.0d0/Pi)*(3*Pi**2)**(1.0d0/3.0d0))
    parameter(nu=15.75592034948314d0)
    parameter(cc0=0.004235d0,cx=-0.001667212d0)
    parameter(c1=1.0d-3*2.568d0, c2=1.0d-3*23.266d0)
    parameter(c3=1.0d-3*7.389d-3)
    parameter(c4=8.723d0, c5=0.472d0, c6=10*7.389d-3)
    parameter(d1=-100.0d0)
    !
    real(kind=8):: p1, p2, p3, p4
    integer:: i
    !
!$omp parallel do private(p1,p2,p3,p4)
    do i=1,np
       p1 = exp(d1*g(i)**4*q2**2*rs(i)*t(i)**2/q1**2)
       p2 = c1 + c2*rs(i) + c3*rs(i)**2
       p3 = 1.0d0 + c4*rs(i) + c5*rs(i)**2 + c6*rs(i)**3
       p4 = -cc0 - 10.0d0*cx/7.0d0 + p2/p3
       h1(i) = p1*g(i)**3*nu*p4*t(i)**2
       h1rs(i) = p1*g(i)**3*nu* &
            (-(p2*(c4 + 2*c5*rs(i) + 3*c6*rs(i)**2)/p3**2) + &
            (c2 + 2*c3*rs(i))/p3)*t(i)**2 + &
            d1*p1*g(i)**7*nu*q2**2*p4*t(i)**4/q1**2
       h1t(i) = 2.0d0*p1*g(i)**3*nu*p4*t(i) + &
            2.0d0*d1*p1*g(i)**7*nu*q2**2*rs(i)*p4*t(i)**3/q1**2
       h1g(i) = 3.0d0*p1*g(i)**2*nu*p4*t(i)**2 + &
            4.0d0*d1*p1*g(i)**6*nu*q2**2*rs(i)*p4*t(i)**4/q1**2
    end do
    return
  end subroutine H1pw91

  subroutine pw91_exchcor(np, n, dn, exc, fxcn, fxcdn, mssg)
    implicit none
    integer:: np
    real(kind=8),dimension(np),intent(in):: n, dn
    real(kind=8),dimension(np)::  exc, fxcn, fxcdn
    integer,intent(in):: mssg
    !
    real(kind=8):: Pi,neps
    parameter(Pi=3.14159265358979323846d0)
    parameter(neps=1.0d-8)
    !
    real(kind=8),dimension(:),allocatable:: rs,kf,s,g,zeta,ks,t
    real(kind=8),dimension(:),allocatable:: ex, exrs, exs
    real(kind=8),dimension(:),allocatable:: ecl,eclrs,eclzet,a,aecl,ag
    real(kind=8),dimension(:),allocatable:: h0,h0a,h0t,h0g
    real(kind=8),dimension(:),allocatable:: h1,h1rs,h1t,h1g
    !
    integer:: i
    if (mssg.gt.0) then
       write(6,*) 'using PW91'
    end if
    allocate(rs(np),kf(np),s(np),g(np),zeta(np),ks(np),t(np), &
         ex(np),exrs(np),exs(np), &
         ecl(np),eclrs(np),eclzet(np),a(np),aecl(np),ag(np), &
         h0(np),h0a(np),h0t(np),h0g(np),h1(np),h1rs(np),h1t(np),h1g(np))
!$omp parallel do
    do i=1,np
       zeta(i) = 0.0d0
       rs(i) = (3.0d0/(4.0d0*Pi*max(n(i),neps)))**(1.0d0/3.0d0)
       kf(i) = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/rs(i)
       s(i) = dn(i)/(2.0d0*kf(i)*max(n(i),neps))
       g(i) = ((1.0d0+zeta(i))**(2.0d0/3.0d0) &
            + (1.0d0-zeta(i))**(2.0d0/3.0d0))/2.0d0
       ks(i) = sqrt(4.0d0*kf(i)/Pi)
       t(i) = dn(i)/(2.0d0*g(i)*ks(i)*max(n(i),neps))
    end do
    !
    call Expw91(np,rs,s,ex,exrs,exs)
    call Ecpw91(np,rs,zeta,ecl,eclrs,eclzet)
    call Apw91(np,ecl,g,a,aecl,ag)
    call H0pw91(np,a,t,g,h0,h0a,h0t,h0g)
    call H1pw91(np,rs,t,g,h1,h1rs,h1t,h1g)
    !
!$omp parallel do
    do i=1,np
       exc(i) = ex(i) + ecl(i) + h0(i) + h1(i)
       fxcn(i) = exc(i)-(exrs(i) + eclrs(i) + h0a(i)*aecl(i)*eclrs(i) &
            + h1rs(i))/3.0d0*rs(i) &
            - 4.0d0/3.0d0*s(i)*exs(i) - 7.0d0/6.0d0*t(i)*(h0t(i) + h1t(i))
       fxcdn(i) = exs(i)/(2.0d0*kf(i)) + (h0t(i) + h1t(i))/(2.0d0*g(i)*ks(i))
    end do
    !
    deallocate(rs,kf,s,g,zeta,ks,t,ex,exrs,exs, &
         ecl,eclrs,eclzet,a,aecl,ag,h0,h0a,h0t,h0g,h1,h1rs,h1t,h1g)
    return
  end subroutine pw91_exchcor

  subroutine pw91_exchcorsp(np, pn, pdn, dn, exc, excor, &
       pfxcn, pfxdn, fcdn, mssg)
    implicit none
    integer:: np
    real(kind=8),intent(in):: pn(np,2), pdn(np,2), dn(np)
    real(kind=8),intent(out):: exc
    real(kind=8),intent(out):: excor(np)
    real(kind=8),intent(out):: pfxcn(np,2), pfxdn(np,2), fcdn(np)
    integer,intent(in):: mssg
    !
    real(kind=8):: Pi,neps,eta
    parameter(Pi=3.14159265358979323846d0)
    parameter(neps=1.0d-8)
    parameter(eta=1.0d-8) ! stop gzeta from blowing up at |zeta| = 1
    !
    real(kind=8),dimension(:),allocatable:: n,rs,kf,s,zeta,ks,t,g,gzeta
    real(kind=8),dimension(:,:),allocatable:: pex, pexrs, pexs
    real(kind=8),dimension(:),allocatable:: ecl,eclrs,eclzet,a,aecl,ag
    real(kind=8),dimension(:),allocatable:: h0,h0a,h0t,h0g
    real(kind=8),dimension(:),allocatable:: h1,h1rs,h1t,h1g
    real(kind=8),dimension(:,:),allocatable:: qrs,qkf,qs
    integer:: i
    real(kind=8):: fcn,fczeton,pfxn(2)
    !
    if (mssg.gt.0) then
       write(6,*) 'using PW91(LSDA)'
    end if
    allocate(n(np),rs(np),kf(np),s(np),zeta(np), &
         ks(np),t(np),g(np), &
         pex(np,2),pexrs(np,2),pexs(np,2), &
         ecl(np),eclrs(np),eclzet(np), &
         qrs(np,2),qkf(np,2),qs(np,2))
!$omp parallel do
    do i=1,np
       n(i) = pn(i,1)+pn(i,2)
       zeta(i) = (pn(i,1)-pn(i,2))/max(n(i),neps)
       rs(i) = (3.0d0/(4.0d0*Pi*max(n(i),neps)))**(1.0d0/3.0d0)
       kf(i) = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/rs(i)
       s(i) = dn(i)/(2.0d0*kf(i)*max(n(i),neps))
       g(i) = ((1.0d0+zeta(i))**(2.0d0/3.0d0) &
            + (1.0d0-zeta(i))**(2.0d0/3.0d0))/2.0d0
       ks(i) = sqrt(4.0d0*kf(i)/Pi)
       t(i) = dn(i)/(2.0d0*g(i)*ks(i)*max(n(i),neps))
    end do
    !
    ! qn := 1/2*pn, qdn := 1/2*pdn
    ! qn, qdn -> qrs, qkf, qs
    !
!$omp parallel do
    do i=1,np
       qrs(i,1) = (3.0d0/(4.0d0*Pi*max(2.0d0*pn(i,1),neps)))**(1.0d0/3.0d0)
       qkf(i,1) = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/qrs(i,1)
       qs(i,1) = (2.0d0*pdn(i,1))/(2.0d0*qkf(i,1)*max(2.0d0*pn(i,1),neps))

       qrs(i,2) = (3.0d0/(4.0d0*Pi*max(2.0d0*pn(i,2),neps)))**(1.0d0/3.0d0)
       qkf(i,2) = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/qrs(i,2)
       qs(i,2) = (2.0d0*pdn(i,2))/(2.0d0*qkf(i,2)*max(2.0d0*pn(i,2),neps))
    end do
    !
    call Expw91(np*2,qrs,qs,pex,pexrs,pexs)
    call Ecpw91(np,rs,zeta,ecl,eclrs,eclzet)
    !
    allocate(a(np),aecl(np),ag(np), &
         h0(np),h0a(np),h0t(np),h0g(np),h1(np),h1rs(np),h1t(np),h1g(np))
    !
    call Apw91(np,ecl,g,a,aecl,ag)
    call H0pw91(np,a,t,g,h0,h0a,h0t,h0g)
    call H1pw91(np,rs,t,g,h1,h1rs,h1t,h1g)
    !
    allocate(gzeta(np))
!$omp parallel do
    do i=1,np
       gzeta(i) = (((1+zeta(i))**2+eta)**(-1.0d0/6.0d0) &
            - ((1-zeta(i))**2 + eta)**(-1.0d0/6.0d0))/3.0d0
    end do

    exc = 0.0d0
!$omp parallel do private(fcn,fczeton,pfxn),reduction(+:exc)
    do i=1,np
       exc = exc + pn(i,1)*pex(i,1) + pn(i,2)*pex(i,2) &
            + n(i)*(ecl(i) + h0(i) + h1(i))
       excor(i) = pn(i,1)*pex(i,1) + pn(i,2)*pex(i,2) &
            + n(i)*(ecl(i) + h0(i) + h1(i))
       pfxn(1) = pex(i,1) - pexrs(i,1)/3.0d0*qrs(i,1) &
            - 4.0d0/3.0d0*qs(i,1)*pexs(i,1)
       pfxn(2) = pex(i,2) - pexrs(i,2)/3.0d0*qrs(i,2) &
            - 4.0d0/3.0d0*qs(i,2)*pexs(i,2)
       fcn = ecl(i) + h0(i) + h1(i) &
            - (eclrs(i) + h0a(i)*aecl(i)*eclrs(i))/3.0d0*rs(i) &
            - 7.0d0/6.0d0*t(i)*(h0t(i) + h1t(i))
       fczeton = eclzet(i) + h0a(i)*(aecl(i)*eclzet(i)+ag(i)*gzeta(i)) &
         + ((h0t(i) + h1t(i))*(-t(i)/g(i)) + h0g(i) + h1g(i) )*gzeta(i)
       pfxcn(i,1) = pfxn(1) + fcn + fczeton*(1.0d0-zeta(i))
       pfxcn(i,2) = pfxn(2) + fcn - fczeton*(1.0d0+zeta(i))
    end do
!$omp parallel do
    do i=1,np
       fcdn(i) = (h0t(i) + h1t(i))/(2.0d0*g(i)*ks(i))
       pfxdn(i,1) = pexs(i,1)/(2.0d0*qkf(i,1))
       pfxdn(i,2) = pexs(i,2)/(2.0d0*qkf(i,2))
    end do
    !
    deallocate(n,rs,kf,s,zeta,ks,t,g,gzeta,pex,pexrs,pexs, &
         ecl,eclrs,eclzet,qrs,qkf,qs, &
         a,aecl,ag,h0,h0a,h0t,h0g,h1,h1rs,h1t,h1g)
    return
  end subroutine pw91_exchcorsp
end module pw91_xcgga
