!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
program testallzsum
  use mpi
  use xmpi
  implicit none
  integer,parameter:: n = 800*800
  real(kind=8),parameter:: eps = 1.0d-8
  complex(kind=8):: a(n),pa(n),ta(n)
  integer:: comm,myrank,nproc,ierr,iseed,i
  real(kind=8):: wt0,wt1,wtick,nop
  integer:: niter, iter

  call MPI_INIT(ierr)
  comm = MPI_COMM_WORLD
  
  call MPI_COMM_RANK(comm,myrank,ierr)
  call MPI_COMM_SIZE(comm,nproc,ierr)

  iseed = 1234*myrank + 567
  do i=1,n
     pa(i) = ran0(iseed)
  end do

  niter = nint(10.0d0*200.0d6/(8.0d0*dble(n)))+1
  nop = 16.0d0*n*niter

  wtick = MPI_WTICK()
  if (myrank .eq. 0) then
     write(6,*) 'number of MPI proc. = ', nproc
     write(6,*) 'wtick = ', wtick
  end if

  call MPI_BARRIER(comm,ierr)
  wt0 = MPI_WTIME()
  do iter = 1, niter
     call MPI_ALLREDUCE(pa,a,n,MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
  end do
  call MPI_BARRIER(comm,ierr)
  wt1 = MPI_WTIME()
  if (myrank .eq. 0) then
     write(6,*) 'MPI_ALLREDUCE'
     write(6,*) (wt1-wt0)*1.0d6, ' micro s in total'
     write(6,*) (wt1-wt0)/niter*1.0d6, ' micro s / iter'
     write(6,*) nop/((wt1-wt0))*1.0d-6, ' MB/s'
  end if

  call MPI_BARRIER(comm,ierr)
  wt0 = MPI_WTIME()
  do iter = 1, niter
     call xmpi_allzsum2(pa,ta,n,comm,ierr)
  end do
  call MPI_BARRIER(comm,ierr)
  wt1 = MPI_WTIME()
  if (myrank .eq. 0) then
     write(6,*) 'xmpi_allzsum2'
     write(6,*) (wt1-wt0)*1.0d6, ' micro s in total'
     write(6,*) (wt1-wt0)/niter*1.0d6, ' micro s / iter'
     write(6,*) nop/((wt1-wt0))*1.0d-6, ' MB/s'
  end if

  do i=1,n
     if (abs(a(i)-ta(i)).gt.eps) then
        write(6,*) myrank,i,a(i),ta(i)
     end if
  end do

  call MPI_BARRIER(comm,ierr)
  wt0 = MPI_WTIME()
  do iter = 1, niter
     call xmpi_allzsum(pa,ta,n,comm,ierr)
  end do
  call MPI_BARRIER(comm,ierr)
  wt1 = MPI_WTIME()
  if (myrank .eq. 0) then
     write(6,*) 'xmpi_allzsum'
     write(6,*) (wt1-wt0)*1.0d6, ' micro s in total'
     write(6,*) (wt1-wt0)/niter*1.0d6, ' micro s / iter'
     write(6,*) nop/((wt1-wt0))*1.0d-6, ' MB/s'
  end if

  do i=1,n
     if (abs(a(i)-ta(i)).gt.eps) then
        write(6,*) myrank,i,a(i),ta(i)
     end if
  end do

  call MPI_FINALIZE(ierr)

  stop

  contains
  real(kind=8) function ran0(idum)
    implicit none
    integer idum,IA,IM,IQ,IR
    real(kind=8) AM
    parameter (IA=16807,IM=2147483647,AM=1./IM,IQ=127773,IR=2836)
    integer k
    if (idum .eq. 0) then
       write(6,*) 'ran0: idum = 0'
       stop
    end if
    k=idum/IQ
    idum=IA*(idum-k*IQ)-IR*k
    if (idum.lt.0) idum=idum+IM
    ran0=2.0d0*(AM*idum-0.5d0)
    return
  end function ran0
end program testallzsum

