!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
program testallzsum
  use mpi
  implicit none
  real(kind=8),allocatable:: sendbuf(:),recvbuf(:)
  integer:: comm,myrank,nproc,ierr,i,dst,src
  real(kind=8):: wt0,wt1,wtick,nop
  integer:: n,niter,iter
  integer,parameter:: tag = 1000
  integer:: status(MPI_STATUS_SIZE)
  integer,parameter:: ninit = 128, nstep = 16

  call MPI_INIT(ierr)
  comm = MPI_COMM_WORLD
  
  call MPI_COMM_RANK(comm,myrank,ierr)
  call MPI_COMM_SIZE(comm,nproc,ierr)

  wtick = MPI_WTICK()
  if (myrank .eq. 0) then
     write(6,*) 'number of MPI proc. = ', nproc
     write(6,*) 'wtick = ', wtick
  end if

  n = ninit
  dst = modulo(myrank+1,nproc)
  src = modulo(myrank-1,nproc)
  do i=1,nstep
     allocate(sendbuf(n),recvbuf(n))
     call MPI_BARRIER(comm,ierr)
     niter = nint(10.0d0*200.0d6/(8.0d0*dble(n)))+1
     wt0 = MPI_WTIME()
     do iter = 1, niter
        call MPI_SENDRECV(sendbuf,n, MPI_DOUBLE_PRECISION, &
             dst, tag, &
             recvbuf, n, MPI_DOUBLE_COMPLEX, &
             src, tag, &
             comm, status, ierr)
     end do
     call MPI_BARRIER(comm,ierr)
     wt1 = MPI_WTIME()
     if (myrank .eq. 0) then
        nop = 8.0d0*n*niter
        write(6,*) 'buffer size = ', 8*n, ' Byte'
        write(6,*) (wt1-wt0)*1.0d6, ' micro s in total'
        write(6,*) (wt1-wt0)/niter*1.0d6, ' micro s / iter'
        write(6,*) nop/((wt1-wt0))*1.0d-6, ' MB/s'
     end if
     deallocate(sendbuf,recvbuf)
     n = n*2
  end do
  call MPI_FINALIZE(ierr)
  stop
end program testallzsum

